
/*******************************************************/
/**                                                   **/
/**       annon.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**         Parsing connections to telnet port        **/
/** It's not yet known whether the connection is from **/
/**    a bot or from a user who tries to enter the    **/
/**                     partyline                     **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>


#include "annon.h"
#include "lines.h"
#include "misc.h"
#include "net.h"
#include "bots.h"
#include "partyline.h"
#include "config.h"
#include "session.h"
#include "console.h"
#include "gencmd.h"
#include "on.h"
#include "out_console.h"


struct annonlist {
  struct annoninfo *first;
  struct annoninfo *tail;
  struct annoninfo *current;
} anlist={ NULL , NULL , NULL };

struct annoninfo {
  char nick[10];
  short int status;   /* 0 is waiting for nick, 1 checking botpassword, 2 checking linepassword, */
                                       /* 3 checking consolepasswd */
  int sockptr;
  struct annoninfo *next;
  struct annoninfo *prev;
  char *from;
  time_t time;        /* When was the annonymous connection created? Needed for timeout */
};

extern char  botnick[10];
extern int   default_outside_partyline;
extern char *motd_file;
extern char *su_password;
extern time_t time_null;

void addannon(int sockptr) {

  char *hostname;
  int   linenr, script_takes_care;

  anlist.current = (struct annoninfo *) xmalloc (sizeof(struct annoninfo));
  if (anlist.first == NULL) {
    anlist.first = anlist.current;
    anlist.current->prev = NULL;
  }
  else {
    anlist.current->prev = anlist.tail;
    anlist.tail->next = anlist.current;
  }
  anlist.current->next = NULL;
  anlist.tail = anlist.current;

  strcpy(anlist.current->nick, "");
  anlist.current->status = 0;
  anlist.current->sockptr = sockptr;
  anlist.current->from = NULL;
  anlist.current->time = time_null;

  hostname = give_connectedhost_name(sockptr);
  script_takes_care = on_connect(hostname);
  if (!script_takes_care) {
    sendtospies("Connection from: %s", hostname);
    console_send2snoopwindow("[net] Connection from: %s", hostname);
  }
  if (!(does_this_match('x', hostname))) {
    if (!script_takes_care) {
      sendtospies("%s is not matching with X-line", hostname);
      console_send2snoopwindow("[net] No X line for %s", hostname);
    }
    send_to_socketptr(sockptr, "You don't have access. (No X-line)\r");
    close_socketptr(sockptr);
    annonconnectionlost();
    return;
  }
  if ((linenr = does_this_match('n', hostname))) {
    if (!script_takes_care) {
      sendtospies("%s is matching with N-line. (N%d: %s)", hostname, linenr, lines_comment('n'));
      console_send2snoopwindow("[net] N%d: %s (%s)", linenr, lines_mask('n'), lines_comment('n'));
    }
    send_to_socketptr(sockptr, "You don't have access. (N%d: %s)\r", linenr, lines_comment('n'));
    close_socketptr(sockptr);
    annonconnectionlost();
    return;
  }
  anlist.current->from = (char *)xmalloc(strlen(hostname) + 1);
  strcpy(anlist.current->from, hostname);
  send_to_socketptr(sockptr, "\r\n\r\n\r\nPlease enter your nick\r");
}

void annonconnectionlost() {

  if (anlist.current == NULL) return;

  xfree(anlist.current->from);
  if ((anlist.current == anlist.first) && (anlist.current == anlist.tail)) {
    xfree(anlist.current);
    anlist.current = NULL;
    anlist.tail = NULL;
    anlist.first = NULL;
    return;
  }

  if (anlist.current->next == NULL) {
    anlist.tail = anlist.current->prev;
    anlist.tail->next = NULL;
    xfree(anlist.current);
    anlist.current = anlist.tail;
    return;
  }

  if (anlist.current->prev == NULL) {
    anlist.first = anlist.current->next;
    anlist.first->prev = NULL;
    xfree(anlist.current);
    anlist.current = anlist.first;
    return;
  }

  anlist.current->prev->next = anlist.current->next;
  anlist.current->next->prev = anlist.current->prev;
  xfree(anlist.current);
  anlist.current = anlist.first;
  return;
}

int annon_findfirst() { return (int)(anlist.current = anlist.first); }
int annon_findnext()  { return (int)(anlist.current = anlist.current->next); }

int annon_findsockptr(int sockptr) {

  anlist.current = anlist.first;
  while (anlist.current != NULL) {
    if (anlist.current->sockptr == sockptr) return 1;
    anlist.current = anlist.current->next;
  }
  return 0;
}

void parse_annon(char *message) {

  int i=0, linenr;

  switch(anlist.current->status) {
  case 0: /* Checking for nickname */
    if (strlen(message) > 9) {
      send_to_socketptr(anlist.current->sockptr, "Nickname is too long, enter a shorter nick\r");
      return;
    }
    if (strlen(message) == 0) {
      send_to_socketptr(anlist.current->sockptr, "Nickname is too short, enter a longer nick\r");
      return;
    }
    for (i=0; message[i]!='\0'; i++) {
      if ((isspace(message[i])) || (message[i]=='@') || (message[i]==':') || (!(isgraph(message[i])))) {
	send_to_socketptr(anlist.current->sockptr, "Nickname contains an illegal character, enter a new nick\r");
	return;
      }
    }
    if (strcasecmp(message, botnick) == 0) {
      send_to_socketptr(anlist.current->sockptr, "Illegal nickname\r");
      return;
    }
    strcpy(anlist.current->nick, message);
    if (strcasecmp(message, "console") == 0) {
      anlist.current->status = 3;
      return;
    }
    else if (botlist_findbot(anlist.current->nick)) {
      if (botlist_allowconnection()) {
	send_to_socketptr(anlist.current->sockptr, "passreq\r");
	anlist.current->status = 1;
	return;
      }
      else {
	sendtospies("Connection from %s@%s, a bot who is currently not allowed to link to me", anlist.current->nick,
		    anlist.current->from);
	console_send2snoopwindow("[net] %s@%s (bot): connection (currently) not allowed", anlist.current->nick,
				 anlist.current->from);
	send_to_socketptr(anlist.current->sockptr, "You don't have access\r");
	close_socketptr(anlist.current->sockptr);
	annonconnectionlost();
	return;
      }
    }
    else {
      if ((session_findnick(anlist.current->nick)) || (party_findnick(anlist.current->nick))) {
	send_to_socketptr(anlist.current->sockptr, "Nickname is in use, enter a new nick\r");
	anlist.current->status = 0;
	return;
      }
      else {
	if (!(linenr = does_this_match('x', anlist.current->from))) {
	  sendtospies("%s is not matching with X-line", anlist.current->from);
	  console_send2snoopwindow("[net] No X line for %s (%s)", anlist.current->from, anlist.current->nick);
	  send_to_socketptr(anlist.current->sockptr, "You don't have access. (No X-line)\r");
	  close_socketptr(anlist.current->sockptr);
	  annonconnectionlost();
	  return;
	}
	if (strcasecmp("", lines_comment('x')) == 0) {
	  catfile(motd_file, annon_send2current, anlist.current->nick);
	  send_to_socketptr(anlist.current->sockptr, "Welcome on the partyline of %s (%s)\r", botnick, LONGVERSION);
	  changesocketptrtype(anlist.current->sockptr, T_PARTY);
	  create_party(anlist.current->nick, anlist.current->sockptr, anlist.current->from);
	  if (!(party_joinchannel(default_outside_partyline))) { 
	    close_socketptr(anlist.current->sockptr); 
	    delete_party();
	  }
	  annonconnectionlost();
	  return;
	}
	else {
	  send_to_socketptr(anlist.current->sockptr, "Please enter password of line X%d\r", linenr);
	  anlist.current->status = 2;
	  return;
	}
      }
    }
  case 1: /* Checking for password of bot */
    if ((!(botlist_findbot(anlist.current->nick))) || (!(botlist_allowconnection()))) {
      sendtospies("%s@%s is not allowed to link to me anymore", anlist.current->nick, anlist.current->from);
      console_send2snoopwindow("[net] %s@%s (bot): connection (currently) not allowed", anlist.current->nick,
			       anlist.current->from);
      send_to_socketptr(anlist.current->sockptr, "You don't have access\r");
      close_socketptr(anlist.current->sockptr);
      annonconnectionlost();
      return;
    }
    if (strcasecmp(message, botlist_password())==0) {
      botlist_incominglink(anlist.current->sockptr);
      changesocketptrtype(anlist.current->sockptr, T_BOT);
      annonconnectionlost();
      return;
    }
    else {
      sendtospies("Bad password attempt from bot %s@%s", anlist.current->nick, anlist.current->from);
      console_send2snoopwindow("[net] Bad password (%s) from %s@%s (bot)", message, anlist.current->nick, 
			       anlist.current->from);
      send_to_socketptr(anlist.current->sockptr, "badpass\r");
      close_socketptr(anlist.current->sockptr);
      annonconnectionlost();
      return;
    }
  case 2:  /* Checking line-password */
    if (!(linenr = does_this_match('x', give_connectedhost_name(anlist.current->sockptr)))) {
      sendtospies("%s@%s is not matching with X-line", anlist.current->nick, anlist.current->from);
      console_send2snoopwindow("[net] No X line for %s (%s)", anlist.current->from, anlist.current->nick);
      send_to_socketptr(anlist.current->sockptr, "You don't have access. (No X-line)\r");
      close_socketptr(anlist.current->sockptr);
      annonconnectionlost();
      return;
    }
    if (strcasecmp(message, lines_comment('x')) == 0) {
      if (party_findnick(anlist.current->nick) || session_findnick(anlist.current->nick)) {
	send_to_socketptr(anlist.current->sockptr, "Your nickname has been taken...\r\nPlease enter a new one\r");
	anlist.current->status = 0;
	return;
      }
      send_to_socketptr(anlist.current->sockptr, "Welcome on the partyline of %s (%s)\r", botnick, 
			LONGVERSION);
      changesocketptrtype(anlist.current->sockptr, T_PARTY);
      create_party(anlist.current->nick, anlist.current->sockptr, anlist.current->from);
      if (!(party_joinchannel(default_outside_partyline))) { 
	close_socketptr(anlist.current->sockptr); 
	delete_party();
      }
      annonconnectionlost();
      return;
    }
    else {
      sendtospies("Bad password from user %s@%s (Line X%d)", anlist.current->nick, anlist.current->from, linenr);
      sendtoadminspies("Given password: %s  Should be: %s", message, lines_comment('x'));
      console_send2snoopwindow("[net] Bad password (%s) from %s@%s (user)", message, anlist.current->nick, 
					  anlist.current->from);
      send_to_socketptr(anlist.current->sockptr, "Sorry\r");
      close_socketptr(anlist.current->sockptr);
      annonconnectionlost();
      return;
    }
  case 3:  /* checking console-password ( = superuser password ) */
    if (strcmp(message, su_password) == 0) {
      changesocketptrtype(anlist.current->sockptr, T_OUTCONSOLE);
      out_console_add(anlist.current->sockptr);
      annonconnectionlost();
    }
    else {
      send_to_socketptr(anlist.current->sockptr, "Sorry\r");
      close_socketptr(anlist.current->sockptr);
      annonconnectionlost();
    }
    return;
  }
}

void check_annon_timeout() {

  if (annon_findfirst()) {
    do
      {
	if (anlist.current->time + ANNON_TIMEOUT < time_null) {
	  send_to_socketptr(anlist.current->sockptr, "Timeout\r");
	  close_socketptr(anlist.current->sockptr);
	  annonconnectionlost();                      /* current of annon list is not defined now */
	  if (!(annon_findfirst())) return;           /* set current to the first position */
	}
      }
    while (annon_findnext());
  }
}

void annon_send2current(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  send_to_socketptr(anlist.current->sockptr, "%s", line);
}
