
/*******************************************************/
/**                                                   **/
/**       assoc.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**      Database containing party-channel-names      **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "misc.h"
#include "assoc.h"

struct assoclist {
  struct assocnode *current;
  struct assocnode *head;
  struct assocnode *tail;
} asclist = { NULL , NULL , NULL };

struct assocnode {
  int channel;
  char *alias;

  struct assocnode *prev;
  struct assocnode *next;
};

void addassoc(int channel, char *alias) {

  struct assocnode *tmp;

  asclist.current = (struct assocnode *) xmalloc (sizeof(struct assocnode));
  asclist.current->channel = channel;
  asclist.current->alias = (char *) xmalloc (strlen(alias) + 1);
  strcpy(asclist.current->alias, alias);
  
  if (asclist.head == NULL) {
    asclist.head = asclist.current;
    asclist.tail = asclist.current;
    asclist.current->next = NULL;
    asclist.current->prev = NULL;
    return;
  }
  
  tmp = asclist.head;
  while ((tmp != NULL) && (channel > tmp->channel)) {
    tmp = tmp->next;
  }
  if (tmp == NULL) {
    asclist.current->next = NULL;
    asclist.current->prev = asclist.tail;
    asclist.tail->next = asclist.current;
    asclist.tail = asclist.current;
    return;
  }
  if (channel == tmp->channel) {
    xfree(tmp->alias);
    asclist.current->next = tmp->next;
    asclist.current->prev = tmp->prev;    
    if (tmp != asclist.head) {
      asclist.current->prev->next = asclist.current;
    }
    else {
      asclist.head = asclist.current;
    }
    if (tmp != asclist.tail) {
      asclist.current->next->prev = asclist.current;
    }
    else {
      asclist.tail = asclist.current;
    }
    xfree(tmp);
    return;
  }
  if (channel < tmp->channel) {
    asclist.current->next = tmp;
    asclist.current->prev = tmp->prev;
    if (tmp != asclist.head) {
      asclist.current->prev->next = asclist.current;
    }
    else {
      asclist.head = asclist.current;
    }
    tmp->prev = asclist.current;
    return;
  }
}

void delassoc(int channel) {

  asclist.current = asclist.head;

  while ((asclist.current != NULL) && (asclist.current->channel < channel))
    asclist.current = asclist.current->next;

  if ((asclist.current == NULL) || (asclist.current->channel != channel))
    return;

  xfree(asclist.current->alias);

  if (asclist.head == asclist.tail) {
    xfree(asclist.current);
    asclist.head = NULL;
    asclist.current = NULL;
    asclist.tail = NULL;
    return;
  }

  if (asclist.current->next == NULL) {
    asclist.tail = asclist.current->prev;
    asclist.tail->next = NULL;
    xfree(asclist.current);
    asclist.current = asclist.tail;
    return;
  }

  if (asclist.current->prev == NULL) {
    asclist.head = asclist.current->next;
    asclist.head->prev = NULL;
    xfree(asclist.current);
    asclist.current = asclist.head;
    return;
  }

  asclist.current->prev->next = asclist.current->next;
  asclist.current->next->prev = asclist.current->prev;
  xfree(asclist.current);
  asclist.current = asclist.head;
  return;
}
  
int assoc_channelnumber(char *alias) {

  asclist.current = asclist.head;

  while (asclist.current != NULL) {
    if (strcasecmp(asclist.current->alias, alias) == 0) {
      return asclist.current->channel;
    }
    asclist.current = asclist.current->next;
  }
  return -1;
}

char *assoc_channelname(int nummer) {

  asclist.current = asclist.head;

  while (asclist.current != NULL) {
    if (asclist.current->channel == nummer) {
      return asclist.current->alias;
    }
    asclist.current = asclist.current->next;
  }
  return NULL;
}
