
/*******************************************************/
/**                                                   **/
/**        bots.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**      The botlist + bot-I/O + start of parsing     **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "session.h"
#include "bots.h"
#include "bottree.h"
#include "config.h"
#include "parsebot.h"
#include "net.h"
#include "partyline.h"
#include "misc.h"
#include "console.h"
#include "on.h"


static struct botlist {
  struct botinfo *head;
  struct botinfo *tail;
  struct botinfo *current;
} blist= { NULL , NULL , NULL };

struct botinfo {
  char nick[10];

  char *hostname;            /* Address */
  int   port;

  char  *nickuserhost;        /* Protection */
  char   protection;
  time_t last_pong;
  char   connected;          /* Connection-state: 0 no connection, 1 connecting, 2 unlinked connection, 
			      * 3 linked connection */

  char   autoconnect;        /* Should the bot always be linked in the botnet? */

  char *password;
  char *version;
  int   socketptr;

  struct botinfo *next;
  struct botinfo *prev;
};


extern char  botnick[10];
extern time_t time_null;

void addbot(char *nick, char *hostname, int port) {

  blist.current = (struct botinfo *) xmalloc (sizeof(struct botinfo));

  if (blist.head == NULL) {
    blist.head = blist.current;
    blist.current->prev = NULL;
  }
  else {
    blist.current->prev = blist.tail;
    blist.tail->next = blist.current;
  }
  blist.current->next = NULL;
  blist.tail = blist.current;

  blist.current->hostname = (char *) xmalloc (strlen(hostname) + 1);

  if (strlen(nick) > 9) nick[9] = '\0';
  strcpy(blist.current->nick, nick);
  strcpy(blist.current->hostname, hostname);
  blist.current->port = port;

  blist.current->nickuserhost = (char *) xmalloc (strlen("")+1);
  strcpy(blist.current->nickuserhost, "");

  blist.current->password = (char *) xmalloc (strlen("default")+1);
  strcpy(blist.current->password, "default");
  blist.current->version = (char *) xmalloc (strlen("")+1);
  strcpy(blist.current->version, "");

  blist.current->connected = 0;
  blist.current->socketptr = 0;
  blist.current->protection = 0;
  blist.current->autoconnect = 0;
}

void delbot() {

  if (blist.current == NULL) {
    return;
  }
  xfree(blist.current->version);
  xfree(blist.current->password);
  xfree(blist.current->hostname);
  xfree(blist.current->nickuserhost);

  botlist_breakconnection();

  if ((blist.current == blist.head) && (blist.current == blist.tail)) {
    xfree(blist.current);
    blist.current = NULL;
    blist.tail = NULL;
    blist.head = NULL;
    return;
  }

  if (blist.current->next == NULL) {
    blist.tail = blist.current->prev;
    blist.tail->next = NULL;
    xfree(blist.current);
    blist.current = blist.tail;
    return;
  }

  if (blist.current->prev == NULL) {
    blist.head = blist.current->next;
    blist.head->prev = NULL;
    xfree(blist.current);
    blist.current = blist.head;
    return;
  }

  blist.current->prev->next = blist.current->next;
  blist.current->next->prev = blist.current->prev;
  xfree(blist.current);
  blist.current = blist.head;
  return;
}

void deleteallbots() {

  blist.current = blist.head;
  while (blist.current != NULL) {
    delbot();
    blist.current = blist.head;
  }
}

int  botlist_findbot(char *nick) {

  blist.current = blist.head;
  while (blist.current != NULL) {
    if (strcasecmp(nick, botlist_nick()) == 0) return 1;
    blist.current = blist.current->next;
  }
  return 0;
}

int  botlist_findsockptr(int sockptr) {

  blist.current = blist.head;
  while (blist.current != NULL) {
    if ((blist.current->connected) && (blist.current->socketptr == sockptr)) return 1;
    blist.current = blist.current->next;
  }
  return 0;
}

int  botlist_findfirst() {

  blist.current = blist.head;
  return (blist.current != NULL);
}

int  botlist_findnext() {

  if (blist.current->next == NULL) return 0;
  blist.current = blist.current->next;
  return 1;
}

int  botlist_length() {

  struct botinfo *tmp;
  int length=0;

  tmp = blist.current;
  blist.current = blist.head;
  while (blist.current != NULL) {
    length++;
    blist.current = blist.current->next;
  }
  blist.current = tmp;
  return length;
}

char *botlist_nick() { return blist.current->nick; }
char *botlist_host() { return blist.current->hostname; }
char *botlist_password() { return blist.current->password; }
char *botlist_nickuserhost() { return blist.current->nickuserhost; }
char *botlist_version() { return blist.current->version; }
int   botlist_port() { return blist.current->port; }
int   botlist_protection() { return blist.current->protection; }
int   botlist_autoconnect() { return blist.current->autoconnect; }

int botlist_allowconnection() { return blist.current->connected == 0; }
  
void botlist_setaddress(char *host, int port) {

  xfree(blist.current->hostname);
  blist.current->hostname = (char *)xmalloc(strlen(host) + 1);
  strcpy(blist.current->hostname, host);
  blist.current->port = port;
}

void botlist_setprot()          { blist.current->protection = 1; }
void botlist_resetprot()        { blist.current->protection = 0; }
void botlist_setautoconnect()   { blist.current->autoconnect = 1; }
void botlist_resetautoconnect() { blist.current->autoconnect = 0; }

void botlist_setpong()    { blist.current->last_pong = time_null; }

void botlist_setnickuserhost(char *nickuserhost) {

  xfree(blist.current->nickuserhost);
  blist.current->nickuserhost = (char *)xmalloc(strlen(nickuserhost)+1);
  strcpy(blist.current->nickuserhost, nickuserhost);
}

void botlist_setpassword(char *password) {

  xfree(blist.current->password);
  blist.current->password = (char *)xmalloc(strlen(password)+1);
  strcpy(blist.current->password, password);
}

void botlist_setversion(char *version) {

  xfree(blist.current->version);
  blist.current->version = (char *)xmalloc(strlen(version)+1);
  strcpy(blist.current->version, version);
}

int loadbots(FILE *fp, int version) {

  int nr_bots, i;
  char *nick=NULL, *host=NULL, *passwd=NULL, *nuh=NULL;
  int   port, prot, autoconnect=0;

  deleteallbots();

  freadint(fp, &nr_bots);
  for (i=0; i<nr_bots; i++) {
    fgetstrfrombin(&nick, fp);             /* nick */
    fgetstrfrombin(&host, fp);             /* host */
    freadint(fp, &port);                   /* port */
    fgetstrfrombin(&passwd, fp);           /* password */
    fgetstrfrombin(&nuh, fp);              /* nick!user@host */
    freadint(fp, &prot);                   /* protection */
    if (version >= 32) {
      freadint(fp, &autoconnect);
    }
    addbot(nick, host, port);
    botlist_setpassword(passwd);
    botlist_setnickuserhost(nuh);
    if (prot)
      botlist_setprot();
    else
      botlist_resetprot();
    if (autoconnect)
      botlist_setautoconnect();
    else
      botlist_resetautoconnect();

    xfree(nick); nick = NULL;
    xfree(host); host = NULL;
    xfree(passwd); passwd = NULL;
    xfree(nuh); nuh = NULL;
  }
  freadint(fp, &i);
  return (i == 346785);
}

void savebots(FILE *fp) {

  fwriteint(fp, botlist_length());

  blist.current = blist.head;
  if (botlist_findfirst()) {
    do
      {
	fwritestr2bin(botlist_nick(), fp);
	fwritestr2bin(botlist_host(), fp);
	fwriteint(fp, botlist_port());
	fwritestr2bin(botlist_password(), fp);
	fwritestr2bin(botlist_nickuserhost(), fp);
	fwriteint(fp, botlist_protection());
	fwriteint(fp, botlist_autoconnect());
      }
    while (botlist_findnext());
  }
  fwriteint(fp, 346785);
}

void botlist_send2all(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct botinfo *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = blist.current;
  blist.current = blist.head;
  while (blist.current != NULL) {
    if (blist.current->connected == 3) send_to_socketptr(blist.current->socketptr, "%s", line);
    blist.current = blist.current->next;
  }
  blist.current = tmp;
}

void botlist_send2allexceptcurrent(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct botinfo *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = blist.current;
  blist.current = blist.head;
  while (blist.current != NULL) {
    if ((blist.current->connected == 3) && (tmp != blist.current)) 
      send_to_socketptr(blist.current->socketptr, "%s", line);
    blist.current = blist.current->next;
  }
  blist.current = tmp;
}

void botlist_send2current(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  if (blist.current->connected < 2) return;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  send_to_socketptr(blist.current->socketptr, "%s", line);
}

int botlist_startconnection() {

  switch(blist.current->connected) {
  case 1: return 0;
  case 2: return 1;
  case 3: return 2;
  }
  /* We are sure the bot is really unconnected now */
  botlist_setpong();    /* pong is set for timeout checking */
  switch(start_connecting(T_BOT, &blist.current->socketptr, blist.current->hostname, blist.current->port)) {
  case -1: return 3;   /* Connection refused */
  case  0: blist.current->connected = 1; return 4;
  case  1: blist.current->connected = 2; return 5;
  }
  return 5;
}

void botlist_connectionlost() {

  switch (blist.current->connected) {
  case 1:
    blist.current->connected = 0;
    party_send2party("*** Link to %s failed", botlist_nick());
    session_send2party("*** Link to %s failed", botlist_nick());
    botlist_send2all("chat %s Link to %s failed", botnick, botlist_nick());
    console_send2snoopwindow("(all) *** Link to %s failed", botlist_nick());
    break;
  case 2:
    blist.current->connected = 0;
    party_send2party("*** Link to %s broken", botlist_nick());
    session_send2party("*** Link to %s broken", botlist_nick());
    botlist_send2all("chat %s Link to %s broken", botnick, botlist_nick());
    console_send2snoopwindow("(all) *** Link to %s broken", botlist_nick());
    break;
  case 3:
    blist.current->connected = 0;
    dellinkeggdropcompatible(botlist_nick());
    if (on_unlink(botlist_nick())) return;
    party_send2party("*** Lost bot: %s", botlist_nick());
    session_send2party("*** Lost bot: %s", botlist_nick());
    botlist_send2all("chat %s Lost bot: %s", botnick, botlist_nick());
    console_send2snoopwindow("(all) *** Lost bot: %s", botlist_nick());
    break;
  }
}

void botlist_connection_established() {

  botlist_setpong();     /* pong is set for timeout */
  blist.current->connected = 2;
}

void botlist_parse_input(char *text) {

  switch (blist.current->connected) {
  case 2: parse_linking_bot(text); return;
  case 3: parse_linked_bot(text); return;
  default: int_err("BUG IN botlist_parse_input"); return;
  }
}

void botlist_breakconnection() {

  if ((blist.current->connected != 2) && (blist.current->connected != 3)) return;

  if (blist.current->connected == 3) {
    botlist_send2current("bye");
    close_socketptr(blist.current->socketptr);
    blist.current->socketptr = -1;
    blist.current->connected = 0;
    dellinkeggdropcompatible(botlist_nick());
    if (on_unlink(botlist_nick())) return;
    party_send2party("*** Unlinked from: %s", botlist_nick());
    session_send2party("*** Unlinked from: %s", botlist_nick());
    botlist_send2all("chat %s Unlinked from: %s", botnick, botlist_nick());
    console_send2snoopwindow("(all) *** Unlinked from: %s", botlist_nick());
  }
  else {
    close_socketptr(blist.current->socketptr);
    blist.current->socketptr = -1;
    blist.current->connected = 0;
  }
}

void botlist_linkestablished() {

  botlist_send2all("nlinked %s %s", botlist_nick(), botnick);
  if (!on_link(botnick, botlist_nick())) {
    party_send2party("*** Linked to %s", botlist_nick());
    session_send2party("*** Linked to %s", botlist_nick());
    console_send2snoopwindow("(all) *** Linked to %s", botlist_nick());
    botlist_send2all("chat %s Linked to %s", botnick, botlist_nick());
  }
  botlist_send2current("*hello!");
  botlist_send2current("version %s", VERSION);
  botlist_send2current("thisbot %s", botnick);
  tellinks2bot();
  addcloselink(botlist_nick(), blist.current->socketptr);
  blist.current->connected = 3;
  botlist_setpong();
}

void botlist_incominglink(int sockptr) {

  blist.current->socketptr = sockptr;
  blist.current->connected = 2;
  botlist_linkestablished();
}

void botlist_check_and_send_pings() {

  static time_t last_called=0;

  if (botlist_findfirst()) {
    do
      {
	if (blist.current->connected == 3) {
	  if (blist.current->last_pong < last_called) {
	    botlist_send2current("bye");
	    close_socketptr(blist.current->socketptr);
	    blist.current->socketptr = -1;
	    blist.current->connected = 0;
	    dellinkeggdropcompatible(botlist_nick());
	    if (on_unlink(botlist_nick()) == 0) {
	      party_send2party("*** Ping timeout: %s", botlist_nick());
	      session_send2party("*** Ping timeout: %s", botlist_nick());
	      botlist_send2all("chat %s Ping timeout: %s", botnick, botlist_nick());
	      console_send2snoopwindow("(all) *** Ping timeout: %s", botlist_nick());
	    }
	  } 
	  else {
	    botlist_send2current("ping");
	  }
	}
	if (blist.current->connected == 2) {
	  if (blist.current->last_pong + 200 < time_null) {
	    botlist_send2current("Timeout");
	    close_socketptr(blist.current->socketptr);
	    blist.current->socketptr = -1;
	    blist.current->connected = 0;
	    party_send2party("*** Connection timeout: %s", botlist_nick());
	    session_send2party("*** Connection timeout: %s", botlist_nick());
	    console_send2snoopwindow("(all) *** Connection timeout: %s", botlist_nick());
	  }
	}
	if ((blist.current->connected == 0) && (botlist_autoconnect()) && (!(part_of_botnet(botlist_nick())))) {
	  botlist_startconnection();
	}
      }
    while (botlist_findnext());
  }
  last_called = time_null;
}
