
/*******************************************************/
/**                                                   **/
/**      bottree.c (c) 1996, 97 Emiel Bruijntjes      **/
/**                                                   **/
/**      Bottree, Database containing the botnet      **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <limits.h>


#include "bottree.h"
#include "misc.h"
#include "config.h"
#include "bots.h"
#include "net.h"
#include "session.h"


struct bottree {
  char nick[10];
  int socket;
  struct connectlist *conlist;
} bttree={ "eembot", -1, NULL };

struct connectlist {
  struct bottree *tree;
  struct connectlist *next;
};



void newnicktopnode(char *nick) {

  strcpy(bttree.nick, nick);
}


int addfarlink_(struct bottree *list, char *parent, char *child) {

  struct connectlist *tmp;

  if (strcasecmp(list->nick, parent) == 0) {
    tmp = list->conlist;
    if (tmp == NULL) {
      list->conlist = (struct connectlist *) xmalloc (sizeof(struct connectlist));
      tmp = list->conlist;
    }
    else {
      while (tmp->next != NULL)
	tmp = tmp->next;

      tmp->next = (struct connectlist *) xmalloc (sizeof(struct connectlist));
      tmp = tmp->next;
    }

    /* Nu wijst tmp naar een lege ge-mallocte plaats in de connectlist */
    
    tmp->next = NULL;
    tmp->tree = (struct bottree *) xmalloc (sizeof(struct bottree));
    
    strcpy(tmp->tree->nick, child);
    tmp->tree->socket = list->socket;
    tmp->tree->conlist = NULL;
    return 1;
  }
  /* Niet aan deze node hangen. misschien een child hiervan */
  else {
    tmp = list->conlist;
    while (tmp != NULL) {
      if (addfarlink_(tmp->tree, parent, child))
	return 1;
      tmp = tmp->next;
    }
    return 0;
  }
}


int addfarlink(char *parent, char *child) {

  struct connectlist *tmp;

  if (strcasecmp(bttree.nick, parent) == 0) {
    tmp = bttree.conlist;
    if (tmp == NULL) {
      bttree.conlist = (struct connectlist *) xmalloc (sizeof(struct connectlist));
      tmp = bttree.conlist;
    }
    else {
      while (tmp->next != NULL)
	tmp = tmp->next;

      tmp->next = (struct connectlist *) xmalloc (sizeof(struct connectlist));
      tmp = tmp->next;
    }

    /* Nu wijst tmp naar een lege ge-mallocte plaats in de connectlist */
    
    tmp->next = NULL;
    tmp->tree = (struct bottree *) xmalloc (sizeof(struct bottree));
    
    strcpy(tmp->tree->nick, child);
    tmp->tree->socket = bttree.socket;
    tmp->tree->conlist = NULL;
    return 1;
  }
  /* Niet aan deze node hangen. misschien een child hiervan */
  else {
    tmp = bttree.conlist;
    while (tmp != NULL) {
      if (addfarlink_(tmp->tree, parent, child))
	return 1;
      tmp = tmp->next;
    }
    return 0;
  }
}


void addcloselink(char *nick, int sock) {

  struct connectlist *tmp;

  if (bttree.conlist == NULL) {
    bttree.conlist = (struct connectlist *) xmalloc (sizeof(struct connectlist));
    tmp = bttree.conlist;
  }
  else {
    tmp = bttree.conlist;
    while (tmp->next != NULL) {
      tmp = tmp->next;
    }
    tmp->next = (struct connectlist *) xmalloc (sizeof(struct connectlist));
    tmp = tmp->next;
  }
      
  tmp->next = NULL;
  tmp->tree = (struct bottree *) xmalloc (sizeof(struct bottree));
  strcpy(tmp->tree->nick, nick);
  tmp->tree->socket = sock;
  tmp->tree->conlist = NULL;
}


int dellinkeggdropcompatible_(struct bottree *tree, char *nick) {

  struct connectlist *tmp;
  struct connectlist *tmper;

  if (strcasecmp(nick, tree->nick) == 0) {
    botlist_send2all("unlinked %s", tree->nick);
    tmp = tree->conlist;
    while (tmp != NULL) {
      dellinkeggdropcompatible_(tmp->tree, tmp->tree->nick);
      tmper = tmp;
      tmp = tmp->next;
      xfree(tmper);
    }
    xfree(tree);
    return 1;
  }
  tmp = tree->conlist;
  tmper = NULL;
  while (tmp != NULL) {
    if (dellinkeggdropcompatible_(tmp->tree, nick)) {
      if (tmper == NULL) {
	tree->conlist = tmp->next;
      }
      else {
	tmper->next = tmp->next;
      }
      xfree(tmp);
      return 0;
    }
    tmper = tmp;
    tmp = tmp->next;
  }
  return 0;
}


int dellinkeggdropcompatible(char *nick) {

  struct connectlist *tmp;
  struct connectlist *tmper;

  if (strcasecmp(nick, bttree.nick) == 0) {
    botlist_send2all("unlinked %s", bttree.nick);
    tmp = bttree.conlist;
    while (tmp != NULL) {
      dellinkeggdropcompatible_(tmp->tree, tmp->tree->nick);
      tmper = tmp;
      tmp = tmp->next;
      xfree(tmper);
    }
    /*    xfree(bttree); */
    int_err("BOTTREE.C IS REALLY FUCKED UP!");
    return 1;
  }
  tmp = bttree.conlist;
  tmper = NULL;
  while (tmp != NULL) {
    if (dellinkeggdropcompatible_(tmp->tree, nick)) {
      if (tmper == NULL) {
	bttree.conlist = tmp->next;
      }
      else {
	tmper->next = tmp->next;
      }
      xfree(tmp);
      return 0;
    }
    tmper = tmp;
    tmp = tmp->next;
  }
  return 0;
}


int dellink_(struct bottree *tree, char *nick) {        

  struct connectlist *tmp;
  struct connectlist *tmper;

  if (strcasecmp(nick, tree->nick) == 0) {
    tmp = tree->conlist;
    while (tmp != NULL) {
      dellink_(tmp->tree, tmp->tree->nick);
      tmper = tmp;
      tmp = tmp->next;
      xfree(tmper);
    }
    xfree(tree);
    return 1;
  }
  tmp = tree->conlist;
  tmper = NULL;
  while (tmp != NULL) {
    if (dellink_(tmp->tree, nick)) {
      if (tmper == NULL) {
	tree->conlist = tmp->next;
      }
      else {
	tmper->next = tmp->next;
      }
      xfree(tmp);
      return 0;
    }
    tmper = tmp;
    tmp = tmp->next;
  }
  return 0;
}



int dellink(char *nick) {   /* Better and faster, but not compatible with eggdrops so i do not use this */

  struct connectlist *tmp;
  struct connectlist *tmper;

  if (strcasecmp(nick, bttree.nick) == 0) {
    tmp = bttree.conlist;
    while (tmp != NULL) {
      dellink_(tmp->tree, tmp->tree->nick);
      tmper = tmp;
      tmp = tmp->next;
      xfree(tmper);
    }
    /*    xfree(tree); */
    int_err("BOTTREE.C CONTAINS BUGS!");
    return 1;
  }
  tmp = bttree.conlist;
  tmper = NULL;
  while (tmp != NULL) {
    if (dellink_(tmp->tree, nick)) {
      if (tmper == NULL) {
	bttree.conlist = tmp->next;
      }
      else {
	tmper->next = tmp->next;
      }
      xfree(tmp);
      return 0;
    }
    tmper = tmp;
    tmp = tmp->next;
  }
  return 0;
}


void tellinks_(struct bottree *tree) {

  struct connectlist *tmp;
  
  tmp = tree->conlist;
  while (tmp != NULL) {
    botlist_send2current("nlinked %s %s", tmp->tree->nick, tree->nick);
    tellinks_(tmp->tree);
    tmp = tmp->next;
  }
}


void tellinks2bot() {

  struct connectlist *tmp;
  
  tmp = bttree.conlist;
  while (tmp != NULL) {
    botlist_send2current("nlinked %s %s", tmp->tree->nick, bttree.nick);
    tellinks_(tmp->tree);
    tmp = tmp->next;
  }
}

int zoekbotintree_(struct bottree *tree, char *nick, int *sock) {

  struct connectlist *tmp;

  if (strcasecmp(tree->nick, nick) == 0) {
    *sock = tree->socket;
    return 1;
  }
  tmp = tree->conlist;
  while (tmp != NULL) {
    if (zoekbotintree_(tmp->tree, nick, sock))
      return 1;
    tmp = tmp->next;
  }
  return 0;
}

int zoekbotintree(char *nick, int *sock) {

  struct connectlist *tmp;

  if (strcasecmp(bttree.nick, nick) == 0) {
    *sock = bttree.socket;
    return 1;
  }
  tmp = bttree.conlist;
  while (tmp != NULL) {
    if (zoekbotintree_(tmp->tree, nick, sock))
      return 1;
    tmp = tmp->next;
  }
  return 0;
}

void printbottree_(struct bottree *tree, int depth, unsigned int lines, void function(char *, ...)) {

  int i=1;
  struct connectlist *tmp;
  char bericht[MAXMSGSIZE]="";

  if (depth == 1) {
    function("   %s", tree->nick);
  }
  else {
    while (i != depth) {
      strcat(bericht, "     ");
      if (lines & (unsigned int)(pow(2, (i-1)))) {
	strcat(bericht, "|");
      }
      else {
	if ((i+1) == depth)
	  strcat(bericht, "`");
	else 
	  strcat(bericht, " ");      
      }
      i++;
    }
    function("%s---%s", bericht, tree->nick);
  }
  lines = (lines | (unsigned int)(pow(2, (depth - 1))));
  tmp = tree->conlist;
  while (tmp != NULL) {
    if (tmp->next == NULL)
      lines = (lines & (UINT_MAX - (unsigned int)(pow(2, (depth - 1)))));
    printbottree_(tmp->tree, depth + 1, lines, function);
    tmp = tmp->next;
  }
}

void printbottree(void function(char *, ...)) {

  struct connectlist *tmp;
  unsigned int lines = 0;

  function("   %s", bttree.nick);
  lines = (lines | (unsigned int)(pow(2, (0))));
  tmp = bttree.conlist;
  while (tmp != NULL) {
    if (tmp->next == NULL)
      lines = (lines & (UINT_MAX - (unsigned int)(pow(2, (0)))));
    printbottree_(tmp->tree, 2, lines, function);
    tmp = tmp->next;
  }
}

char *geefparent_(struct bottree *tree, char *nick) {

  struct connectlist *tmp;
  char *retvalue;

  tmp = tree->conlist;
  while (tmp != NULL) {
    if (strcasecmp(tmp->tree->nick, nick) == 0)
      return tree->nick;
    tmp = tmp->next;
  }
  tmp = tree->conlist;
  while (tmp != NULL) {
    if (retvalue = geefparent_(tmp->tree, nick), retvalue != NULL)
      return retvalue;
    tmp = tmp->next;
  }
  return NULL;
}

char *geefparent(char *nick) {

  struct connectlist *tmp;
  char *retvalue;

  tmp = bttree.conlist;
  while (tmp != NULL) {
    if (strcasecmp(tmp->tree->nick, nick) == 0)
      return bttree.nick;
    tmp = tmp->next;
  }
  tmp = bttree.conlist;
  while (tmp != NULL) {
    if (retvalue = geefparent_(tmp->tree, nick), retvalue != NULL)
      return retvalue;
    tmp = tmp->next;
  }
  return NULL;
}

int  part_of_botnet(char *nick) {

  int tmp;

  return (zoekbotintree(nick, &tmp));
}

int  same_branche(char *bot1, char *bot2) {

  int tmp1, tmp2;

  if (!(zoekbotintree(bot1, &tmp1))) return 0;
  if (!(zoekbotintree(bot2, &tmp2))) return 0;

  return (tmp1 == tmp2);
}

void bottree_send2bot(char *bot, char *format, ...) {

  int socketptr;
  char    line[WAYTOBIG_OUT];
  va_list msg;

  if (zoekbotintree(bot, &socketptr)) {
    if(!format) return;
    va_start(msg, format);
    vsprintf(line, format, msg);
    line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
    va_end(msg);
    send_to_socketptr(socketptr, "%s", line);
  }
}

