
/*******************************************************/
/**                                                   **/
/**      cfgfile.c (c) 1996, 97 Emiel Bruijntjes      **/
/**                                                   **/
/**        Reading and creating the config-file       **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <ctype.h>
#include <stdlib.h>


#include "misc.h"
#include "server.h"
#include "bots.h"
#include "values.h"
#include "userlist.h"
#include "channel.h"
#include "sescmd.h"
#include "lines.h"
#include "session.h"
#include "partycmd.h"
#include "net.h"
#include "cfgfile.h"
#include "tcl.h"

extern char  botnick[10];
extern char *progname;

int  write_cfg_file() {

  FILE *fp;

  if ((fp = fopen(botnick, "w")) == NULL) return 0;
  fwriteint(fp, CONFIG_VERSION);
  save_values(fp);
  savebots(fp);
  save_channels(fp);
  save_sescmds(fp);
  save_all_lines(fp);
  save_partycmds(fp);
  save_servers(fp);
  save_allow_settings(fp);
  save_userlist(fp);
  save_scriptlist(fp);
  fclose(fp);
  return 1;
}

void cfgcorrupt() {
  
  fprintf(stderr, "Config file corrupted\n");
  exit(1);
}

int read_cfg_file() {

  FILE *fp;
  int version;

  if ((fp = fopen(botnick, "r")) == NULL) return 0;
  if (!(freadint(fp, &version))) { cfgcorrupt(); fclose(fp); return 0; }
  if (load_values(fp, version) == 0) { /* fprintf(stderr, "1\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (loadbots(fp, version) == 0) { /* fprintf(stderr, "2\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_channels(fp) == 0) { /* fprintf(stderr, "3\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_sescmds(fp, version) == 0) { /* fprintf(stderr, "4\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_all_lines(fp) == 0) { /* fprintf(stderr, "5\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_partycmds(fp) == 0) { /* fprintf(stderr, "6\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_servers(fp) == 0) { /* fprintf(stderr, "7\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_allow_settings(fp) == 0) { /* fprintf(stderr, "8\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_userlist(fp, version) == 0) { /* fprintf(stderr, "9\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  if (load_scriptlist(fp, version) == 0) { /* fprintf(stderr, "10\n"); */ cfgcorrupt(); fclose(fp); return 0; }
  fclose(fp);
  return 1;
}

void noconfigfile() {

  char buffer[1000];
  int  tmp;
  char *tmpstr;

  printf("\nYou did not specify a configuration-file.\n");
noconfigfile_enternick:
  printf("\nEnter the nick of the bot: ");
  scanf("%1000s", buffer);
  if (strlen(buffer) > 9) {
    printf("Nickname is too long\n");
    goto noconfigfile_enternick;
  }
  strcpy(botnick, buffer);
  printf("\nEnter the server: (<server> <port>) ");
  scanf("%1000s %i", buffer, &tmp);
  setserver(buffer, tmp);

  if (getenv("USER")) {
    setbotlogin(getenv("USER"));
    if (my_gethostname()) {
      tmpstr = (char *) xmalloc (strlen(getenv("USER"))+strlen(my_gethostname())+2);
      sprintf(tmpstr, "%s@%s", getenv("USER"), my_gethostname());
      setmaintainer(tmpstr);
      xfree(tmpstr);
    }
  }
  tmpstr = (char *) xmalloc (strlen("/msg ")+strlen(botnick)+strlen(" help")+1);
  sprintf(tmpstr, "/msg %s help", botnick);
  setircname(tmpstr);
  xfree(tmpstr);
  
  tmpstr = (char *) xmalloc (strlen(botnick)+strlen(".notes")+1);
  sprintf(tmpstr, "%s.notes", botnick);
  setnotefile(tmpstr);
  xfree(tmpstr);

  addline('x', "*", "", returnmaintainer());

  if (!(write_cfg_file())) {
    printf("Error: I could not create configuration file: ./%s\n", botnick);
    exit(1);
  }
  printf("\nI have created a configuration file: ./%s\n", botnick);
  printf("You can start this bot with \"%s %s\"\n", progname, botnick);
  printf("or with \"%s -c %s\" if you want to use the console.\n\n", progname, botnick);
  printf("All you have to do is:\n");
  printf("  1. Go to IRC\n");
  printf("  2. Become superuser:                 /msg %s su password\n", botnick);
  printf("  3. Change the superuser password:    /msg %s set su_passwd <anything you want>\n", botnick);
  printf("  4. Let me join a channel:            /msg %s addchan <channel> +nt\n", botnick);
  printf("  5. Get more information:             /msg %s help\n\n\n", botnick);
}
