
/*******************************************************/
/**                                                   **/
/**       channel.c (c) 1996, 97 Emiel Bruijntjes     **/
/**                                                   **/
/**  Database with channels I'm on and I should join  **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#include "channel.h"
#include "chanuser.h"
#include "misc.h"
#include "server.h"
#include "config.h"

int convert_mode_string(const char *mode_string, int *mode, char *key, int *limit) {

  char *argv[3];
  int   argc=3;
  char *workstring;
  int current=1;
  int i=0;

  workstring = (char *) xmalloc (strlen(mode_string) + 1);
  strcpy(workstring, mode_string);
  
  split_string(workstring, argv, &argc);
  
  while ((argv[0][i] != '\0') && (!isspace(argv[0][i]))) {
    switch (argv[0][i]) {
    case 'i': *mode = *mode | CHAN_MODE_I; break;
    case 'k':
      if (argc <= current) {
	xfree(workstring);
        return 0;
      }
      *mode = *mode | CHAN_MODE_K;
      strncpy(key, argv[current], MAXKEYLEN);
      current++;
      break;
    case 'l':
      if (argc < current) {
	xfree(workstring);
        return 0;
      }
      *mode = *mode | CHAN_MODE_L;
      *limit = atoi(argv[current]);
      current++;
      break;
    case 'm': *mode = *mode | CHAN_MODE_M; break;
    case 'n': *mode = *mode | CHAN_MODE_N; break;
    case 'p': *mode = *mode | CHAN_MODE_P; break;
    case 's': *mode = *mode | CHAN_MODE_S; break;
    case 't': *mode = *mode | CHAN_MODE_T; break;
    case '+': break;
    default:
      xfree(workstring);
      return 0;
    }
    i++;
  }
  xfree(workstring);
  return 1;
}


/********************************************************/
/*                                                      */
/*                     chans2watch                      */ 
/*             channels the bot should be on            */
/*                                                      */
/********************************************************/

struct {
  char *name;
  int   mode;
  int   limit;
  char *key;
} chans2watch[MAX_IRC_CHANNELS];

static int nr_chans2watch=0;

void init_channels() {

  int i;

  for (i=0; i < MAX_IRC_CHANNELS; i++) {
    chans2watch[i].name  = NULL;
    chans2watch[i].mode  = 0;
    chans2watch[i].limit = 0;
    chans2watch[i].key   = NULL;
  }
}

extern char current_nick[10];

void save_channels(FILE *fp) {

  int i;

  fwriteint(fp, nr_chans2watch);
  for (i=0; i<nr_chans2watch; i++) {
    fwritestr2bin(chans2watch[i].name, fp);
    fwriteint(fp, chans2watch[i].mode);
    fwriteint(fp, chans2watch[i].limit);
    fwritestr2bin(chans2watch[i].key, fp);
  }
  fwriteint(fp, 238945);
}

int load_channels(FILE *fp) {

  int i;

  for (i=0; i<nr_chans2watch; i++) {
    xfree(chans2watch[i].name); chans2watch[i].name = NULL;
    xfree(chans2watch[i].key);  chans2watch[i].key  = NULL;
  }

  freadint(fp, &nr_chans2watch);
  for (i=0; i<nr_chans2watch; i++) {
    fgetstrfrombin(&chans2watch[i].name, fp);
    freadint(fp, &chans2watch[i].mode);
    freadint(fp, &chans2watch[i].limit);
    fgetstrfrombin(&chans2watch[i].key, fp);
  }
  freadint(fp, &i);
  return (i == 238945);
}

int addchan2watch(char *name, int mode, char *key, int limit) {

  int i;

  if (nr_chans2watch >= MAX_IRC_CHANNELS) return 0;

  for (i=0; i<nr_chans2watch; i++)
    if (strcasecmp(chans2watch[i].name, name) == 0) return 0;

  chans2watch[nr_chans2watch].name = (char *)xmalloc(strlen(name)+1);
  chans2watch[nr_chans2watch].key  = (char *)xmalloc(strlen(key) +1);
  strcpy(chans2watch[nr_chans2watch].name, name);
  strcpy(chans2watch[nr_chans2watch].key , key);
  chans2watch[nr_chans2watch].mode  = mode;
  chans2watch[nr_chans2watch].limit = limit;
  nr_chans2watch++;
  return nr_chans2watch;
}

int delchan2watch(char *name) {

  int i;

  for (i=0; i<nr_chans2watch; i++) {
    if (strcasecmp(chans2watch[i].name, name) == 0) {
      xfree(chans2watch[i].name);
      xfree(chans2watch[i].key);
      nr_chans2watch--;
      chans2watch[i].name  = chans2watch[nr_chans2watch].name;
      chans2watch[i].mode  = chans2watch[nr_chans2watch].mode;
      chans2watch[i].limit = chans2watch[nr_chans2watch].limit;
      chans2watch[i].key   = chans2watch[nr_chans2watch].key;
      chans2watch[nr_chans2watch].name  = NULL;
      chans2watch[nr_chans2watch].mode  = 0;
      chans2watch[nr_chans2watch].limit = 0;
      chans2watch[nr_chans2watch].key   = NULL;
      return 1;
    }
  }
  return 0;
}

int  is_chan2watch(char *name) {

  int i;

  for (i=0; i<nr_chans2watch; i++) {
    if (strcasecmp(chans2watch[i].name, name) == 0)
      return 1;
  }
  return 0;
}

char * channel_i_should_be_on() {

  int i, j;

  for (i=0; i<nr_chans2watch; i++) {
    if (!(channellist_findchan(chans2watch[i].name))) {
      while (1) {
	j = rand() % nr_chans2watch;
	if (!(channellist_findchan(chans2watch[j].name))) return chans2watch[j].name;
      }
    }
  }
  return NULL;
}

void printchans2watch(void output(char *, ...)) {

  int i;
  char atsign;
  char limit[30];

  output("Channel             Mode");
  output("---------------     -----");
  
  for (i=0; i<nr_chans2watch; i++) {
    sprintf(limit, "%d", chans2watch[i].limit);
    if (ischanneloperator(chans2watch[i].name, current_nick))
      atsign = '@';
    else
      atsign = ' ';
    output("%c%-18s %s%s%s%s%s%s%s%s %s %s %s", atsign, chans2watch[i].name, chans2watch[i].mode & 1 ? "i" : 
	   "", chans2watch[i].mode & 2 ? "k" : "", chans2watch[i].mode & 4 ? "l" : "", chans2watch[i].mode & 
	   8 ? "m" : "", chans2watch[i].mode & 16 ? "n" : "",chans2watch[i].mode & 32 ? "p" : "", 
	   chans2watch[i].mode & 64 ? "s" : "", chans2watch[i].mode & 128 ? "t" : "", chans2watch[i].mode & 2
	   ? chans2watch[i].key : "", chans2watch[i].mode & 4 ? limit : "", 
	   channellist_findchan(chans2watch[i].name) ? "" : "Joining");
  }
  output("---------------     -----");
}


char *channel2watch(int i) {

  if (i > nr_chans2watch) return NULL;

  return chans2watch[i-1].name;
}
      


/********************************************************/
/*                                                      */
/*                     channellist                      */ 
/*                channels the bot is on                */
/*                                                      */
/********************************************************/



struct channellist {
  struct channel *head;
  struct channel *tail;
  struct channel *current;
  int length;
} chanlist={ NULL , NULL , NULL, 0 };


struct channel {
  char *channelname;
  unsigned mode_i:1; /* 1 */
  unsigned mode_k:1; /* 2 */
  unsigned mode_l:1; /* 4 */
  unsigned mode_m:1; /* 8 */
  unsigned mode_n:1; /* 16 */
  unsigned mode_p:1; /* 32 */
  unsigned mode_s:1; /* 64 */
  unsigned mode_t:1; /* 128 */
  int  limit;
  char key[25];
  chanusers_t chusers;
  char *ban[MAX_BANS];
  char  ops[3][10];    /* Users I have to op in the future */
  int   nr_ops;        /* number of users I have to op */
  struct channel *prev;
  struct channel *next;
};

void addchannel2chanlist(char *name) {

  int   i;

  chanlist.current = (struct channel *) xmalloc (sizeof(struct channel));

  chanlist.current->next = chanlist.head;
  chanlist.current->prev = NULL;

  if (chanlist.length == 0) {
    chanlist.tail = chanlist.current;
  }
  else {
    chanlist.head->prev = chanlist.current;
  }
  chanlist.head = chanlist.current;
  chanlist.length = chanlist.length + 1;

  chanlist.current->channelname = (char *) xmalloc ((strlen(name)+1));

  strcpy(chanlist.current->channelname, name);
  chanlist.current->mode_i = chanlist.current->mode_k = chanlist.current->mode_l = chanlist.current->mode_m =
    chanlist.current->mode_n = chanlist.current->mode_p = chanlist.current->mode_s = chanlist.current->mode_t =
    chanlist.current->limit = 0;
  strcpy(chanlist.current->key, "");
  chanlist.current->nr_ops = 0;
  init_chanusers_t(&chanlist.current->chusers);
  for (i=0; i<MAX_BANS; i++) chanlist.current->ban[i] = NULL;

  for (i=0; i<nr_chans2watch; i++) {
    if (strcasecmp(chans2watch[i].name, name) == 0) {
      chanlist.current->mode_i = getalnaarbin((chans2watch[i].mode & CHAN_MODE_I));
      chanlist.current->mode_k = getalnaarbin((chans2watch[i].mode & CHAN_MODE_K));
      chanlist.current->mode_l = getalnaarbin((chans2watch[i].mode & CHAN_MODE_L));
      chanlist.current->mode_m = getalnaarbin((chans2watch[i].mode & CHAN_MODE_M));
      chanlist.current->mode_n = getalnaarbin((chans2watch[i].mode & CHAN_MODE_N));
      chanlist.current->mode_p = getalnaarbin((chans2watch[i].mode & CHAN_MODE_P));
      chanlist.current->mode_s = getalnaarbin((chans2watch[i].mode & CHAN_MODE_S));
      chanlist.current->mode_t = getalnaarbin((chans2watch[i].mode & CHAN_MODE_T));
      chanlist.current->limit = chans2watch[i].limit;
      strncpy(chanlist.current->key, chans2watch[i].key, 25);
      return;
    }
  }
}

int channellist_makefirst() {

  if ((chanlist.current == NULL) || (chanlist.current == chanlist.head))
    return 0;

  if (chanlist.current == chanlist.tail) {
    chanlist.tail->prev->next = NULL;
    chanlist.tail = chanlist.tail->prev;
  }
  else {
    chanlist.current->next->prev = chanlist.current->prev;
    chanlist.current->prev->next = chanlist.current->next;
  }
  chanlist.current->next = chanlist.head;
  chanlist.current->prev = NULL;
  chanlist.head->prev = chanlist.current;
  chanlist.head = chanlist.current;
  return 1;
}

int channellist_findchan(char *name) {

  if ((chanlist.current != NULL) && (strcasecmp(chanlist.current->channelname, name) == 0)) {
    return 1;
  }

  chanlist.current = chanlist.head;
  while (chanlist.current != NULL) {
    if (strcasecmp(chanlist.current->channelname, name) == 0) {
      return 1;
    }
    chanlist.current = chanlist.current->next;
  }
  return 0;
}

void delete_channel() {

  short int i;

  xfree(chanlist.current->channelname);
  empty_chanusers_t(&chanlist.current->chusers);
  for (i=0; i<MAX_BANS; i++) xfree (chanlist.current->ban[i]);

  if (chanlist.length <= 0) return;

  if (chanlist.length == 1) {
    xfree(chanlist.current);
    chanlist.current = NULL;
    chanlist.length = 0;
    chanlist.head = NULL;
    chanlist.tail = NULL;
    return;
  }

  if (chanlist.current == chanlist.head) {
    chanlist.head = chanlist.current->next;
    xfree(chanlist.current);
    chanlist.current = chanlist.head;
    chanlist.current->prev = NULL;
    chanlist.length = chanlist.length - 1;
    return;
  }
  if (chanlist.current == chanlist.tail) {
    chanlist.tail = chanlist.tail->prev;
    chanlist.tail->next = NULL;
    xfree(chanlist.current);
    chanlist.current = chanlist.head;
    chanlist.length = chanlist.length - 1;
    return;
  }
  chanlist.current->next->prev = chanlist.current->prev;
  chanlist.current->prev->next = chanlist.current->next;
  xfree(chanlist.current);
  chanlist.length = chanlist.length - 1;
  chanlist.current = chanlist.head;
  return;
}

chanusers_t * channellist_chanusers() { return &chanlist.current->chusers; }

int is_on_a_channel(char *nick) {

  chanlist.current = chanlist.head;

  while (chanlist.current != NULL) {
    if (chanusers_findnick(channellist_chanusers(), nick)) return 1;
    chanlist.current = chanlist.current->next;
  }
  return 0;
}

void changenickonchannels(char *old, char *new) {

  char oper;

  chanlist.current = chanlist.head;

  while (chanlist.current != NULL) {
    if (chanusers_findnick(channellist_chanusers(), old)) {
      oper = chanusers_ischanop(channellist_chanusers());
      chanusers_part_channel(channellist_chanusers());
      chanusers_join_channel(channellist_chanusers(), new);
      if (oper)
        chanusers_mark_chanop(channellist_chanusers());
    }
    chanlist.current = chanlist.current->next;
  }
}

void delete_all_channels() {
  
  chanlist.current = chanlist.head;
  while (chanlist.current != NULL) {
    delete_channel();
    chanlist.current = chanlist.head;
  }
}

void removenickfromchannels(char *nick) {

  chanlist.current = chanlist.head;
  while (chanlist.current != NULL) {
    if (chanusers_findnick(channellist_chanusers(), nick))
      chanusers_part_channel(channellist_chanusers());
    chanlist.current = chanlist.current->next;
  }
}

int add_ban_to_channel(char *ban) {

  int i=0, j=-1;

  for (i=MAX_BANS-1; i>=0; i--) {
    if (chanlist.current->ban[i]) {
      if (strcasecmp(ban, chanlist.current->ban[i]) == 0)
	return 1;                                          /* Ban was already placed on channel */
    }
    else
      j=i;
  }
  if (j < 0) return 0;          /* List is full */

  chanlist.current->ban[j] = (char *) xmalloc ((strlen(ban)+1));
  strcpy(chanlist.current->ban[j], ban);
  return 1;
}

int del_ban_from_channel(char *ban) {

  int i;

  for (i=0; i<MAX_BANS; i++) {
    if (strcasecmp(chanlist.current->ban[i] ? chanlist.current->ban[i] : "", ban) == 0) {
      xfree (chanlist.current->ban[i]);
      chanlist.current->ban[i] = NULL;
      return 1;
    }
  }
  return 0;
}

int channellist_musthavemode(char mode_ch) {

  switch (mode_ch) {
  case 'i': return chanlist.current->mode_i; break;
  case 'k': return chanlist.current->mode_k; break;
  case 'l': return chanlist.current->mode_l; break;
  case 'm': return chanlist.current->mode_m; break;
  case 'n': return chanlist.current->mode_n; break;
  case 'p': return chanlist.current->mode_p; break;
  case 's': return chanlist.current->mode_s; break;
  case 't': return chanlist.current->mode_t; break;
  };
  return 0;
}

char *channellist_channelkey()  { return chanlist.current->key; }
int   channellist_limit()       { return chanlist.current->limit; }
char *channellist_channelname() { return chanlist.current->channelname; }

int fullbanlist() {

  int i;

  for (i=0; i<MAX_BANS; i++) {
    if (chanlist.current->ban[i] == NULL) {
      return 0;
    }
  }
  return 1;
}

char *returnrandomban() {

  int i;

  while (1) {
    i = rand() % MAX_BANS;
    if (chanlist.current->ban[i] != NULL)
      return chanlist.current->ban[i];
  }
}

int  is_banned(char *nickuserhost) {

  int i;

  for (i=0; i<MAX_BANS; i++) {
    if (chanlist.current->ban[i])
      if (match(chanlist.current->ban[i], nickuserhost)) return 1;
  }
  return 0;
}

int channellist_findfirst() {
  
  if (chanlist.head == NULL)
    return 0;
  chanlist.current = chanlist.head;
  return 1;
}

int channellist_findlast() {

  return (int)(chanlist.current = chanlist.tail);
}

int channellist_next() {
  
  if (chanlist.current->next == NULL)
    return 0;
  chanlist.current = chanlist.current->next;
  return 1;
}

void channellist_flushops() {

  if (chanlist.current->nr_ops == 0) return;
  send_to_server("MODE %s +o%s%s %s %s %s", channellist_channelname(), chanlist.current->nr_ops > 1 ? "o" : "",
		 chanlist.current->nr_ops > 2 ? "o" : "", chanlist.current->ops[0], chanlist.current->nr_ops > 1 ? 
		 chanlist.current->ops[1] : "", chanlist.current->nr_ops > 2 ? chanlist.current->ops[2] : "");
  chanlist.current->nr_ops = 0;
}

void channellist_adduser2op(char *nick) {

  int i;

  for (i=0; i<chanlist.current->nr_ops; i++) {
    if (strcasecmp(chanlist.current->ops[i], nick) == 0) return;
  }
  strcpy(chanlist.current->ops[chanlist.current->nr_ops], nick);
  chanlist.current->nr_ops++;
  if (chanlist.current->nr_ops >= 3) channellist_flushops();
}

int  channels_does_have_op_queue() {

  if (channellist_findfirst()) {
    do
      {
	if (chanlist.current->nr_ops > 0) return 1;
      }
    while (channellist_next());
  }
  return 0;
}

char *returnban(int bannr) {

  return chanlist.current->ban[bannr];
}

int ischanneloperator(char *channel, char *nick) {

  return ((channellist_findchan(channel)) && (chanusers_findnick(channellist_chanusers(), nick)) && 
          (chanusers_ischanop(channellist_chanusers())));
}

int isonchannel(char *channel, char *nick) {

  return ((channellist_findchan(channel)) && (chanusers_findnick(channellist_chanusers(), nick)));
}

int    channellist_numberofusers() { return chanusers_length(channellist_chanusers()); }

int    channellist_numberofops() {

  int nr=0;

  if (chanusers_findfirst(channellist_chanusers())) {
    do
      {
	if (chanusers_ischanop(channellist_chanusers())) nr++;
      }
    while (chanusers_findnext(channellist_chanusers()));
  }
  return nr;
}
