
/*******************************************************/
/**                                                   **/
/**       channel.h (c) 1996, 97 Emiel Bruijntjes     **/
/**                                                   **/
/**  Database with channels I'm on and I should join  **/
/**                                                   **/
/*******************************************************/

#ifndef CHANNEL_H_
#define CHANNEL_H_

#include <stdio.h>

#include "chanuser.h"
#include "reg.h"

#define MAX_IRC_CHANNELS     20       /* Maximum number of channels I can join */
#define MAX_BANS             20       /* Maximum number of bans on channel */

#define CHAN_MODE_I           1       /* Channelmodes */
#define CHAN_MODE_K           2
#define CHAN_MODE_L           4
#define CHAN_MODE_M           8
#define CHAN_MODE_N          16
#define CHAN_MODE_P          32
#define CHAN_MODE_S          64
#define CHAN_MODE_T         128


/* Functions every module should have: */

void init_channels();
void save_channels(FILE *);
int  load_channels(FILE *);

/* Handy function to convert "mode +olk 23 secret" into representation the bot understands ... */

int convert_mode_string(const char *mode_string, int *mode, char *key, int *limit);


/* Database of channels I should watch: */

int addchan2watch(char *name, int mode, char *key, int limit);
int delchan2watch(char *name);

char * channel_i_should_be_on();       /* returns a channelname of a channel I should join. Returns NULL when there's no
					* channel to join */

void printchans2watch(void output(char *, ...));
int  is_chan2watch(char *name);        /* Returns TRUE when name is in the chans2watch-list */

char *channel2watch(int i);    /* Returns the name of channel "i" in chan2watch-array */

/* Database of channels I am currently active on: */

void addchannel2chanlist(char *name);
void delete_channel();
void delete_all_channels();

int channellist_findfirst();
int channellist_findlast();
int channellist_next();
int channellist_makefirst();
int channellist_findchan(char *name);

int    channellist_numberofusers();
int    channellist_numberofops();
int    channellist_limit();
int    channellist_musthavemode(char);                /* Return true when the channel should have mode specified */
char * channellist_channelkey();
char * channellist_channelname();

chanusers_t * channellist_chanusers();

int  isonchannel(char *channel, char *nick);
int  is_on_a_channel(char *nick);                  /* Returns true when nick is on a channel where I am also */
int  ischanneloperator(char *channel, char *nick); /* Returns true when nick has ops on channel specified */

int  fullbanlist();      /* Return true when banlistis full */
char *returnrandomban();
char *returnban(int);
int  is_banned(char *);

void changenickonchannels(char *old, char *new);
void removenickfromchannels(char *nick);

int  add_ban_to_channel(char *ban);
int  del_ban_from_channel(char *ban);

void channellist_flushops();
void channellist_adduser2op(char *nick);
int  channels_does_have_op_queue();

#endif
