
/*******************************************************/
/**                                                   **/
/**      chanuser.c (c) 1996, 97 Emiel Bruijntjes     **/
/**                                                   **/
/**        Database with users active on channel      **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "chanuser.h"
#include "misc.h"

void  init_chanusers_t(chanusers_t *c) { c->current = c->first = c->last = NULL; }
void empty_chanusers_t(chanusers_t *c) { while (chanusers_findfirst(c)) chanusers_part_channel(c); }

void chanusers_join_channel(chanusers_t *chusers, char *nick) {

  chusers->current = (ch_user_t *) xmalloc (sizeof(ch_user_t));
  
  chusers->current->nick = (char *) xmalloc (strlen(nick) + 1);
  strcpy(chusers->current->nick, nick);
  chusers->current->chanop = 0;

  if (chusers->first == NULL) {
    chusers->first = chusers->current;
    chusers->current->prev = NULL;
  }
  else {
    chusers->current->prev = chusers->last;
    chusers->last->next = chusers->current;
  }
  chusers->current->next = NULL;
  chusers->last = chusers->current;
}

void chanusers_part_channel(chanusers_t *chusers) {

  xfree(chusers->current->nick);
  
  if ((chusers->current == chusers->first) && (chusers->current == chusers->last)) {
    xfree(chusers->current);
    chusers->current = NULL;
    chusers->last = NULL;
    chusers->first = NULL;
    return;
  }

  if (chusers->current->next == NULL) {
    chusers->last = chusers->current->prev;
    chusers->last->next = NULL;
    xfree(chusers->current);
    chusers->current = chusers->last;
    return;
  }

  if (chusers->current->prev == NULL) {
    chusers->first = chusers->current->next;
    chusers->first->prev = NULL;
    xfree(chusers->current);
    chusers->current = chusers->first;
    return;
  }

  chusers->current->prev->next = chusers->current->next;
  chusers->current->next->prev = chusers->current->prev;
  xfree(chusers->current);
  chusers->current = chusers->first;
  return;
}

int  chanusers_length(chanusers_t *chusers) {

  ch_user_t *current;
  int i=0;

  current = chusers->current;
  chusers->current = chusers->first;
  while (chusers->current != NULL) {
    i++;
    chusers->current = chusers->current->next;
  }
  chusers->current = current;
  return i;
}

int  chanusers_findnick (chanusers_t *chusers, char *nick) {

  for (chusers->current = chusers->first; chusers->current != NULL; chusers->current = chusers->current->next)
    if (strcasecmp(chusers->current->nick, nick) == 0) return 1;
  return 0;
}

int  chanusers_findfirst(chanusers_t *chusers) { return (int)(chusers->current = chusers->first); }
int  chanusers_findlast (chanusers_t *chusers) { return (int)(chusers->current = chusers->last);  }
int  chanusers_findnext (chanusers_t *chusers) { return (int)(chusers->current = chusers->current->next); }
int  chanusers_findprev (chanusers_t *chusers) { return (int)(chusers->current = chusers->current->prev); }

void chanusers_mark_chanop    (chanusers_t *chusers) { chusers->current->chanop = 1; }
void chanusers_mark_not_chanop(chanusers_t *chusers) { chusers->current->chanop = 0; }

char *chanusers_nick(chanusers_t *c)     { return c->current->nick;   }
int   chanusers_ischanop(chanusers_t *c) { return c->current->chanop; }
