
/********************************************************/
/**                                                    **/
/**       eembot.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                    **/
/**   IRC Bot/IRC client/HTTP server/Partyline-server  **/
/**                                                    **/
/********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "misc.h"
#include "net.h"
#include "cfgfile.h"
#include "console.h"
#include "bottree.h"
#include "session.h"
#include "lines.h"
#include "notes.h"
#include "values.h"
#include "channel.h"
#include "server.h"
#include "parse.h"
#include "annon.h"
#include "bots.h"
#include "http.h"
#include "xwin.h"
#include "tcl.h"
#include "term.h"
#include "tcl_procs.h"
#include "on.h"


extern char *maintainer;
extern int   starttime;
extern char  botnick[10];
extern char *progname;
extern int   httpport;
extern int   listenport;
extern int   httpsockptr;
extern int   listensockptr;

       time_t time_null;

int my_daemon(int nochdir, int noclose) {   /* copied from C-library and changed to (emacs-)indentation I like */

  int fd;
  
  switch (fork()) {
  case -1:
    return (-1);
  case 0:
    break;
  default:
    exit(0);
  }
  
  if (setsid() == -1)
    return (-1);
  
  if (!nochdir)
    (void)chdir("/");
  
  if (!noclose && (fd = open("/dev/null", O_RDWR, 0)) != -1) {
    (void)dup2(fd, STDIN_FILENO);
    (void)dup2(fd, STDOUT_FILENO);
    (void)dup2(fd, STDERR_FILENO);
    if (fd > 2)
      (void)close (fd);
  }
  return (0);
}


void remove_file() {
  
  char *file;
  
  file = (char *) xmalloc (strlen(botnick) + 5);
  sprintf(file, "%s.pid", botnick);
  unlink(file);
  xfree(file);
}


void check_file() {    /* Check if the file botnick.pid exists, if so, call exit() */

  struct stat sb;
  char *file;

  file = (char *) xmalloc (strlen(botnick) + 5);
  sprintf(file, "%s.pid", botnick);

  if (stat(file, &sb) == 0) {   /* File exists */
    fprintf(stderr, "Check if this bot is not already active, if not,\n");
    fprintf(stderr, "remove the file %s, and start the bot again.\n", file);
    exit(1);
  }
  xfree(file);
}

void create_file() {   /* Create the file botnick.pid */

  char *file;
  FILE *fp;
  
  file = (char *) xmalloc (strlen(botnick) + 5);
  sprintf(file, "%s.pid", botnick);
  
  if ((fp = fopen(file, "w")) == NULL) {
    fprintf(stderr, "I could not open %s for writing. Exiting...\n", file);
    exit(1);
  }
  fprintf(fp, "%lu\n", (unsigned long)getpid());
  fclose(fp);
  atexit(remove_file);
  xfree(file);
}


void every_hour() {

  static time_t last_visit=0;

  if ((time_null - 3600) < last_visit) return;
  last_visit = time_null;

  auto_remove_lines('a');
  auto_remove_lines('b');
  auto_remove_lines('d');
  auto_remove_lines('h');
  auto_remove_lines('i');
  auto_remove_lines('o');
  auto_remove_lines('x');
  auto_remove_lines('n');
  write_cfg_file();
  updatenotefile();
  if (give_localport(listensockptr) != listenport) {
    close_socketptr(listensockptr);
    listensockptr = start_listen(T_ANNON, listenport);
  }
  if (give_localport(httpsockptr) != httpport) {
    close_socketptr(httpsockptr);
    httpsockptr = start_listen(T_HTTP, httpport);
  }
}

void every_5_minutes() {

  static time_t last_visit=0;

  if ((time_null - 300) < last_visit) return;
  last_visit = time_null;

  check_annon_timeout();
  botlist_check_and_send_pings();
  check_connection();                /* check server connection */
  check_http_timeout();
  check_channels();
}

void every_60_seconds() {

  check_nick();
  increase_ontime_sessions();
  random_delete_away_sessions();
  session_startnotify();
  console_print_status(); /* Update the console */
  join_channels();
  on_timer();
}

void every_5_seconds() {

  if (channellist_findfirst()) {
    do 
      {
	channellist_flushops();
      }
    while (channellist_next());
  }
}


void eembot_init() {   /* Call init-function in every module */

  starttime = time(NULL);
  time_null = starttime;
  net_init();
  init_values();
  init_channels();
  init_allow();
  init_X();
  init_tcl();
}


int main(int ARGC, char *ARGV[]) {

  struct timeval timev;
  time_t last5, last60;
  int    timeout;

  progname = ARGV[0];

  signal(SIGPIPE, SIG_IGN);  /* Broken pipe, ignore */

  eembot_init();

  switch (ARGC) {
  case 1:
    noconfigfile();
    return 0;
  case 2:
    if (strcmp(ARGV[1], "-c") == 0) {
      noconfigfile();
      return 0;
    }
    else {
      if (strlen(ARGV[1]) > 9) {
	fprintf(stderr, "%s: Name of config file is too long.\n", progname);
	return 0;
      }
      strcpy(botnick, ARGV[1]);
      newnicktopnode(botnick);
      check_file();
      if (!(read_cfg_file())) {
	fprintf(stderr, "I could not load config-file %s\n", ARGV[1]);
	return 0;
      }
      my_daemon(1, 0);
      create_file();
    }
    break;
  case 3:
    if (strcmp(ARGV[1], "-c") != 0) {
      fprintf(stderr, "%s: Usage: %s [-c] [<config-file>]\n", progname, progname);
      return 1;
    }
    if (strlen(ARGV[2]) > 9) {
      fprintf(stderr, "%s: Name of config file is too long.\n", progname);
      return 0;
    }
    strcpy(botnick, ARGV[2]);
    newnicktopnode(botnick);
    check_file();
    if (!(read_cfg_file())) {
      fprintf(stderr, "I could not load config-file %s\n", ARGV[2]);
      return 0;
    }
    create_file();
    init_terminal();
    atexit(stop_terminal);
    break;
  default:
    fprintf(stderr, "%s: Usage: %s [-c] [<config-file>]\n", progname, progname);
    return 1;
  }

  eval_scriptlist();

  last5 = last60 = time_null;

  while (1) {

    /* The elements of the timev struct should be set every time before the wait_for_io() call, because it
     * COULD be (it's operating system dependent) that those values are changed by select().
     */
    
    if (server_has_queue())
      timeout = 1;
    else if (channels_does_have_op_queue())
      timeout = last5 + 5 - time_null;
    else
      timeout = last60 + 60 - time_null;

    if (timeout < 0)
      timev.tv_sec = 1;
    else
      timev.tv_sec = timeout;
    
    timev.tv_usec = 0;
    
    wait_for_io(&timev);
    if (last5 + 5 <= time_null) {
      last5 = time_null;
      every_5_seconds();
    }
    if (last60 + 60 <= time_null) {
      last60 = time_null;
      every_60_seconds();
    }
    
    sendnextmessage();      /* Send stored message to the server. Messages could be stored to prevent flood */
    
    /* The following function do their own checking to see when they were last called */
    every_hour();
    every_5_minutes();
  }
}
