
/*******************************************************/
/**                                                   **/
/**      history.c (c) 1996, 97 Emiel Bruijntjes      **/
/**                                                   **/
/**   Database used by the console to store messages  **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "history.h"
#include "misc.h"

struct history {
  char *text;
  
  struct history *next;
  struct history *prev;
};


static struct history *history_first   = NULL;   /* First message in history */
static struct history *history_last    = NULL;   /* Last  message in history */
static struct history *history_current = NULL;   /* Last  message returned   */

static int nr_history=0;                         /* Number of messages stored */

static char * empty_history="";

void add_history(char *text) {                   /* Adds text to the history. history_current will be NULL */

  int len;

  if ((len = strlen(text)) == 0) return;

  history_current = (struct history *) xmalloc (sizeof(struct history));
  history_current->text = (char *) xmalloc (len+1);
  strcpy(history_current->text, text);

  if (history_first == NULL) {
    history_first = history_current;
    history_current->prev = NULL;
  }
  else {
    history_last->next = history_current;
    history_current->prev = history_last;
  }
  history_current->next = NULL;
  history_last = history_current;
  nr_history++;

  if (nr_history > HISTORY_MAX) {
    history_first = history_first->next;
    xfree(history_first->prev->text);
    xfree(history_first->prev);
    history_first->prev = NULL;
    nr_history--;
  }
  history_current = NULL;
}

char *history_next() {                /* Returns next message from history. */

  if (history_current == NULL)        /* There is not yet a next message */
    return empty_history;

  if (history_current->next) {                 /* There is a next message */
    history_current = history_current->next;
    return history_current->text;
  }                                            /* We already have returned the last message */
  else {
    history_current = NULL;
    return empty_history;
  }
}

char *history_prev() {

  if (history_last == NULL)          /* No messages in buffer */
    return empty_history;

  if (history_current == NULL) {
    history_current = history_last;
    return history_current->text;
  }
  if (history_current->prev) {
    history_current = history_current->prev;
  }
  return history_current->text;
}


static struct history *msg_history_first   = NULL;   /* First msg in history */
static struct history *msg_history_last    = NULL;   /* Last  msg in history */
static struct history *msg_history_current = NULL;   /* Last  msg returned   */

static int nr_msg_history=0;                         /* Number of messages stored */


void add_msg_history(char *text) {                   /* Adds text to the history. history_current will be NULL */

  msg_history_current = (struct history *) xmalloc (sizeof(struct history));
  msg_history_current->text = (char *) xmalloc (strlen(text)+1);
  strcpy(msg_history_current->text, text);

  if (msg_history_first == NULL) {
    msg_history_first = msg_history_current;
    msg_history_current->prev = NULL;
  }
  else {
    msg_history_last->next = msg_history_current;
    msg_history_current->prev = msg_history_last;
  }
  msg_history_current->next = NULL;
  msg_history_last = msg_history_current;
  nr_msg_history++;

  msg_history_current = msg_history_first;
  while (msg_history_current != msg_history_last) {
    if (strcasecmp(msg_history_current->text, text) == 0) {
      if (msg_history_current->prev == NULL) {
	msg_history_first = msg_history_current->next;
	msg_history_first->prev = NULL;
      }
      else {
	msg_history_current->prev->next = msg_history_current->next;
	msg_history_current->next->prev = msg_history_current->prev;
      }
      xfree(msg_history_current->text);
      xfree(msg_history_current);
      nr_msg_history--;
      break;
    }
    msg_history_current = msg_history_current->next;
  }

  if (nr_msg_history > HISTORY_MAX) {
    msg_history_first = msg_history_first->next;
    xfree(msg_history_first->prev->text);
    xfree(msg_history_first->prev);
    msg_history_first->prev = NULL;
    nr_msg_history--;
  }
  msg_history_current = msg_history_first;
}

char *msg_history_prev() {

  if (nr_msg_history == 0)
    return empty_history;

  if (msg_history_current == msg_history_first)
    msg_history_current = msg_history_last;
  else
    msg_history_current = msg_history_current->prev;

  return msg_history_current->text;
}
