
/*******************************************************/
/**                                                   **/
/**        http.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**      Parsing messages sent to the HTTP server     **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <strings.h>


#include "http.h"
#include "userlist.h"
#include "misc.h"
#include "values.h"
#include "config.h"
#include "net.h"
#include "lines.h"
#include "session.h"
#include "partyline.h"
#include "channel.h"
#include "gencmd.h"
#include "toptree.h"


struct httplist {
  struct httpinfo *first;
  struct httpinfo *tail;
  struct httpinfo *current;
} httplist={ NULL , NULL , NULL };

struct httpinfo {
  int    sockptr;
  time_t created;

  struct httpinfo *next;
  struct httpinfo *prev;
};

extern char * httpdir;
extern char   botnick[10];
extern int    listensockptr;
extern char * ftp_root_dir;
extern time_t time_null;
extern char * whoisdir;
extern char   my_nickuserhost[];
extern char   current_nick[];

static char * wrong_nick=NULL;


void addhttp(int sockptr) {

  httplist.current = (struct httpinfo *) xmalloc (sizeof(struct httpinfo));
  if (httplist.first == NULL) {
    httplist.first = httplist.current;
    httplist.current->prev = NULL;
  }
  else {
    httplist.current->prev = httplist.tail;
    httplist.tail->next = httplist.current;
  }
  httplist.current->next = NULL;
  httplist.tail = httplist.current;

  httplist.current->sockptr = sockptr;
  httplist.current->created = time_null;
}

void httpconnectionlost() {

  if (httplist.current == NULL) return;

  if ((httplist.current == httplist.first) && (httplist.current == httplist.tail)) {
    xfree(httplist.current);
    httplist.current = NULL;
    httplist.tail = NULL;
    httplist.first = NULL;
    return;
  }

  if (httplist.current->next == NULL) {
    httplist.tail = httplist.current->prev;
    httplist.tail->next = NULL;
    xfree(httplist.current);
    httplist.current = httplist.tail;
    return;
  }

  if (httplist.current->prev == NULL) {
    httplist.first = httplist.current->next;
    httplist.first->prev = NULL;
    xfree(httplist.current);
    httplist.current = httplist.first;
    return;
  }

  httplist.current->prev->next = httplist.current->next;
  httplist.current->next->prev = httplist.current->prev;
  xfree(httplist.current);
  httplist.current = httplist.first;
  return;
}

int http_findsockptr(int sockptr) {

  httplist.current = httplist.first;
  while (httplist.current != NULL) {
    if (httplist.current->sockptr == sockptr) return 1;
    httplist.current = httplist.current->next;
  }
  return 0;
}

int http_findfirst() { return (int)(httplist.current = httplist.first); }
int http_findnext()  { return (int)(httplist.current = httplist.current->next); }

void check_http_timeout() {

  char *file=NULL;

  if (http_findfirst()) {
    do
      {
	if (httplist.current->created + HTTP_TIMEOUT < time_null) {
	  file = (char *) xmalloc (strlen(httpdir) + strlen("timeout.html") + 1);
	  sprintf(file, "%stimeout.html", httpdir);
	  http_sendfile_replace_vars(file);
	  xfree(file);
	  close_socketptr(httplist.current->sockptr);
	  httpconnectionlost();                      /* current of http list is not defined now */
	  if (!(http_findfirst())) return;           /* set current to the first position */
	}
      }
    while (http_findnext());
  }
}

void nosuchfile() {
 
  char buffer[WAYTOBIG];
  int  nbytes;
  FILE *fp;
  char *file=NULL;

  file = (char *) xmalloc (strlen(httpdir) + strlen("nosuchfile.html") + 1);
  sprintf(file, "%s%s", httpdir, "nosuchfile.html");
 
  if ((fp = fopen(file, "r")) == NULL) {
    xfree(file);
    return;
  }
  while ((nbytes = fread(buffer, sizeof(char), WAYTOBIG, fp))) {
    send_to_socketptr_raw(httplist.current->sockptr, buffer, nbytes);
  }
  xfree(file);
  fclose(fp);
  return;
}

void http_sendfile_raw(char *file) {
 
  char buffer[WAYTOBIG];
  int  nbytes;
  FILE *fp;
 
  if ((fp = fopen(file, "r")) == NULL) {
    nosuchfile();
    return;
  }
  while ((nbytes = fread(buffer, sizeof(char), WAYTOBIG, fp))) {
    send_to_socketptr_raw(httplist.current->sockptr, buffer, nbytes);
  }
  fclose(fp);
  return;
}

/* A number of functions needed for replace_var */

static char *http_wrongnick() { return wrong_nick; }
static char *bot_nick() { return botnick; }
static char *http_ulist_lastseen() { return time2string(userlist_lastseen()); }
static char *http_ulist_ontime() { 
  
  static char buffer[15]; 

  sprintf(buffer, "%d:%d", userlist_ontime()/60, userlist_ontime()%60);
  return buffer;
}

static char * http_aline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('a', nickuserhost))
    return lines_mask('a');
  else
    return "No match";
}

static char * http_bline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('b', nickuserhost))
    return lines_mask('b');
  else
    return "No match";
}

static char * http_dline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('d', nickuserhost))
    return lines_mask('d');
  else
    return "No match";
}

static char * http_hline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('h', nickuserhost))
    return lines_mask('h');
  else
    return "No match";
}

static char * http_iline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('i', nickuserhost))
    return lines_mask('i');
  else
    return "No match";
}

static char * http_oline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('o', nickuserhost))
    return lines_mask('o');
  else
    return "No match";
}

static char * http_xline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('x', nickuserhost))
    return lines_mask('x');
  else
    return "No match";
}

static char * http_nline_mask() {

  char *nickuserhost;

  nickuserhost = (char *) xmalloc (strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
  sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
  if (does_this_match('n', nickuserhost))
    return lines_mask('n');
  else
    return "No match";
}

static int http_listen() { return give_localport(listensockptr); }

void replace_var(char *var) {

  int i;
  struct {
    char *var;
    char *(*charvalue)();
    int   (*intvalue)();
  } http_vars[] = {
    { "bot", bot_nick, NULL },
    { "maintainer", returnmaintainer, NULL },
    { "hostname", my_gethostname, NULL },
    { "listen", NULL, http_listen },
    { "ulist_nick", userlist_nick, NULL },
    { "ulist_lastseen", http_ulist_lastseen, NULL },
    { "ulist_lasthost", userlist_lasthost, NULL },
    { "ulist_prot", NULL, userlist_protection },
    { "ulist_publics", NULL, userlist_publics },
    { "ulist_kicks", NULL, userlist_kicks },
    { "ulist_kickeds", NULL, userlist_kickeds },
    { "ulist_joins", NULL, userlist_joins },
    { "ulist_leaves", NULL, userlist_leaves },
    { "ulist_quits", NULL, userlist_quits },
    { "ulist_nicks", NULL, userlist_nicks },
    { "ulist_modes", NULL, userlist_modes },
    { "ulist_manip", NULL, userlist_manipulation },
    { "ulist_ontime", http_ulist_ontime, NULL },
    { "aline_mask", http_aline_mask, NULL },
    { "bline_mask", http_bline_mask, NULL },
    { "dline_mask", http_dline_mask, NULL },
    { "hline_mask", http_hline_mask, NULL },
    { "iline_mask", http_iline_mask, NULL },
    { "oline_mask", http_oline_mask, NULL },
    { "xline_mask", http_xline_mask, NULL },
    { "nline_mask", http_nline_mask, NULL }, 
    { "wrong_nick", http_wrongnick, NULL },
    { NULL, NULL, NULL }
  };
  for (i=0; http_vars[i].var; i++) {
    if (strcasecmp(var, http_vars[i].var) == 0) {
      if (http_vars[i].charvalue)
	send_to_socketptr_nonewline(httplist.current->sockptr, "%s", http_vars[i].charvalue() ? http_vars[i].charvalue()
				    : "");
      else
	send_to_socketptr_nonewline(httplist.current->sockptr, "%d", http_vars[i].intvalue());
      return;
    }
  }
  send_to_socketptr_nonewline(httplist.current->sockptr, "%s", var);
}

void http_sendfile_replace_vars(char *file) {

  char *buffer=NULL;
  int  nbytes;
  FILE *fp;
  char var[200];

  if ((fp = fopen(file, "r")) == NULL) {
    nosuchfile();
    return;
  }
  
  while ((nbytes = my_getdelim(&buffer, fp, '%'))) {
    send_to_socketptr_raw(httplist.current->sockptr, buffer, nbytes);
    strcpy(var, "");
    if (!(fscanf(fp, "%200s", var))) { xfree(buffer); fclose(fp); return; }
    fgetc(fp);      /* Read space */
    replace_var(var);
  }
  xfree(buffer);
  fclose(fp);
  return;
}

void http_send_userlist() {

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");
  
  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>The userlist </H1>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><TABLE>");

  if (userlist_findfirst()) {
    do
      {
	send_to_socketptr(httplist.current->sockptr, "<TR><TD><A HREF=\"~%s\">%s</A></TD><TD>%s</TD></TR>", 
			  userlist_nick(), userlist_nick(), userlist_lasthost());
      }
    while (userlist_findnext());
  }
  send_to_socketptr(httplist.current->sockptr, "</TABLE></P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);

  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}

void http_sendtop(char type, int length) {

  struct toptree *tree;
  int (*function)() = userlist_publics;
  int pos = 1;

  switch(type) {
  case 'a': function = userlist_publics; break;
  case 'b': function = userlist_kicks; break;
  case 'c': function = userlist_kickeds; break;
  case 'd': function = userlist_joins; break;
  case 'e': function = userlist_leaves; break;
  case 'f': function = userlist_quits; break;
  case 'g': function = userlist_nicks; break;
  case 'h': function = userlist_modes; break;
  case 'i': function = userlist_ontime; break;
  case 'j': function = userlist_protection; break;
  case 'k': function = userlist_manipulation; break;
  case 'l': function = userlist_msgpminute; break;
  case 'm': function = userlist_msgpminute; break;
  default: function = userlist_publics; break;
  }

  if (userlist_findfirst()) {
    createtree(&tree, userlist_nick(), function());
    while (userlist_findnext()) {
      add_to_tree(tree, userlist_nick(), function());
    }
  }

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");
  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>Top</H1>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  
  send_to_socketptr(httplist.current->sockptr, "<FORM METHOD=GET ACTION=\"\">");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Publics</TD><TD><TD><input type=radio name=top value=\"a\" %s></TD><TD>Modes</TD><TD><input type=radio name=top value=\"h\" %s></TD></TR>", type < 'b' || type > 'm' ? "checked" : "",
		    type == 'h' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Kicks</TD><TD><TD><input type=radio name=top value=\"b\" %s></TD><TD>Ontime</TD><TD><input type=radio name=top value=\"i\" %s></TD></TR>", type == 'b' ? "checked" : "",
		    type == 'i' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Kickeds</TD><TD><TD><input type=radio name=top value=\"c\" %s></TD><TD>Protection</TD><TD><input type=radio name=top value=\"j\" %s></TD></TR>", type == 'c' ? "checked" : "",
		    type == 'j' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Joins</TD><TD><TD><input type=radio name=top value=\"d\" %s></TD><TD>Manipulation</TD><TD><input type=radio name=top value=\"k\" %s></TD></TR>", type == 'd' ? "checked" : "",
		    type == 'k' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Leaves</TD><TD><TD><input type=radio name=top value=\"e\" %s></TD><TD>Msgs/Minute</TD><TD><input type=radio name=top value=\"l\" %s></TD></TR>", type == 'e' ? "checked" : "",
		    type == 'l' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Quits</TD><TD><TD><input type=radio name=top value=\"f\" %s></TD><TD>Msgs/Hour</TD><TD><input type=radio name=top value=\"m\" %s></TD></TR>", type == 'f' ? "checked" : "",
		    type == 'm' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>Nicks</TD><TD><TD><input type=radio name=top value=\"g\" %s></TD><TD></TD></TR>", type == 'g' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "</TABLE>");
  send_to_socketptr(httplist.current->sockptr, "<input type=text name=length size=10 maxlength=20 value=\"%d\">", length);
  send_to_socketptr(httplist.current->sockptr, "<input type=submit value=print>");
  send_to_socketptr(httplist.current->sockptr, "</FORM></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
  send_to_socketptr(httplist.current->sockptr, "<TR><TH>Position</TH><TH>Nick</TH><TH>Score</TH></TR>");
  if (function == userlist_ontime)
    printtimetoptree_html(tree, &pos, 1, length, send_to_socketptr, httplist.current->sockptr);
  else
    printtoptree_html(tree, &pos, 1, length, send_to_socketptr, httplist.current->sockptr);
  send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);

  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
  deltree(tree);
}


void http_send_party() {

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");
  
  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>The partyline</H1>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  if (listensockptr >= 0) {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Do not click in this window</P></CENTER>");
    send_to_socketptr(httplist.current->sockptr, "<CENTER><applet code=\"tcp.class\" width=600 height=400><param name=server value=\"%s\"><param name=port value=\"%d\"></applet></CENTER>", give_localhost_name(listensockptr), 
		      give_localport(listensockptr));
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Click in this input-bar to activate it</P></CENTER>");
  }
  else {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P>The partyline is now not open</P></CENTER>");
  }
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);

  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}


void send_edit_file(char *format, ...) {

  char    line[2000];
  va_list msg;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  send_to_socketptr(httplist.current->sockptr, "<CENTER>%s</CENTER>", line);
}


void http_send_userinfo(char *nick) {

  char user[10];
  int  i=0, j=0;
  char *tmp=NULL;
  char *whoisfile;
  char *file;

  while (nick[i] != '\0' && j < 10) {
    if (nick[i] == '%') {
      if (nick[i+1] == '\0' || nick[i+2] == '\0') { break; }
      user[j] = hex2dec(&nick[i+1], 2);
      i += 2;
    }
    else {
      user[j] = nick[i];
    }
    j++; i++;
  }
  user[j] = '\0';

  if (strcasecmp(user, botnick) == 0) {
    file = (char *)xmalloc(strlen(httpdir) + strlen("index.html") + 1);
    sprintf(file, "%s%s", httpdir, "index.html");
    http_sendfile_replace_vars(file);
    xfree(file);
    return;
  }

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");

  if (userlist_findnick(user)) {
    send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>Information from the userlist</H1>");
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

    send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Nick</TH><TD>%s</TD></TR>", userlist_nick());
    tmp = userlist_lasthost();
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Last user@host</TH><TD>%s</TD></TR>", 
		      *tmp == '\0' ? "---" : tmp);
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Last seen</TH><TD>%s</TD></TR>",
		      time2string(userlist_lastseen())); 
    send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER><BR>");

    send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Publics</TH><TD> %d</TD><TH ALIGN=left>Kicks</TH><TD> %d</TD><TH ALIGN=left>Quits</TH><TD> %d</TD></TR>", userlist_publics(), userlist_kicks(), userlist_quits());
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Kicked</TH><TD> %d</TD><TH ALIGN=left>Joins</TH><TD> %d</TD><TH ALIGN=left>Modes</TH><TD> %d</TD></TR>", userlist_kickeds(), userlist_joins(), userlist_modes());
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Leaves</TH><TD> %d</TD><TH ALIGN=left>Nicks</TH><TD> %d</TD><TH ALIGN=left>Ontime</TH><TD> %d:%02d</TD></TR>", userlist_leaves(), userlist_nicks(), userlist_ontime()/60, 
		      userlist_ontime()%60);
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Manip</TH><TD> %d</TD><TH ALIGN=left>Protection</TH><TD> %d</TD></TR>", userlist_manipulation(), userlist_protection());
    send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  }
  if (session_findnick(user)) {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
    send_to_socketptr(httplist.current->sockptr, "<TR><TD>%s</TD><TH>Superuser</TH><TD>%s</TD><TH>Administrator</TH><TD>%s</TD><TH>Operator</TH></TR>", session_superuser() ? "<IMG SRC=\"yes.gif\">" : "---", session_administrator() ? 
		      "<IMG SRC=\"yes.gif\">" : "---", session_operator() ? "<IMG SRC=\"yes.gif\">" : "---");
    send_to_socketptr(httplist.current->sockptr, "<TR><TD>%s</TD><TH>Identified</TH><TD>%s</TD><TH>Spy</TH><TD>%s</TD><TH>Away</TH></TR>", session_identified() ? "<IMG SRC=\"yes.gif\">" : "---", session_isspy() ? 
		      "<IMG SRC=\"yes.gif\">" : "---", session_isaway() ? "<IMG SRC=\"yes.gif\">" : "---");
    send_to_socketptr(httplist.current->sockptr, "<TR><TD>%s</TD><TH>Chat active</TH><TD>%s</TD><TH>Send Active</TH><TD>%s</TD><TH>Get active</TH></TR>", session_chatstatus() == 3 ? "<IMG SRC=\"yes.gif\">" : "---", session_sendstatus() == 2 ? 
		      "<IMG SRC=\"yes.gif\">" : "---", session_getstatus() == 2 ? "<IMG SRC=\"yes.gif\">" : "---");
    send_to_socketptr(httplist.current->sockptr, "<TR><TD>%s</TD><TH>Chat request</TH><TD>%s</TD><TH>Send request</TH><TD>%s</TD><TH>Get request</TH></TR>", session_chatstatus() == 1 || session_chatstatus() == 2 ? "<IMG SRC=\"yes.gif\">" : 
		      "---", session_sendstatus() == 1 ? "<IMG SRC=\"yes.gif\">" : "---", session_getstatus() == 1 ? 
		      "<IMG SRC=\"yes.gif\">" : "---");
    send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
    send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>Matching lines</H1>");
    send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Line</TH><TH ALIGN=left>Mask</TH><TH ALIGN=left>Locked</TH></TR>");
    printmatchinglines_html('a', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('b', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('d', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('h', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('i', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('o', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('x', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    printmatchinglines_html('n', session_nickuserhost(), send_to_socketptr, httplist.current->sockptr);
    send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
  }
  else {
    if (tmp && *tmp) {
      tmp = (char *) xmalloc (strlen(user) + strlen(tmp) + 2);
      sprintf(tmp, "%s!%s", user, userlist_lasthost());
      send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
      send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>Matching lines</H1>");
      send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
      send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Line</TH><TH ALIGN=left>Mask</TH><TH ALIGN=left>Locked</TH></TR>");
      printmatchinglines_html('a', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('b', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('d', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('h', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('i', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('o', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('x', tmp, send_to_socketptr, httplist.current->sockptr);
      printmatchinglines_html('n', tmp, send_to_socketptr, httplist.current->sockptr);
      xfree(tmp);
    }
  }
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  hoofdletter(user);
  whoisfile = (char *) xmalloc (strlen(user) + strlen(whoisdir) + 1);
  sprintf(whoisfile, "%s%s", whoisdir, user);
  if (catfile(whoisfile, send_edit_file, "websurfer")) {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  }
  xfree(whoisfile);
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);
  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}

void http_send_irc() {

  char *channel;

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");

  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>IRC</H1>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<CENTER><P>%s is now on the following channels:</P></CENTER>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
  send_to_socketptr(httplist.current->sockptr, "<TR><TH>Channel</TH><TH ALIGN=center>#users</TH><TH ALIGN=center>#ops</TH></TR>");
  if (channellist_findfirst()) {
    do
      {
	channel = channellist_channelname();
	send_to_socketptr(httplist.current->sockptr, "<TR><TD><A HREF=\"?chan=%c%s\">%s</A></TD><TD ALIGN=left>%d</TD><TD ALIGN=left>%d</TD></TR>",
			  channel[0] == '#' ? '1' : channel[0] == '&' ? '2' : '3', channel+1, channel, 
			  channellist_numberofusers(), channellist_numberofops());
      }
    while (channellist_next());
  }
  send_to_socketptr(httplist.current->sockptr, "</TABLE></CHANNEL>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);
  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}


void http_sendchannel(char *channel) {

  switch (channel[0]) {
  case '1': channel[0] = '#'; break;
  case '2': channel[0] = '&'; break;
  case '3': channel[0] = '+'; break;
  }

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");

  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>%s</H1>", channel);
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  if (channellist_findchan(channel)) {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
    send_to_socketptr(httplist.current->sockptr, "<TR><TH>Operator</TH><TH>Nick</TH><TH>user@host</TH></TR>");
    if (chanusers_findfirst(channellist_chanusers())) {
      do
	{
	  send_to_socketptr(httplist.current->sockptr, "<TR><TD>%s</TD><TD><A HREF=\"~%s\">%s</A></TD><TD>%s</TD></TR>",
			    chanusers_ischanop(channellist_chanusers()) ? "<IMG SRC=\"yes.gif\">" : "",
			    chanusers_nick(channellist_chanusers()), chanusers_nick(channellist_chanusers()),
			    (strcasecmp(current_nick, chanusers_nick(channellist_chanusers())) == 0) ? 
			    index(my_nickuserhost, '!') ? index(my_nickuserhost, '!') + 1 : "-" :
			    session_findnick(chanusers_nick(channellist_chanusers())) ? 
			    (index(session_nickuserhost(), '!') ? index(session_nickuserhost(), '!') + 1 : "-") : "-");
	}
      while (chanusers_findnext(channellist_chanusers()));
    }
    send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
  }
  else {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P>I'm not on channel %s</P></CENTER>", channel);
  }
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);
  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}



void http_send_lines(char type) {

  int nr=1;

  type = tolower(type);

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");

  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>Lines</H1>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<FORM METHOD=GET ACTION=\"\">");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
  send_to_socketptr(httplist.current->sockptr, "<TR><TD>A</TD><TD>B</TD><TD>D</TD><TD>H</TD><TD>I</TD><TD>O</TD><TD>X</TD><TD>N</TD>");
  send_to_socketptr(httplist.current->sockptr, "<TD ROWSPAN=2><input type=submit value=print></TD></TR>");
  send_to_socketptr(httplist.current->sockptr, "<TR>");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"a\" %s></TD>",
		    type == 'a' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"b\" %s></TD>",
		    type == 'b' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"d\" %s></TD>",
		    type == 'd' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"h\" %s></TD>",
		    type == 'h' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"i\" %s></TD>",
		    type == 'i' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"o\" %s></TD>",
		    type == 'o' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"x\" %s></TD>",
		    type == 'x' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "<TD><input type=radio name=lines value=\"n\" %s></TD>",
		    type == 'n' ? "checked" : "");
  send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "</form>");

  switch (type) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n': case '-':
    break;
  default:
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Illegal line-type</P></CENTER>");
    return;
  }

  if (type != '-' && lines_findfirst(type)) {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
    send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Line</TH><TH ALIGN=left>Mask</TH><TH ALIGN=left>Locked</TH></TR>");
    do
      {
	send_to_socketptr(httplist.current->sockptr, "<TR><TD><A HREF=\"?line=%c%-4d\">%c%-4d</A></TD><TD>%35s</TD><TD>%s</TD></TR>", toupper(type), nr, toupper(type), nr, lines_mask(type), lines_islocked(type) ? "<IMG SRC=\"yes.gif\">" : "");
	nr++;
      }
    while (lines_findnext(type));
    send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
  }
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);
  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}

void http_send_maskinfo(char *line) {

  char type;
  int linenr;

  send_to_socketptr(httplist.current->sockptr, "<HTML><HEAD><TITLE>%s</TITLE></HEAD>", botnick);
  send_to_socketptr(httplist.current->sockptr, "<BODY TEXT=\"#99CCFF\" BGCOLOR=\"#000000\" LINK=\"#FFFF00\" VLINK=\"#CCCC66\" ALINK=\"#CCCC66\">");

  send_to_socketptr(httplist.current->sockptr, "<H1 ALIGN=CENTER>Maskinfo</H1>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Enter linenumber:</P></CENTER>");

  send_to_socketptr(httplist.current->sockptr, "<CENTER><FORM ACTION=\"\" METHOD=GET>");
  send_to_socketptr(httplist.current->sockptr, "<INPUT TYPE=TEXT name=line size=30 maxlength=40></FORM></CENTER>");

  type = tolower(line[0]);
  linenr = atoi(&line[1]);
  if (linenr < 1) {
    send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Linenumbers should be at least one</P></CENTER>");
  }
  else {
    switch(type) {
    case 'a': case 'o': case 'x':
      if (!(lines_findnr(type, linenr))) {
	send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Line %c%d not found</P></CENTER>", toupper(type), 
			  linenr);
      }
      else {
	send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Line</TH><TD>%c%d</TD></TR>", toupper(type), linenr);
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Mask</TH><TD>%s</TD></TR>", lines_mask(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Created by</TH><TD>%s</TD></TR>", lines_from(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Created on</TH><TD>%s</TD></TR>", 
			  time2string(lines_time(type)));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Locked</TH><TD>%s</TD></TR>", lines_islocked(type) ?
			  "<IMG SRC=\"yes.gif\">" : "");
	send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
      }
      break;
    case 'b': case 'd': case 'i': case 'n':
      if (!(lines_findnr(type, linenr))) {
	send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Line %c%d not found</P></CENTER>", toupper(type), 
			  linenr);
      }
      else {
	send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Line</TH><TD>%c%d</TD></TR>", toupper(type), linenr);
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Mask</TH><TD>%s</TD></TR>", lines_mask(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Comment</TH><TD>%s</TD></TR>", lines_comment(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Created by</TH><TD>%s</TD></TR>", lines_from(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Created on</TH><TD>%s</TD></TR>", 
			  time2string(lines_time(type)));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Locked</TH><TD>%s</TD></TR>", lines_islocked(type) ?
			  "<IMG SRC=\"yes.gif\">" : "");
	send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
      }
      break;
    case 'h':
      if (!(lines_findnr(type, linenr))) {
	send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Line %c%d not found</P></CENTER>", toupper(type), 
			  linenr);
      }
      else {
	send_to_socketptr(httplist.current->sockptr, "<CENTER><TABLE>");
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Line</TH><TD>%c%d</TD></TR>", toupper(type), linenr);
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Mask</TH><TD>%s</TD></TR>", lines_mask(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Comment</TH><TD>%s</TD></TR>", lines_comment(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Created by</TH><TD>%s</TD></TR>", lines_from(type));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Created on</TH><TD>%s</TD></TR>", 
			  time2string(lines_time(type)));
	send_to_socketptr(httplist.current->sockptr, "<TR><TH ALIGN=left>Former type</TH><TD>");
	switch(lines_islocked(type)) {
	case 0: send_to_socketptr(httplist.current->sockptr, "A"); break;
	case 1: send_to_socketptr(httplist.current->sockptr, "B"); break;
	case 2: send_to_socketptr(httplist.current->sockptr, "D"); break;
	case 3: send_to_socketptr(httplist.current->sockptr, "I"); break;
	case 4: send_to_socketptr(httplist.current->sockptr, "O"); break;
	case 5: send_to_socketptr(httplist.current->sockptr, "X"); break;
	case 6: send_to_socketptr(httplist.current->sockptr, "N"); break;
	default: send_to_socketptr(httplist.current->sockptr, "?"); break;
	}
	send_to_socketptr(httplist.current->sockptr, "</TD></TR>");
	send_to_socketptr(httplist.current->sockptr, "</TABLE></CENTER>");
      }
      break;
    default:
      send_to_socketptr(httplist.current->sockptr, "<CENTER><P>Unknown line-type</P></CENTER>");
      break;
    }
  }
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><IMG SRC=\"rule12.gif\" HEIGHT=22 WIDTH=468></P></CENTER>");
  send_to_socketptr(httplist.current->sockptr, "<CENTER><P><A HREF=\"/\">%s</A></P></CENTER>", botnick);
  send_to_socketptr(httplist.current->sockptr, "</BODY></HTML>");
}




void http_get(char *argv[], int argc) {

  char *file=NULL;

  if (argc == 1) return;

  /* make sure the current of the userlist points to something,
  ** so the %ulist_* variables do have a value and will not lead
  ** to Segmentation Fault when used.
  */
  if (!(userlist_findfirst())) { createunode("Eem"); }

  if (strcasecmp(argv[1], "/") == 0) {
    file = (char *)xmalloc(strlen(httpdir) + strlen("index.html") + 1);
    sprintf(file, "%s%s", httpdir, "index.html");
    http_sendfile_replace_vars(file);
    close_socketptr(httplist.current->sockptr);
    httpconnectionlost();
    xfree(file);
    return;
  }

  if (argv[1][0] == '/') {
    switch (argv[1][1]) {
    case '?':
      if (strncmp(&argv[1][2], "user=", 5) == 0) {
	http_send_userinfo(&argv[1][7]);
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
	return;
      }
      if (strncmp(&argv[1][2], "lines=", 6) == 0) {
	http_send_lines(argv[1][8]);
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
	return;
      }
      if (strncmp(&argv[1][2], "line=", 5) == 0) {
	http_send_maskinfo(&argv[1][7]);
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
	return;
      }
      if (strncmp(&argv[1][2], "chan=", 5) == 0) {
	http_sendchannel(&argv[1][7]);
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
	return;
      }
      if (strncmp(&argv[1][2], "top=", 4) == 0 && strncmp(&argv[1][7], "&length=", 8) == 0) {
	http_sendtop(argv[1][6], atoi(&argv[1][15]));
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
	return;
      }
      nosuchfile();
      close_socketptr(httplist.current->sockptr);
      httpconnectionlost();
      return;
    case '~':
      http_send_userinfo(&argv[1][2]);
      close_socketptr(httplist.current->sockptr);
      httpconnectionlost();
      return;
    case '!':
      if (strcasecmp("userlist", &argv[1][2]) == 0) {
	http_send_userlist();
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
      }
      else if (strcasecmp("irc", &argv[1][2]) == 0) {
	http_send_irc();
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
      }
      else if (strcasecmp("party", &argv[1][2]) == 0) {
	http_send_party();
	close_socketptr(httplist.current->sockptr);
	httpconnectionlost();
      }
      return;
    default:
      file = (char *)xmalloc(strlen(httpdir) + strlen(argv[1]));
      sprintf(file, "%s%s", httpdir, &argv[1][1]);
      if (match("*.html", file))
	http_sendfile_replace_vars(file);
      else
	http_sendfile_raw(file);
      close_socketptr(httplist.current->sockptr);
      httpconnectionlost();
      xfree(file);
      return;
    }
  }
}

struct {
  char *command;
  void (*function)(char *argv[], int argc);
} http_cmds[] = 
  { 
    { "GET",      http_get },
    { NULL,       NULL     }
  };

void http_parseinput(char *string) {

  int argc=7;
  char *argv[7];

  int i;

  split_string(string, argv, &argc);

  for (i=0; http_cmds[i].command; i++) {
    if (strcasecmp(http_cmds[i].command, argv[0]) == 0) {
      http_cmds[i].function(argv, argc);
      return;
    }
  }
}
