
/*******************************************************/
/**                                                   **/
/**    irc_console.c (c) 1996, 97 Emiel Bruijntjes    **/
/**                                                   **/
/** Messages entered on the console while in IRC mode **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#include "irc_console.h"
#include "channel.h"
#include "server.h"
#include "console.h"
#include "misc.h"
#include "gencmd.h"
#include "session.h"
#include "chanuser.h"
#include "config.h"
#include "reg.h"

extern char  current_nick[10];
extern char  commandchar3;
extern char *helpdir;
extern time_t time_null;


static char *query=NULL;

void irc_cmd_query(char *argv[], int argc) {

  if (argc == 1) {
    if (query) {
      console_send2ircwindow("*** Ending conversation with %s", query);
      xfree(query);
      query = NULL;
    }
    else
      console_send2ircwindow("*** You aren't querying anyone!");
  }
  else {
    xfree (query);
    query = (char *) xmalloc (strlen(argv[1]) + 1);
    strcpy(query, argv[1]);
    console_send2ircwindow("*** Starting conversation with %s", query);
  }
}

void irc_cmd_msg(char *argv[], int argc) {

  if (argc == 1) {
    console_send2ircwindow("*** You must specify a nickname or channel!");
    return;
  }
  if (argc == 2) {
    return;
  }
  if (argc >= 3) {
    repairstring(&argv[2], argc - 2);
    gen_cmd_msg(argv[1], argv[2], send_now_to_server);
  }
}

void irc_cmd_notice(char *argv[], int argc) {

  if (argc == 1) {
    console_send2ircwindow("*** You must specify a nickname or channel!");
    return;
  }
  if (argc == 2) {
    return;
  }
  if (argc >= 3) {
    repairstring(&argv[2], argc - 2);
    gen_cmd_notice(argv[1], argv[2], send_now_to_server);
  }
}

void irc_cmd_join(char *argv[], int argc) {

  char *channel;

  if (argc >= 2) {
    if (!(is_channel(argv[1]))) {
      channel = (char *) xmalloc (strlen(argv[1])+2);
      sprintf(channel, "#%s", argv[1]);
    }
    else {
      channel = (char *) xmalloc (strlen(argv[1])+1);
      strcpy(channel, argv[1]);
    }
    if (channellist_findchan(channel)) {
      channellist_makefirst();
      console_print_status();
    }
    else {
      addchan2watch(channel, 144, "nokey", 1997);
      if (argc == 2)
	send_now_to_server("JOIN %s", channel);
      else
	send_now_to_server("JOIN %s %s", channel, argv[2]);
    }
    xfree(channel);
  }
}

void irc_cmd_leavejoin(char *argv[], int argc) {

  char *channel;

  if (argc == 1) {
    if (channellist_findfirst()) {
      send_now_to_server("PART %s\nJOIN %s", channellist_channelname(), channellist_channelname());
    }
    return;
  }
  if (is_channel(argv[1])) {
    send_now_to_server("PART %s\nJOIN %s", argv[1], argv[1]);
  }
  else {
    channel = (char *) xmalloc (strlen(argv[1])+1);
    sprintf(channel, "#%s", argv[1]);
    send_now_to_server("PART %s\nJOIN %s", channel, channel);
    xfree(channel);
  }
  console_print_status();
}

void irc_cmd_leave(char *argv[], int argc) {

  char *channel;

  if (argc == 1) {
    if (channellist_findfirst()) {
      send_now_to_server("PART %s", channellist_channelname());
      delchan2watch(channellist_channelname());
    }
    return;
  }
  if (is_channel(argv[1])) {
    send_now_to_server("PART %s", argv[1]);
    delchan2watch(argv[1]);
  }
  else {
    channel = (char *) xmalloc (strlen(argv[1])+1);
    sprintf(channel, "#%s", argv[1]);
    send_now_to_server("PART %s", channel);
    delchan2watch(channel);
    xfree(channel);
  }
  console_print_status();
}


void irc_cmd_me(char *argv[], int argc) {

  if ((argc > 1) && (channellist_findfirst())) {
    repairstring(&argv[1], argc - 1);
    gen_cmd_me(channellist_channelname(), argv[1], send_now_to_server);
  }
}
      
void irc_cmd_mode(char *argv[], int argc) {

  if (argc == 1) return;
  if ((argc == 2) && (channellist_findfirst()))
    send_now_to_server("MODE %s :%s", channellist_channelname(), argv[1]);
  else {
    if (channellist_findfirst()) {
      if ((argv[1][0] == '+') || (argv[1][0] == '-')) {
	repairstring(&argv[1], argc - 1);
	send_now_to_server("MODE %s %s", channellist_channelname(), argv[1]);
      }
      else {
	repairstring(&argv[2], argc - 2);
	send_now_to_server("MODE %s %s", argv[1], argv[2]);
      }
    }
    else {
      repairstring(&argv[2], argc - 2);
      send_now_to_server("MODE %s %s", argv[1], argv[2]);
    }
  }
}

void irc_cmd_op(char *argv[], int argc) {

  char *nick[3]= { NULL, NULL, NULL };
  int i, j=0;

  if (!(channellist_findfirst()))
    return;

  for (i=1; i<argc; i++) {
    if (session_findnick(argv[i])) {
      if (!(session_ismatching('d'))) {
	nick[j] = argv[i];
	j++;
	if (j >= 3) break;
      }
      else {
	console_send2ircwindow("*** %s is not allowed to become channel operator", argv[i]);
      }
    }
  }
  switch (j) {
  case 0:
    return;
  case 1:
    send_now_to_server("MODE %s +o %s", channellist_channelname(), nick[0]);
    return;
  case 2:
    send_now_to_server("MODE %s +oo %s %s", channellist_channelname(), nick[0], nick[1]);
    return;
  default:
    send_now_to_server("MODE %s +ooo %s %s %s", channellist_channelname(), nick[0], nick[1], nick[2]);
    return;
  }
}

void irc_cmd_deop(char *argv[], int argc) {
  
  if (!(channellist_findfirst()))
    return;

  switch (argc) {
  case 1:
    return;
  case 2:
    send_now_to_server("MODE %s -o %s", channellist_channelname(), argv[1]);
    return;
  case 3:
    send_now_to_server("MODE %s -oo %s %s", channellist_channelname(), argv[1], argv[2]);
    return;
  default:
    send_now_to_server("MODE %s -ooo %s %s %s", channellist_channelname(), argv[1], argv[2], argv[3]);
    return;
  }
}

void irc_cmd_ban(char *argv[], int argc) {

  int i;
  char *banmatch=NULL;

  if (!(channellist_findfirst()))
    return;

  if (argc == 1) {
    for (i=0; i<MAX_BANS; i++) {
      if (returnban(i)) {
	console_send2ircwindow("*** %s: %s", channellist_channelname(), returnban(i));
      }
    }
    return;
  }
  if (session_findnick(argv[1])) {
    banmatch = (char *) xmalloc (strlen(session_nickuserhost()) + 1);
    maakbanmatch(session_nickuserhost(), banmatch);
    send_now_to_server("MODE %s %s+b %s %s", channellist_channelname(), ischanneloperator(channellist_channelname(), argv[1]) ? "-o" : 
		       "", ischanneloperator(channellist_channelname(), argv[1]) ? argv[1] : "", banmatch);
    xfree(banmatch);
  }
  else
    console_send2ircwindow("*** Cannot find %s", argv[1]);
}

void irc_cmd_bankick(char *argv[], int argc) {

  int i;
  char *banmatch=NULL;

  if (!(channellist_findfirst()))
    return;

  if (argc == 1) {
    for (i=0; i<MAX_BANS; i++) {
      if (returnban(i)) {
	console_send2ircwindow("*** %s: %s", channellist_channelname(), returnban(i));
      }
    }
    return;
  }
  if (session_findnick(argv[1])) {
    banmatch = (char *) xmalloc (strlen(session_nickuserhost()) + 1);
    maakbanmatch(session_nickuserhost(), banmatch);
    send_now_to_server("MODE %s %s+b %s %s\nKICK %s %s :- = bankick = -", channellist_channelname(), 
		       ischanneloperator(channellist_channelname(), argv[1]) ? "-o" : "",
		       ischanneloperator(channellist_channelname(), argv[1]) ? argv[1] : "", banmatch, channellist_channelname(),
		       argv[1]);
    xfree(banmatch);
  }
  else
    console_send2ircwindow("*** Cannot find %s", argv[1]);
}

void irc_cmd_unban(char *argv[], int argc) {

  int i, j=0;
  char *bans[3];

  if (!(channellist_findfirst()))
    return;

  if (argc > 1) {
    if (session_findnick(argv[1])) {
      for (i=0; i<MAX_BANS; i++) {
	if (match(returnban(i) ? returnban(i) : "", session_nickuserhost())) {
	  bans[j] = returnban(i);
	  j++;
	  if (j == 3) {
	    send_now_to_server("MODE %s -bbb %s %s %s", channellist_channelname(), bans[0], bans[1], bans[2]);
	    j = 0;
	  }
	}
      }
      if (j)
	send_now_to_server("MODE %s -b%c %s %s", channellist_channelname(), j >= 2 ? 'b' : ' ', bans[0], j >= 2 ? bans[1] : "");
    }
    else
      console_send2ircwindow("*** Cannot find %s", argv[1]);
  }
}

void irc_cmd_massunban(char *argv[], int argc) {
  
  char *channel;
  int   i, j;
  char *bans[3];

  if (argc >= 2)
    channel = argv[1];
  else {
    if (!(channellist_findfirst()))
      return;
    channel = channellist_channelname();
  }
  j = 0;
  for (i = 0; i < MAX_BANS; i++) {
    if (returnban(i)) {
      bans[j] = returnban(i);
      j++;
      if (j == 3) {
	send_now_to_server("MODE %s -bbb %s %s %s", channel, bans[0], bans[1], bans[2]);
	j = 0;
      }
    }
  }
  if (j)
    send_now_to_server("MODE %s -b%c %s %s", channel, j >= 2 ? 'b' : ' ', bans[0], j >= 2 ? bans[1] : "");
}

void irc_cmd_massop(char *argv[], int argc) {

  char *channel;
  int i;
  char *ops[3];

  if (argc >= 2) {
    channel = argv[1];
    if (!(channellist_findchan(channel))) {
      console_send2ircwindow("*** %s: You're not on that channel", channel);
      return;
    }
  }
  else {
    if (!(channellist_findfirst())) {
      return;
    }
    channel = channellist_channelname();
  }
  i = 0;
  if (chanusers_findfirst(channellist_chanusers())) {
    do {
      if ((!(chanusers_ischanop(channellist_chanusers()))) && (session_findnick(chanusers_nick(channellist_chanusers()))) &&
	  (!(session_ismatching('d')))) {
	ops[i] = chanusers_nick(channellist_chanusers());
	i++;
	if (i == 3) {
	  send_now_to_server("MODE %s +ooo %s %s %s", channel, ops[0], ops[1], ops[2]);
	  i = 0;
	}
      } 
    } while (chanusers_findnext(channellist_chanusers()));
  }
  if (i)
    send_now_to_server("MODE %s +o%c %s %s", channel, i >= 2 ? 'o' : ' ', ops[0], i >= 2 ? ops[1] : "");
}
  
void irc_cmd_massdeop(char *argv[], int argc) {

  char *channel;
  int i;
  char *ops[3];

  if (argc >= 2) {
    channel = argv[1];
    if (!(channellist_findchan(channel))) {
      console_send2ircwindow("*** %s: You're not on that channel", channel);
      return;
    }
  }
  else {
    if (!(channellist_findfirst())) {
      return;
    }
    channel = channellist_channelname();
  }
  i = 0;
  if (chanusers_findfirst(channellist_chanusers())) {
    do {
      if (chanusers_ischanop(channellist_chanusers()) && 
	  (strcasecmp(chanusers_nick(channellist_chanusers()), current_nick) != 0)) {
	ops[i] = chanusers_nick(channellist_chanusers());
	i++;
	if (i == 3) {
	  send_now_to_server("MODE %s -ooo %s %s %s", channel, ops[0], ops[1], ops[2]);
	  i = 0;
	}
      } 
    } while (chanusers_findnext(channellist_chanusers()));
  }
  if (i)
    send_now_to_server("MODE %s -o%c %s %s", channel, i >= 2 ? 'o' : ' ', ops[0], i >= 2 ? ops[1] : "");
}


void irc_cmd_topic(char *argv[], int argc) {

  if (argc == 1) {
    if (channellist_findfirst()) {
      send_now_to_server("TOPIC %s", channellist_channelname());
    }
    return;
  }
  if (argc == 2) {
    if (is_channel(argv[1])) {
      send_now_to_server("TOPIC %s", argv[1]);
      return;
    }
    if (channellist_findfirst()) {
      send_now_to_server("TOPIC %s :%s", channellist_channelname(), argv[1]);
    }
    return;
  }
  if (argc > 2) {
    if (is_channel(argv[1])) {
      repairstring(&argv[2], argc - 2);
      send_now_to_server("TOPIC %s :%s", argv[1], argv[2]);
    }
    else {
      if (channellist_findfirst()) {
	repairstring(&argv[1], argc - 1);
	send_now_to_server("TOPIC %s :%s", channellist_channelname(), argv[1]);
      }
    }
  }
}

void irc_cmd_invite(char *argv[], int argc) {

  if (argc == 2) {
    if (channellist_findfirst()) {
      send_now_to_server("INVITE %s %s", argv[1], channellist_channelname());
    }
  }
  if (argc > 2) {
    send_now_to_server("INVITE %s %s", argv[1], argv[2]);
  }
}

void irc_cmd_kick(char *argv[], int argc) {
  
  switch (argc) {
  case 1:
    return;
  case 2:
    if (channellist_findfirst()) {
      send_now_to_server("KICK %s %s :Console", channellist_channelname(), argv[1]);
    }
    return;
  case 3:
    if (is_channel(argv[1])) {
      send_now_to_server("KICK %s %s :Console", argv[1], argv[2]);
      return;
    }
    /* No break or return */
  default:
    if (is_channel(argv[1])) {
      repairstring(&argv[3], argc - 3);
      send_now_to_server("KICK %s %s :%s", argv[1], argv[2], argv[3]);
    }
    else {
      if (channellist_findfirst()) {
	repairstring(&argv[2], argc - 2);
	send_now_to_server("KICK %s %s :%s", channellist_channelname(), argv[1], argv[2]);
      }
    }
  }
}

void irc_cmd_masskick(char *argv[], int argc) {

  char *channel;
  int   i;
  char message[WAYTOBIG]="";

  if (argc >= 2) {
    channel = argv[1];
    if (!(channellist_findchan(channel))) {
      console_send2ircwindow("*** %s: You're not on that channel", channel);
      return;
    }
  }
  else {
    if (!(channellist_findfirst())) {
      return;
    }
    channel = channellist_channelname();
  }
  i = 0;
  if (chanusers_findfirst(channellist_chanusers())) {
    do {
      if (strcasecmp(chanusers_nick(channellist_chanusers()), current_nick) != 0) {
	strcat(message, chanusers_nick(channellist_chanusers()));
	strcat(message, ",");
	i++;
	if (i == 20) {
	  send_now_to_server("KICK %s %s :Masskick from console", channel, message);
	  i = 0;
	  strcpy(message, "");
	}
      }
    } while (chanusers_findnext(channellist_chanusers()));
  }
  if (i)
    send_now_to_server("KICK %s %s :Masskick from console", channel, message);
}

void irc_cmd_whois(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("WHOIS %s", argv[1]);
  }
}

void irc_cmd_whowas(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("WHOWAS %s", argv[1]);
  }
}

void irc_cmd_list(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("LIST %s", argv[1]);
  }
  else {
    send_now_to_server("LIST");
  }
}

void irc_cmd_names(char *argv[], int argc) {

  char buf[WAYTOBIG]="";
  int i=0;
  chanusers_t *chs;

  if (argc == 1) {
    if (!(channellist_findfirst())) return;
  }
  else {
    if (!(channellist_findchan(argv[1]))) {
      console_send2ircwindow("*** %s: You're not on that channel", argv[1]);
      return;
    }
  }
  chs = channellist_chanusers();
  if (chanusers_findfirst(chs)) {
    do
      {
	if (chanusers_ischanop(chs)) 
	  strcat(buf, "@");
	strcat(buf, chanusers_nick(chs));
	strcat(buf, " ");
	i++;
	if (i == 50) {
	  console_send2ircwindow("*** Users on %s: %s", channellist_channelname(), buf);
	  i=0;
	  strcpy(buf, "");
	}
      }
    while (chanusers_findnext(chs));
  }
  if (i)
    console_send2ircwindow("*** Users on %s: %s", channellist_channelname(), buf);
}

void irc_cmd_lusers(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("LUSERS %s", argv[1]);
  }
  else {
    send_now_to_server("LUSERS");
  }
}

void irc_cmd_time(char *argv[], int argc) {
  
  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("TIME %s", argv[1]);
  }
  else {
    send_now_to_server("TIME");
  }
}

void irc_cmd_version(char *argv[], int argc) {
  
  if (argc > 1) {
    send_now_to_server("PRIVMSG %s :VERSION", argv[1]);
  }
  else {
    send_now_to_server("VERSION");
    console_send2ircwindow("*** Client: %s (console)", LONGVERSION);
  }
}

void irc_cmd_ping(char *argv[], int argc) {

  if (argc > 1) {
    send_now_to_server("PRIVMSG %s :PING %lu", argv[1], (unsigned long int)time_null);
  }
}

void irc_cmd_ctcp(char *argv[], int argc) {

  int i=0;

  if (argc >= 3) {
    repairstring(&argv[2], argc - 2);
    while (argv[2][i] != ' ' && argv[2][i] != '\0') {
      argv[2][i] = toupper(argv[2][i]);
      i++;
    }
    send_now_to_server("PRIVMSG %s :%s", argv[1], argv[2]);
  }
}

void irc_cmd_stats(char *argv[], int argc) {
  
  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("STATS %s", argv[1]);
  }
  else {
    send_now_to_server("STATS");
  }
}

void irc_cmd_links(char *argv[], int argc) {
  
  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("LINKS %s", argv[1]);
  }
  else {
    send_now_to_server("LINKS");
  }
}

void irc_cmd_admin(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("ADMIN %s", argv[1]);
  }
  else {
    send_now_to_server("ADMIN");
  }
}

void irc_cmd_motd(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    send_now_to_server("MOTD %s", argv[1]);
  }
  else {
    send_now_to_server("MOTD");
  }
}

void irc_cmd_help(char *argv[], int argc) {
  
  char *file;
  int i;
  struct {
    char *command;
    char *file;
  } helpfile_argc1[] = {
    {"admin", "ihelp.admin"},
    {"b", "ihelp.ban"},
    {"ban", "ihelp.ban"},
    {"bankick", "ihelp.bankick"},
    { "bk", "ihelp.bankick"},
    { "bkick", "ihelp.bankick"},
    { "clearban", "ihelp.massunban"},
    { "ctcp", "ihelp.ctcp"},
    { "deop", "ihelp.deop"},
    { "help", "ihelp.help"},
    { "i", "ihelp.invite"},
    { "invite", "ihelp.invite"},
    { "j", "ihelp.join"},
    { "join", "ihelp.join"},
    { "k", "ihelp.kick"},
    { "kick", "ihelp.kick"},
    { "l", "ihelp.leave"},
    { "leave", "ihelp.leave"},
    { "leavejoin", "ihelp.leavejoin"},
    { "links", "ihelp.links"},
    { "list", "ihelp.list"},
    { "lj", "ihelp.leavejoin"},
    { "luser", "ihelp.lusers"},
    { "lusers", "ihelp.lusers"},
    { "m", "ihelp.msg"},
    { "massdeop", "ihelp.massdeop"},
    { "masskick", "ihelp.masskick"},
    { "massop", "ihelp.massop"},
    { "massunban", "ihelp.massunban"},
    { "me", "ihelp.me"},
    { "mode", "ihelp.mode"},
    { "motd", "ihelp.motd"},
    { "msg", "ihelp.msg"},
    { "n", "ihelp.names"},
    { "names", "ihelp.names"},
    { "notice", "ihelp.notice"},
    { "op", "ihelp.op"},
    { "ping", "ihelp.ping"},
    { "query", "ihelp.query"},
    { "quit",  "ihelp.signoff"},
    { "s", "ihelp.signoff"},
    { "signoff", "ihelp.signoff"},
    { "stats", "ihelp.stats"},
    { "t", "ihelp.topic"},
    { "time", "ihelp.time"},
    { "topic", "ihelp.topic"},
    { "unban", "ihelp.unban"},
    { "version", "ihelp.version"},
    { "whois", "ihelp.whois"},
    { "whowas", "ihelp.whowas"},
    { NULL, NULL }
  };

  if (argc == 1) {
    file = (char *) xmalloc (strlen(helpdir)+strlen("ihelp")+1);
    sprintf(file, "%sihelp", helpdir);
    catfile(file, console_send2ircwindow2, "console");
    xfree(file);
    return;
  }
  if (argc >= 2) {
    for (i = 0; helpfile_argc1[i].command; i++) {
      if (strcasecmp(argv[1], helpfile_argc1[i].command) == 0) {
	file = (char *)xmalloc(strlen(helpdir) + strlen(helpfile_argc1[i].file) + 1);
	sprintf(file, "%s%s", helpdir, helpfile_argc1[i].file);
	catfile(file, console_send2ircwindow2, "console");
	xfree(file);
	return;
      }
    }
    console_send2ircwindow("*** %s: No such topic", argv[1]);
    return;
  }
}

void irc_cmd_signoff(char *argv[], int argc) {

  if (argc > 1) {
    repairstring(&argv[1], argc - 1);
    gen_cmd_signoff(argv[1]);
  }
  else
    gen_cmd_signoff("Signoff from console");
}


struct {
  char *command;
  void (*function)(char *argv[], int argc);
} irc_commands[] = {
  { "admin",      irc_cmd_admin     },
  { "b",          irc_cmd_ban       },
  { "ban",        irc_cmd_ban       },
  { "bankick",    irc_cmd_bankick   },
  { "bk",         irc_cmd_bankick   }, 
  { "bkick",      irc_cmd_bankick   },
  { "clearban",   irc_cmd_massunban },
  { "ctcp",       irc_cmd_ctcp      },
  { "deop",       irc_cmd_deop      },
  { "help",       irc_cmd_help      },
  { "i",          irc_cmd_invite    },
  { "invite",     irc_cmd_invite    },
  { "j",          irc_cmd_join      },
  { "join",       irc_cmd_join      },
  { "k",          irc_cmd_kick      },
  { "kick",       irc_cmd_kick      },
  { "l",          irc_cmd_leave     },
  { "leave",      irc_cmd_leave     },
  { "leavejoin",  irc_cmd_leavejoin },
  { "links",      irc_cmd_links     },
  { "list",       irc_cmd_list      },
  { "lj",         irc_cmd_leavejoin },
  { "luser",      irc_cmd_lusers    },
  { "lusers",     irc_cmd_lusers    },
  { "m",          irc_cmd_msg       },
  { "massdeop",   irc_cmd_massdeop  },
  { "masskick",   irc_cmd_masskick  },
  { "massop",     irc_cmd_massop    },
  { "massunban",  irc_cmd_massunban },
  { "me",         irc_cmd_me        },
  { "mode",       irc_cmd_mode      },
  { "motd",       irc_cmd_motd      },
  { "msg",        irc_cmd_msg       },
  { "n",          irc_cmd_names     }, 
  { "names",      irc_cmd_names     },
  { "notice",     irc_cmd_notice    },
  { "op",         irc_cmd_op        },
  { "ping",       irc_cmd_ping      },
  { "query",      irc_cmd_query     },
  { "quit",       irc_cmd_signoff   },
  { "s",          irc_cmd_signoff   },
  { "signoff",    irc_cmd_signoff   },
  { "stats",      irc_cmd_stats     },
  { "t",          irc_cmd_topic     },
  { "time",       irc_cmd_time      },
  { "topic",      irc_cmd_topic     },
  { "unban",      irc_cmd_unban     },
  { "version",    irc_cmd_version   },
  { "whois",      irc_cmd_whois     },
  { "whowas",     irc_cmd_whowas    },
  { NULL,         NULL,             }
};

void parse_irc_command(char *command) {

  int argc=7, i;
  char *argv[7];

  split_string(command, argv, &argc);
  if (argc == 0) return;

  for (i=0; irc_commands[i].command; i++) {
    if (strcasecmp(argv[0], irc_commands[i].command) == 0) {
      irc_commands[i].function(argv, argc);
      return;
    }
  }
  console_send2ircwindow("*** Unknown command: %s", argv[0]);
}


void parse_irc_console(char *text) {

  if ((text == NULL) || (text[0] == '\0'))
    return;

  if (text[0] == commandchar3) {
    if (text[1] == ' ')
      text += 2;
    else {
      text++;
      parse_irc_command(text);
      return;
    }
  }

  if (query) {
    gen_cmd_msg(query, text, send_now_to_server);
  }
  else if (channellist_findfirst()) {
    gen_cmd_msg(channellist_channelname(), text, send_now_to_server);
  }
}
