
/*******************************************************/
/**                                                   **/
/**       lines.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**          Database containing the "lines"          **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "reg.h"
#include "lines.h"
#include "misc.h"

struct lines {
  int    length;
  struct linenode *current;
  struct linenode *head;
  struct linenode *tail;
  char   required;      /* 0 = nothing, 1 = operator, 2 = administrator, 3 = superuser */
  char   lock_required; /* 0 = nothing, 1 = operator, 2 = administrator, 3 = superuser */
  time_t timeout;
};

struct linenode {
  char    *mask;
  char    *from;
  char    *comment;
  time_t   created;
  int      lock;
  
  struct linenode *next;
  struct linenode *prev;
};

extern time_t time_null;

struct lines alines={ 0 , NULL , NULL , NULL , 2 , 3 , 0 };
struct lines blines={ 0 , NULL , NULL , NULL , 1 , 2 , 0 };
struct lines dlines={ 0 , NULL , NULL , NULL , 1 , 2 , 0 };
struct lines hlines={ 0 , NULL , NULL , NULL , 0 , 0 , 604800 };    /* timeout one week */
struct lines ilines={ 0 , NULL , NULL , NULL , 1 , 2 , 0 };
struct lines olines={ 0 , NULL , NULL , NULL , 1 , 2 , 0 };
struct lines plines={ 0 , NULL , NULL , NULL , 2 , 3 , 0 };
struct lines xlines={ 0 , NULL , NULL , NULL , 2 , 3 , 0 };
struct lines nlines={ 0 , NULL , NULL , NULL , 2 , 3 , 0 };
struct lines *curlines=NULL;

void setcurlines(char linetype) {

  switch (linetype) {
  case 'a': curlines = &alines; break;
  case 'b': curlines = &blines; break;
  case 'd': curlines = &dlines; break;    
  case 'h': curlines = &hlines; break;				  
  case 'i': curlines = &ilines; break;
  case 'o': curlines = &olines; break;
  case 'x': curlines = &xlines; break;
  case 'n': curlines = &nlines; break;
  }
}

void save_lines(char type, FILE *fp) {

  setcurlines(type);

  fwritechar(fp, curlines->required);
  fwritechar(fp, curlines->lock_required);
  fwritetime_t(fp, curlines->timeout);
  fwriteint(fp, curlines->length);
  if (lines_findfirst(type)) {
    do
      {
	fwritestr2bin(lines_mask(type), fp);
	fwritestr2bin(lines_comment(type),fp);
	fwritestr2bin(lines_from(type), fp);
	fwritetime_t(fp, lines_time(type));
	fwriteint(fp, lines_islocked(type));
      }
    while (lines_findnext(type));
  }
}

void save_all_lines(FILE *fp) {

  save_lines('a', fp); save_lines('b', fp); save_lines('d', fp); save_lines('h', fp);
  save_lines('i', fp); save_lines('o', fp); save_lines('x', fp); save_lines('n', fp);

  fwriteint(fp, 516093);
}

void load_lines(char type, FILE *fp) {

  int i;

  setcurlines(type);

  deletealllines(type);
  freadchar(fp, &curlines->required);
  freadchar(fp, &curlines->lock_required);
  freadtime_t(fp, &curlines->timeout);
  freadint(fp, &curlines->length);
  
  for (i=0; i<curlines->length; i++) {
    curlines->current = (struct linenode *)xmalloc(sizeof(struct linenode));
    if (curlines->head == NULL) {
      curlines->head = curlines->current;
      curlines->current->prev = NULL;
    } 
    else {
      curlines->tail->next = curlines->current;
      curlines->current->prev = curlines->tail;
    }
    curlines->current->next = NULL;
    curlines->tail = curlines->current;
    curlines->current->comment = curlines->current->from = curlines->current->mask = NULL;
    fgetstrfrombin(&curlines->current->mask, fp);
    fgetstrfrombin(&curlines->current->comment, fp);
    fgetstrfrombin(&curlines->current->from, fp);
    freadtime_t(fp, &curlines->current->created);
    freadint(fp, &curlines->current->lock);
  }
}

int load_all_lines(FILE *fp) {

  int test;

  load_lines('a', fp); load_lines('b', fp); load_lines('d', fp); load_lines('h', fp);
  load_lines('i', fp); load_lines('o', fp); load_lines('x', fp); load_lines('n', fp);

  freadint(fp, &test);
  return (test == 516093);
}

void setalinetimeout(int newvalue) { if (newvalue >= 0) alines.timeout = newvalue; }
void setblinetimeout(int newvalue) { if (newvalue >= 0) blines.timeout = newvalue; }
void setdlinetimeout(int newvalue) { if (newvalue >= 0) dlines.timeout = newvalue; }
void sethlinetimeout(int newvalue) { if (newvalue >= 0) hlines.timeout = newvalue; }
void setilinetimeout(int newvalue) { if (newvalue >= 0) ilines.timeout = newvalue; }
void setolinetimeout(int newvalue) { if (newvalue >= 0) olines.timeout = newvalue; }
void setxlinetimeout(int newvalue) { if (newvalue >= 0) xlines.timeout = newvalue; }
void setnlinetimeout(int newvalue) { if (newvalue >= 0) nlines.timeout = newvalue; }

int alinetimeout() { return alines.timeout; }
int blinetimeout() { return blines.timeout; }
int dlinetimeout() { return dlines.timeout; }
int hlinetimeout() { return hlines.timeout; }
int ilinetimeout() { return ilines.timeout; }
int olinetimeout() { return olines.timeout; }
int xlinetimeout() { return xlines.timeout; }
int nlinetimeout() { return nlines.timeout; }


int  required_to_change(char linetype) {

  setcurlines(linetype);
  return curlines->required;
}

int  required_to_change_when_locked(char linetype) {

  setcurlines(linetype);
  return curlines->lock_required;
}

void change_require_to_change(char type, int normal, int lock) {

  setcurlines(type);
  curlines->required = normal;
  curlines->lock_required = lock;
}

int  numberoflines(char type) {

  setcurlines(type);
  return curlines->length;
}

int  numberoflockedlines(char type) {

  int i=0;

  if (lines_findfirst(type)) {
    do {
      if (lines_islocked(type)) i++;
    } while (lines_findnext(type));
  }
  return i;
}


int addline(char linetype, char *mask, char *comment, char *from) {

  int i=0;

  setcurlines(linetype);
  curlines->current = curlines->head;
  while (curlines->current != NULL) {
    i++;
    if (strcasecmp(curlines->current->mask, mask) == 0) return i;
    curlines->current = curlines->current->next;
  }
  i++;

  curlines->current = (struct linenode *)xmalloc(sizeof(struct linenode));

 if (curlines->length == 0) {
    curlines->head = curlines->current;
    curlines->current->prev = NULL;
  } 
  else {
    curlines->tail->next = curlines->current;
    curlines->current->prev = curlines->tail;
  }

  curlines->current->next = NULL;
  curlines->tail = curlines->current;
  curlines->length = curlines->length + 1;

  curlines->current->mask = (char *) xmalloc (strlen(mask)+1);
  strcpy (curlines->current->mask, mask);

  curlines->current->comment = (char *) xmalloc (strlen(comment)+1);
  strcpy (curlines->current->comment, comment);

  curlines->current->from = (char *) xmalloc (strlen(from)+1);
  strcpy (curlines->current->from, from);

  curlines->current->created = time_null;
  curlines->current->lock = 0;
  return i;
}

void moveline(char dest, char source, int type, char *from) {

  struct lines *sourcelist=&alines;
  struct lines *destlist=&alines;

  switch (source) {
  case 'a': sourcelist = &alines; break;
  case 'b': sourcelist = &blines; break;
  case 'd': sourcelist = &dlines; break;    
  case 'h': sourcelist = &hlines; break;				  
  case 'i': sourcelist = &ilines; break;
  case 'o': sourcelist = &olines; break;
  case 'x': sourcelist = &xlines; break;
  case 'n': sourcelist = &nlines; break;
  default: fprintf(stderr, "Bug ... Crash ...\n"); exit(1);
  }

  switch (dest) {
  case 'a': destlist = &alines; break;
  case 'b': destlist = &blines; break;
  case 'd': destlist = &dlines; break;
  case 'h': destlist = &hlines; break;
  case 'i': destlist = &ilines; break;
  case 'o': destlist = &olines; break;
  case 'x': destlist = &xlines; break;
  case 'n': destlist = &nlines; break;
  default: fprintf(stderr, "Bug ... Crash ...\n"); exit(1);
  }

  destlist->current = sourcelist->current;

  if (sourcelist->length == 1) {
    sourcelist->current = NULL;
    sourcelist->tail = NULL;
    sourcelist->head = NULL;
    sourcelist->length = 0;
  }
  else {
    if (sourcelist->current->next == NULL) {
      sourcelist->tail = sourcelist->current->prev;
      sourcelist->tail->next = NULL;
      sourcelist->current = sourcelist->tail;
      sourcelist->length = sourcelist->length - 1;
    }
    else {
      if (sourcelist->current->prev == NULL) {
	sourcelist->head = sourcelist->current->next;
	sourcelist->head->prev = NULL;
	sourcelist->current = sourcelist->head;
	sourcelist->length = sourcelist->length - 1;
      }
      else {
	sourcelist->current->prev->next = sourcelist->current->next;
	sourcelist->current->next->prev = sourcelist->current->prev;
	sourcelist->length = sourcelist->length - 1;
	sourcelist->current = sourcelist->head;
      }
    }
  }
  if (destlist->head == NULL) {
    destlist->head = destlist->current;
    destlist->tail = destlist->current;
    destlist->current->prev = NULL;
    destlist->current->next = NULL;
  }
  else {
    destlist->current->prev = destlist->tail;
    destlist->tail->next = destlist->current;
    destlist->tail = destlist->current;
    destlist->current->next = NULL;
  }
  destlist->length = destlist->length + 1;
  xfree(destlist->current->from);
  destlist->current->from = (char *) xmalloc (strlen(from)+1);
  strcpy (destlist->current->from, from);
  destlist->current->created = time_null;
  destlist->current->lock = type;
}


void delline(char linetype) {

  setcurlines(linetype);

  xfree (curlines->current->mask);
  xfree (curlines->current->comment);
  xfree (curlines->current->from);

  if (curlines->length == 1) {
    xfree(curlines->current);
    curlines->current = NULL;
    curlines->tail = NULL;
    curlines->head = NULL;
    curlines->length = 0;
    return;
  }

  if (curlines->current->next == NULL) {
    curlines->tail = curlines->current->prev;
    curlines->tail->next = NULL;
    xfree(curlines->current);
    curlines->current = curlines->tail;
    curlines->length = curlines->length - 1;
    return;
  }

  if (curlines->current->prev == NULL) {
    curlines->head = curlines->current->next;
    curlines->head->prev = NULL;
    xfree(curlines->current);
    curlines->current = curlines->head;
    curlines->length = curlines->length - 1;
    return;
  }

  curlines->current->prev->next = curlines->current->next;
  curlines->current->next->prev = curlines->current->prev;
  xfree(curlines->current);
  curlines->length = curlines->length - 1;
  curlines->current = curlines->head;
  return;
}

void deletealllines(char linetype) {

  setcurlines(linetype);

  while ((curlines->current = curlines->head)) {
    delline(linetype);
  }
}


int does_this_match(char linetype, char *string) {

  int i=0;

  setcurlines(linetype);

  curlines->current = curlines->head;

  while (curlines->current != NULL) {
    i++;
    if (match(curlines->current->mask, string))
      return i;
    curlines->current = curlines->current->next;
  }
  return 0;
}

int next_match(char linetype, char *format, int i) {
  
  if (!(lines_findnr(linetype, i))) return 0;
  i--;
  while (curlines->current != NULL) {
    i++;
    if (match(curlines->current->mask, format))
      return i;
    curlines->current = curlines->current->next;
  }
  return 0;
}

int lines_findfirst(char linetype) {
  
  setcurlines(linetype);
  return (int)(curlines->current = curlines->head);
}

int lines_findnext (char linetype) {

  setcurlines(linetype);
  return (int)(curlines->current = curlines->current->next);
}

int lines_findnr   (char linetype, int linenr) {

  int i=1;

  if (linenr <= 0) return 0;

  setcurlines(linetype);
  curlines->current = curlines->head;

  while (curlines->current != NULL) {
    if (i == linenr) return 1;
    i++;
    curlines->current = curlines->current->next;
  }
  return 0;
}

char *lines_comment(char linetype) {
  
  setcurlines(linetype);
  return curlines->current->comment;
}

char *lines_mask(char linetype) {

  setcurlines(linetype);
  return curlines->current->mask;
}

char *lines_from(char linetype) {
  
  setcurlines(linetype);
  return curlines->current->from;
}

time_t lines_time(char linetype) {

  setcurlines(linetype);
  return curlines->current->created;
}

int lines_islocked(char linetype) {

  setcurlines(linetype);
  return curlines->current->lock;
}

void lines_setlock(char linetype) {

  setcurlines(linetype);
  curlines->current->lock = 1;
}

void lines_setunlock(char linetype) {

  setcurlines(linetype);
  curlines->current->lock = 0;
}

void lines_setcomment(char linetype, char *comment) {

  setcurlines(linetype);
  xfree(curlines->current->comment);
  
  curlines->current->comment = (char *) xmalloc (strlen(comment)+1);
  strcpy (curlines->current->comment, comment);
}

void auto_remove_lines(char linetype) {

  setcurlines(linetype);
  if (curlines->timeout == 0) return;
  while ((lines_findfirst(linetype)) && ((curlines->current->created + curlines->timeout) < time_null)) delline(linetype);
} 

void printlines(char type, int begin, int end, void output(char *, ...)) {

  int i=1;

  setcurlines(type);
  if (type != 'h') {
    if (lines_findfirst(type)) {
      do {
	if (i > end) return;
	if (i >= begin)
	  output("%c%-4d %35s %s", toupper(type), i, lines_mask(type), lines_islocked(type) ? "-=locked=-" : "");
	i++;
      } while (lines_findnext(type));
    }
    return;
  }
  if (lines_findfirst(type)) {
    do {
      if (i > end) return;
      if (i >= begin) {
        switch (lines_islocked(type)) {
        case 0:
          output("H%-4d %35s (A-line)", i, lines_mask(type));
          break;
        case 1:
          output("H%-4d %35s (B-line)", i, lines_mask(type));
          break;
        case 2:
          output("H%-4d %35s (D-line)", i, lines_mask(type));
          break;  
        case 3:
          output("H%-4d %35s (I-line)", i, lines_mask(type));
          break;
        case 4:
          output("H%-4d %35s (O-line)", i, lines_mask(type));
          break;
	case 5:
          output("H%-4d %35s (X-line)", i, lines_mask(type));
          break;
	case 6:
          output("H%-4d %35s (N-line)", i, lines_mask(type));
          break;  
        default:
          output("H%-4d %35s (Unknown)", i, lines_mask(type));
          break;
        }
      }
      i++;
    } while (lines_findnext(type));
  }
}


void printmatchinglines(char type, char *nickuserhost, void output(char *, ...)) {

  int i=1;

  setcurlines(type);
  if (type != 'h') {
    if (lines_findfirst(type)) {
      do {
	if (match(lines_mask(type), nickuserhost))
	  output("%c%-4d %35s %s", toupper(type), i, lines_mask(type), lines_islocked(type) ? "-=locked=-" : "");
	i++;
      } while (lines_findnext(type));
    }
    return;
  }
  if (lines_findfirst(type)) {
    do {
      if (match(lines_mask(type), nickuserhost)) {
        switch (lines_islocked(type)) {
        case 0:
          output("H%-4d %35s (A-line)", i, lines_mask(type));
          break;
        case 1:
          output("H%-4d %35s (B-line)", i, lines_mask(type));
          break;
        case 2:
          output("H%-4d %35s (D-line)", i, lines_mask(type));
          break;  
        case 3:
          output("H%-4d %35s (I-line)", i, lines_mask(type));
          break;
        case 4:
          output("H%-4d %35s (O-line)", i, lines_mask(type));
          break;
	case 5:
          output("H%-4d %35s (X-line)", i, lines_mask(type));
          break;
	case 6:
          output("H%-4d %35s (N-line)", i, lines_mask(type));
          break;  
        default:
          output("H%-4d %35s (Unknown)", i, lines_mask(type));
          break;
        }
      }
      i++; 
    } while (lines_findnext(type));
  }
}


void printmatchinglines_html(char type, char *nickuserhost, void output(int, char *, ...), int sockptr) {

  int i=1;

  setcurlines(type);
  if (lines_findfirst(type)) {
    do {
      if (match(lines_mask(type), nickuserhost))
	output(sockptr, "<TR><TD><A HREF=\"?line=%c%-4d\">%c%-4d</A></TD><TD>%35s</TD><TD>%s</TD></TR>", toupper(type), i, toupper(type), i, lines_mask(type), lines_islocked(type) ? "<IMG SRC=\"yes.gif\">" : "");
      i++;
    } while (lines_findnext(type));
  }
}
