
/*******************************************************/
/**                                                   **/
/**       lines.h (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**          Database containing the "lines"          **/
/**                                                   **/
/*******************************************************/


#ifndef LINES_H_
#define LINES_H_

#include <time.h>
#include <stdio.h> 

void save_all_lines(FILE *);
int  load_all_lines(FILE *);

int  required_to_change(char);             /* 0 = nothing, 1 = operator, 2 = administrator, 3 = superuser */
int  required_to_change_when_locked(char); /* 0 = nothing, 1 = operator, 2 = administrator, 3 = superuser */
void change_require_to_change(char type, int normal, int lock);

int  addline (char type, char *mask, char *comment, char *from);  /* Create a line */
void moveline(char src , char dest, int lock, char *from);        /* Move a line to another type, this is only used
								  ** with 'h' as second argument. */

void delline (char type);      /* Really delete a line. (Line will not be stored in H-lines) */

void deletealllines(char type);

int does_this_match(char type, char *string);  /* Checks if string matches with one of the line of type TYPE, and if so,
					       ** returns the number of the matching line. */
int next_match(char type, char *string, int nr); /* as does_this_match(), but now checks only the lines with number n and
						 ** higher. */

int lines_findfirst(char type);
int lines_findnext (char type);
int lines_findnr   (char type, int nr);

char  *lines_comment (char type);
char  *lines_mask    (char type);
char  *lines_from    (char type);
time_t lines_time    (char type);
int    lines_islocked(char type);

void lines_setlock   (char type);
void lines_setunlock (char type);
void lines_setcomment(char type, char *comment);  /* Set comment or password of a line */

void auto_remove_lines(char type);  /* Remove lines older than value set with set?linetimeout() */

void printlines(char type, int begin, int end, void output(char *, ...));
void printmatchinglines(char type, char *nickuserhost, void output(char *, ...));
void printmatchinglines_html(char type, char *nickuserhost, void output(int, char *, ...), int sockptr);

int  numberoflines(char type);
int  numberoflockedlines(char type);

void setalinetimeout(int);
void setblinetimeout(int);
void setdlinetimeout(int);
void sethlinetimeout(int);
void setilinetimeout(int);
void setolinetimeout(int);
void setxlinetimeout(int);
void setnlinetimeout(int);

int alinetimeout();
int blinetimeout();
int dlinetimeout();
int hlinetimeout();
int ilinetimeout();
int olinetimeout();
int xlinetimeout();
int nlinetimeout();

#endif
