
/*******************************************************/
/**                                                   **/
/**        misc.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**              Miscellaneous functions              **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif


#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <strings.h>


#include "misc.h"


void putlog(char *file, char *format, ...) {

  FILE    *fp;
  char    line[2000];
  va_list msg;

  if ((fp = fopen(file, "a")) == NULL) return;
  
  if(!format) { fclose(fp); return; }
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[1999] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  fprintf(fp, "[%s] %s\n", time2string(0), line);
  fclose(fp);
}

char * repairstring(char *argv[], int argc) {

  int i=0;
  char *ch;

  for (i=0; i < argc - 1; i++) {
    ch = index(argv[i], '\0');
    *ch = ' ';
  }
  return argv[0];
}

void split_string(char *string, char *argv[], int *argc) {

  int max;

  max = *argc;

  *argc = 0;
  while (*string == ' ') string++;
  if (*string == '\0') return;

  argv[0] = string; *argc = 1;
  if (*argc >= max) return;
  while (*string != '\0') {
    if (*string == ' ') {
      *string = '\0';
      string++;
      while (*string == ' ') string++;
      if (*string == '\0') return;
      argv[*argc] = string; (*argc)++;
      if (*argc >= max) return;
    }    
    string++;
  }
}

char * time2string(time_t curtime) {

  register char *ch1;
  register char *ch2;

  if (curtime == 0) curtime = time(NULL);
  ch1 = ctime(&curtime);
  ch2 = index(ch1, '\n');
  *ch2 = '\0';
  return ch1;
}


void *
xmalloc (size_t size)
{
  register void *value = malloc (size);
  if (value == 0) {
    fprintf(stderr, "virtual memory exhausted\n");
    exit (1);
  }
  return value;
}

void *
xrealloc (void *ptr, size_t size)
{
  register void *value;
  
  if (ptr == NULL)
    return xmalloc(size);
  else {
    if ((value = realloc (ptr, size)) == NULL) {
      fprintf(stderr, "virtual memory exhausted\n");
      exit(1);
    }
  }
  return value;
}

/*
void 
xfree(void *ptr)
{
  if (ptr != NULL) free(ptr);
}
*/

int nickonly(char *nick, char *nickuserhost) {
  
  int i=0;

  while (nickuserhost[i] != '!') {
    if (nickuserhost[i] == '\0') {
      return 0;
    }
    nick[i] = nickuserhost[i];
    i++;
  }
  nick[i]='\0';
  return 1;
}

int is_user(char *line) {

  int i = 0;

  while ((line[i] != '!') && (line[i] != '.') && (line[i] != '\0')) {
    i++;
  }
  if (line[i] == '.' || line[i] == '\0') {
    return 0;
  }
  else {
    return 1;
  }
}

int getalnaarbin(int i) {
  
  if (i==0)
    return 0;
  else
    return 1;
}

void maakbanmatch(char *nickuserhost, char *match) {

  int i=0;
  int j=0;
  int laatstepunt=-1;

  match[j] = '*';
  j++;

  while (nickuserhost[i] != '!')
    i++;

  match[j] = '!';
  j++;
  match[j] = '*';
  j++;

  i++;
  i++;

  while (nickuserhost[i] != '@') {
    match[j] = nickuserhost[i];
    i++; j++;
  }

  match[j] = '@';
  j++;i++;
  if ((nickuserhost[i] < '0') || (nickuserhost[i] > '9')) {
    match[j] = '*';
    j++;
    while (nickuserhost[i] != '.') {
      i++;
    }
    while (nickuserhost[i] != '\0') {
      match[j] = nickuserhost[i];
      i++; j++;
    }
    match[j] = '\0';
  }
  else {
    while (nickuserhost[i] != '\0') {
      if (nickuserhost[i] == '.') {
	laatstepunt = j;
      }
      match[j]=nickuserhost[i];
      i++;j++;
    }
    if (laatstepunt != -1) {
      match[laatstepunt + 1] = '*';
      match[laatstepunt + 2] = '\0';
    }
    else {
      match[j] = '\0';
    }
  }
}

char *botonly(char *nickbot) {

  int i=0;

  while (nickbot[i] != '@') {
    if ((nickbot[i] == '\0') || (nickbot[i] == '\n') || (nickbot[i] == '
')) {
      return &nickbot[0];
    }
    i++;
  }
  return &nickbot[i+1];
}

int is_bot(char *string) {

  int i=0;

  while (1) {
    if (string[i] == '\0')
      return 1;
    if (string[i] == '@')
      return 0;
    i++;
  }
}

void hoofdletter(char *string) {

  int i=0;

  while(string[i] != '\0') {
    string[i] = toupper(string[i]);
    i++;
  }
}

int    freadint(FILE *fp, int *tmpint) {

  return fread(tmpint, sizeof(int), 1, fp);
}

int    fwriteint(FILE *fp, int tmpint) {

  return fwrite(&tmpint, sizeof(int), 1, fp);
}

int    freadtime_t(FILE *fp, time_t *tmpint) {

  return fread(tmpint, sizeof(time_t), 1, fp);
}

int    fwritetime_t(FILE *fp, time_t tmpint) {

  return fwrite(&tmpint, sizeof(time_t), 1, fp);
}

int    fwritechar(FILE *fp, char tmpchar) {

  return fwrite(&tmpchar, sizeof(char), 1, fp);
}

int    freadchar(FILE *fp, char *tmpchar) {

  return fread(tmpchar, sizeof(char), 1, fp);
}

size_t fgetstrfrombin(char **buf, FILE *fp) {

  register int ch='\0';
  size_t size;

  xfree(*buf);
  size = 0;
  *buf = NULL;
  while (((ch = fgetc(fp)) != EOF) && (ch != '\0')) {
    size++;
    *buf = xrealloc(*buf, size);
    (*buf)[size-1] = ch;
  }
  size++;
  *buf = xrealloc(*buf, size);
  (*buf)[size-1]=ch;
  return size;
}

void fwritestr2bin(char *buf, FILE *fp) {

  fprintf(fp, "%s", buf);
  fwritechar(fp, '\0');
}

size_t my_getdelim(char **buf, FILE *fp, int delim) {

  register int ch;
  size_t size;

  xfree(*buf);
  size = 0;
  *buf = NULL;
  while (((ch = fgetc(fp)) != EOF) && (ch != delim)) {
    size++;
    *buf = xrealloc(*buf, size);
    (*buf)[size-1] = ch;
  }
  return size;
}

int my_feof(FILE *fp) {

  int ch;

  ungetc(ch = fgetc(fp), fp);
  return (ch == EOF);
}


int is_channel(char *string) {

  return ((string[0] == '#') || (string[0] == '&') || (string[0] == '+'));
}

char * spaces(int n) {

  static char *buf=NULL;
  int i;

  if (n < 0) n = 0;

  buf = (char *) xrealloc (buf, n+1);
  for (i=0; i<n; i++)
    buf[i] = ' ';
  buf[n] = '\0';
  return buf;
}


static int hex2dec_(char c) {

  if (isdigit(c))
    return c-'0';
  else
    return tolower(c) - 'a' + 10;
}
  

int hex2dec(char *str, int n) {

  int i, retvalue=0;

  for (i=0; i<n; i++) {
    retvalue *= 16;
    retvalue += hex2dec_(str[i]);
  }
  return retvalue;
}
