
/*******************************************************/
/**                                                   **/
/**        net.c (c) 1996, 97 Emiel Bruijntjes        **/
/**                                                   **/
/**   Internet connections and I/O via the internet   **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif


#include <stdarg.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>


#include "net.h"
#include "config.h"
#include "http.h"
#include "bots.h"
#include "session.h"
#include "server.h"
#include "partyline.h"
#include "annon.h"
#include "misc.h"
#include "parse.h"
#include "term.h"
#include "xwin.h"
#include "out_console.h"


#ifdef SOLARIS
extern  int     gethostname(char *, int);
#endif

#ifdef SUNOS
extern void   bzero(void *, size_t);
extern int    socket(int, int, int);
extern int    accept(int, struct sockaddr *, int *);
extern int    bind(int, const struct sockaddr *, int);
extern int    listen(int, int);
extern int    gethostname(char *, int);
extern int    getsockname(int, struct sockaddr *, int *);
extern int    getpeername(int, struct sockaddr *, int *);
extern int    connect(int, struct sockaddr *, int);
extern int    send(int, const char *, int, int);
#endif

struct io {
  char *buffer;            /* Already read characters */
  char *queue;             /* Data we still need to send */
  size_t bytes_read;       /* Current length of buffer (not size) */
  size_t queue_length;     /* Number of bytes stored (length of queue) */ 
  int sock;                /* The filedescriptor of the connection */
  char flags;              /* Flags, see below */
  char type;               /* Type, see net.h */
  time_t flood_time;       /* How long are we checking? */
  size_t flood_bytes;      /* How many bytes were sent in FLOOD_TIME seconds? */
};

/* flags: */
#define F_LISTEN            1
#define F_CONNECTING        2
#define F_CONNECT           4
#define F_CLOSING           8

static fd_set read_set;
static fd_set write_set;
static int io_array_ub = 0;
static struct io *io_array[MAX_CONNECTIONS];

extern int internetbytessent;
extern int internetbytesreceived;

extern time_t time_null;

void net_init() {       /* Call this procedure before you start to use this module */

  int i;

  FD_ZERO(&read_set);
  FD_ZERO(&write_set);
  for (i=0; i < MAX_CONNECTIONS; i++) {
    io_array[i] = NULL;
  }
}

void empty_io(int ptr) {    /* Deletes io_array[ptr] */

  int i;

  xfree (io_array[ptr]->buffer);
  xfree (io_array[ptr]->queue);
  if (io_array[ptr]->sock >= 0) {
    FD_CLR(io_array[ptr]->sock, &read_set);
    FD_CLR(io_array[ptr]->sock, &write_set);
    close(io_array[ptr]->sock);
  }
  xfree (io_array[ptr]);
  io_array[ptr] = NULL;
  if (io_array_ub == ptr) {
    for (i = ptr; i >= 0; i--) {
      if (io_array[i] != NULL) { io_array_ub = i; break; }
    }
  }
}

int getsock() {             /* Creates a new entry in io_array and returns element-number in array, or -1 on failure */

  int i;

  for (i=0; i < MAX_CONNECTIONS; i++) {
    if (io_array[i] == NULL) {
      
      io_array[i] = (struct io *) xmalloc (sizeof(struct io));
      if (io_array_ub <= i) io_array_ub = i;
      io_array[i]->flags = 0;
      io_array[i]->type = 0;
      io_array[i]->buffer = NULL;
      io_array[i]->queue = NULL;
      io_array[i]->queue_length = 0;
      io_array[i]->bytes_read  = 0;
      io_array[i]->flood_time  = 0;
      io_array[i]->flood_bytes = 0;
      if ((io_array[i]->sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	empty_io(i);
	return -1;
      }
      else {
	fcntl(io_array[i]->sock,F_SETFL,O_NONBLOCK);
	return i;
      }
    }
  }
  return -1;
}

int get_stdin() {

  int i;

  for (i=0; i < MAX_CONNECTIONS; i++) {
    if (io_array[i] == NULL) {
      io_array[i] = (struct io *) xmalloc (sizeof(struct io));
      if (io_array_ub <= i) io_array_ub = i;
      io_array[i]->flags = F_CONNECT;
      io_array[i]->type = T_CONSOLE;
      io_array[i]->buffer = NULL;
      io_array[i]->queue = NULL;
      io_array[i]->queue_length = 0;
      io_array[i]->bytes_read = 0;
      io_array[i]->sock = fileno(stdin);
      io_array[i]->flood_time  = 0;
      io_array[i]->flood_bytes = 0;
      FD_SET(io_array[i]->sock, &read_set);
      return i;
    }
  }
  return -1;
}


int add_X_socket(int sock) {

  int i;

  for (i=0; i < MAX_CONNECTIONS; i++) {
    if (io_array[i] == NULL) {
      io_array[i] = (struct io *) xmalloc (sizeof(struct io));
      if (io_array_ub <= i) io_array_ub = i;
      io_array[i]->flags = F_CONNECT;
      io_array[i]->type = T_XCONSOLE;
      io_array[i]->buffer = NULL;
      io_array[i]->queue = NULL;
      io_array[i]->queue_length = 0;
      io_array[i]->bytes_read = 0;
      io_array[i]->sock = sock;
      io_array[i]->flood_time  = 0;
      io_array[i]->flood_bytes = 0;
      FD_SET(io_array[i]->sock, &read_set);
      return i;
    }
  }
  return -1;
}


void del_X_sockptr(int ptr) {

  int i;

  if (io_array[ptr]->sock >= 0) {
    FD_CLR(io_array[ptr]->sock, &read_set);
    FD_CLR(io_array[ptr]->sock, &write_set);
  }
  xfree (io_array[ptr]);
  io_array[ptr] = NULL;
  if (io_array_ub == ptr) {
    for (i = ptr; i >= 0; i--) {
      if (io_array[i] != NULL) { io_array_ub = i; break; }
    }
  }
}



int my_accept(int ptr) {  /* Creates a new entry in io_array and returns element-number in array, or -1 on failure */

  int i;
  size_t size;
  struct sockaddr_in addr;

  for (i=0; i < MAX_CONNECTIONS; i++) {
    if (io_array[i] == NULL) {
      io_array[i] = (struct io *) xmalloc (sizeof(struct io));
      if (io_array_ub <= i) io_array_ub = i;
      io_array[i]->flags = F_CONNECT;
      io_array[i]->type = io_array[ptr]->type;
      io_array[i]->buffer = NULL;
      io_array[i]->queue = NULL;
      io_array[i]->queue_length = 0;
      io_array[i]->bytes_read = 0;
      io_array[i]->flood_time = 0;
      io_array[i]->flood_bytes = 0;
      size = sizeof(struct sockaddr_in);
      if ((io_array[i]->sock = accept(io_array[ptr]->sock, (struct sockaddr *)&addr, &size)) < 0) {
	empty_io(i);
	return -1;
      }
      else {
	fcntl(io_array[i]->sock,F_SETFL,O_NONBLOCK);
	return i;
      }
    }
  }
  size = sizeof(struct sockaddr_in);
  if ((i = accept(io_array[ptr]->sock, (struct sockaddr *)&addr, &size)) != -1) { close(i); }
  return -1;
}

int my_bind(int ptr, struct sockaddr_in *name) {     /* binds an internet addres to a socket, return 0 on failure */

  if (bind(io_array[ptr]->sock, (struct sockaddr *)name, sizeof(struct sockaddr_in)) == -1) {
    empty_io(ptr);
    return 0;
  }
  else {
    return 1;
  }
}

int my_listen(int ptr) {            /* Starts listening on a port, returns 0 on failure */

  if ((io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) {
    empty_io(ptr);
    return 0;
  }
  if (listen(io_array[ptr]->sock, 1) < 0) {
    empty_io(ptr);
    return 0;
  }
  return 1;
}

char * my_gethostname() {

  static char myhostname[256]="";
  struct hostent *host;

  if ((myhostname[0]) && (index(myhostname, '.'))) return myhostname;
  gethostname(myhostname, 256);
  if (index(myhostname, '.') == NULL) {
    if ((host = gethostbyname(myhostname)) == NULL) { return myhostname; }
    if (strcasecmp(host->h_name, "localhost") == 0) {
      return myhostname;
    }
    return host->h_name;
  }
  return myhostname;
}

unsigned long int my_ip() {

  struct in_addr *in;
  struct hostent *hp;

  hp=gethostbyname(my_gethostname());
  if (hp==NULL) exit(1);
  in=(struct in_addr *)(hp->h_addr_list[0]);
  return ntohl(in->s_addr);
}

int start_listen(char type, unsigned int port) {

  struct sockaddr_in name;
  int ptr;

  if ((ptr = getsock()) < 0) return -1;
  name.sin_family = AF_INET;
  name.sin_port = htons(port);
  name.sin_addr.s_addr = INADDR_ANY;
  if (!(my_bind(ptr, &name))) return -1;
  io_array[ptr]->flags = F_LISTEN;
  io_array[ptr]->type = type;
  if (!(my_listen(ptr))) return -1;
  FD_SET(io_array[ptr]->sock, &read_set);
  return ptr;
}

int start_random_listen(char type) {

  return start_listen(type, 0);
}

void stop_listen(int ptr) {

  empty_io(ptr);
}

int my_getsockname(int ptr, struct sockaddr_in *addr) {

  size_t size;

  if ((io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) {
    return 0;
  }

  size = sizeof(struct sockaddr_in);
  if (getsockname(io_array[ptr]->sock, (struct sockaddr *)addr, &size) < 0) {
    return 0;
  }
  return 1;
}

int my_getpeername(int ptr, struct sockaddr_in *addr) {

  size_t size;

  if ((io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) {
    return 0;
  }

  size = sizeof(struct sockaddr_in);
  if (getpeername(io_array[ptr]->sock, (struct sockaddr *)addr, &size) < 0) {
    return 0;
  }
  return 1;
}


struct hostent * my_gethostbyaddr(unsigned long int ip) {

  return gethostbyaddr((char *)&ip, sizeof(ip), AF_INET);
}


char * give_localhost_name(int ptr) {

  struct sockaddr_in s_address;
  struct hostent *hp;

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) return give_localhost_ip(ptr);
  if (!(my_getsockname(ptr, &s_address))) return give_localhost_ip(ptr);
  if ((hp=my_gethostbyaddr(s_address.sin_addr.s_addr)) == NULL) return my_gethostname();
  if (strcasecmp(hp->h_name, "localhost") == 0) {
    return my_gethostname();
  }
  return hp->h_name;
}

char * give_localhost_ip(int ptr) {

  unsigned char no[4];
  unsigned long int addr;
  static char name[16];

  addr = give_localip(ptr);
  no[0]=(addr>>24)&0xff; 
  no[1]=(addr>>16)&0xff; 
  no[2]=(addr>>8)&0xff;
  no[3]=addr&0xff;
  sprintf(name, "%u.%u.%u.%u", no[0], no[1], no[2], no[3]);
  return name;
}
  

unsigned long int give_localip(int ptr) {

  struct sockaddr_in s_address;

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) return 0;

  if (!(my_getsockname(ptr, &s_address))) return 0;
  return ntohl(s_address.sin_addr.s_addr);
}


unsigned short int give_localport(int ptr) {

  struct sockaddr_in s_address;

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) return 0;

  if (!(my_getsockname(ptr, &s_address))) return 0;
  return ntohs(s_address.sin_port);
}

char * give_connectedhost_name(int ptr) {

  struct sockaddr_in s_address;
  struct hostent *hp;

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) return give_connectedhost_ip(ptr);
  if (!(my_getpeername(ptr, &s_address))) return give_connectedhost_ip(ptr);
  if ((hp=my_gethostbyaddr(s_address.sin_addr.s_addr)) == NULL) return give_connectedhost_ip(ptr);
  if (strcasecmp("localhost", hp->h_name)==0) {
    return my_gethostname();
  }
  return hp->h_name;
}

char * give_connectedhost_ip(int ptr) {

  unsigned char no[4];
  unsigned long int addr;
  static char name[16];

  addr = give_connectedip(ptr);
  no[0]=(addr>>24)&0xff; 
  no[1]=(addr>>16)&0xff; 
  no[2]=(addr>>8)&0xff;
  no[3]=addr&0xff;
  sprintf(name, "%u.%u.%u.%u", no[0], no[1], no[2], no[3]);
  return name;
}
  

unsigned long int give_connectedip(int ptr) {

  struct sockaddr_in s_address;

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) return 0;

  if (!(my_getpeername(ptr, &s_address))) return 0;
  return ntohl(s_address.sin_addr.s_addr);
}


unsigned short int give_connectedport(int ptr) {

  struct sockaddr_in s_address;

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) return 0;

  if (!(my_getpeername(ptr, &s_address))) return 0;
  return ntohs(s_address.sin_port);
}


int my_connect(int ptr, struct sockaddr_in *addr) {

  if ((ptr < 0) || (io_array[ptr] == NULL) || (io_array[ptr]->sock == -1)) {
    empty_io(ptr);
    return -1;
  }
  if (connect(io_array[ptr]->sock, (struct sockaddr *)addr, sizeof(struct sockaddr_in)) < 0) {
    if (errno == EINPROGRESS) {
      return 0;
    }
    empty_io(ptr);
    return -1;
  }
  return 1;
}

int start_connecting(char type, int *ptr, char *hostname, unsigned short int port) {

  struct sockaddr_in name;
  struct hostent *host;

  if ((*ptr = getsock()) < 0) return -1;
  if ((host = gethostbyname(hostname)) == NULL) { empty_io(*ptr); return -1; }
  name.sin_family = AF_INET;
  name.sin_port = htons(port);
  memcpy(&name.sin_addr.s_addr, host->h_addr, host->h_length);
  switch(my_connect(*ptr, &name)) {
  case -1: 
    return -1;
  case  0:
    FD_SET(io_array[*ptr]->sock, &write_set);
    io_array[*ptr]->flags = F_CONNECTING;
    io_array[*ptr]->type = type;
    return 0;
  case 1:
    FD_SET(io_array[*ptr]->sock, &read_set);
    io_array[*ptr]->flags = F_CONNECT;
    io_array[*ptr]->type = type;
    return 1;
  default:
    return -1;
  }
}


int start_connecting2(char type, int *ptr, unsigned long int addr, unsigned short int port) {

  struct sockaddr_in name;

  if ((*ptr = getsock()) < 0) return -1;
  name.sin_family = AF_INET;
  name.sin_port = htons(port);
  name.sin_addr.s_addr = htonl(addr);
  switch(my_connect(*ptr, &name)) {
  case -1: 
    return -1;
  case  0:
    FD_SET(io_array[*ptr]->sock, &write_set);
    io_array[*ptr]->flags = F_CONNECTING;
    io_array[*ptr]->type = type;
    return 0;
  case 1:
    FD_SET(io_array[*ptr]->sock, &read_set);
    io_array[*ptr]->flags = F_CONNECT;
    io_array[*ptr]->type = type;
    return 1;
  default:
    return -1;
  }
}

int  socketptr_is_flooding(int ptr, size_t nr_bytes) {

  switch (io_array[ptr]->type) {
  case T_CHAT: case T_PARTY: case T_ANNON: case T_HTTP: case T_OUTCONSOLE:
    if ((io_array[ptr]->flood_time + 60) < time_null) {
      io_array[ptr]->flood_time  = time_null;
      io_array[ptr]->flood_bytes = 0;
    }
    io_array[ptr]->flood_bytes += nr_bytes;
    return (int)(io_array[ptr]->flood_bytes > FLOOD);
  default:
    return 0;
  }
}

int  read_and_parse(int ptr, void function(char *)) {
  
  ssize_t bytes;
  
  if (io_array[ptr]->buffer == NULL) { io_array[ptr]->buffer = (char *) xmalloc (WAYTOBIG); }
  if (io_array[ptr]->bytes_read >= WAYTOBIG) { empty_io(ptr); return 0; }
  while ((bytes = read(io_array[ptr]->sock, &io_array[ptr]->buffer[io_array[ptr]->bytes_read], 1)) > 0) {
    internetbytesreceived += bytes;
    if (socketptr_is_flooding(ptr, bytes)) { empty_io(ptr); return 0; }
    if (io_array[ptr]->buffer[io_array[ptr]->bytes_read] == '\r') io_array[ptr]->buffer[io_array[ptr]->bytes_read] = '\0';
    if (io_array[ptr]->buffer[io_array[ptr]->bytes_read] == '\n') {
      io_array[ptr]->buffer[io_array[ptr]->bytes_read] = '\0';
      io_array[ptr]->bytes_read = 0;
      function(io_array[ptr]->buffer);
      return 1;
    }
    io_array[ptr]->bytes_read = io_array[ptr]->bytes_read + bytes;
    if (io_array[ptr]->bytes_read >= WAYTOBIG) { empty_io(ptr); return 0; }
  }
  if ((bytes == 0) || ((bytes == -1) && (errno != EAGAIN))) { empty_io(ptr); return 0; }
  return 1;
}

static int  is_connected(int ptr) {   /* Function called to see if non-blocking connecting socket is now connected */

  if (read(io_array[ptr]->sock, &io_array[ptr]->buffer[io_array[ptr]->bytes_read], 0) < 0) {
    switch (errno) {
    case EWOULDBLOCK: return 1;
    default:
      empty_io(ptr);
      return 0;
    }
  }
  return 1;
}

int read_buffer_raw(int ptr) {

  ssize_t bytes;

  if (io_array[ptr]->buffer == NULL) {
    io_array[ptr]->buffer = (char *) xmalloc (WAYTOBIG * 10);
  }
  if ((bytes = read(io_array[ptr]->sock, io_array[ptr]->buffer, WAYTOBIG * 10)) <= 0) {
    empty_io(ptr);
    return 0;
  }
  internetbytesreceived += bytes;
  if (socketptr_is_flooding(ptr, bytes)) { empty_io(ptr); return 0; }
  return bytes;
}

int read_int32bit(int ptr, unsigned int *nbytes) {

  int bytes;

  if ((bytes = read(io_array[ptr]->sock, nbytes, sizeof(unsigned int))) < sizeof(unsigned int)) {
    empty_io(ptr);
    return 0;
  }
  internetbytesreceived += bytes;
  if (socketptr_is_flooding(ptr, bytes)) { empty_io(ptr); return 0; }
  return bytes;
}

void read_socket(int ptr) {

  int i;
  unsigned int nbytes;

  switch(io_array[ptr]->flags) {
  case F_LISTEN:
    switch(io_array[ptr]->type) {
    case T_CHAT:
      if ((i = my_accept(ptr)) != -1) {
	FD_SET(io_array[i]->sock, &read_set);
	empty_io(ptr);
	if (!(session_findsockptr(T_CHAT, ptr))) { empty_io(i); return; }
	session_chat_connection_established(i);
      }
      return;
    case T_ANNON:
      if ((i = my_accept(ptr)) != -1) {
	FD_SET(io_array[i]->sock, &read_set);
	addannon(i);   /* It's possible the connection will be closed in addannon() */
	if (io_array[i] != NULL) io_array[i]->type = T_ANNON;
      }
      return;
    case T_HTTP:
      if ((i = my_accept(ptr)) != -1) {
	FD_SET(io_array[i]->sock, &read_set);
	addhttp(i);
	if (io_array[i] != NULL) io_array[i]->type = T_HTTP;
      }
      return;
    case T_SEND:
      if ((i = my_accept(ptr)) != -1) {
	FD_SET(io_array[i]->sock, &read_set);
	empty_io(ptr);
	if (!(session_findsockptr(T_SEND, ptr))) { empty_io(i); return; }
	session_sendestablished(i);
      }
      return;
    default:
      return;
    }
  case F_CONNECT:
    switch(io_array[ptr]->type) {
    case T_CHAT:
      if (!(session_findsockptr(T_CHAT, ptr))) { empty_io(ptr); return; }
      if (!(read_and_parse(ptr, session_parse_input))) { session_chat_connection_lost(); }
      return;
    case T_SERVER:
      if (!(read_and_parse(ptr, parse_server_msg))) { server_lost(); }
      return;
    case T_ANNON:
      if (!(annon_findsockptr(ptr))) { empty_io(ptr); return; }
      if (!(read_and_parse(ptr, parse_annon))) { annonconnectionlost(); }
      return;
    case T_BOT:
      if (!(botlist_findsockptr(ptr))) { empty_io(ptr); return; }
      if (!(read_and_parse(ptr, botlist_parse_input))) { botlist_connectionlost(); }
      return;
    case T_PARTY:
      if (!(party_findsockptr(ptr))) { empty_io(ptr); return; }
      if (!(read_and_parse(ptr, party_parse_input))) { delete_party(); }
      return;
    case T_HTTP:
      if (!(http_findsockptr(ptr))) { empty_io(ptr); return; }
      if (!(read_and_parse(ptr, http_parseinput))) { httpconnectionlost(); }
      return;
    case T_SEND:
      if (!(session_findsockptr(T_SEND, ptr))) { empty_io(ptr); return; }
      if (!(read_int32bit(ptr, &nbytes))) { session_closesend(); }
      else { session_parsesend(ntohl(nbytes)); }
      return;
    case T_GET:
      if (!(session_findsockptr(T_GET, ptr))) { empty_io(ptr); return; }
      if (!((i = read_buffer_raw(ptr)))) { session_getlost(); }
      else { session_parseget(io_array[ptr]->buffer, i); }
      return;
    case T_CONSOLE:
      read_and_parse_terminal();
      return;
    case T_XCONSOLE:
      if (!X_session_findsockptr(ptr)) { empty_io(ptr); return; }
      read_X_events();
      return;
    case T_OUTCONSOLE:
      if (!out_console_findsockptr(ptr)) { empty_io(ptr); return; }
      if (!(read_and_parse(ptr, parse_out_console))) { out_console_del(); }
      return;
    default:
      return;
    }
  case F_CLOSING:
    read_buffer_raw(ptr);
    return;
  default:
    return;
  }
}

void establish_connection(int ptr) {
  
  io_array[ptr]->flags = F_CONNECT;
  FD_CLR(io_array[ptr]->sock, &write_set);
    
  switch(io_array[ptr]->type) {
  case T_CHAT:
    if (!(session_findsockptr(T_CHAT, ptr))) { empty_io(ptr); return; }
    if (is_connected(ptr)) {
      session_chat_connection_established(ptr);
      FD_SET(io_array[ptr]->sock, &read_set);
    }
    else { session_chat_connection_lost(); }
    return;
  case T_GET:
    if (!(session_findsockptr(T_GET, ptr))) { empty_io(ptr); return; }
    if (is_connected(ptr)) {
      session_get_connection_established();
      FD_SET(io_array[ptr]->sock, &read_set);
    } 
    else { session_getlost(); }
    return;
  case T_SERVER:
    if (is_connected(ptr)) {
      init_server();
      FD_SET(io_array[ptr]->sock, &read_set);
    }
    else { server_lost(); }
    return;
  case T_BOT:
    if (!(botlist_findsockptr(ptr))) { empty_io(ptr); return; }
    if (is_connected(ptr)) {
      botlist_connection_established();
      FD_SET(io_array[ptr]->sock, &read_set);
    }
    else { botlist_connectionlost(); }
    return;
  default:
    return;
  }
}

static void dequeue_sockptr(int sockptr);

int wait_for_io(struct timeval *timeout) {

  int fds, i;
  fd_set tmp_read_set, tmp_write_set;

  tmp_read_set = read_set;
  tmp_write_set = write_set;

  fds = select(FD_SETSIZE, &tmp_read_set, &tmp_write_set, NULL, timeout);
  time_null = time(NULL);

  switch (fds) {
  case -1: break;
  case  0: return 0;
  default:
    for (i = 0; i <= io_array_ub; i++) {
      if (fds == 0) return 1;
      if (io_array[i] == NULL) continue;
      if (FD_ISSET(io_array[i]->sock, &tmp_write_set)) {
	if (io_array[i]->flags == F_CONNECTING)
	  establish_connection(i);
	else
	  dequeue_sockptr(i);
	fds--;
      }
      if ((io_array[i] != NULL) && (io_array[i]->sock != -1) && (FD_ISSET(io_array[i]->sock, &tmp_read_set))) {
	FD_CLR(io_array[i]->sock, &tmp_read_set);
	read_socket(i);
	fds--;
      }
    }
  }
  return 1;
}

void changesocketptrtype(int ptr, char type) {

  if ((ptr >= 0) && (io_array[ptr] != NULL)) io_array[ptr]->type = type;
}

void close_socketptr(int ptr) {

  if ((ptr >= 0) && (io_array[ptr] != NULL) && (io_array[ptr]->sock != -1)) {
    if (io_array[ptr]->queue)
      io_array[ptr]->flags = F_CLOSING;
    else
      empty_io(ptr);
  }
}

static int my_send(int sockptr, char *buf, size_t n) {

  int nr_bytes_sent=0;

  if (io_array[sockptr]->queue != NULL) {
    io_array[sockptr]->queue = (char *) xrealloc (io_array[sockptr]->queue, io_array[sockptr]->queue_length + n);
    memcpy(io_array[sockptr]->queue + io_array[sockptr]->queue_length, buf, n);
    io_array[sockptr]->queue_length += n;
  }
  else {
    nr_bytes_sent = send(io_array[sockptr]->sock, buf, n, 0);
    if (nr_bytes_sent < 0) nr_bytes_sent = 0;
    if (nr_bytes_sent < n) {
      io_array[sockptr]->queue = (char *) xmalloc (n - nr_bytes_sent);
      memcpy(io_array[sockptr]->queue, buf + nr_bytes_sent, n - nr_bytes_sent);
      io_array[sockptr]->queue_length += n - nr_bytes_sent;
      FD_SET(io_array[sockptr]->sock, &write_set);
    }
  }
  return nr_bytes_sent;
}

static void dequeue_sockptr(int sockptr) {

  char *buf;
  int   n;

  buf = io_array[sockptr]->queue;
  n   = io_array[sockptr]->queue_length;
  io_array[sockptr]->queue = NULL;
  io_array[sockptr]->queue_length = 0;
  FD_CLR(io_array[sockptr]->sock, &write_set);
  my_send(sockptr, buf, n);
  if ((io_array[sockptr]->queue == NULL) && (io_array[sockptr]->flags = F_CLOSING)) empty_io(sockptr);
  xfree(buf);
}


void send_to_socketptr(int sockptr, char *format, ...) {

  char    bigbuf[WAYTOBIG_OUT];          
  va_list msg;

  if ((sockptr < 0) || (io_array[sockptr] == NULL) || (io_array[sockptr]->sock == -1) || 
      (io_array[sockptr]->flags == F_CLOSING)) return;
  /* This next line should fix the problem with sending NULLs w/ SunOS */
  /* It should probably never happen to begin with, but SunOS will put */
  /* the process to sleep indefinately if you try. Linux works. -- CJS */
  if(!format) return;
  va_start(msg, format);
  vsprintf(bigbuf, format, msg);

  strcat(bigbuf, "\n");

  bigbuf[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);
  internetbytessent += my_send(sockptr, bigbuf, strlen(bigbuf));
}

void send_to_socketptr_raw(int sockptr, char *buffer, size_t size) {

  if ((sockptr < 0) || (io_array[sockptr] == NULL) || (io_array[sockptr]->sock == -1) ||
      (io_array[sockptr]->flags == F_CLOSING)) return;
  internetbytessent += my_send(sockptr, buffer, size);
}

void send_to_socketptr_nonewline(int sockptr, char *format, ...) {

  char    bigbuf[WAYTOBIG_OUT];          
  va_list msg;

  if ((sockptr < 0) || (io_array[sockptr] == NULL) || (io_array[sockptr]->sock == -1) ||
      (io_array[sockptr]->flags == F_CLOSING)) return;
  /* This next line should fix the problem with sending NULLs w/ SunOS */
  /* It should probably never happen to begin with, but SunOS will put */
  /* the process to sleep indefinately if you try. Linux works. -- CJS */
  if(!format) return;
  va_start(msg, format);
  vsprintf(bigbuf, format, msg);

  bigbuf[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);
  internetbytessent += my_send(sockptr, bigbuf, strlen(bigbuf));
}

void print_connections(void output(char *, ...)) {

  int i;

  output("entry  fd  ");
  output("---------------------------------------------------------");
  for (i = 0; i <= io_array_ub; i++) {
    if (io_array[i] == NULL)
      output("%3d    --  Not used", i);
    else {
      switch (io_array[i]->type) {
      case T_CHAT:
	switch (io_array[i]->flags) {
	case F_LISTEN:
	  output("%3d   %3d  DCC CHAT: Listening at port %d", i, io_array[i]->sock, give_localport(i));
	  break;
	case F_CONNECTING:
	  output("%3d   %3d  DCC CHAT: Trying %s %d", i, io_array[i]->sock, give_connectedhost_name(i), 
		 give_connectedport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  DCC CHAT: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  DCC CHAT: Closing connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_SEND:
	switch (io_array[i]->flags) {
	case F_LISTEN:
	  output("%3d   %3d  DCC SEND: Listening at port %d", i, io_array[i]->sock, give_localport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  DCC SEND: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  DCC SEND: Closing connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_GET:
	switch (io_array[i]->flags) {
	case F_CONNECTING:
	  output("%3d   %3d  DCC GET: Trying %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  DCC GET: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  DCC GET: Closing connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_PARTY:
	switch (io_array[i]->flags) {
	case F_CONNECT:
	  output("%3d   %3d  Outside partyline: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	case F_CLOSING:
	  output("%3d   %3d  Outside partyline: Closing connection with %s %d", i, io_array[i]->sock, 
		 give_connectedhost_name(i), give_connectedport(i));
	  break;
	}
	break;
      case T_SERVER:
	switch(io_array[i]->flags) {
	case F_CONNECTING:
	  output("%3d   %3d  IRC server: Trying %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  IRC server: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  IRC server: Closing connection with %s %d", i, io_array[i]->sock, 
		 give_connectedhost_name(i), give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_ANNON:
	switch(io_array[i]->flags) {
	case F_LISTEN:
	  output("%3d   %3d  Anonymous: Listening at port %d", i, io_array[i]->sock, give_localport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  Anonymous: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  Anonymous: Closing connection with %s %d", i, io_array[i]->sock, 
		 give_connectedhost_name(i), give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_BOT:
	switch(io_array[i]->flags) {
	case F_CONNECTING:
	  output("%3d   %3d  Bot: Trying %s %d", i, io_array[i]->sock, give_connectedhost_name(i), 
		 give_connectedport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  Bot: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  Bot: Closing connection with %s %d", i, io_array[i]->sock, 
		 give_connectedhost_name(i), give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_HTTP:
	switch(io_array[i]->flags) {
	case F_LISTEN:
	  output("%3d   %3d  HTTP-server: Listening at port %d", i, io_array[i]->sock, give_localport(i));
	  break;
	case F_CONNECT:
	  output("%3d   %3d  HTTP-server: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  HTTP-server: Closing connection with %s %d", i, io_array[i]->sock, 
		 give_connectedhost_name(i), give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      case T_CONSOLE:
	output("%3d   %3d  Console", i, io_array[i]->sock);
	break;
      case T_XCONSOLE:
	output("%3d   %3d  Xwindow on %s %d", i, io_array[i]->sock, give_connectedhost_name(i), give_connectedport(i));
	break;
      case T_OUTCONSOLE:
	switch(io_array[i]->flags) {
	case F_CONNECT:
	  output("%3d   %3d  Outside-console: Connection with %s %d", i, io_array[i]->sock, give_connectedhost_name(i),
		 give_connectedport(i));
	  break;
	case F_CLOSING:
	  output("%3d   %3d  Outside-console: Closing connection with %s %d", i, io_array[i]->sock, 
		 give_connectedhost_name(i), give_connectedport(i));
	  break;
	default:
	  output("Bug detected");
	  break;
	}
	break;
      default:
	output("Bug detected");
      }
    }
  }
}
