
/*******************************************************/
/**                                                   **/
/**        net.h (c) 1996, 97 Emiel Bruijntjes        **/
/**                                                   **/
/**   Internet connections and I/O via the internet   **/
/**                                                   **/
/*******************************************************/

#ifndef NET_H_
#define NET_H_

#include <sys/time.h>

#define MAX_CONNECTIONS 200       /* maximum number of sockets (including listen-ports, and stdin) */

#define FLOOD          4000       /* number of bytes we should at most accept per minute from user-sockets
				  ** T_CHAT, T_PARTY, T_ANNON, T_HTTP, T_OUTCONSOLE */

/* types: */
#define T_CHAT               1
#define T_SEND               2
#define T_GET                3
#define T_PARTY              4
#define T_SERVER             5
#define T_ANNON              6
#define T_BOT                7
#define T_HTTP               8
#define T_CONSOLE            9
#define T_XCONSOLE           10
#define T_OUTCONSOLE         11

void net_init();

int start_listen(char type, unsigned int port);   /* return -1 on failure */
int start_random_listen(char type);
int get_stdin();                                  /* Also check for input on stdin */
int add_X_socket(int);
void del_X_sockptr(int ptr);

char * my_gethostname();

char * give_localhost_name(int);
char * give_localhost_ip(int);
unsigned long int give_localip(int);
unsigned short int give_localport(int);
unsigned long int my_ip();

char * give_connectedhost_name(int);
char * give_connectedhost_ip(int);
unsigned long int give_connectedip(int);
unsigned short int give_connectedport(int);

int start_connecting(char type, int *ptr, char *hostname, unsigned short int port);
/* Returns 1 on succes, 0 when connection is not YET established and -1 on failure */
int start_connecting2(char type, int *ptr, unsigned long int addr, unsigned short int port);
/* Returns 1 on succes, 0 when connection is not YET established and -1 on failure */


void print_connections(void output(char *, ...));
/* Print the current settings in de io_array datastructure, using output-command "output" */

int wait_for_io(struct timeval *);

void changesocketptrtype(int socketptr, char type);
void close_socketptr(int);

void send_to_socketptr(int ptr, char *, ...);
void send_to_socketptr_raw(int ptr, char *, size_t size);
void send_to_socketptr_nonewline(int ptr, char *, ...);

#endif
