
/*******************************************************/
/**                                                   **/
/**       notes.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**                  Notes - database                 **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>


#include "notes.h"
#include "misc.h"

extern char   * notefile;
extern int      maxnotefilesize;
extern time_t   notetimeout;
extern time_t   time_null;

int addnote(char *to, char *from, char *msg) {

  FILE *fp;
  struct stat file_info;
  
  if ((stat(notefile, &file_info) && (errno != ENOENT))< 0) return 0;     /* stat() returns error */
  if (file_info.st_size >= maxnotefilesize) return 0;                     /* File is too large */
  if ((fp = fopen(notefile, "a")) == NULL) return 0;                      /* Couldn't open file */

  fwritestr2bin(to, fp);
  fwritestr2bin(from, fp);
  fwritestr2bin(msg, fp);
  fwritetime_t(fp, time_null);
  fwritechar(fp, 0);
  fclose(fp);
  return 1;
}

int user_has_note(char *nick) {

  FILE   *fp;
  char   *to=NULL, *from=NULL, *msg=NULL;
  char    note_already_read=0;
  time_t  note_created;

  if ((fp = fopen(notefile, "r")) == NULL) return 0;

  while (!(my_feof(fp))) {
    fgetstrfrombin(&to, fp);
    fgetstrfrombin(&from, fp);
    fgetstrfrombin(&msg, fp);
    freadtime_t(fp, &note_created);
    freadchar(fp, &note_already_read);
    if ((!(note_already_read)) && (strcasecmp(nick, to) == 0)) {
      xfree(to); xfree(from); xfree(msg);
      fclose(fp);
      return 1;
    }
  }
  xfree(to); xfree(from); xfree(msg);
  fclose(fp);
  return 0;
}

int return_note(char *nick, char **from, char **msg) {

  FILE   *fp;
  char   *to=NULL;
  char    note_already_read=0;
  time_t  note_created;

  if ((fp = fopen(notefile, "r+")) == NULL) return 0;

  while (!(my_feof(fp))) {
    fgetstrfrombin(&to, fp);
    fgetstrfrombin(from, fp);
    fgetstrfrombin(msg, fp);
    freadtime_t(fp, &note_created);
    freadchar(fp, &note_already_read);
    if ((!(note_already_read)) && (strcasecmp(nick, to) == 0)) {
      fseek(fp, -1, SEEK_CUR);
      fwritechar(fp, 1);
      xfree(to);
      fclose(fp);
      return 1;
    }
  }
  xfree(to);
  fclose(fp);
  return 0;
}

void updatenotefile() {

  FILE *fporg;
  FILE *fpnew;
  char *tmpfile;
  char *to=NULL, *from=NULL, *msg=NULL;
  time_t note_created=0;
  char note_already_read=0;

  tmpfile = (char *) xmalloc (strlen(notefile) + 10);
  strcpy(tmpfile, notefile);
  strcat(tmpfile, "&!@#$");

  if (link(notefile, tmpfile) == -1) {
    xfree(tmpfile);
    return;
  }
  if (unlink(notefile) == -1) {
    xfree(tmpfile);
    return;
  }
  if ((fporg = fopen(tmpfile, "r")) == NULL) {
    unlink(tmpfile);
    xfree(tmpfile);
    return;
  }
  if ((fpnew = fopen(notefile, "w")) == NULL) {
    fclose(fporg);
    unlink(tmpfile);
    xfree(tmpfile);
    return;
  }
  while (!(my_feof(fporg))) {
    fgetstrfrombin(&to, fporg);
    fgetstrfrombin(&from, fporg);
    fgetstrfrombin(&msg, fporg);
    freadtime_t(fporg, &note_created);
    freadchar(fporg, &note_already_read);
    if ((!(note_already_read)) && ((time_null - note_created) < notetimeout)) {
      fwritestr2bin(to, fpnew);
      fwritestr2bin(from, fpnew);
      fwritestr2bin(msg, fpnew);
      fwritetime_t(fpnew, note_created);
      fwritechar(fpnew, 0);
    }
  }
  fclose(fpnew);
  fclose(fporg);
  unlink(tmpfile);
  xfree(tmpfile);
  xfree(to); xfree(from); xfree(msg);
}
