
/*********************************************************/
/**                                                     **/
/**           on.c (c) 1997, Emiel Bruijntjes           **/
/**                                                     **/
/**           database with tcl-on procedures           **/
/**                                                     **/
/*********************************************************/

#include "config.h"

#ifdef NO_TCL

int add_on(int type, int nr, char *mask, char *proc) { return 0; }
int del_on(int type, int nr, char *mask) { return 0; }


int on_msg(char *nickuserhost, char *message) { return 0; }
int on_public(char *channel, char *nickuserhost, char *message) { return 0; }
int on_public_notice(char *channel, char *nickuserhost, char *message) { return 0; }
int on_notice(char *nickuserhost, char *message) { return 0; }
int on_join(char *nickuserhost, char *channel) { return 0; }
int on_part(char *nickuserhost, char *channel) { return 0; }
int on_nick(char *oldnick, char *newnick) { return 0; }
int on_kick(char *from, char *kicked, char *channel, char *comment) { return 0; }
int on_quit(char *nickuserhost, char *comment) { return 0; }
int on_mode(char *from, char *channel, char *mode) { return 0; }
int on_connect(char *from) { return 0; }
int on_console(char *msg) { return 0; }
int on_link(char *parent, char *bot) { return 0; }
int on_unlink(char *bot) { return 0; }
int on_ctcp(char *from, char *command, char *rest) { return 0; }
int on_ctcp_reply(char *from, char *command) { return 0; }
int on_botnet(char *bot, char *message) { return 0; }
int on_outside(char *nick, char *message) { return 0; }
int on_irc_input(char *nick, char *message) { return 0; }
int on_irc_cmd(char *nick, char *message) { return 0; }
int on_outside_cmd(char *nick, char *message) { return 0; }
int on_timer() { return 0; }

#else /* not NO_TCL */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <tcl.h>
#include <time.h>

#include "on.h"
#include "misc.h"
#include "reg.h"
#include "session.h"

struct proc_list {
  struct proc_list_node *first;
  struct proc_list_node *last;
  struct proc_list_node *current;
};

struct proc_list_node {
  struct proc_list_node *next;
  struct proc_list_node *prev;

  char *mask;
  char *proc;
};

struct on_list {
  struct on_list_node *first;
  struct on_list_node *last;
  struct on_list_node *current;
};

struct on_list_node {
  struct on_list_node *next;
  struct on_list_node *prev;

  int nr;
  struct proc_list proclist;
};

extern time_t      time_null;
extern Tcl_Interp *interp;

static struct on_list on_msg_list           = { NULL , NULL , NULL };
static struct on_list on_public_list        = { NULL , NULL , NULL };
static struct on_list on_public_notice_list = { NULL , NULL , NULL };
static struct on_list on_notice_list        = { NULL , NULL , NULL };
static struct on_list on_join_list          = { NULL , NULL , NULL };
static struct on_list on_part_list          = { NULL , NULL , NULL };
static struct on_list on_nick_list          = { NULL , NULL , NULL };
static struct on_list on_kick_list          = { NULL , NULL , NULL };
static struct on_list on_quit_list          = { NULL , NULL , NULL };
static struct on_list on_mode_list          = { NULL , NULL , NULL };
static struct on_list on_connect_list       = { NULL , NULL , NULL };
static struct on_list on_console_list       = { NULL , NULL , NULL };
static struct on_list on_link_list          = { NULL , NULL , NULL };
static struct on_list on_unlink_list        = { NULL , NULL , NULL };
static struct on_list on_ctcp_list          = { NULL , NULL , NULL };
static struct on_list on_ctcp_reply_list    = { NULL , NULL , NULL };
static struct on_list on_botnet_list        = { NULL , NULL , NULL };
static struct on_list on_outside_list       = { NULL , NULL , NULL };
static struct on_list on_irc_input_list     = { NULL , NULL , NULL };
static struct on_list on_irc_cmd_list       = { NULL , NULL , NULL };
static struct on_list on_outside_cmd_list   = { NULL , NULL , NULL };
static struct on_list on_timer_list         = { NULL , NULL , NULL };


static void create_on_entry(struct on_list *list, int nr) {

  struct on_list_node *tmp;

  for (list->current = list->first; list->current; list->current = list->current->next) {
    if (list->current->nr == nr) return;
    if (list->current->nr > nr) {
      tmp = (struct on_list_node *) xmalloc (sizeof(struct on_list_node));
      tmp->next = list->current;
      tmp->prev = list->current->prev;
      tmp->nr   = nr;
      tmp->proclist.first   = NULL;
      tmp->proclist.last    = NULL;
      tmp->proclist.current = NULL;

      if (list->current == list->first) {
	list->first = tmp;
      }
      else {
	list->current->prev->next = tmp;
      }
      list->current->prev = tmp;
      list->current = tmp;
      return;
    }
  }
  tmp = (struct on_list_node *) xmalloc (sizeof(struct on_list_node));
  tmp->next = NULL;
  tmp->prev = list->last;
  tmp->nr   = nr;
  tmp->proclist.first   = NULL;
  tmp->proclist.last    = NULL;
  tmp->proclist.current = NULL;

  if (list->first == NULL) {
    list->first   = tmp;
    list->last    = tmp;
    list->current = tmp;
  }
  else {
    list->last->next = tmp;
    list->last       = tmp;
    list->current    = tmp;
  }
}

static void delete_on_entry(struct on_list *list, int nr) {

  for (list->current = list->first; list->current; list->current = list->current->next) {
    if (list->current->nr == nr) {
      if (list->first == list->last) {
	xfree(list->current);
	list->first   = NULL;
	list->last    = NULL;
	list->current = NULL;
	return;
      }
      if (list->current->next == NULL) {
	list->last = list->current->prev;
	list->last->next = NULL;
	xfree(list->current);
	list->current = list->last;
	return;
      }
      if (list->current->prev == NULL) {
	list->first = list->current->next;
	list->first->prev = NULL;
	xfree(list->current);
	list->current = list->first;
	return;
      }
      list->current->prev->next = list->current->next;
      list->current->next->prev = list->current->prev;
      xfree(list->current);
      list->current = list->first;
    }
  }
}

static int on_list_findnr(struct on_list *list, int nr) {

  for (list->current = list->first; list->current; list->current = list->current->next) {
    if (list->current->nr == nr) return 1;
    if (list->current->nr > nr) return 0;
  }
  return 0;
}


static struct proc_list *return_proc_list(struct on_list *list) { return &list->current->proclist; }

static struct on_list *convert_type(int type) {

  switch (type) {
  case ON_MSG:
    return &on_msg_list;
  case ON_PUBLIC:
    return &on_public_list;
  case ON_PUBLIC_NOTICE:
    return &on_public_notice_list;
  case ON_NOTICE:
    return &on_notice_list;
  case ON_JOIN:
    return &on_join_list;
  case ON_PART:
    return &on_part_list;
  case ON_NICK:
    return &on_nick_list;
  case ON_KICK:
    return &on_kick_list;
  case ON_QUIT:
    return &on_quit_list;
  case ON_MODE:
    return &on_mode_list;
  case ON_CONNECT:
    return &on_connect_list;
  case ON_CONSOLE:
    return &on_console_list;
  case ON_LINK:
    return &on_link_list;
  case ON_UNLINK:
    return &on_unlink_list;
  case ON_CTCP:
    return &on_ctcp_list;
  case ON_CTCP_REPLY:
    return &on_ctcp_reply_list;
  case ON_BOTNET:
    return &on_botnet_list;
  case ON_OUTSIDE:
    return &on_outside_list;
  case ON_IRC_INPUT:
    return &on_irc_input_list;
  case ON_IRC_CMD:
    return &on_irc_cmd_list;
  case ON_OUTSIDE_CMD:
    return &on_outside_cmd_list;
  case ON_TIMER:
    return &on_timer_list;
  default:
    return NULL;
  }
}


int add_on(int type, int nr, char *mask, char *proc) {

  struct on_list *on_list;
  struct proc_list *list;

  on_list = convert_type(type);

  create_on_entry(on_list, nr);
  list = return_proc_list(on_list);

  for (list->current = list->first; list->current; list->current = list->current->next) {
    if (strcasecmp(list->current->mask, mask) == 0) {
      list->current->mask = (char *) xrealloc (list->current->mask, strlen(mask) + 1);
      strcpy(list->current->mask, mask);
      list->current->proc = (char *) xrealloc (list->current->proc, strlen(proc) + 1);
      strcpy(list->current->proc, proc);
      return 0;
    }
  }

  list->current = (struct proc_list_node *) xmalloc (sizeof(struct proc_list_node));
  if (list->first == NULL) {
    list->first = list->current;
  }
  else {
    list->last->next = list->current;
  }
  list->current->next = NULL;
  list->current->prev = list->last;
  list->last = list->current;
  
  list->current->mask = (char *) xmalloc (strlen(mask) + 1);
  strcpy(list->current->mask, mask);
  list->current->proc = (char *) xmalloc (strlen(proc) + 1);
  strcpy(list->current->proc, proc);
  return 1;
}  


int del_on(int type, int nr, char *mask) {

  struct on_list *on_list;
  struct proc_list *list;

  on_list = convert_type(type);

  if (!on_list_findnr(on_list, nr)) return 0;
  list = return_proc_list(on_list);

  for (list->current = list->first; list->current; list->current = list->current->next) {
    if (strcasecmp(list->current->mask, mask) == 0) {
      xfree(list->current->mask);
      xfree(list->current->proc);
      if (list->last == list->first) {
	xfree(list->current);
	list->current = NULL;
	list->last = NULL;
	list->first = NULL;
	break;
      }
      if (list->current->next == NULL) {
	list->last = list->current->prev;
	list->last->next = NULL;
	xfree(list->current);
	list->current = list->last;
	break;
      }
      if (list->current->prev == NULL) {
	list->first = list->current->next;
	list->first->prev = NULL;
	xfree(list->current);
	list->current = list->first;
	break;
      }
      list->current->prev->next = list->current->next;
      list->current->next->prev = list->current->prev;
      xfree(list->current);
      list->current = list->first;
      break;
    }
  }
  if (list->first == NULL) {
    delete_on_entry(on_list, nr);
  }
  return 1;
}


int check_on(struct on_list *list, char *args, char *string) {

  int   ret_value=0;

  if (list->first == NULL) return 0;

  for (list->current = list->first; list->current; list->current = list->current->next) {
    for (list->current->proclist.current = list->current->proclist.first; 
	 list->current->proclist.current; list->current->proclist.current = 
	   list->current->proclist.current->next) {
      if (match(list->current->proclist.current->mask, string)) {
	if (Tcl_VarEval(interp, list->current->proclist.current->proc, " ", args, NULL) != TCL_OK) {
	  int_err("error in tcl: %s", interp->result);
	}
	if (list->current->nr == 0) {
	  ret_value = 1;
	}
	break;
      }
    }
  }
  xfree(args);
  xfree(string);
  return ret_value;
}


int on_msg(char *nickuserhost, char *message) {

  char *argv[2];
  char *string;

  if (on_msg_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nickuserhost) + strlen(message) + 2);
  sprintf(string, "%s %s", nickuserhost, message);

  argv[0] = nickuserhost;
  argv[1] = message;

  return check_on(&on_msg_list, Tcl_Merge(2, argv), string);
}



int on_public(char *channel, char *nickuserhost, char *message) {

  char *argv[3];
  char *string;

  if (on_public_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(channel) + strlen(nickuserhost) + strlen(message) + 3);
  sprintf(string, "%s %s %s", channel, nickuserhost, message);

  argv[0] = channel;
  argv[1] = nickuserhost;
  argv[2] = message;

  return check_on(&on_public_list, Tcl_Merge(3, argv), string);
}


int on_public_notice(char *channel, char *nickuserhost, char *message) {

  char *argv[3];
  char *string;

  if (on_public_notice_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(channel) + strlen(nickuserhost) + strlen(message) + 3);
  sprintf(string, "%s %s %s", channel, nickuserhost, message);

  argv[0] = channel;
  argv[1] = nickuserhost;
  argv[2] = message;

  return check_on(&on_public_notice_list, Tcl_Merge(3, argv), string);
}


int on_notice(char *nickuserhost, char *message) {

  char *argv[2];
  char *string;
  
  if (on_notice_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nickuserhost) + strlen(message) + 2);
  sprintf(string, "%s %s", nickuserhost, message);

  argv[0] = nickuserhost;
  argv[1] = message;

  return check_on(&on_notice_list, Tcl_Merge(2, argv), string);
}


int on_join(char *nickuserhost, char *channel) {

  char *argv[2];
  char *string;

  if (on_join_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nickuserhost) + strlen(channel) + 2);
  sprintf(string, "%s %s", nickuserhost, channel);

  argv[0] = nickuserhost;
  argv[1] = channel;

  return check_on(&on_join_list, Tcl_Merge(2, argv), string);
}


int on_part(char *nickuserhost, char *channel) {

  char *argv[2];
  char *string;

  if (on_part_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nickuserhost) + strlen(channel) + 2);
  sprintf(string, "%s %s", nickuserhost, channel);

  argv[0] = nickuserhost;
  argv[1] = channel;

  return check_on(&on_part_list, Tcl_Merge(2, argv), string);
}


int on_nick(char *oldnick, char *newnick) {

  char *argv[2];
  char *string;

  if (on_nick_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(oldnick) + strlen(newnick) + 2);
  sprintf(string, "%s %s", oldnick, newnick);

  argv[0] = oldnick;
  argv[1] = newnick;

  return check_on(&on_nick_list, Tcl_Merge(2, argv), string);
}


int on_kick(char *from, char *kicked, char *channel, char *comment) {

  char *argv[4];
  char *string;

  if (on_kick_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(from) + strlen(kicked) + strlen(channel) + strlen(comment) + 4);
  sprintf(string, "%s %s %s %s", from, kicked, channel, comment);

  argv[0] = from;
  argv[1] = kicked;
  argv[2] = channel;
  argv[3] = comment;

  return check_on(&on_kick_list, Tcl_Merge(4, argv), string);
}


int on_quit(char *from, char *comment) {

  char *argv[2];
  char *string;

  if (on_quit_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(from) + strlen(comment) + 2);
  sprintf(string, "%s %s", from, comment);

  argv[0] = from;
  argv[1] = comment;

  return check_on(&on_quit_list, Tcl_Merge(2, argv), string);
}


int on_mode(char *from, char *channel, char *modes) {

  char *argv[3];
  char *string;

  if (on_mode_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(from) + strlen(channel) + strlen(modes) + 3);
  sprintf(string, "%s %s %s", from, channel, modes);

  argv[0] = from;
  argv[1] = channel;
  argv[2] = modes;

  return check_on(&on_mode_list, Tcl_Merge(3, argv), string);
}


int on_connect(char *from) {

  char *argv[1];
  char *string;

  if (on_connect_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(from) + 1);
  sprintf(string, "%s", from);

  argv[0] = from;

  return check_on(&on_connect_list, Tcl_Merge(1, argv), string);
}


int on_console(char *msg) {
  
  char *argv[1];
  char *string;

  if (on_console_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(msg) + 1);
  sprintf(string, "%s", msg);

  argv[0] = msg;

  return check_on(&on_console_list, Tcl_Merge(1, argv), string);
}


int on_link(char *parent, char *bot) {

  char *argv[2];
  char *string;

  if (on_link_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(parent) + strlen(bot) + 2);
  sprintf(string, "%s %s", parent, bot);

  argv[0] = parent;
  argv[1] = bot;

  return check_on(&on_link_list, Tcl_Merge(2, argv), string);
}


int on_unlink(char *bot) {

  char *argv[1];
  char *string;

  if (on_unlink_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(bot) + 1);
  sprintf(string, "%s", bot);

  argv[0] = bot;

  return check_on(&on_unlink_list, Tcl_Merge(1, argv), string);
}


int on_ctcp(char *from, char *command, char *rest) {

  char *argv[3];
  char *string;

  if (on_ctcp_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(from) + strlen(command) + strlen(rest) + 3);
  sprintf(string, "%s %s %s", from, command, rest);

  argv[0] = from;
  argv[1] = command;
  argv[2] = rest;

  return check_on(&on_ctcp_list, Tcl_Merge(3, argv), string);
}


int on_ctcp_reply(char *from, char *command) {

  char *argv[2];
  char *string;

  if (on_ctcp_reply_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(from) + strlen(command) + 2);
  sprintf(string, "%s %s", from, command);

  argv[0] = from;
  argv[1] = command;

  return check_on(&on_ctcp_reply_list, Tcl_Merge(2, argv), string);
}


int on_botnet(char *bot, char *message) {

  char *argv[2];
  char *string;

  if (on_botnet_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(bot) + strlen(message) + 2);
  sprintf(string, "%s %s", bot, message);

  argv[0] = bot;
  argv[1] = message;

  return check_on(&on_botnet_list, Tcl_Merge(2, argv), string);
}


int on_outside(char *nick, char *message) {

  char *argv[2];
  char *string;

  if (on_outside_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nick) + strlen(message) + 2);
  sprintf(string, "%s %s", nick, message);

  argv[0] = nick;
  argv[1] = message;

  return check_on(&on_outside_list, Tcl_Merge(2, argv), string);
}


int on_irc_input(char *nick, char *message) {

  char *argv[2];
  char *string;

  if (on_irc_input_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nick) + strlen(message) + 2);
  sprintf(string, "%s %s", nick, message);

  argv[0] = nick;
  argv[1] = message;

  return check_on(&on_irc_input_list, Tcl_Merge(2, argv), string);
}


int on_irc_cmd(char *nick, char *message) {

  char *argv[2];
  char *string;

  if (on_irc_cmd_list.first == NULL) return 0;
  
  string = (char *) xmalloc (strlen(nick) + strlen(message) + 2);
  sprintf(string, "%s %s", nick, message);

  argv[0] = nick;
  argv[1] = message;

  return check_on(&on_irc_cmd_list, Tcl_Merge(2, argv), string);
}


int on_outside_cmd(char *nick, char *message) {

  char *argv[2];
  char *string;

  if (on_outside_cmd_list.first == NULL) return 0;

  string = (char *) xmalloc (strlen(nick) + strlen(message) + 2);
  sprintf(string, "%s %s", nick, message);

  argv[0] = nick;
  argv[1] = message;

  return check_on(&on_outside_cmd_list, Tcl_Merge(2, argv), string);
}


int on_timer() {

  char *string;
  struct tm *tm;
  char *argv[1];

  if (on_timer_list.first == NULL) return 0;

  tm = localtime(&time_null);

  string = (char *) xmalloc (10);
  sprintf(string, "%2d:%02d", tm->tm_hour, tm->tm_min);

  argv[0] = string;

  return check_on(&on_timer_list, Tcl_Merge(1, argv), string);
}


#endif /* NO_TCL */
