
/*********************************************************/
/**                                                     **/
/**      out_console.c (c) 1997 Emiel Bruijntjes        **/
/**                                                     **/
/** Parsing and sending msg's from/to outside consoles  **/
/**     (JAVA consoles, MicroSoft Windows consoles      **/
/**    and other consoles that are not yet but might    **/
/**         become available in the future)             **/
/**                                                     **/
/*********************************************************/


#include <stdlib.h>
#include <string.h>


#include "misc.h"
#include "console.h"
#include "channel.h"
#include "party_console.h"
#include "irc_console.h"
#include "cmd_console.h"
#include "net.h"


struct {
  struct console_info *first;
  struct console_info *last;
  struct console_info *current;
} out_console_list = { NULL , NULL , NULL };

struct console_info {
  struct console_info *prev;
  struct console_info *next;
  
  int sockptr;
};

       int  nr_out_consoles = 0;

extern char botnick[10];
extern char current_nick[10];

void out_console_add(int sockptr) {

  char buffer[300], *channel;

  nr_out_consoles++;
  out_console_list.current = (struct console_info *) xmalloc (sizeof(struct console_info));
  if (out_console_list.first == NULL) {
    out_console_list.first = out_console_list.current;
    out_console_list.current->prev = NULL;
  }
  else {
    out_console_list.current->prev = out_console_list.last;
    out_console_list.last->next = out_console_list.current;
  }
  out_console_list.current->next = NULL;
  out_console_list.last = out_console_list.current;

  out_console_list.current->sockptr = sockptr;

  switch (console_mode()) {
  case CONSOLE_IRC_MODE:
    if (channellist_findfirst())
      channel = channellist_channelname();
    else
      channel = "";
    sprintf(buffer, "%s: IRC-mode   (channel: %c%s)", botnick, 
	    ischanneloperator(channel, current_nick) ? '@' : ' ', channel);
    break;
  case CONSOLE_PARTY_MODE:
    sprintf(buffer, "%s: Party-mode   (channel: %s)", botnick, current_party_channel());
    break;
  case CONSOLE_COMMAND_MODE:
    sprintf(buffer, "%s: Command-mode", botnick);
    break;
  }
  send_to_socketptr(out_console_list.current->sockptr, "mode %s", buffer);
}


void out_console_del() {

  if (out_console_list.current == NULL) return;

  nr_out_consoles--;
  if ((out_console_list.current == out_console_list.first) && (out_console_list.current == out_console_list.last)) {
    xfree(out_console_list.current);
    out_console_list.current = NULL;
    out_console_list.last = NULL;
    out_console_list.first = NULL;
    return;
  }

  if (out_console_list.current->next == NULL) {
    out_console_list.last = out_console_list.current->prev;
    out_console_list.last->next = NULL;
    xfree(out_console_list.current);
    out_console_list.current = out_console_list.last;
    return;
  }

  if (out_console_list.current->prev == NULL) {
    out_console_list.first = out_console_list.current->next;
    out_console_list.first->prev = NULL;
    xfree(out_console_list.current);
    out_console_list.current = out_console_list.first;
    return;
  }

  out_console_list.current->prev->next = out_console_list.current->next;
  out_console_list.current->next->prev = out_console_list.current->prev;
  xfree(out_console_list.current);
  out_console_list.current = out_console_list.first;
  return;
}


int out_console_findsockptr(int sockptr) {

  for (out_console_list.current = out_console_list.first; out_console_list.current; 
       out_console_list.current = out_console_list.current->next) {
    if (sockptr == out_console_list.current->sockptr) return 1;
  }
  return 0;
}


void send2out_consoles_win1(char *text) {

  struct console_info *tmp;

  tmp = out_console_list.current;
  for (out_console_list.current = out_console_list.first; out_console_list.current; 
       out_console_list.current = out_console_list.current->next) {
    send_to_socketptr(out_console_list.current->sockptr, "win1 %s", text);
  }
  out_console_list.current = tmp;
}


void send2out_consoles_win2(char *text) {

  struct console_info *tmp;

  tmp = out_console_list.current;
  for (out_console_list.current = out_console_list.first; out_console_list.current; 
       out_console_list.current = out_console_list.current->next) {
    send_to_socketptr(out_console_list.current->sockptr, "win2 %s", text);
  }
  out_console_list.current = tmp;
}


void sendmode2out_consoles() {

  char buffer[300], *channel;
  struct console_info *tmp;

  switch (console_mode()) {
  case CONSOLE_IRC_MODE:
    if (channellist_findfirst())
      channel = channellist_channelname();
    else
      channel = "";
    sprintf(buffer, "%s: IRC-mode   (channel: %c%s)", botnick, 
	    ischanneloperator(channel, current_nick) ? '@' : ' ', channel);
    break;
  case CONSOLE_PARTY_MODE:
    sprintf(buffer, "%s: Party-mode   (channel: %s)", botnick, current_party_channel());
    break;
  case CONSOLE_COMMAND_MODE:
    sprintf(buffer, "%s: Command-mode", botnick);
    break;
  }
  tmp = out_console_list.current;
  for (out_console_list.current = out_console_list.first; out_console_list.current; 
       out_console_list.current = out_console_list.current->next) {
    send_to_socketptr(out_console_list.current->sockptr, "mode %s", buffer);
  }
  out_console_list.current = tmp;
}


void parse_out_console(char *text) {

  if (strcasecmp(text, "mode") == 0) {
    switch (console_mode()) {
    case CONSOLE_IRC_MODE:
      console_setmode(CONSOLE_COMMAND_MODE);
      break;
    case CONSOLE_COMMAND_MODE:
      console_setmode(CONSOLE_PARTY_MODE);
      break;
    case CONSOLE_PARTY_MODE:
      console_setmode(CONSOLE_IRC_MODE);
      break;
    }
  }
  else if (strncasecmp(text, "input ", 6) == 0) {
    switch (console_mode()) {
    case CONSOLE_IRC_MODE:
      parse_irc_console(&text[6]);
      break;
    case CONSOLE_PARTY_MODE:
      parse_party_console(&text[6]);
      break;
    case CONSOLE_COMMAND_MODE:
      parse_cmd_console(&text[6]);
      break;
    }
  }
}

