
/*******************************************************/
/**                                                   **/
/**       parse.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                   **/
/**          Messages from the IRC server             **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <strings.h>
#include <ctype.h>


#include "reg.h"
#include "parse.h"
#include "session.h"
#include "channel.h"
#include "misc.h"
#include "userlist.h"
#include "server.h"
#include "chanuser.h"
#include "bots.h"
#include "console.h"
#include "bottree.h"
#include "config.h"
#include "ctcp.h"
#include "history.h"
#include "lines.h"
#include "on.h"

extern char  botnick[10];
extern char  current_nick[10];
extern char  commandchar1;
extern char *irc_logfile;
extern time_t time_null;

       char   my_nickuserhost[MAXNICKUSERHOSTLEN]="EemBot!erbruijn@keg.cs.vu.nl";
static char   correct_nick=1;     /* Boolean */
static time_t last_msg;


/* leave and join channel if we are alone and not operator, or when there are only bots
** on the channel and no channel operators.
*/
void cyclechannel(char *channel) {

  chanusers_t *chusers;

  if (channellist_findchan(channel) == 0) return;        /* I am not on the channel specified */

  if (ischanneloperator(channel, current_nick)) return;  /* I am channel operator */

  if (channellist_numberofusers() == 1) {                /* I am alone */
    send_fast_to_server("PART %s\nJOIN %s", channel, channel);
    return;
  }
  
  chusers = channellist_chanusers();
  if (chanusers_findfirst(chusers)) {
    do
      {
	if (chanusers_ischanop(chusers)) return;        /* There is an chanop on the channel */

	/* if it's not an protected bot, return */
	if (!(botlist_findbot(chanusers_nick(chusers)) && botlist_protection() &&
	      part_of_botnet(chanusers_nick(chusers)))) return;
      }
    while (chanusers_findnext(chusers));
  }
  send_fast_to_server("PART %s", channel);
  send_fast_to_server("JOIN %s", channel);              /* message will be sent one second after PART */
}

void server_doesnt_follow_protocol() {

  send_now_to_server("QUIT :Internal database does not match with server");
  console_send2ircwindow("*** Signoff: %s (Internal database does not match with server)", current_nick);
  nextserver();
}

static void dump_saved_modes(char *modes, char *argv[], int *argc) {

  /* This function is used by check_channel() */

  if (*argc == 3) {
    send_to_server("MODE %s %s %s %s %s", channellist_channelname(), modes, argv[0], argv[1], argv[2]);
    *argc = 0;
    strcpy(modes, "");
  }
}

static void really_dump_saved_modes(char *modes, char *argv[], int *argc) {

  /* This function is used by check_channel() */

  if (*argc >= 1) {
    send_to_server("MODE %s %s %s %s %s", channellist_channelname(), modes, argv[0], *argc > 1 ? argv[1] : "",
		   *argc > 2 ? argv[2] : "");
    *argc = 0;
    strcpy(modes, "");
  }
}

static void check_channel() {
  
  chanusers_t *chusers;
  char  mode_buffer[10]="";
  char *arg_buffer[3];
  char *kick[50];         /* 50 should be enough */
  char *comment[50];
  int   linenr_array[50];
  int   linenr;
  int   nr_modes=0, nr_kicks=0, i;

  chusers = channellist_chanusers();

  if ((!chanusers_findnick(chusers, current_nick)) || (!chanusers_ischanop(chusers))) return;

  if (chanusers_findfirst(chusers)) {
    do
      {
	if (!(session_findnick(chanusers_nick(chusers)))) continue;
	if (chanusers_ischanop(chusers)) {
	  if ((!session_isallowed(GETOPS_ALLOWED)) || session_ismatching('d') || session_ismatching('b')) {
	    strcat(mode_buffer, "-o");
	    arg_buffer[nr_modes] = session_nick();
	    nr_modes++;
	    dump_saved_modes(mode_buffer, arg_buffer, &nr_modes);
	  }
	} 
	else {  /* user is not channel-operator */
	  if ((!session_isallowed(GETOPS_ALLOWED)) || session_ismatching('d') || session_ismatching('b'))
	    continue;
	  if (session_isallowed(ALWAYSOP_ALLOWED)) {
	    strcat(mode_buffer, "+o");
	    arg_buffer[nr_modes] = session_nick();
	    nr_modes++;
	    dump_saved_modes(mode_buffer, arg_buffer, &nr_modes);
	  }
	}
      }
    while (chanusers_findnext(chusers));
    really_dump_saved_modes(mode_buffer, arg_buffer, &nr_modes);
    chanusers_findlast(chusers);
    do
      {
	if (nr_kicks > 45) break;   /* Without this there could be a segmentation fault once in a millenium */
	if (!(session_findnick(chanusers_nick(chusers)))) continue;
	if (session_ismatching('b') && (linenr = does_this_match('b', session_nickuserhost())) != 0) {
	  if (is_banned(session_nickuserhost()))
	    send_to_server("KICK %s %s :B%d: %s", channellist_channelname(), session_nick(), linenr, lines_comment('b'));
	  else {
	    strcat(mode_buffer, "+b");
	    arg_buffer[nr_modes] = lines_mask('b');
	    nr_modes++;
	    dump_saved_modes(mode_buffer, arg_buffer, &nr_modes);
	    kick[nr_kicks] = session_nick();
	    comment[nr_kicks] = lines_comment('b');
	    linenr_array[nr_kicks] = linenr;
	    nr_kicks++;
	  }
	}
      }
    while (chanusers_findprev(chusers));
    really_dump_saved_modes(mode_buffer, arg_buffer, &nr_modes);
    for (i=0; i<nr_kicks; i++) {
      send_to_server("KICK %s %s :B%d %s", channellist_channelname(), kick[i], linenr_array[i], comment[i]);
    }
  }
}
      
	

void check_channels() {

  if (channellist_findfirst()) {
    do
      {
	check_channel();
      }
    while (channellist_next());
  }
}


void join_channels() {

  if (channel_i_should_be_on()) {
    send_to_server("JOIN %s", channel_i_should_be_on());
  }
}


void check_nick() {

  if (!(correct_nick))
    send_to_server("NICK %s", botnick);
}

void parse_001(char *from, char *rest) {

  /* Welcome to the Internet Relay Network nick */

  char *nick;
  int i;

  nick = rindex(rest, '\0');
  nick--;
  while (isspace(*nick)) nick--;
  *(nick+1)='\0';
  while (*nick != ' ') nick--;
  nick++;

  strcpy(current_nick, nick);

  console_send2ircwindow("Welcome to the Internet Relay Network %s", current_nick);

  send_fast_to_server("MODE %s -sw+i", current_nick);

  for (i=1; channel2watch(i); i++) {
    send_fast_to_server("JOIN %s", channel2watch(i));
  }
}


void parse_error(char *from, char *rest) {

  char *chptr;

  /* I check first for the "@", altough I'm searching for the "(". It's possible to
  ** include a ( in the user part of the nick!user@host, so immediately searching for
  ** a ( can cause strange messages on the console if nickuserhost is nick!us(er@host.nl
  */
  if ((chptr = index(rest, '@')) == NULL) console_send2ircwindow("*** Signoff: %s (unknown)", current_nick);
  else {
    if ((chptr = index(chptr, '(')) == NULL) console_send2ircwindow("*** Signoff %s (unknown)", current_nick);
    else console_send2ircwindow("*** Signoff %s: %s", current_nick, chptr);
  }
  nextserver();
}

void parse_ping(char *from, char *rest) {

  send_now_to_server("PONG %s", rest);
}

void send_alternative_nick() {

  /* Send another nick to the server and store that nick in the variable "current_nick" */

  char *char_number;
  int   int_number;
  char  number_buffer[20];

  char_number = rindex(current_nick, '\0');
  while (isdigit(*(char_number-1))) char_number--;
	
  int_number = atoi(char_number);
  int_number++;

  sprintf(number_buffer, "%d", int_number);

  if (strlen(number_buffer) > strlen(char_number)) {
    if (strlen(current_nick) >= 9) {
      strcpy(current_nick, botnick);
      strcpy(char_number - 1, number_buffer);
    }
    else {
      strcpy(current_nick, botnick);
      strcpy(char_number, number_buffer);
    }
  }
  else {
    strcpy(current_nick, botnick);
    strcpy(char_number, number_buffer);
  }

  send_now_to_server("NICK :%s", current_nick);
  correct_nick = 0;
}


void parse_privmsg(char *from, char *rest) {    /* A privmsg sent to us or channel */

  char nick[10];       /* The nick who sent the message */
  char *to;            /* A channel, our nick or a server */
  char *message;       /* The message sent */
  char *tmp;
  char its_me=0;       /* Boolean to indicate that I'm sending messages to myself */

  if (!(is_user(from))) { /* notice from server (this could happen on undernet?) */
    console_send2ircwindow("*** [%s] %s", from, (tmp = index(rest, ':')) == NULL ? "" : tmp + 1);
    return;
  }

  nickonly(nick, from);

  if (strcasecmp(nick, current_nick) == 0)         /* It's a message from myself */
    its_me = 1;

  if (its_me == 0) {
    if (!session_findnick(nick)) { create_session(from); session_marknotify(); }
    if (session_ismatching('i')) return;
  }

  while (*rest == ' ') rest++;
  to = rest;
  if ((message = index(rest, ':')) == NULL) message = index(rest, '\0');
  else message++;
  if ((tmp = index(to, ' '))) *tmp = '\0';

  switch (to[0]) {
  case '#': case '&': case '+':                           /* Message sent to a channel */
    if (on_public(to, from, message)) return;
    if (!(userlist_findnick(nick))) createunode(nick);    /* Update the userlist */
    userlist_incpublics();

    if (session_is_flooding() && ischanneloperator(to, current_nick))   /* Check for flooding */
      send_now_to_server("KICK %s %s :Stop flooding", to, nick);

    if (message[0] == '') {                             /* Public CTCP */
      parse_ctcp(from, message, 1);                       /* The console will see this message from ctcp.c */
      return;
    }

    if (channellist_findfirst() && (strcasecmp(to, channellist_channelname()) == 0))    /* Current (=1st) channel */
      console_send2ircwindow("<%s> %s", nick, message);
    else
      console_send2ircwindow("<%s:%s> %s", nick, to, message);
    
    if (message[0] == commandchar1) {                      /* A command */
      session_parse_input(&message[1]);
      return;
    }
    return;
  case '$':                                                /* Message sent to IRC server */
    console_send2ircwindow("-%s:%s- %s", nick, to, message);
    return;
  default:                                                 /* Message sent to us */
    if (on_msg(from, message)) return;
    if (message[0] == '') {                              /* Private CTCP */
      if (!(its_me))
	parse_ctcp(from, message, 0);                      /* The console will see this message from ctcp.c */
      return;
    }
    console_send2ircwindow("*%s* %s", nick, message);
    add_msg_history(nick);
    if (its_me == 0) session_parse_input(message);
    return;
  }
}

void parse_notice(char *from, char *rest) {    /* A notice sent to us or channel */

  char nick[10];       /* The nick who sent the message */
  char *to;            /* A channel, our nick or a server */
  char *message;       /* The message sent */
  char *tmp;
  char  its_me=0;

  if (!is_user(from)) { /* notice from server (this happens on undernet) */
    console_send2ircwindow("*** [%s] %s", from, (tmp = index(rest, ':')) == NULL ? "" : tmp + 1);
    return;
  }

  nickonly(nick, from);

  if (strcasecmp(nick, current_nick) == 0)         /* It's a message from myself */
    its_me = 1;

  if (its_me == 0) {
    if (!session_findnick(nick)) { create_session(from); session_marknotify(); }
    if (session_ismatching('i')) return;
  }
  
  while (*rest == ' ') rest++;
  to = rest;
  if ((message = index(rest, ':')) == NULL) message = index(rest, '\0');
  else message++;
  if ((tmp = index(to, ' '))) *tmp = '\0';

  switch (to[0]) {
  case '#': case '&': case '+':                           /* Message sent to a channel */
    if (on_public_notice(to, from, message)) return;
    if (!(userlist_findnick(nick))) createunode(nick);    /* Update the userlist */
    userlist_incpublics();
    
    if (session_is_flooding() && ischanneloperator(to, current_nick))   /* Check for flooding */
      send_now_to_server("KICK %s %s :Stop flooding", to, nick);
    
    if (channellist_findfirst() && (strcasecmp(to, channellist_channelname()) == 0))    /* Current (=1st) channel */
      console_send2ircwindow("-%s- %s", nick, message);
    else
      console_send2ircwindow("-%s:%s- %s", nick, to, message);
    break;
  case '$':                                                /* Message sent to IRC server */
    console_send2ircwindow("-%s:%s- %s", nick, to, message);
    return;
  default:                                                 /* Message sent to us */
    if (on_notice(from, message)) return;
    if (message[0] == '') {                              /* Private CTCP reply */
      message++;
      if ((tmp = rindex(message, ''))) *tmp = '\0';
      parse_ctcp_reply(from, message);
      return;
    }
    add_msg_history(nick);
    console_send2ircwindow("-%s- %s", nick, message);
    return;
  }
}

void parse_302(char *from, char *rest) {  /*  rest = my_nick :nick=?user@host nick=?user@host and so on ... */

  /* USERHOST reply. Needed for notify */

  /* Every time session_startnotify() is called, a number of userhosts are requested,
  ** the answer is sent to the bot with this message, the bot calls session_marknotify,
  ** to indicate that the notify-message was received. When the bot has not received
  ** an /USERHOST answer for a couple (2) of minutes for a user who's not on one of
  ** the channels the bot is also on, the session will be deleted.
  */

  char *nickisuserhost[10];         /* Will never need more than 5 providing session.c will not be hacked */
  int   counter=0, i;
  char *nickuserhost, *nick, *userhost;

  if ((rest = index(rest, ':')) == NULL) return;
  rest++;

  while (*rest == ' ') rest++;
  if (*rest == '\0') return;

  nickisuserhost[0] = rest;
  counter = 1;

  while ((rest = index(rest, ' ')) != NULL) {
    *rest = '\0';
    rest++;
    while (*rest == ' ') rest++;
    if (*rest != '\0') {
      nickisuserhost[counter] = rest;
      counter++;
    }
  }

  for (i=0; i<counter; i++) {
    nick     = nickisuserhost[i];
    userhost = index(nickisuserhost[i], '=');
    *userhost = '\0'; userhost += 2;
    nickuserhost = (char *) xmalloc (strlen(nick) + strlen(userhost) + 2);
    sprintf(nickuserhost, "%s!%s", nick, userhost);
    if (session_findnick(nick)) {
      if (strcasecmp(session_nickuserhost(), nickuserhost) == 0)
	session_marknotify();
      else
	delete_session();
    }
    xfree(nickuserhost);
  }
}
      
void parse_433(char *from, char *rest) {

  /* Nickname in use */

  console_send2ircwindow("*** %s: Nickname is already in use", current_nick);

  while (isspace(*rest)) rest++;

  if (*rest == '*') {    /* We received the message when we were registering */
    send_alternative_nick();
  }
}

void parse_437(char *from, char *rest) {  /* * nick/chan :Nick/channel is temporarily unavailable */

  console_send2ircwindow("*** %s", rest);

  while (*rest == ' ') rest++;

  if (*rest == '*') {       /* We received the message when we were registering */
    rest++;
    while (*rest == ' ') rest++;
    if (strncasecmp(current_nick, rest, strlen(current_nick)) == 0)
      send_alternative_nick();
  }
}

void parse_join(char *from, char *rest) {

  char nick[10];
  char *channel;
  chanusers_t *chusers;
  int i=0, line_nr, script_takes_care=0;
  
  nickonly(nick, from);
  channel = &rest[1];

  while ((channel[i] != '') && (channel[i] != '\0')) {  
    /* Sometimes a bell appears in the channelname... weird */
    i++;
  }
  channel[i] = '\0';

  script_takes_care = on_join(from, channel);

  if (!script_takes_care) {
    console_send2ircwindow("*** %s (%s) has joined channel %s", nick, index(from, '!')+1, channel);
    
    /* logfile */
    putlog(irc_logfile, "%s: JOIN %s", channel, from);
  }    

  if (strcasecmp(nick, current_nick) == 0) {                      /* Bot joins a channel */
    strcpy(my_nickuserhost, from);                                /* Set my_nickuserhost to the correct value */
    addchannel2chanlist(channel);                                 /* Update the channel-list */
    /* I do not yet add myself to the chanusers, that will happen
    ** in the function parse_353(), msg 353 is sent to clients when
    ** they join a channel, together with the join-message
    */
    send_now_to_server("WHO %s\nMODE %s b", channel, channel);    /* Request information about this channel */
    console_print_status();                            /* Update the console */
    return;
  }

  if (session_findnick(nick)) {                               /* User already had an sessionlist-entry */
    if (strcasecmp(from, session_nickuserhost()) != 0) {           /* Not the right nick!user@host */
      delete_session();                                               /* Delete the other session */
      create_session(from);                                           /* And create a new one */
    }
  }
  else {
    create_session(from);                                             /* Create a session */
  }

  if (!(channellist_findchan(channel))) {                       /* We have just created this channel */
    int_err("Bug detected in parse_join. Please contact Eem - erbruijn@cs.vu.nl");
    server_doesnt_follow_protocol();
    return;
  }
  else {
    chusers = channellist_chanusers();
    chanusers_join_channel(chusers, nick);                                  /* Adds the user to the channel */
  }
  
  if (script_takes_care) return;

  if ((chanusers_findnick(chusers, current_nick)) && (chanusers_ischanop(chusers))) { /* Am I channel operator? */
    if (session_ismatching('b') && (line_nr = does_this_match('b', session_nickuserhost())) != 0) { 
      /* should the user be banned? */
      if (fullbanlist()) {                       /* Remove a ban when there are already 20 bans active on the channel */
	send_now_to_server("MODE %s -b+b %s %s\nKICK %s %s :B%d: %s", channel, returnrandomban(), 
			   lines_mask('b'), channel, nick, line_nr, lines_comment('b'));
      }
      else {
	send_now_to_server("MODE %s +b %s\nKICK %s %s :B%d: %s", channel, lines_mask('b'), channel, 
			   nick, line_nr, lines_comment('b'));
      }
    }
    else {
      if (session_isallowed(AUTOOP_ALLOWED) || session_isallowed(ALWAYSOP_ALLOWED)) {
	if (!session_ismatching('d'))
	  channellist_adduser2op(nick);
      }
    }
  }
  if (session_ismatching('i')) return;       /* User should be ignored. Do not update his userlist-entry */
  
  if (!(userlist_findnick(nick))) createunode(nick);
  userlist_incjoins();
}

/* initialize an empty channel after joining it.
** called from the list of users on the channel.
*/
void initchannel() {

  char modes[15]="";
  char arg[40]="";

  send_to_server("TOPIC %s :Hello and welcome on channel %s! Try /msg %s help.", channellist_channelname(), 
		 channellist_channelname(), current_nick);
  if (channellist_musthavemode('i')) strcat(modes, "+i");
  if (channellist_musthavemode('k')) {
    strcat(modes, "+k");
    sprintf(arg, "%s%s ", arg, channellist_channelkey());
  }
  if (channellist_musthavemode('l')) {
    strcat(modes, "+l");
    sprintf(arg, "%s%d ", arg, channellist_limit());
  }
  if (channellist_musthavemode('m')) strcat(modes, "+m");
  if (channellist_musthavemode('n')) strcat(modes, "+n");
  if (channellist_musthavemode('p')) strcat(modes, "+p");
  if (channellist_musthavemode('s')) strcat(modes, "+s");
  if (channellist_musthavemode('t')) strcat(modes, "+t");
  
  if (strlen(modes) > 0)
    send_fast_to_server("MODE %s %s %s", channellist_channelname(), modes, arg);
}


void parse_353(char *from, char *rest) {    /* The list of users on the channel. Sent to us when we joined a channel */

  /* rest:  nick = #channel :@nick1 *nick2 +nick3 nick4 ... */

  chanusers_t *chusers;
  char *channel, *nicklist, *nick;
  char oper = 0;

  if ((channel = index(rest, '=')) == NULL) {
    if ((channel = index(rest, '#')) == NULL) {
      if ((channel = index(rest, '&')) == NULL)
        return;   
    }
  }
  else {  
    channel++;
  }
  while (*channel == ' ') channel++;
  if ((nicklist = index(rest, ':')) == NULL) return;
  while (*nicklist == ':' || *nicklist == ' ') nicklist--;
  nicklist++; *nicklist = '\0';
  while (*nicklist != ':') nicklist++;

  /* channel has right value, nicklist points to ":nick1 nick2 nick3" */

  if (channellist_findchan(channel) == 0) return;     /* information about channel the bot is not active on */
  chusers = channellist_chanusers();

  nicklist++;        /* ignore ':' */
  while (*nicklist == ' ') nicklist++;

  console_send2ircwindow("*** Users on %s: %s", channel, nicklist);

  while (*nicklist != '\0') {

    nick = nicklist;

    if ((nicklist = index(nicklist, ' ')) == NULL) { 
      nicklist = nick;
      nicklist = index(nicklist, '\0'); 
    }
    else {
      *nicklist = '\0';
      nicklist++;
      while (*nicklist == ' ') nicklist++;
    }

    /* nick contains the nick we can parse now, nicklist points to the shortened nicklist */

    while (*nick == '*' || *nick == '+') nick++;
    if (*nick == '@') {
      oper = 1;
      nick++;
      while (*nick == '*' || *nick == '+') nick++;
    }
    else {
      oper = 0;
    }

    if (!(chanusers_findnick(chusers, nick))) { chanusers_join_channel(chusers, nick); }
    if (oper)
      chanusers_mark_chanop(chusers);
    else
      chanusers_mark_not_chanop(chusers);
  }
  if (channellist_numberofusers() == 1 && ischanneloperator(channel, current_nick)) {
    initchannel();
    console_print_status();
  }
}


void parse_352(char *from, char *rest) {        /* /WHO reply   The bot requests /who when he joins a channel */

  /*  rest = current_nick #channel user host server nick H@ :hubs ircname */

  char *channel;
  char *user;
  char *host;
  char *nick;
  char *tmp;
  char *nickuserhost;

  while (*rest == ' ') rest++;
  if ((channel = index(rest, ' ')) == NULL) return;
  while (isspace(*channel)) channel++;
  if ((user = index(channel, ' ')) == NULL) return;
  *user = '\0'; user++;
  while (isspace(*user)) user++;

  if (!(channellist_findchan(channel))) return;

  if ((host = index(user, ' ')) == NULL) return;
  *host = '\0'; host++;
  while (isspace(*host)) host++;
  
  if ((nick = index(host, ' ')) == NULL) return;
  *nick = '\0'; nick++;
  while (isspace(*nick)) nick++;
  if ((nick = index(nick, ' ')) == NULL) return;
  while (isspace(*nick)) nick++;
  if ((tmp = index(nick, ' ')) != NULL) *tmp = '\0';

  /* All variables are initialized now */

  if (!(is_on_a_channel(nick))) return;  /* user is not on a channel the bot is also active on, we don't need a session */
  if (strcasecmp(nick, current_nick) == 0) return;

  nickuserhost = (char *) xmalloc (strlen(nick) + strlen(user) + strlen(host) + 3);
  sprintf(nickuserhost, "%s!%s@%s", nick, user, host);

  if (session_findnick(nick)) {          /* We already have a session with this nick */
    if (strcasecmp(nickuserhost, session_nickuserhost()) != 0) {
      delete_session();
      create_session(nickuserhost);
    }
  }
  else {
    create_session(nickuserhost);
  }

  xfree(nickuserhost);
}



void botleaveschannel(char *channel) {        /* Function is called when bot leaves a channel (from parse_part()
					       * or parse_kick()
					       */

  char nick[10];
  chanusers_t *chusers;

  if (!(channellist_findchan(channel))) { server_doesnt_follow_protocol(); return; }
  chusers = channellist_chanusers();

  while (chanusers_findfirst(chusers)) {             /* Remove every nick from the chanusers */
    strcpy(nick, chanusers_nick(chusers));
    chanusers_part_channel(chusers);
    if (session_findnick(nick)) {          /* We have a session with this nick */
      if (!(is_on_a_channel(nick))) {       /* But we can't see the user on one of the channels we're on */
	if (important_session()) {             /* This user should be notified */
	  session_marknotify();
	}
	else {                                 /* User is not important */
	  delete_session();
	}
      }
    }
  }
  if (!(channellist_findchan(channel))) { server_doesnt_follow_protocol(); return; }
  delete_channel();
  console_print_status();       /* Update the console */
}

void parse_part(char *from, char *rest) {

  char nick[10];
  chanusers_t *chusers;
  short int i=0, script_takes_care;

  nickonly(nick, from);

  while ((rest[i] != ' ') && (rest[i] != '\0')) i++;
  rest[i] = '\0';   /* rest contains the channel-name */

  script_takes_care = on_part(from, rest);
  
  if (!script_takes_care) {
    /* logfile */
    putlog(irc_logfile, "%s: LEAVE %s", rest, from);
    
    console_send2ircwindow("*** %s has left channel %s", nick, rest);
  }

  if (!(channellist_findchan(rest))) {         /* Called to get current in channellist on the right place */
    int_err("1.Bug detected in parse_part. Please contact Eem - erbruijn@cs.vu.nl");
    server_doesnt_follow_protocol(); 
    return;
  }
  
  chusers = channellist_chanusers();
  if (!(chanusers_findnick(chusers, nick))) {         /* Called to get current of chusers right */
    int_err("2.Bug detected in parse_part. Please contact Eem - erbruijn@cs.vu.nl");
    server_doesnt_follow_protocol();
    return;
  }
  chanusers_part_channel(chusers);

  if (strcasecmp(nick, current_nick) == 0) {  /* The bot leaves a channel ! */
    botleaveschannel(rest);
    return;
  }

  if (session_findnick(nick)) {              /* An ordinary user has left the channel */
    if (!session_ismatching('i')) {
      if (!script_takes_care) {
	if (!(userlist_findnick(nick))) createunode(nick);   /* Update the userlist */
	userlist_incleaves();
	userlist_setlasthost(from);
      }
    }
    if (!(is_on_a_channel(nick))) {           /* We are not on the same channel anymore. */
      if (important_session()) {                 /* User should be notified */
	session_marknotify(from);                        
      }
      else {                                     /* User is not important */
	delete_session();
      }
    }
  }
  cyclechannel(rest);                            /* Cycle channel if needed */
}

void changenickuserhost(char *newnick) {

  char tmp[MAXNICKUSERHOSTLEN];

  strcpy(tmp, newnick);
  strcat(tmp, index(my_nickuserhost, '!'));
  strcpy(my_nickuserhost, tmp);
}

void parse_nick(char *from, char *rest) {

  char oldnick[10];
  char newnick[10];
  int i, script_takes_care;

  nickonly(oldnick, from);

  i = 1;
  while ((rest[i] != ' ') && (rest[i] != '\0')) {
    newnick[(i-1)] = rest[i];
    i++;
  }
  newnick[(i-1)] = '\0';
  
  script_takes_care = on_nick(oldnick, newnick);

  if (!script_takes_care) {
    putlog(irc_logfile, "NICK: %s -> %s", oldnick, newnick);
    
    console_send2ircwindow("*** %s is now known as %s", oldnick, newnick);
  }

  if ((session_findnick(newnick)) && (strcasecmp(newnick, oldnick) != 0)) {
    if (is_on_a_channel(newnick)) {         /* We already had a session with the new nick, delete that session */
      int_err("Bug detected in parse_nick. Please contact Eem - erbruijn@cs.vu.nl");
      server_doesnt_follow_protocol();
      return;
    }
    delete_session();
  }

  changenickonchannels(oldnick, newnick);
  if (session_findnick(oldnick)) 
    session_newnick(newnick);              /* Change the nick in the sessionlist */
  else {
    create_session(from);
    session_newnick(newnick);
  }

  if (strcasecmp(oldnick, current_nick) == 0) {     /* The bot changes his nick */
    strcpy(current_nick, newnick);
    console_print_status();               /* Update the console */
    changenickuserhost(newnick);
    if (strcasecmp(newnick, botnick) == 0) correct_nick=1;  /* The bot now has the right nick */
    return;
  }

  if (script_takes_care) return;

  if (session_ismatching('i')) return;   /* return if we have to ignore the user */

  if (!(userlist_findnick(oldnick))) createunode(oldnick);    /* Update userlist */
  userlist_incnicks();
  userlist_setlasthost(from);
  
  if (!(userlist_findnick(newnick))) createunode(newnick);
  userlist_incnicks();
}


void parse_kick(char *from, char *rest) {

  char kicknick[10];
  int i=0, script_takes_care;
  char *kickednick;
  char *channel;
  char *comment;
  chanusers_t *chusers;
  
  channel = rest;
  while (rest[i] != ' ') i++;
  rest[i] = '\0';                  /* channel has right value */

  i++;
  while	(rest[i] == ' ') i++;
  kickednick = &rest[i];

  if ((comment = index(&rest[i], ':')) == NULL) comment = index(&rest[i], '\0');
  else comment++;

  while (rest[i] != ' ') i++;
  rest[i] = '\0';                  /* kickednick has right value */

  if (is_user(from)) nickonly(kicknick, from);

  script_takes_care = on_kick(from, kickednick, channel, comment);

  if (!script_takes_care) {
    /* logfile */
    putlog(irc_logfile, "%s: KICK %s -> %s", channel, from, kickednick);
    
    console_send2ircwindow("*** %s has been kicked off channel %s by %s (%s)", kickednick, channel, kicknick, comment);
  }

  if (!(channellist_findchan(channel))) {
    int_err("1.Bug detected in parse_kick. Please contact Eem - erbruijn@cs.vu.nl");
    server_doesnt_follow_protocol();
    return;
  }
  chusers = channellist_chanusers();
  if (!(chanusers_findnick(chusers, kickednick))) {
    int_err("2.Bug detected in parse_kick. Please contact Eem - erbruijn@cs.vu.nl");
    server_doesnt_follow_protocol();
    return;
  }
  chanusers_part_channel(chusers);          /* channellist has been updated */

  if (strcasecmp(current_nick, kickednick) == 0) {     /* The bot has been kicked */
    if (!script_takes_care) {
      if (is_user(from) && session_findnick(kicknick) && !session_ismatching('i')) {
	if (!(userlist_findnick(kicknick))) createunode(kicknick);
	userlist_inckicks();
      }
    }
    botleaveschannel(channel);
    if (!script_takes_care) {
      send_to_server("JOIN %s", channel);
    }
    return;
  }

  if (strcasecmp(current_nick, kicknick) == 0) {
    if (!script_takes_care) {
      if (session_findnick(kickednick) && !session_ismatching('i')) {
	if (!(userlist_findnick(kickednick))) createunode(kickednick);
	userlist_inckickeds();
	userlist_setlasthost(session_nickuserhost());
      }
    }
    return;
  }

  if (!script_takes_care) {
    if (is_user(from)) {
      if (!session_findnick(kicknick)) create_session(from);
      if (botlist_findbot(kicknick) && botlist_protection() && part_of_botnet(kicknick)) {
	/* Do nothing, just tolerate */
      }
      else {
	if (session_better_protected(kickednick, kicknick)) {
	  if (no_revenge_flood()) {
	    if ((ischanneloperator(channel, current_nick)) && (revenge_should_be_hard()))
	      send_to_server("KICK %s %s :Don't mess with my friends!", channel, kicknick);
	    else
	      session_sendnotice("Don't mess with my friends!");
	  }
	}
      }
      if (!session_ismatching('i')) {
	if (!userlist_findnick(kicknick)) createunode(kicknick);
	userlist_inckicks();
      }
    }
  }
  
  if (session_findnick(kickednick)) {
    if (!script_takes_care) {
      if (!session_ismatching('i')) {
	if (!(userlist_findnick(kickednick))) createunode(kickednick);
	userlist_inckickeds();
	userlist_setlasthost(session_nickuserhost());
      }
    }
    if (!is_on_a_channel(kickednick)) {
      if (important_session())
	session_marknotify(session_nickuserhost());
      else
	delete_session();
    }
  }
  cyclechannel(channel);
}

void parse_quit(char *from, char *rest) {

  char nick[10];

  nickonly(nick, from);
  
  removenickfromchannels(nick);
  
  if (on_quit(nick, index(rest, ':')+1) == 0) {
    /* logfile */
    putlog(irc_logfile, "all: SIGNOFF %s", from);
    
    console_send2ircwindow("*** Signoff: %s (%s)", nick, index(rest, ':')+1);
    
    if (session_findnick(nick)) {
      if (!session_ismatching('i')) {
	if (!(userlist_findnick(nick))) createunode(nick);
	userlist_incquits();
	userlist_setlasthost(from);
      }
      delete_session();
    }
  }

  if (channellist_findfirst()) {
    do 
      {
	cyclechannel(channellist_channelname());
      } 
    while (channellist_next());
  }
}

int plusmode (char *from, char mode, char *arg, char *modeback, char *argback) {

  char nick[10];
  chanusers_t *chusers;

  switch (mode) {
  case 'b':
    add_ban_to_channel(arg);
    if (!is_user(from)) {         /* Server ban. Remove the ban */
      strcat(modeback, "-b");
      strcat(argback, arg);
      strcat(argback, " ");
      return 0;
    }
    nickonly(nick, from);
    if (strcasecmp(nick, current_nick) == 0)        /* Ban was set by the bot */
      return 0;
    if ((botlist_findbot(nick)) && (botlist_protection()) &&        /* Ban was set by other bot */
	(part_of_botnet(nick))) {
      return 0;
    }
    if (match(arg, my_nickuserhost)) {   /* Ban is matching with my nick!user@host */
      strcat(modeback, "-b");
      strcat(argback, arg);
      strcat(argback, " ");
      return 2;
    }
    if (botlist_findfirst()) {
      do 
	{
	  if (match(arg, botlist_nickuserhost()) &&
	      botlist_protection() && part_of_botnet(botlist_nick())) {    /* Ban is matching with other bot */
	    strcat(modeback, "-b");
	    strcat(argback, arg);
	    strcat(argback, " ");
	    return 0;
	  }
	}
      while (botlist_findnext());
    }
    if (session_findfirst()) {     /* Check the entire sessionlist to see if someone is matching with ban */
      do 
	{
	  if ((session_identified()) && (match(arg, session_nickuserhost()))) {
	    if (session_better_protected(session_nick(), nick)) {
	      strcat(modeback, "-b");
	      strcat(argback, arg);
	      strcat(argback, " ");
	      return 2;
	    }
	  }
	}
      while (session_next());
    }
    if ((session_findnick(nick)) && (session_isallowed(PLACEBANS_ALLOWED))) return 0;
    else {
      strcat(modeback, "-b");
      strcat(argback, arg);
      strcat(argback, " ");
      return 0;
    }
    
  case 'i':
    if (!(channellist_musthavemode('i'))) {
      strcat(modeback, "-i");
    }
    break;

  case 'k':
    if (!(channellist_musthavemode('k'))) {
      strcat(modeback, "-k");
      strcat(argback, arg);
      strcat(argback, " ");
    }
    else {
      if (strcasecmp(arg, channellist_channelkey()) != 0) {    /* Wrong channelkey */
	strcat(modeback, "-k");
	strcat(argback, arg);
	strcat(argback, " ");
      }
    }
    break;
 
  case 'l':
    if (!(channellist_musthavemode('l'))) {
      strcat(modeback, "-l");
    }
    else {
      if (atoi(arg) != channellist_limit()) {
	strcat(modeback, "-l");
      }
    }
    break;

  case 'm':
    if (!(channellist_musthavemode('m'))) {
      strcat(modeback, "-m");
    }
    break;

  case 'n':
    if (!(channellist_musthavemode('n'))) {
      strcat(modeback, "-n");
    }
    break;

  case 'o':
    chusers = channellist_chanusers();
    
    if (strcasecmp(arg, current_nick) == 0) {    /* The bot received operator status !! */
      if (!(chanusers_findnick(chusers, arg))) {
	int_err("Bug detected in plusmode. Please contact Eem - erbruijn@cs.vu.nl");
	server_doesnt_follow_protocol();
	return -99999;
      }
      if (chanusers_ischanop(chusers)) return 0;   /* We already had ops */
      else {
	send_fast_to_server("MODE %s", channellist_channelname());   /* Check the current channelmode(s) */
	chanusers_mark_chanop(chusers);
	check_channel();
	console_print_status();                     /* Update the console */
	return 0;
      }
    }
    if (!(chanusers_findnick(chusers, arg))) {
      int_err("Bug detected in plusmode. Please contact Eem - erbruijn@cs.vu.nl");
      server_doesnt_follow_protocol();
      return -999999;
    }	
    if (chanusers_ischanop(chusers)) return 0;     /* User already had ops */
    chanusers_mark_chanop(chusers);
    
    if (!(is_user(from))) {                      /* Server-op */
      if (!(session_findnick(arg))) {        /* Nick not found in sessionlist, remove chan-op status */
	strcat(modeback, "-o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 0;
      }
      else {                                     /* Nick found in sessionlist */
	if (botlist_findbot(arg) && botlist_protection() && part_of_botnet(arg)) {
	  /* nothing */
	}
	else {
	  if ((!session_isallowed(GETOPS_ALLOWED)) || session_ismatching('d')) {   /* Nick should not receive auto-ops */
	    strcat(modeback, "-o");
	    strcat(argback, arg);
	    strcat(argback, " ");
	  }
	}
	return 0;
      }
    }
    else {                                       /* User-op */
      nickonly(nick, from);
      if (strcasecmp(nick, current_nick) == 0) return 0;
      if (botlist_findbot(nick) && botlist_protection() && part_of_botnet(nick)) return 0;
      if (botlist_findbot(arg) && botlist_protection() && part_of_botnet(arg)) return 0;

      if ((!session_findnick(nick)) || (!session_isallowed(GIVEOPS_ALLOWED)) || 
	  (!session_findnick(arg)) || session_ismatching('d') || (!session_isallowed(GETOPS_ALLOWED))) {
	strcat(modeback, "-o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 0;
      }
    }
    break;
  
  case 'p': 
    if (!(channellist_musthavemode('p'))) {
      strcat(modeback, "-p");
    }
    break;

  case 's': 
    if (!(channellist_musthavemode('s'))) {
      strcat(modeback, "-s");
    }
    break;

  case 't': 
    if (!(channellist_musthavemode('t'))) {
      strcat(modeback, "-t");
    }
    break;

  case 'v':
    break;
    
  }
  return 0;
}

int minmode (char *from, char mode, char *arg, char *modeback, char *argback) {

  char nick[10];
  chanusers_t *chusers;
  int i=0;

  switch (mode) {
  case 'b':
    del_ban_from_channel(arg);
    nickonly(nick, from);
    if (strcasecmp(current_nick, nick) == 0) return 0;
    if ((botlist_findbot(nick)) && (botlist_protection()) && (part_of_botnet(nick))) return 0;
    if ((!(session_findnick(nick))) || (!(session_isallowed(UNBAN_ALLOWED)))) {
      strcat(modeback, "+b");
      strcat(argback, arg);
      strcat(argback, " ");
      return 0;
    }
    break;
  
  case 'i':
    if (channellist_musthavemode('i')) {
      strcat(modeback, "+i");
    }
    break;

  case 'k':
    if (channellist_musthavemode('k')) {
      strcat(modeback, "+k");
      strcat(argback, channellist_channelkey());
      strcat(argback, " ");
    }
    break;

  case 'l':
    if (channellist_musthavemode('l')) {
      strcat(modeback, "+l");
      while (argback[i] != '\0')
	i++;
      sprintf(&argback[i], "%d", channellist_limit());
      strcat(argback, " ");
    }
    break;

  case 'm':
    if (channellist_musthavemode('m')) {
      strcat(modeback, "+m");
    }
    break;

  case 'n':
    if (channellist_musthavemode('n')) {
      strcat(modeback, "+n");
    }
    break;  

  case 'o':	
    chusers = channellist_chanusers();
    if (!(chanusers_findnick(chusers, arg))) {
      int_err("Bug detected in minmode. Please contact Eem - erbruijn@cs.vu.nl");
      server_doesnt_follow_protocol();
      return -99999;
    }
    if (!chanusers_ischanop(chusers)) return 0;
    chanusers_mark_not_chanop(chusers);
    
    if (is_user(from)) {           /* Deop by user */
      nickonly(nick, from); 
      if (strcasecmp(nick, current_nick) == 0) {       /* The bot deops someone */
	return 0;
      }
      if (strcasecmp(arg, current_nick) == 0) {       /* The bot has been de-opped */
        console_print_status();             /* Update the console */
	return 0;
      }
      if ((botlist_findbot(nick)) && (botlist_protection()) && (part_of_botnet(nick))) {  /* An other bot deops someone */
	return 0;
      }
      if ((botlist_findbot(arg)) && (botlist_protection()) && (part_of_botnet(arg))) { /* An other bot has been deopped */
	strcat(modeback, "+o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 0;
      }
      /* User had to be deopped */
      if ((!session_findnick(arg)) || (session_ismatching('d')) || (!session_isallowed(GETOPS_ALLOWED))) return 0;  
      if (session_better_protected(arg, nick)) {                                             /* User was protected */
	strcat(modeback, "+o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 1;
      }
      if (session_isallowed(ALWAYSOP_ALLOWED)) {
	strcat(modeback, "+o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 0;
      }
      return 0;
    }
    else {                                            /* server deops */
      if (strcasecmp(arg, current_nick) == 0) {       /* The bot has been de-opped */
        console_print_status();            /* Update the console */
	return 0;
      }
      if ((botlist_findbot(arg)) && (botlist_protection()) && (part_of_botnet(arg))) {   /* A bot was deopped */
	strcat(modeback, "+o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 0;
      }
      if (!session_findnick(arg) || (session_ismatching('d')) || (!session_isallowed(GETOPS_ALLOWED))) return 0;
      if (session_identified() || session_isallowed(ALWAYSOP_ALLOWED)) {     
	/* Victim was protected */
	strcat(modeback, "+o");
	strcat(argback, arg);
	strcat(argback, " ");
	return 0;
      }
    }
    break;
    
  case 'p':
    if (channellist_musthavemode('p')) {
      strcat(modeback, "+p");
    }
    break;

  case 's':
    if (channellist_musthavemode('s')) {
      strcat(modeback, "+s");
    }
    break;  
    
  case 't':
    if (channellist_musthavemode('t')) {
      strcat(modeback, "+t");
    }
    break;

  case 'v':
    break;
  }
  return 0;
}

void parse_mode(char *from, char *rest) {

  char *channel, *modes, *args, *tmp;
  char *argvector[4]={ NULL, NULL, NULL, NULL};     /* I belief 4 is the maximum number of arguments */
  int   argcounter=0, revenge=0;
  char  nick[10];
  int (*function)(char *, char, char *, char *, char *)=plusmode;
  char  modeback[20]="";
  char  argback[400]="";

  while (*rest == '\0') rest++;

  if (is_user(from)) nickonly(nick, from);

  if (strcasecmp(from, current_nick) == 0) {         /* User-modes are changed. I don't care about those modes */
    if ((modes = index(rest, ':')) == NULL) return;
    modes--; *modes = '\0'; modes += 2;
    console_send2ircwindow("*** Mode change \"%s\" for user %s by %s", modes, rest, from);
    return;
  }

  channel = rest;
  if ((modes = index(channel, ' ')) == NULL) {
    server_doesnt_follow_protocol();
    return;
  }                                               /* modes = "+ntkl hello 365" */
  *modes = '\0'; modes++;

  tmp = index(modes, '\0');
  tmp--;
  if (*tmp == ' ') *tmp = '\0';
  
  if (on_mode(modes, channel, from)) return;
  console_send2ircwindow("*** Mode change \"%s\" on channel %s by %s", modes, channel, is_user(from) ? nick : from);
  if (*tmp == '\0') *tmp = ' ';

  if ((args = index(modes, ' ')) != NULL) {
    *args = '\0'; args++;
    argvector[0] = args;
    for (;*args != '\0';args++) {
      if (*args == ' ') {
	argcounter++;
	*args = '\0';
	args++;
	argvector[argcounter]=args;
      }
    }                                           /* modes = "+ntkl", argvector[0] = "hello", argvector[1] = "365" */
  }

  if (is_user(from)) {
    if (strcasecmp(nick, current_nick) != 0) {
      if (!session_findnick(nick)) create_session(from);
      if (!session_ismatching('i')) {
	if (!(userlist_findnick(nick))) createunode(nick);
	userlist_incmodes();
      }
      if (strcasecmp(modes, "-ooo") == 0)
	send_now_to_server("MODE %s -o %s", channel, nick);
    }
  }

  if (!(channellist_findchan(channel))) {
    server_doesnt_follow_protocol();
    return;
  }

  argcounter=0;
  while (*modes != '\0') {
    if (revenge < -10) return;
    switch (*modes) {
    case '+': function = plusmode; break;
    case '-': function = minmode; break;
    case 'b': revenge += function(from, *modes, argvector[argcounter], modeback, argback); argcounter++; break;
    case 'i': revenge += function(from, *modes, NULL, modeback, argback); break;
    case 'k': revenge += function(from, *modes, argvector[argcounter], modeback, argback); argcounter++; break;
    case 'l': 
      if (function == plusmode) {
	revenge += function(from, *modes, argvector[argcounter], modeback, argback); argcounter++; break;
      }
      else {
	revenge += function(from, *modes, NULL, modeback, argback); break;
      }
    case 'm': revenge += function(from, *modes, NULL, modeback, argback); break;
    case 'n': revenge += function(from, *modes, NULL, modeback, argback); break;
    case 'o': revenge += function(from, *modes, argvector[argcounter], modeback, argback); argcounter++; break;
    case 'p': revenge += function(from, *modes, NULL, modeback, argback); break;
    case 's': revenge += function(from, *modes, NULL, modeback, argback); break;
    case 't': revenge += function(from, *modes, NULL, modeback, argback); break;
    case 'v': revenge += function(from, *modes, argvector[argcounter], modeback, argback); break;
    }
    modes++;
  }
  if ((ischanneloperator(channel, current_nick)) && (strlen(modeback) > 0)) {
    send_now_to_server("MODE %s %s %s", channel, modeback, argback);
  }
  if (!is_user(from)) return;
  if (strcasecmp(current_nick, nick) == 0) return;

  if ((session_findnick(nick)) && (no_revenge_flood()) && (revenge > 0) && (is_user(from))) {
    if ((revenge_should_be_hard()) && (ischanneloperator(channel, current_nick))) {
      send_fast_to_server("KICK %s %s :Don't mess with my friends!", channel, nick);
    }
    else {
      if (!(session_findnick(nick))) { create_session(from); }
      session_sendnotice("Don't mess with my friends!");
    }
  }
}

void parse_324(char *from, char *rest) {          /* MODE Reply */

  char *channel;
  int i=0;
  char modes[20]="";
  char args[50]="";
  int j=0;
  char tmp[40];
  int k=0;

  while ((rest[i] != '#') && (rest[i] != '&')) i++;

  channel = &rest[i];
  while (rest[i] != ' ') i++;
  rest[i]='\0';                    /* Channel has right value now */

  if (!(channellist_findchan(channel))) return;
  if (!(chanusers_findnick(channellist_chanusers(), current_nick))) return;
  if (!(chanusers_ischanop(channellist_chanusers()))) return;

  i=i+2; /* i should point to first mode-character now */
  j=i;

  while ((rest[j] != ' ') && (rest[i] != '\0')) j++;
  j++;     /* j should point to first argument */

  if (rest[i] == 'p') {
    i++;
    if (!(channellist_musthavemode('p')))
      strcat(modes, "-p");
  }
  else {
    if (channellist_musthavemode('p'))
      strcat(modes, "+p");
  }

  if (rest[i] == 's') {
    i++;
    if (!(channellist_musthavemode('s')))
      strcat(modes, "-s");
  }
  else {
    if (channellist_musthavemode('s'))
      strcat(modes, "+s");
  }

  if (rest[i] == 'm') {
    i++;
    if (!(channellist_musthavemode('m')))
      strcat(modes, "-m");
  }
  else {
    if (channellist_musthavemode('m'))
      strcat(modes, "+m");
  }

  if (rest[i] == 't') {
    i++;
    if (!(channellist_musthavemode('t')))
      strcat(modes, "-t");
  }
  else {
    if (channellist_musthavemode('t'))
      strcat(modes, "+t");
  }

  if (rest[i] == 'i') {
    i++;
    if (!(channellist_musthavemode('i')))
      strcat(modes, "-i");
  }
  else {
    if (channellist_musthavemode('i'))
      strcat(modes, "+i");
  }

  if (rest[i] == 'n') {
    i++;
    if (!(channellist_musthavemode('n')))
      strcat(modes, "-n");
  }
  else {
    if (channellist_musthavemode('n'))
      strcat(modes, "+n");
  }

  if (rest[i] == 'l') {
    i++;
    if (!(channellist_musthavemode('l')))
      strcat(modes, "-l");
    else {
      k=0;
      while ((rest[j] != ' ') && (rest[j] != '\0')) {
	tmp[k] = rest[j];
	j++; k++;
      }
      tmp[k] = '\0';
      j++;
      if (atoi(tmp) != channellist_limit()) {
	strcat(modes, "-l");
      }
    }
  }
  else {
    if (channellist_musthavemode('l')) {
      strcat(modes, "+l");
      sprintf(args, "%d", channellist_limit());
      strcat(args, " ");
    }
  }

  if (rest[i] == 'k') {
    i++;
    k=0;
    while ((rest[j] != ' ') && (rest[j] != '\n') && (rest[j] != '
') && (rest[j] != '\0')) {
      tmp[k] = rest[j];
      j++; k++;
    }
    tmp[k] = '\0';
    if (!(channellist_musthavemode('k'))) {
      strcat(modes, "-k");
      strcat(args, tmp);
    }
    else {
      if (strcasecmp(tmp, channellist_channelkey()) != 0) {
	strcat(modes, "-k");
	strcat(args, tmp);
      }
    }
  }
  else {
    if (channellist_musthavemode('k')) {
      strcat(modes, "+k");
      strcat(args, channellist_channelkey());
    }
  }

  if (strlen(modes) > 1) {
    send_fast_to_server("MODE %s %s %s", channel, modes, args);
  }
}

void parse_367(char *from, char *rest) {     /* Existing bans on channel */

  char *channel;
  int i=0;
  char *ban;

  while ((rest[i] != '#') && (rest[i] != '&')) i++;
  channel = &rest[i];

  while(rest[i] != ' ') i++;
  rest[i] = '\0';
  i++;
  ban = &rest[i];
  while ((rest[i] != ' ') && (rest[i] != '\0')) i++;
  rest[i] = '\0';

  if (channellist_findchan(channel)) add_ban_to_channel(ban);	
}

void parse_332(char *from, char *rest) {      /* Current topic */

  char *chptr;

  chptr = index(rest, ':');
  console_send2ircwindow("*** Topic: %s", chptr == NULL ? "" : chptr+1);
}

void parse_topic(char *from, char *rest) {    /* Function only called when bot has console */

  char *chptr, *channel, *topic;
  char nick[10];

  nickonly(nick, from);

  while (*rest == ' ') rest++;
  if ((topic = index(rest, ':')) == NULL)
    topic = index(rest, '\0');
  else
    topic++;

  channel = rest;
  if ((chptr = index(channel, ' ')) != NULL)
    *chptr = '\0';

  console_send2ircwindow("*** %s has changed the topic on channel %s to %s", nick, channel, topic);
}

void parse_341(char *from, char *rest) {    /* INVITE reply:   from 341 current_nick invited_nick channel */

  char *nick, *channel;

  while (*rest == ' ') rest++;

  if ((nick = index(rest, ' ')) == NULL) return;
  nick++;
  if ((channel = index(nick, ' ')) == NULL) return;
  *channel = '\0'; channel++;

  console_send2ircwindow("*** Inviting %s to channel %s", nick, channel);
}

void parse_invite(char *from, char *rest) {     /*   from INVITE current_nick :channel  */

  char nick[10];
  char *channel;

  if ((channel = index(rest, ':')) == NULL) return;
  channel++;

  nickonly(nick, from);
  console_send2ircwindow("*** %s (%s) invites you to channel %s", nick, index(from, '!')+1, channel);

  if (is_chan2watch(channel)) {
    send_to_server("JOIN %s", channel);
  }
}

void parse_311(char *from, char *rest) {    /* /WHOIS 1st line reply.  rest = nick user host * :ircname */

  char *nick, *user, *host, *ircname, *tmp;

  while (*rest == ' ') rest++;

  if ((nick = index(rest, ' ')) == NULL) return;
  nick++;
  if ((user = index(nick, ' ')) == NULL) return;
  *user = '\0'; user++;
  if ((host = index(user, ' ')) == NULL) return;
  *host = '\0'; host++;
  if ((tmp = index(host, ' ')) == NULL) return;
  *tmp = '\0'; tmp++;
  if ((ircname = index(tmp, ':')) == NULL) return;
  ircname++;

  console_send2ircwindow("*** %s is %s@%s (%s)", nick, user, host, ircname);
}

void parse_319(char *from, char *rest) {  /* /WHOIS 2nd line. rest = nick :channels */

  char *channels;

  if ((channels = index(rest, ':')) == NULL) return;
  channels++;

  console_send2ircwindow("*** on channels: %s", channels);
}

void parse_312(char *from, char *rest) {    /* /WHOIS 3rd line. rest = nick server :server-description *
					     * also /WHOWAS-line */
  char *server;
  char *description;

  while (*rest == ' ') rest++;
  if ((server = index(rest, ' ')) == NULL) return;    /* points to nick now */
  server++;
  if ((server = index(server, ' ')) == NULL) return;
  server++;
  if ((description = index(rest, ':')) == NULL) return;
  description--; *description = '\0'; description += 2;

  console_send2ircwindow("*** on IRC via server %s (%s)", server, description);
}

void parse_317(char *from, char *rest) {     /* /WHOIS 4th line. rest = nick seconds :seconds idle */

  int seconds;
  char *chptr, *nick;

  while (*rest == ' ') rest++;
  if ((rest = index(rest, ' ')) == NULL) return;
  rest++;
  nick = rest;
  if ((chptr = index(rest, ' ')) == NULL) return;
  *chptr = '\0'; chptr++;
  seconds = atoi(chptr);
  if (seconds > 180)
    console_send2ircwindow("*** %s has been idle %d minutes", nick, seconds / 60);
  else
    console_send2ircwindow("*** %s has been idle %d seconds", nick, seconds);
}

void parse_314(char *from, char *rest) {    /* /WHOWAS 1st line reply.  rest = nick user host * :ircname */

  char *nick, *user, *host, *ircname, *tmp;

  while (*rest == ' ') rest++;

  if ((nick = index(rest, ' ')) == NULL) return;
  nick++;
  if ((user = index(nick, ' ')) == NULL) return;
  *user = '\0'; user++;
  if ((host = index(user, ' ')) == NULL) return;
  *host = '\0'; host++;
  if ((tmp = index(host, ' ')) == NULL) return;
  *tmp = '\0'; tmp++;
  if ((ircname = index(tmp, ':')) == NULL) return;
  ircname++;

  console_send2ircwindow("*** %s was %s@%s (%s)", nick, user, host, ircname);
}

void parse_322(char *from, char *rest) {    /* /LIST reply:  rest = current_nick channel nr_users :topic */

  char *channel, *topic, *nr_users;

  while (*rest == ' ') rest++;

  if ((channel = index(rest, ' ')) == NULL) return;
  channel++;
  if ((nr_users = index(channel, ' ')) == NULL) return;
  *nr_users = '\0'; nr_users++;
  if ((topic = index(nr_users, ':')) == NULL) return;
  topic--; *topic = '\0'; topic +=2;

  console_send2ircwindow("*** %-20s %3d %s", channel, atoi(nr_users), topic);
}

void parse_254(char *from, char *rest) {   /* /LUSERS reply:  rest = current_nick nr_channels :channels formed */

  while (*rest == ' ') rest++;
  if ((rest = index(rest, ' ')) == NULL) return;
  rest++;

  console_send2ircwindow("*** %d channels have been formed", atoi(rest));
}

void parse_255(char *from, char *rest) {  /* /LUSERS reply: rest = current_nick :I have NR client, NR services and NR 
					   * servers */

  if ((rest = index(rest, ':')) == NULL) return;
  if ((rest = index(rest, ' ')) == NULL) return;
  rest++;
  if ((rest = index(rest, ' ')) == NULL) return;
  rest++;
  console_send2ircwindow("*** This server has %s connected", rest);
}

void parse_364(char *from, char *rest) {  /* /LINKS reply: rest = current_nick serv1 serv2 :hubs server-info */

  char *serv1, *serv2, *hubs, *serv_info;

  while (*rest == ' ') rest++;
  if ((serv1 = index(rest, ' ')) == NULL) return;
  serv1++;
  if ((serv2 = index(serv1, ' ')) == NULL) return;
  *serv2 = '\0'; serv2++;
  if ((hubs = index(serv2, ':')) == NULL) return;
  hubs--; *hubs = '\0'; hubs += 2;
  if ((serv_info = index(hubs, ' ')) == NULL) return;
  *serv_info = '\0'; serv_info++;

  console_send2ircwindow("*** %-22s %-22s %2s %s", serv1, serv2, hubs, serv_info);
}

void parse_351(char *from, char *rest) {  /*  /VERSION reply: rest = current_nick version_nr servnanme :setting */

  char *servname, *setting, *version_nr;

  while (*rest == ' ') rest++;
  if ((version_nr = index(rest, ' ')) == NULL) return;
  version_nr++;
  if ((servname = index(version_nr, ' ')) == NULL) return;
  *servname = '\0'; servname++;
  if ((setting = index(servname, ':')) == NULL) return;
  setting--; *setting = '\0'; setting += 2;

  console_send2ircwindow("*** Server: %s %s %s", servname, version_nr, setting);
}

void parse_normal_function(char *from, char *rest) {   /* only called when we have a console */

  char *channel, *message;    /* message is part after ':' */

  while (*rest == ' ') rest++;
  message = index(rest, ':');
  channel = index(rest, ' ');
  if (channel) channel++;
  if (channel >= message) channel = NULL;

  if (message && channel) {
    message--;
    *message = '\0'; 
    message++;
  }

  if (message) message++;

  if (channel) {
    if (message)
      console_send2ircwindow("*** %s: %s", channel, message);
    else
      console_send2ircwindow("*** %s", channel);
  }
  else {
    if (message)
      console_send2ircwindow("*** %s", message);
    else
      console_send2ircwindow("*** %s", rest);
  }
}

struct {
  char    *name;
  void    (*function)(char *from, char *rest);
  char    console_flag;
} parse_funcs[] =
  {	
    { "PRIVMSG",    parse_privmsg },
    { "JOIN",       parse_join    },
    { "PART",       parse_part    },
    { "MODE",       parse_mode    },
    { "NICK",       parse_nick    },
    { "KICK",       parse_kick    },
    { "PING",       parse_ping    },
    { "QUIT",       parse_quit    },
    { "ERROR",      parse_error   },
    { "KILL",       NULL          },
    { "NOTICE",     parse_notice  },
    { "TOPIC",      parse_topic   },
    { "INVITE",     parse_invite  },
    { "001",        parse_001     },  /* First message from server */
    { "219",        NULL          },  /* End of /STATS report */
    { "254",        parse_254     },  /* /LUSERS reply, nr of channels formed */
    { "255",        parse_255     },  /* /LUSERS reply, nr of connections */
    { "302",        parse_302     },  /* /USERHOST reply */
    { "311",        parse_311     },  /* /WHOIS reply: User is user@host (ircname) */
    { "312",        parse_312     },  /* /WHOIS reply: On IRC via server server (description) */
    { "314",        parse_314     },  /* /WHOWAS reply: User was user@host (ircname) */
    { "315",        NULL          },  /* End of /WHO list. */
    { "317",        parse_317     },  /* /WHOIS reply: seconds idle */
    { "318",        NULL          },  /* End of /WHOIS list. */
    { "319",        parse_319     },  /* /WHOIS reply: On channels: channels */
    { "322",        parse_322     },  /* /LIST reply */
    { "323",        NULL          },  /* End of /LIST */
    { "324",        parse_324     },  /* Current channel mode */
    { "332",        parse_332     },  /* Current topic */
    { "341",        parse_341     },  /* Inviting NICK to channel CHANNEL */
    { "351",        parse_351     },  /* /VERSION reply from server */
    { "352",        parse_352     },  /* /WHO reply */
    { "353",        parse_353     },  /* /NAMES reply */
    { "364",        parse_364     },  /* /LINKS answer */
    { "365",        NULL          },  /* End of /LINKS list. */
    { "366",        NULL          },  /* End of /NAMES list. */
    { "367",        parse_367     },  /* /MODE b reply */
    { "368",        NULL          },  /* End of Channel Ban List */
    { "369",        NULL          },  /* End of WHOWAS */
    { "376",        NULL          },  /* End of /MOTD command. */
    { "433",        parse_433     },  /* Nickname in use */
    { "437",        parse_437     },  /* Nick/channel is temporarily unavailable */
    { NULL,         NULL          }
  };

void parse_server_msg(char *line) {

  char    *from; 
  char    *command;
  char    *rest;
  int     i;

  if (line[0]==':') {
    i = 1;
    from = &line[1];
    while (line[i] != ' ') i++;
    line[i] = '\0';
    i++;
    command = &line[i];
    while (line[i] != ' ') i++;
    line[i] = '\0';
    i++;
    rest = &line[i];
  }
  else {
    i = 0;
    command = &line[i];
    while (line[i] != ' ') i++;	
    command[i]='\0';
    i++;
    rest = &line[i];
    from = &line[0];
  }

  for(i=0; parse_funcs[i].name; i++) {
    if ( strcasecmp(parse_funcs[i].name, command) == 0 ) {
      last_msg = time_null;
      if (parse_funcs[i].function) parse_funcs[i].function(from, rest);
      return;
    }
  }
  parse_normal_function(from, rest);
}

void check_connection() {

  /* If I haven't heard anything from the server for 10 minutes, it's time to connect to the next server
   * in the server-list. */

  if (last_msg + 600 < time_null) nextserver();
}


