
/*******************************************************/
/**                                                   **/
/**      parsebot.c (c) 1996, 97 Emiel Bruijntjes     **/
/**                                                   **/
/**         Parsing messages from other bots          **/
/**                                                   **/
/*******************************************************/

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>


#include "parsebot.h"
#include "bots.h"
#include "session.h"
#include "bottree.h"
#include "partyline.h"
#include "misc.h"
#include "notes.h"
#include "assoc.h"
#include "config.h"
#include "console.h"
#include "on.h"

extern char botnick[10];

void parsebot_please(char *argv[], int argc) {   /* Please enter your nick */

  botlist_send2current("%s", botnick);
}

void parsebot_you(char *argv[], int argc) {      /* You don't have access */

  botlist_breakconnection();
  repairstring(argv, argc);
  sendtospies("bot %s: %s", botlist_nick(), argv[0]);
  console_send2snoopwindow("[bot] %s: %s", botlist_nick(), argv[0]);
}

void parsebot_passreq(char *argv[], int argc) {  /* Enter a password */

  botlist_send2current("%s", botlist_password());
}

void parsebot_badpass(char *argv[], int argc) {

  botlist_breakconnection();
  sendtospies("I do not know the right password for %s", botlist_nick());
  sendtoadminspies("Password: %s", botlist_password());
  console_send2snoopwindow("[bot] %s: Password (%s) incorrect", botlist_nick(), botlist_password());
}

void parsebot_hello(char *argv[], int argc) {

  botlist_linkestablished();
}

void parsebot_thisbot(char *argv[], int argc) {

  if ((argc != 2) || (strcasecmp(argv[1], botlist_nick()) != 0)) {
    botlist_send2all("chat %s Disconnecting %s (imposter)", botnick, botlist_nick());
    party_send2party("*** Disconnecting %s (imposter)", botlist_nick());
    session_send2party("*** Disconnecting %s (imposter)", botlist_nick());
    console_send2snoopwindow("(all) *** Disconnecting %s (imposter)", botlist_nick());
    botlist_breakconnection();
    return;
  }
}

void parsebot_version(char *argv[], int argc) {

  if (argc < 2) return;
  repairstring(&argv[1], argc - 1);
  botlist_setversion(argv[1]);
}

void parsebot_ping(char *argv[], int argc) {

  botlist_send2current("pong");
}

void parsebot_pong(char *argv[], int argc) {

  botlist_setpong();
}

void parsebot_chat(char *argv[], int argc) {           /* chat <bot> <message> */

  if (argc < 3) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (!(same_branche(botlist_nick(), argv[1]))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  repairstring(&argv[2], argc - 2);
  session_send2party("*** (%s) %s", argv[1], argv[2]);
  party_send2party("*** (%s) %s", argv[1], argv[2]);
  botlist_send2allexceptcurrent("chat %s %s", argv[1], argv[2]);
  console_send2snoopwindow("(all) *** (%s) %s", argv[1], argv[2]);
}

void parsebot_nlinked(char *argv[], int argc) {         /* nlinked <new bot> <parent> */

  if (argc != 3) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (strlen(argv[1]) > 9) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (!(same_branche(argv[2], botlist_nick()))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (part_of_botnet(argv[1])) {
    botlist_send2all("chat %s Loop detected: %s linked to %s", botnick, argv[2], argv[1]);
    party_send2party("*** Loop detected: %s linked to %s", argv[2], argv[1]);
    session_send2party("*** Loop detected: %s linked to %s", argv[2], argv[1]);
    console_send2snoopwindow("(all) *** Loop detected: %s linked to %s", argv[2], argv[1]);
    botlist_breakconnection();
    return;
  }
  on_link(argv[2], argv[1]);
  addfarlink(argv[2], argv[1]);
  botlist_send2allexceptcurrent("nlinked %s %s", argv[1], argv[2]);
}

void parsebot_chan(char *argv[], int argc) {    /* chan [<nick>@]<bot> <channel> <message> */

  int channel;

  if (argc < 4) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  repairstring(&argv[3], argc - 3);
  channel = atoi(argv[2]);
  if ((channel < 0) || (channel > 99999)) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (index(argv[1], '@')) {
    if (!(same_branche(botlist_nick(), index(argv[1], '@')+1))) {
      botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
      return;
    }
    session_send2partychannel(channel, "<%s> %s", argv[1], argv[3]);
    party_send2partychannel(channel, "<%s> %s", argv[1], argv[3]);
    botlist_send2allexceptcurrent("chan %s %d %s", argv[1], channel, argv[3]);
    console_send2snoopwindow("(%03d) <%s> %s", channel, argv[1], argv[3]);
  }
  else {
    if (!(same_branche(botlist_nick(), argv[1]))) {
      botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
      return;
    }
    session_send2partychannel(channel, "*** (%s) %s", argv[1], argv[3]);
    party_send2partychannel(channel, "*** (%s) %s", argv[1], argv[3]);
    botlist_send2allexceptcurrent("chan %s %d %s", argv[1], channel, argv[3]);
    console_send2snoopwindow("(%03d) *** (%s) %s", channel, argv[1], argv[3]);
  }
}

void parsebot_unlinked(char *argv[], int argc) {     /* unlinked <bot> */

  if (argc != 2) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (!(same_branche(argv[1], botlist_nick()))) {
    /* Does happen ...
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    */
    return;
  }
  on_unlink(argv[1]);
  dellink(argv[1]);
  botlist_send2allexceptcurrent("unlinked %s", argv[1]);
}

void parsebot_bye(char *argv[], int argc) {    /* bye */

  botlist_breakconnection();
}

void parsebot_who(char *argv[], int argc) {   /* who <user>@<bot> <destination bot> <channel> */

  int channel;

  if (argc != 4) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (!(index(argv[1], '@'))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (!(same_branche(index(argv[1], '@')+1, botlist_nick()))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (!(part_of_botnet(argv[2]))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  channel = atoi(argv[3]);
  if ((channel < 0) || (channel > 99999)) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (strcasecmp(argv[2], botnick) == 0) {
    botlist_send2current("priv %s %s I'm sorry,", botnick, argv[1]);
    botlist_send2current("priv %s %s %s does not support \".who\" anymore", botnick, argv[1], VERSION);
    return;
  }
  bottree_send2bot(index(argv[1], '@')+1, "who %s %s %d", argv[1], argv[2], channel);
}


void parsebot_priv(char *argv[], int argc) {  /* priv [<nick>@]<srcbot> <nick>@<destbot> <message> */

  char *srcbot, *destbot;

  if (argc < 4) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if ((srcbot  = index(argv[1], '@')) == NULL) srcbot = argv[1]; else srcbot++;
  if ((destbot = index(argv[2], '@')) == NULL) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  *destbot='\0'; destbot++;
  if (!(same_branche(srcbot, botlist_nick()))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (same_branche(destbot, botlist_nick())) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  repairstring(&argv[3], argc - 3);
  if (strcasecmp(destbot, botnick) == 0) {  /* message for one of our users */
    if (party_findnick(argv[2])) {
      if (argv[1] == srcbot)
	party_send2current("*** [%s] %s", srcbot, argv[3]);
      else
	party_send2current("*%s* %s", argv[1], argv[3]);
      return;
    }
    if ((session_findnick(argv[2])) && (session_has_chat())) {
      if (argv[1] == srcbot)
	session_sendnotice("*** [%s] %s", srcbot, argv[3]);
      else {
	session_sendnotice("*%s* %s", argv[1], argv[3]);
	if (session_isaway()) {
	  botlist_send2current("priv %s %s %s is away (%s)", botnick, argv[1], session_nick(), session_isaway());
	  if (addnote(argv[2], argv[1], argv[3]))
	    botlist_send2current("priv %s %s Your note has been stored", botnick, argv[1]);
	  else
	    botlist_send2current("priv %s %s Your note could not be stored", botnick, argv[1]);
	}
      }
      return;
    }
    if (argv[1] == srcbot) return;
    if (addnote(argv[2], argv[1], argv[3]))
      botlist_send2current("priv %s %s %s is not here. Your note could not be stored", botnick, argv[1], argv[2]);
    else
      botlist_send2current("priv %s %s %s is not here. Your note has been stored", botnick, argv[1], argv[2]);
    return;
  }
  bottree_send2bot(destbot, "priv %s %s@%s %s", argv[1], argv[2], destbot, argv[3]);
}


void parsebot_whoq(char *argv[], int argc) {   /* who? [<nick>]@<bot> <channel> */

  int   channel;
  char *srcbot;

  if (argc != 3) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if ((srcbot = index(argv[1], '@')) == NULL)
    srcbot = argv[1];
  else
    srcbot++;
  
  if (!same_branche(botlist_nick(), srcbot)) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  channel = atoi(argv[2]);
  if ((channel < 0) || (channel > 99999)) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (session_findfirst()) {
    do {
      if (session_isonchannel(channel))
	botlist_send2current("whom %s %c%s %s %s", argv[1], session_givechar(), session_nick(), 
			     botnick, index(session_nickuserhost(), '!')+1);
    } while (session_next());
  }
  if (party_findfirst()) {
    do {
      if (party_isonchannel(channel))
	botlist_send2current("whom %s -%s %s %s", argv[1], party_nick(), botnick, party_from());
    } while (party_next());
  }
  botlist_send2allexceptcurrent("who? %s %d", argv[1], channel);
}


void parsebot_actchan(char *argv[], int argc) {   /* actchan <nick>@<bot> <channel> <message> */

  int channel;

  if (argc < 4) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  repairstring(&argv[3], argc - 3);
  channel = atoi(argv[2]);
  if ((channel < 0) || (channel > 99999)) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (index(argv[1], '@')) {
    if (!(same_branche(botlist_nick(), index(argv[1], '@')+1))) {
      botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
      return;
    }
    session_send2partychannel(channel, "* %s %s", argv[1], argv[3]);
    party_send2partychannel(channel, "* %s %s", argv[1], argv[3]);
    botlist_send2allexceptcurrent("actchan %s %d %s", argv[1], channel, argv[3]);
    console_send2snoopwindow("(%03d) * %s %s", channel, argv[1], argv[3]);
  }
  else {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
}


void parsebot_link(char *argv[], int argc) {   /* link [<nick>@]<bot> <destbot> <bot2link2> */

  char *saved_nick;
  int   convalue;
  char *frombot;

  if (argc != 4) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if ((!(part_of_botnet(argv[2]))) || (part_of_botnet(argv[3]))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (index(argv[1], '@'))
    frombot = index(argv[1], '@') + 1;
  else
    frombot = argv[1];
  if ((!(part_of_botnet(frombot))) || same_branche(frombot, argv[2])) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (strcasecmp(argv[2], botnick)==0) {     /* Request was sent to this bot */
    saved_nick = botlist_nick();
    if (!(botlist_findbot(argv[3]))) {
      botlist_findbot(saved_nick);
      if (frombot != argv[1])       /* Only send message when link command does not come from bot */
	botlist_send2current("priv %s %s %s: No such bot in botlist", botnick, argv[1], argv[3]);
      return;
    }
    convalue = botlist_startconnection();
    botlist_findbot(saved_nick); 
    if (frombot != argv[1]) {       /* Only send message when link command does not come from bot */
      switch(convalue) {
      case 0:
	botlist_send2current("priv %s %s I am already trying to connect to %s", botnick, argv[1], argv[3]);
	return;
      case 1:
	botlist_send2current("priv %s %s I am already connected with %s. We are checking passwords ...", botnick, argv[1], 
			     argv[3]);
	return;
      case 2:
	botlist_send2current("priv %s %s I am already linked to %s", botnick, argv[1], argv[3]);
	return;
      case 3:
	botlist_send2current("priv %s %s I could not connect to %s", botnick, argv[1], argv[3]);
      return;
      case 4:
	botlist_send2current("priv %s %s Connection to %s in progress", botnick, argv[1], argv[3]);
	return;
      case 5:
	botlist_send2current("priv %s %s Connection with %s established", botnick, argv[1], argv[3]);
	return;
      }
    }
    return;
  }
  bottree_send2bot(argv[2], "link %s %s %s", argv[1], argv[2], argv[3]);
}


void parsebot_whom(char *argv[], int argc) {     /* whom [<destnick>@]<destbot> <infonick> <srcbot> <infofrom> */

  char *destbot;
  
  if (argc < 5) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }

  if ((destbot = index(argv[1], '@')) == NULL)
    destbot = argv[1];
  else
    destbot++;

  if (same_branche(destbot, botlist_nick()) || (!same_branche(argv[3], botlist_nick()))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }

  repairstring(&argv[4], argc - 4);

  if (strcasecmp(destbot, botnick) == 0) {
    if (destbot == argv[1]) {
      console_send2snoopwindow2("%-12s %-10s %s", argv[2], argv[3], argv[4]);
      return;
    }
    else {
      destbot--; *destbot = '\0';
      if ((session_findnick(argv[1])) && (session_has_chat())) {
	session_sendnotice("%-12s %-10s %s", argv[2], argv[3], argv[4]);
	return;
      }
      if (party_findnick(argv[1])) {
	party_send2current("%-12s %-10s %s", argv[2], argv[3], argv[4]);
	return;
      }
      return;
    }
  }
  bottree_send2bot(destbot, "whom %s %s %s %s", argv[1], argv[2], argv[3], argv[4]);
}
      

void parsebot_unlink(char *argv[], int argc) {   /* unlink [<nick>@]<srcbot> <destbot> <unlink> */

  char *saved_nick;
  char *frombot;

  if (argc != 4) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (index(argv[1], '@'))
    frombot = index(argv[1], '@') + 1;
  else
    frombot = argv[1];

  if ((!(same_branche(frombot, botlist_nick()))) || (same_branche(argv[2], botlist_nick())) || 
      (!(part_of_botnet(argv[3]))) || (same_branche(argv[3], botlist_nick()))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (strcasecmp(argv[2], botnick) == 0) {
    saved_nick = botlist_nick();
    if (!(botlist_findbot(argv[3]))) {
      botlist_findbot(saved_nick);
      if (frombot != argv[1]) /* only send message when command came from user */
	botlist_send2current("priv %s %s I am not linked with %s", botnick, argv[1], argv[3]);
      return;
    }
    botlist_breakconnection();
    botlist_findbot(saved_nick);
    if (frombot != argv[1])  /* only send message when command came from user */
      botlist_send2current("priv %s %s Disconnecting %s", botnick, argv[1], argv[3]);
    return;
  }
  bottree_send2bot(argv[2], "unlink %s %s %s", argv[1], argv[2], argv[3]);
}

void parsebot_trace(char *argv[], int argc) {    /* trace <nick>@<bot> <destbot> <route> */

  if ((argc != 4) || (index(argv[1], '@')==NULL) || (!(same_branche(index(argv[1], '@')+1, botlist_nick()))) ||
      (same_branche(botlist_nick(), argv[2]))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (strcasecmp(botnick, argv[2]) == 0) {
    botlist_send2current("traced %s %s:%s", argv[1], argv[3], botnick);
    return;
  }
  bottree_send2bot(argv[2], "trace %s %s %s:%s", argv[1], argv[2], botnick, argv[3]);
}

void parsebot_traced(char *argv[], int argc) {   /* traced <nick>@<bot> <path> */

  char *destbot;

  if ((argc != 3) || (index(argv[1], '@')==NULL) || (same_branche((destbot = index(argv[1], '@')+1), botlist_nick())) ||
      (!(part_of_botnet(destbot)))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  destbot--; *destbot='\0'; destbot++;
  if (strcasecmp(destbot, botnick)==0) {
    if ((session_findnick(argv[1])) && (session_has_chat())) {
      session_sendnotice("Trace result: %s", argv[2]);
      return;
    }
    if (party_findnick(argv[1])) {
      party_send2current("Trace result: %s", argv[2]);
      return;
    }
    return;
  }
  bottree_send2bot(destbot, "traced %s@%s %s", argv[1], destbot, argv[2]);
}

void parsebot_reject(char *argv[], int argc) {    /* reject <srcnick>@<srcbot> <destnick>@<destbot> <comment> */

  char *srcbot, *destbot;

  if ((argc < 4) || (index(argv[1], '@')==NULL) || (!(same_branche((srcbot=index(argv[1], '@')+1), botlist_nick()))) ||
      (!(part_of_botnet(srcbot))) || (index(argv[2], '@')==NULL) || 
      (same_branche((destbot=index(argv[2], '@')+1), botlist_nick())) || (!(part_of_botnet(destbot)))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  destbot--; *destbot='\0'; destbot++;
  repairstring(&argv[3], argc - 3);
  if (strcasecmp(botnick, destbot)==0) {
    if ((session_findnick(argv[2])) && (session_currentchannel() != -1)) {
      session_boot(argv[1], argv[3]);
      return;
    }
    if (party_findnick(argv[2])) {
      party_boot(argv[1], argv[3]);
      return;
    }
    return;
  }
  bottree_send2bot(destbot, "reject %s %s@%s %s", argv[1], argv[2], destbot, argv[3]);
}

void parsebot_assoc(char *argv[], int argc) {   /* assoc <channel> <name> */

  int channel;

  if ((argc < 3) || ((channel = atoi(argv[1])) < 0) || (channel > 99999)) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  if (isdigit(argv[2][0]))
    delassoc(channel);
  else {
    repairstring(&argv[2], argc-2);
    addassoc(channel, argv[2]);
  }
  botlist_send2allexceptcurrent("assoc %d %s", channel, argv[2]);
}

void parsebot_infoq(char *argv[], int argc) {   /* info? <nick>@<bot> */

  if ((argc != 2) || (index(argv[1], '@')==NULL) || (!(same_branche(index(argv[1], '@')+1, botlist_nick())))) {
    botlist_send2current("chat %s NOTICE: Fake message rejected", botnick);
    return;
  }
  botlist_send2current("priv %s %s %s", botnick, argv[1], LONGVERSION);
  botlist_send2allexceptcurrent("info? %s", argv[1]);
}

struct {
  char *command;
  void (*function)(char *argv[], int argc);
} parse_linking[] = {
  { "Please",     parsebot_please   },
  { "You",        parsebot_you      },
  { "passreq",    parsebot_passreq  },
  { "*hello!",    parsebot_hello    },
  { "badpass",    parsebot_badpass  },
  { NULL,         NULL              }
};

struct {
  char *command;
  void (*function)(char *argv[], int argc);
} parse_linked[] = {
  { "thisbot",    parsebot_thisbot  },
  { "version",    parsebot_version  },
  { "ping",       parsebot_ping     },
  { "pong",       parsebot_pong     },
  { "chat",       parsebot_chat     },
  { "nlinked",    parsebot_nlinked  },
  { "chan",       parsebot_chan     },
  { "unlinked",   parsebot_unlinked },
  { "bye",        parsebot_bye      },
  { "who",        parsebot_who      },
  { "priv",       parsebot_priv     },
  { "who?",       parsebot_whoq     },
  { "actchan",    parsebot_actchan  },
  { "link",       parsebot_link     },
  { "whom",       parsebot_whom     },
  { "unlink",     parsebot_unlink   },
  { "trace",      parsebot_trace    },
  { "traced",     parsebot_traced   },
  { "info?",      parsebot_infoq    },
  { NULL,         NULL              }
};

void parse_linking_bot(char *text) {

  int argc=7;
  char *argv[7];
  int i;

  split_string(text, argv, &argc);
  if (argc == 0) return;

  for (i=0; parse_linking[i].command; i++) {
    if (strcasecmp(argv[0], parse_linking[i].command) == 0) {
      parse_linking[i].function(argv, argc);
      return;
    }
  }
}


void parse_linked_bot(char *text) {

  int argc=7;
  char *argv[7];
  int i;

  if (on_botnet(botlist_nick(), text)) return;

  split_string(text, argv, &argc);
  if (argc == 0) return;

  for (i=0; parse_linked[i].command; i++) {
    if (strcasecmp(argv[0], parse_linked[i].command) == 0) {
      parse_linked[i].function(argv, argc);
      return;
    }
  }
}
