
/*******************************************************/
/**                                                   **/
/**   party_console.c (c) 1996, 97 Emiel Bruijntjes   **/
/**                                                   **/
/**Messages entered on the console while in party mode**/
/**                                                   **/
/*******************************************************/


#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <strings.h>


#include "party_console.h"
#include "bots.h"
#include "partyline.h"
#include "session.h"
#include "console.h"
#include "assoc.h"
#include "misc.h"
#include "bottree.h"
#include "gencmd.h"


static int cur_channel=-1;


extern char  commandchar3;
extern char  botnick[10];
extern char *helpdir;


void party_console_join(char *argv[], int argc) {

  int channel;

  if (argc < 2) return;

  if (isdigit(argv[1][0]) || (argv[1][0] == '-')) { channel = atoi(argv[1]); }
  else {
    repairstring(&argv[1], argc - 1);
    if ((channel = assoc_channelnumber(argv[1])) == -1) {
      console_send2snoopwindow("(---) *** %s: No such channel", argv[1]);
      return;
    }
  }
  if (channel > 99999) {
    console_send2snoopwindow("(---) *** %d: No such channel", channel);
    return;
  }
  if ((cur_channel = channel) < 0) cur_channel = -1;
  console_print_mode();
}

void party_console_all(char *argv[], int argc) {

  cur_channel = -1;
  console_print_mode();
}

void party_console_assoc(char *argv[], int argc) {
  
  int   channel=0;
  char *channelname;
  
  if (argc == 1) {
    console_send2snoopwindow("(---) *** Number  Name");
    console_send2snoopwindow("(---) *** ------  ---------------------------");
    for (channel = 1; channel < 100000; channel++) {
      if ((channelname = assoc_channelname(channel)) != NULL) {
	console_send2snoopwindow("(---) *** %6d  %s", channel, channelname);
      }
    }
    console_send2snoopwindow("(---) *** ------  ---------------------------");
    return;
  }
  if (cur_channel == 0) {
    console_send2snoopwindow("(---) *** You can't change the name of channel 0, that's the main party line");
    return;
  }
  if (cur_channel < 0) {
    console_send2snoopwindow("(---) *** You are not on a channel");
    return;
  }
  repairstring(&argv[1], argc - 1);
  if (argv[1][0] == '-') {
    delassoc(cur_channel);
    botlist_send2all("assoc %d 0", cur_channel);
    botlist_send2all("chan %s %d The name of channel %d has been removed", botnick, cur_channel, cur_channel);
    party_send2partychannel(cur_channel, "*** The name of channel %d has been removed", cur_channel);
    session_send2partychannel(cur_channel, "*** The name of channel %d has been removed", cur_channel);
    console_send2snoopwindow("(%03d) *** The name of channel %d has been removed", cur_channel, cur_channel);
    return;
  }
  if (isdigit(argv[1][0])) {
    console_send2snoopwindow("(---) *** First character of the channelname can't be a digit");
    return;
  }
  if (strlen(argv[1]) > 20) {
    console_send2snoopwindow("(---) *** Channelnames can't be longer than 20 characters");
    return;
  }
  addassoc(cur_channel, argv[1]);
  botlist_send2all("assoc %d %s", cur_channel, argv[1]);
  botlist_send2all("chan %s %d Channel %d is now also known as channel \"%s\"", botnick, cur_channel, cur_channel, 
		   argv[1]);
  party_send2partychannel(cur_channel, "*** Channel %d is now also known as channel \"%s\"", cur_channel, argv[1]);
  session_send2partychannel(cur_channel, "*** Channel %d is now also known as channel \"%s\"", cur_channel, argv[1]);
  console_send2snoopwindow("(%03d) *** Channel %d is now also known as channel \"%s\"", cur_channel, cur_channel, argv[1]);
}

void party_console_resetassoc(char *argv[], int argc) {

  int channel;

  channel = cur_channel;

  cur_channel = 1;
  while (cur_channel < 100000) {
    if (assoc_channelname(cur_channel)) {
      argv[0] = "assoc";
      argv[1] = "-";
      party_console_assoc(argv, 2);
    }
    cur_channel++;
  }
  cur_channel = channel;
}

void party_console_boot(char *argv[], int argc) {

  char *comment, *bot;

  if (argc < 2) return;

  if (argc >= 3) {
    repairstring(&argv[2], argc - 2);
    comment = argv[2];
  }
  else
    comment = "boot from console";

  if (index(argv[1], '@')==NULL) {
    if (session_findnick(argv[1])) { 
      session_boot(argv[1], comment);
      return;
    }
    if (party_findnick(argv[1])) { 
      party_boot(argv[1], comment);
      return;
    }
    return;
  }
  bot = index(argv[1], '@')+1;
  if (strcasecmp(bot, botnick)==0) {
    bot--; *bot='\0'; bot++;
    if (session_findnick(argv[1])) { 
      session_boot(argv[1], comment);
      return;
    }
    if (party_findnick(argv[1])) { 
      party_boot(argv[1], comment);
      return;
    }
    return;
  }
  if (!(part_of_botnet(bot))) {
    console_send2snoopwindow("(---) *** %s is not part of the botnet", bot);
    return;
  }
  bottree_send2bot(bot, "reject %s@%s %s %s", session_nick(), botnick, argv[1], comment);
}

void party_console_bottree(char *argv[], int argc) {

  printbottree(console_send2snoopwindow2);
}

void party_console_help(char *argv[], int argc) {

  char *file;
  int i;
  struct {
    char *command;
    char *file;
  } helpfile_argc1[] = {
    { "all", "pchelp.all" },
    { "assoc", "pchelp.assoc" },
    { "boot", "pchelp.boot" },
    { "bottree", "pchelp.bottree" },
    { "help", "pchelp.help" },
    { "join", "pchelp.join" },
    { "resassoc", "pchelp.resassoc" },
    { "who", "pchelp.who" },
    { NULL, NULL }
  };

  if (argc == 1) {
    file = (char *) xmalloc (strlen(helpdir)+strlen("pchelp")+1);
    sprintf(file, "%spchelp", helpdir);
    catfile(file, console_send2snoopwindow2, "Console");
    xfree(file);
    return;
  }
  if (argc >= 2) {
    for (i = 0; helpfile_argc1[i].command; i++) {
      if (strcasecmp(argv[1], helpfile_argc1[i].command) == 0) {
	file = (char *)xmalloc(strlen(helpdir) + strlen(helpfile_argc1[i].file) + 1);
	sprintf(file, "%s%s", helpdir, helpfile_argc1[i].file);
	catfile(file, console_send2snoopwindow2, "Console");
	xfree(file);
	return;
      }
    }
    console_send2snoopwindow("(---) *** %s: No such topic", argv[1]);
    return;
  }
}

void party_console_who(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    if ((channel = cur_channel) == -1) {
      console_send2snoopwindow2("You are not on a channel");
      return;
    }
  }
  else {
    if (isdigit(argv[1][0])) {
      channel = atoi(argv[1]);
    }
    else {
      repairstring(&argv[1], argc - 1);
      if ((channel = assoc_channelnumber(argv[1])) == -1) {
	console_send2snoopwindow2("*** %s: No such channel", argv[1]);
	return;
      }
    }
  }
  console_send2snoopwindow2("Nick         Bot        From");
  console_send2snoopwindow2("----------   ---------  ------------------------------");
  if (session_findfirst()) {
    do {
      if (session_isonchannel(channel)) {
	console_send2snoopwindow2("%c%-11s %-10s %s", session_givechar(), session_nick(), botnick, 
				  index(session_nickuserhost(), '!')+1) ;
      }
    } while (session_next());
  }
  if (party_findfirst()) {
    do
      {
	if (party_isonchannel(channel)) {
	  console_send2snoopwindow2("-%-11s %-10s %s", party_nick(), botnick, party_from());
	}
      }
    while (party_next());
  }
  botlist_send2all("who? %s %d", botnick, channel);
}




struct {
  char *command;
  void (*function)(char *argv[], int argc);
} party_commands[] = {
  { "all",        party_console_all        },
  { "assoc",      party_console_assoc      },
  { "boot",       party_console_boot       },
  { "bottree",    party_console_bottree    },
  { "help",       party_console_help       },
  { "join",       party_console_join       },
  { "resassoc",   party_console_resetassoc },
  { "who",        party_console_who        },
  { NULL,         NULL                     }
};

void parse_party_command(char *command) {

  int argc=7, i;
  char *argv[7];

  split_string(command, argv, &argc);
  if (argc == 0) return;

  for (i=0; party_commands[i].command; i++) {
    if (strcasecmp(argv[0], party_commands[i].command) == 0) {
      party_commands[i].function(argv, argc);
      return;
    }
  }
  console_send2snoopwindow("(---) *** Unknown command: %s", argv[0]);
}

void parse_party_console(char *text) {

  if ((text == NULL) || (text[0] == '\0'))
    return;

  if (text[0] == commandchar3) {
    if (text[1] == ' ')
      text += 2;
    else {
      text++;
      parse_party_command(text);
      return;
    }
  }
  if (cur_channel == -1) {
    console_send2snoopwindow("(all) *** %s", text);
    party_send2party("*** %s", text);
    session_send2party("*** %s", text);
    botlist_send2all("chat %s %s", botnick, text);
  }
  else {
    console_send2snoopwindow("(%03d) *** %s", cur_channel, text);
    party_send2partychannel(cur_channel, "*** %s", text);
    session_send2partychannel(cur_channel, "*** %s", text);
    botlist_send2all("chan %s %d %s", botnick, cur_channel, text);
  }
}

char *current_party_channel() {

  static char buf[30]="";

  if (cur_channel < 0)
    strcpy(buf, "all");
  else {
    if (assoc_channelname(cur_channel)) {
      sprintf(buf, "\"%s\"", assoc_channelname(cur_channel));
    }
    else {
      sprintf(buf, "%d", cur_channel);
    }
  }
  return buf;
}
