
/*******************************************************/
/**                                                   **/
/**      partycmd.c (c) 1996, 97 Emiel Bruijntjes     **/
/**                                                   **/
/**      Parsing commands from telnet-connections     **/
/**                                                   **/
/*******************************************************/

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>


#include "partycmd.h"
#include "misc.h"
#include "session.h"
#include "partyline.h"
#include "assoc.h"
#include "bots.h"
#include "bottree.h"
#include "notes.h"
#include "gencmd.h"
#include "console.h"
#include "on.h"


extern char *helpdir;
extern char  botnick[10];


void partycmd_help(char *argv[], int argc) {

  char *file;
  int   i;
  struct {
    char *command;
    char *file;
  } party_help[] = {
    { "help",   "phelp"   },
    { "join",   "phelp.join"   },
    { "leave",  "phelp.leave"  },
    { "act",    "phelp.act"    },
    { "assoc",  "phelp.assoc"  },
    { "quit",   "phelp.quit"   },
    { "who",    "phelp.who"    },
    { "bots",   "phelp.bots"   },
    { "bottree","phelp.bottree"},
    { "echo",   "phelp.echo"   },
    { "note",   "phelp.note"   },
    { "whois",  "phelp.whois"  }, 
    {  NULL,     NULL,         }
  };

  if (argc > 2) {
    party_send2current("Usage: help [command]");
    return;
  }
  if (argc == 1) {
    file = (char *)xmalloc(strlen(helpdir)+strlen("phelp")+1);
    sprintf(file, "%s%s", helpdir, "phelp");
    catfile(file, party_send2current, party_nick());
    xfree(file);
    return;
  }
  for (i = 0; party_help[i].command; i++) {
    if (strcasecmp(argv[1], party_help[i].command) == 0) {
      file = (char *)xmalloc(strlen(helpdir) + strlen(party_help[i].file) + 1);
      sprintf(file, "%s%s", helpdir, party_help[i].file);
      catfile(file, party_send2current, party_nick());
      xfree(file);
      return;
    }
  }
  party_send2current("help: No such topic");
}

void partycmd_join(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    party_send2current("Usage: join <channelnumber>|<channelname>");
    return;
  }
  if (isdigit(argv[1][0])) { channel = atoi(argv[1]); }
  else {
    repairstring(&argv[1], argc - 1);
    if ((channel = assoc_channelnumber(argv[1])) == -1) {
      party_send2current("%s: No such channel", argv[1]);
      return;
    }
  }
  if ((channel < 0) || (channel > 99999)) {
    party_send2current("%d: No such channel", channel);
    return;
  }
  if (!(party_joinchannel(channel))) {
    party_send2current("You have joined too many channels");
    return;
  }
}

void partycmd_leave(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    if ((channel = party_currentchannel()) == -1) {
      party_send2current("You are not on a channel");
      return;
    }
  }
  else {
    if (isdigit(argv[1][0])) { channel = atoi(argv[1]); }
    else {
      repairstring(&argv[1], argc - 1);
      if ((channel = assoc_channelnumber(argv[1])) == -1) {
	party_send2current("%s: No such channel", argv[1]);
	return;
      }
    }
    if ((channel < 0) || (channel > 99999)) {
      party_send2current("%d: No such channel", channel);
      return;
    }
  }
  party_leavechannel(channel);
}

void partycmd_act(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    party_send2current("Usage: act <text>");
    return;
  }
  if ((channel = party_currentchannel()) == -1) {
    party_send2current("You are not on a channel");
    return;
  }
  repairstring(&argv[1], argc - 1);
  party_send2partychannel(channel, "* %s %s", party_nick(), argv[1]);
  session_send2partychannel(channel, "* %s %s", party_nick(), argv[1]);
  botlist_send2all("actchan %s@%s %d %s", party_nick(), botnick, channel, argv[1]);
  console_send2snoopwindow("(%03d) * %s %s", channel, party_nick(), argv[1]);
}

void partycmd_assoc(char *argv[], int argc) {

  int   channel=0;
  char *channelname;

  if (argc == 1) {
    party_send2current("Number  Name");
    party_send2current("------  ---------------------------");
    for (channel = 1; channel < 100000; channel++) {
      if ((channelname = assoc_channelname(channel)) != NULL) {
	party_send2current("%6d  %s", channel, channelname);
      }
    }
    party_send2current("------  ---------------------------");
    return;
  }
  if ((channel = party_currentchannel()) == -1) {
    party_send2current("You are not on a channel");
    return;
  }
  if (channel == 0) {
    party_send2current("You can't change the name of channel 0, that's the main party line");
    return;
  }
  repairstring(&argv[1], argc - 1);
  if (argv[1][0] == '-') {
    delassoc(channel);
    botlist_send2all("assoc %d 0", channel);
    botlist_send2all("chan %s %d The name of channel %d has been removed by %s", botnick, channel, channel, party_nick());
    party_send2partychannel(channel, "*** The name of channel %d has been removed by %s", channel, party_nick());
    session_send2partychannel(channel, "*** The name of channel %d has been removed by %s", channel, party_nick());
    console_send2snoopwindow("(%03d) *** The name of channel %d has been removed by %s", channel, channel, party_nick());
    return;
  }
  if (isdigit(argv[1][0])) {
    party_send2current("*** First character of the channelname can't be a digit");
    return;
  }
  if (strlen(argv[1]) > 20) {
    party_send2current("*** Channelnames can't be longer than 20 characters");
    return;
  }
  addassoc(channel, argv[1]);
  botlist_send2all("assoc %d %s", channel, argv[1]);
  botlist_send2all("chan %s %d %s named this channel \"%s\"", botnick, channel, party_nick(), argv[1]);
  party_send2partychannel(channel, "*** %s has named this channel \"%s\"", party_nick(), argv[1]);
  session_send2partychannel(channel, "*** %s has named this channel \"%s\"", party_nick(), argv[1]);
  console_send2snoopwindow("(%03d) *** %s has named this channel \"%s\"", channel, party_nick(), argv[1]);
}

void partycmd_quit(char *argv[], int argc) {

  close_and_delete_party();
}

void partycmd_who(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    if ((channel = party_currentchannel()) == -1) {
      party_send2current("*** You are not on a channel");
      return;
    }
  }
  else {
    if (isdigit(argv[1][0])) {
      channel = atoi(argv[1]);
    }
    else {
      repairstring(&argv[1], argc - 1);
      if ((channel = assoc_channelnumber(argv[1])) == -1) {
	party_send2current("*** %s: No such channel", argv[1]);
	return;
      }
    }
  }
  party_send2current("Nick         Bot        From");
  party_send2current("----------   ---------  ------------------------------");
  if (session_findfirst()) {
    do {
      if (session_isonchannel(channel)) {
	party_send2current("%c%-11s %-10s %s", session_givechar(), session_nick(), botnick, 
			   index(session_nickuserhost(), '!')+1) ;
      }
    } while (session_next());
  }
  party_sendwho2party(channel);
  botlist_send2all("who? %s@%s %d", party_nick(), botnick, channel);
}

void partycmd_bots(char *argv[], int argc) {

  if (argc != 1) {
    party_send2current("Usage: bots");
    return;
  }
  botlist_send2all("info? %s@%s", party_nick(), botnick);
}

void partycmd_bottree(char *argv[], int argc) {

  if (argc != 1) {
    party_send2current("Usage: bottree");
    return;
  }
  printbottree(party_send2current);
}

void partycmd_echo(char *argv[], int argc) {

  if (argc != 2) {
    party_send2current("Usage: echo [on|off]");
    return;
  }
  if (strcasecmp(argv[1], "on")==0) {
    party_echo_on();
    party_send2current("echo turned on");
    return;
  }
  if (strcasecmp(argv[1], "off")==0) {
    party_echo_off();
    party_send2current("echo turned off");
    return;
  }
  party_send2current("Usage: echo [on|off]");
}

void partycmd_note(char *argv[], int argc) {
  
  char *tmp, *bot;

  if (argc < 3) {
    party_send2current("Usage: note <nick>[@<bot>] <message>");
    return;
  }
  repairstring(&argv[2], argc - 2);

  party_send2current("-> *%s* %s", argv[1], argv[2]);
  if (index(argv[1], '@')==NULL) {
  jumpinpartycmd_note:
    if ((session_findnick(argv[1])) && (session_has_chat())) {
      session_sendnotice("*%s* %s", party_nick(), argv[2]);
      if (session_isaway()) {
	party_send2current("%s is away: %s", argv[1], session_isaway());
	if (addnote(argv[1], party_nick(), argv[2]))
	  party_send2current("Your note has been stored");
	else
	  party_send2current("Your note could not be stored");
	return;
      }
      return;
    }
    tmp = party_nick();
    if (party_findnick(argv[1])) { 
      party_send2current("*%s* %s", tmp, argv[2]);
      return;
    }
    party_findnick(tmp);
    party_send2current("I do neither have an IRC- nor telnet-session with %s", argv[1]);
    if (addnote(argv[1], tmp, argv[2]))
      party_send2current("Your note has been stored");
    else
      party_send2current("Your note could not be stored");
    return;
  }
  bot = index(argv[1], '@')+1;
  if (strcasecmp(bot, botnick)==0) {
    bot--; *bot='\0'; bot++;
    goto jumpinpartycmd_note;
  }
  if (!(part_of_botnet(bot))) {
    party_send2current("%s is not part of the botnet", bot);
    return;
  }
  bottree_send2bot(bot, "priv %s@%s %s %s", party_nick(), botnick, argv[1], argv[2]);
}

void partycmd_whois(char *argv[], int argc) {

  char *nick, *chans;

  if (argc != 2) {
    party_send2current("Usage: whois <nick>");
    return;
  }
  nick = party_nick();
  if (party_findnick(argv[1])) {
    chans = party_currentchannels();
    party_findnick(nick);
    party_send2current("*** Party channels: %s (outside)", chans);
    return;
  }
  party_findnick(nick);
  if ((session_findnick(argv[1])) && (session_currentchannel() != -1)) {
    party_send2current("*** Party channels: %s (IRC)", session_currentchannels());
    return;
  }
  party_send2current("*** %s: No such nick", argv[1]);
}

struct {
  char *command;
  void (*function)(char *argv[], int argc);
  char allowed;
} parse_partycmd[] = {
  { "help",    partycmd_help,    1 },
  { "join",    partycmd_join,    1 },
  { "leave",   partycmd_leave,   1 },
  { "act",     partycmd_act,     1 },
  { "assoc",   partycmd_assoc,   1 },
  { "quit",    partycmd_quit,    1 },
  { "who",     partycmd_who,     1 },
  { "bots",    partycmd_bots,    1 },
  { "bottree", partycmd_bottree, 1 },
  { "echo",    partycmd_echo,    1 },
  { "note",    partycmd_note,    1 },
  { "whois",   partycmd_whois,   1 },
  { NULL,      NULL,             0 }
};


void parse_party(char *text) {

  int argc=7;
  char *argv[7];
  int i;

  if (on_outside_cmd(party_nick(), text)) return;

  sendtospies("%s@%s (telnet): %s", party_nick(), party_from(), text);
  split_string(text, argv, &argc);
  if (argc == 0) return;

  for (i=0; parse_partycmd[i].command; i++) {
    if (strcasecmp(argv[0], parse_partycmd[i].command) == 0) {
      if (parse_partycmd[i].allowed)
	parse_partycmd[i].function(argv, argc);
      else
	party_send2current("%s: Command has been disabled");
      return;
    }
  }
  party_send2current("%s: Command not found", argv[0]);
}

int enable_partycmd(char *cmd) {

  int i;

  for (i=0; parse_partycmd[i].command; i++) {
    if (strcasecmp(cmd, parse_partycmd[i].command) == 0) {
      parse_partycmd[i].allowed = 1;
      return 1;
    }
  }
  return 0;
}

int disable_partycmd(char *cmd) {

  int i;

  for (i=0; parse_partycmd[i].command; i++) {
    if (strcasecmp(cmd, parse_partycmd[i].command) == 0) {
      parse_partycmd[i].allowed = 0;
      return 1;
    }
  }
  return 0;
}

void save_partycmds(FILE *fp) {

  int i;

  for (i=0; parse_partycmd[i].command; i++)
    fwritechar(fp, parse_partycmd[i].allowed);
}

int load_partycmds(FILE *fp) {

  int i;

  for (i=0; parse_partycmd[i].command; i++)
    freadchar(fp, &parse_partycmd[i].allowed);

  return 1;
}
