
/*******************************************************/
/**                                                   **/
/**     partyline.c (c) 1996, 97 Emiel Bruijntjes     **/
/**                                                   **/
/**     List of telnet-party users + I/O with them    **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif


#include <stdlib.h>
#include <string.h>
#include <stdarg.h>


#include "partyline.h"
#include "annon.h"
#include "config.h"
#include "net.h"
#include "bots.h"
#include "misc.h"
#include "session.h"
#include "partycmd.h"
#include "assoc.h"
#include "console.h"
#include "on.h"

struct partylist {
  struct partynode *head;
  struct partynode *tail;
  struct partynode *current;
  int length;
} partlist = { NULL , NULL , NULL , 0 };

struct partynode {
  int  socketptr;
  char party;
  int  party_channels[MAX_CHANNELS];
  char nick[10];
  char *from;
  char echo;

  struct partynode *prev;
  struct partynode *next;
};

extern char  botnick[10];
extern char  commandchar2;
extern char *party_logfile;

void create_party(char *nick, int sockptr, char *from) {

  partlist.current = (struct partynode *)xmalloc(sizeof(struct partynode));

 if (partlist.length == 0) {
    partlist.head = partlist.current;
    partlist.current->prev = NULL;
  } 
  else {
    partlist.tail->next = partlist.current;
    partlist.current->prev = partlist.tail;
  }
  partlist.current->next = NULL;
  partlist.tail = partlist.current;
  partlist.length = partlist.length + 1;

  if (strlen(nick) > 9) nick[9] = '\0';
  strcpy(partlist.current->nick, nick);
  partlist.current->socketptr         = sockptr;
  partlist.current->party             = 0;
  partlist.current->echo              = 1;
  partlist.current->from = (char *)xmalloc(strlen(from)+1);
  strcpy(partlist.current->from, from);
}


void close_and_delete_party() {

  party_leaveallchannels();
  close_socketptr(partlist.current->socketptr);
  delete_party();
}

void delete_party() {

  if (partlist.current == NULL) return;

  party_leaveallchannels();

  xfree(partlist.current->from);
  if (partlist.length == 1) {
    xfree(partlist.current);
    partlist.current = NULL;
    partlist.tail    = NULL;
    partlist.head    = NULL;
    partlist.length  = 0;
    return;
  }

  if (partlist.current->next == NULL) {
    partlist.tail = partlist.current->prev;
    partlist.tail->next = NULL;
    xfree(partlist.current);
    partlist.current = partlist.tail;
    partlist.length = partlist.length - 1;
    return;
  }

  if (partlist.current->prev == NULL) {
    partlist.head = partlist.current->next;
    partlist.head->prev = NULL;
    xfree(partlist.current);
    partlist.current = partlist.head;
    partlist.length = partlist.length - 1;
    return;
  }

  partlist.current->prev->next = partlist.current->next;
  partlist.current->next->prev = partlist.current->prev;
  xfree(partlist.current);
  partlist.length = partlist.length - 1;
  partlist.current = partlist.head;
  return;
}

int  party_findfirst() { return (int)(partlist.current = partlist.head); }
int  party_next() { return (int)(partlist.current = partlist.current->next); }

int  party_findnick(char *nick) {

  partlist.current = partlist.head;
  while (partlist.current != NULL) {
    if (strcasecmp(nick, party_nick()) == 0) return 1;
    partlist.current = partlist.current->next;
  }
  return 0;
}

int  party_findsockptr(int ptr) {

  partlist.current = partlist.head;
  while (partlist.current != NULL) {
    if (ptr == partlist.current->socketptr) return 1;
    partlist.current = partlist.current->next;
  }
  return 0;
}

int   party_length() { return partlist.length; }

char * party_from() { return partlist.current->from; }
char * party_nick() { return partlist.current->nick; }
int    party_currentchannel() { return partlist.current->party ? partlist.current->party_channels[0] : -1; }

void  party_echo_on()  { partlist.current->echo = 1; }
void  party_echo_off() { partlist.current->echo = 0; }
int   party_echo()     { return partlist.current->echo; }

void party_nickcollision() {

  party_send2current("Nick collision");
  party_leaveallchannels();
  changesocketptrtype(partlist.current->socketptr, T_ANNON);
  addannon(partlist.current->socketptr);
  delete_party();
}

int party_joinchannel(int channel) {

  int i;

  if (party_isonchannel(channel)) {
    for (i = 0; i < partlist.current->party; i++) {
      if (channel == partlist.current->party_channels[i]) {
	partlist.current->party_channels[i] = partlist.current->party_channels[0];
	partlist.current->party_channels[0] = channel;
	party_send2current("You are now talking to channel %d", channel);
	return 1;
      }
    }
    party_send2current("There seems to be a bug in this bot");
  }
  
  if (partlist.current->party >= MAX_CHANNELS) return 0;
  
  if (partlist.current->party != 0)
    partlist.current->party_channels[(int)partlist.current->party] = partlist.current->party_channels[0];
  partlist.current->party_channels[0] = channel;
  partlist.current->party = partlist.current->party + 1;

  if (channel == 0) {
    session_send2partychannel(0, "*** %s has joined the party line", party_nick());
    botlist_send2all("chan %s 0 %s has joined the party line", botnick, party_nick());
    party_send2partychannel(0, "*** %s has joined the party line", party_nick());
    console_send2snoopwindow("(000) *** %s has joined the party line", party_nick());
    return 1;
  }
  if (assoc_channelname(channel)) {
    session_send2partychannel(channel, "*** %s has joined channel \"%s\"", party_nick(), assoc_channelname(channel));
    botlist_send2all("chan %s %d %s has joined channel \"%s\"", botnick, channel, party_nick(), 
		     assoc_channelname(channel));
    party_send2partychannel(channel, "*** %s has joined channel \"%s\"", party_nick(), assoc_channelname(channel));
    console_send2snoopwindow("(%03d) *** %s has joined channel \"%s\"", channel, party_nick(), 
					assoc_channelname(channel));
    return 1;
  }
  session_send2partychannel(channel, "*** %s has joined channel %d", party_nick(), channel);
  botlist_send2all("chan %s %d %s has joined channel %d", botnick, channel, party_nick(), channel);
  party_send2partychannel(channel, "*** %s has joined channel %d", party_nick(), channel);
  console_send2snoopwindow("(%03d) *** %s has joined channel %d", channel, party_nick(), channel);
  return 1;
}

void party_leavechannel(int channel) {

  int i;
  
  for (i=0; i<partlist.current->party; i++) {
    if (channel == partlist.current->party_channels[i]) {
      if (channel == 0) {
	session_send2partychannel(0, "*** %s has left the party line", party_nick());
	botlist_send2all("chan %s 0 %s has left the party line", botnick, party_nick());
	party_send2partychannel(0, "*** %s has left the party line", party_nick());
	console_send2snoopwindow("(000) *** %s has left the party line", party_nick());
      } 
      else {
	if (assoc_channelname(channel)) {
	  session_send2partychannel(channel, "*** %s has left channel \"%s\"", party_nick(), assoc_channelname(channel));
	  botlist_send2all("chan %s %d %s has left channel \"%s\"", botnick, channel, party_nick(), 
			   assoc_channelname(channel));
	  party_send2partychannel(channel, "*** %s has left channel \"%s\"", party_nick(), assoc_channelname(channel));
	  console_send2snoopwindow("(%03d) *** %s has left channel \"%s\"", channel, party_nick(), 
					      assoc_channelname(channel));
	}
	else {
	  session_send2partychannel(channel, "*** %s has left channel %d", party_nick(), channel);
	  botlist_send2all("chan %s %d %s has left channel %d", botnick, channel, party_nick(), channel);
	  party_send2partychannel(channel, "*** %s has left channel %d", party_nick(), channel);
	  console_send2snoopwindow("(%03d) *** %s has left channel %d", channel, party_nick(), channel);
	}
      }
      partlist.current->party_channels[i] = partlist.current->party_channels[partlist.current->party - 1];
      (partlist.current->party)--;
      return;
    }
  }
}

void party_boot(char *from, char *comment) {

  int i, channel;
  
  for (i=0; i<partlist.current->party; i++) {
    channel = partlist.current->party_channels[i];
    if (channel == 0) {
      session_send2partychannel(0, "*** %s has been kicked off the party line by %s (%s)", party_nick(), from, comment);
      botlist_send2all("chan %s 0 %s has been kicked off the party line by %s (%s)", botnick, party_nick(), from, 
		       comment);
      party_send2partychannel(0, "*** %s has been kicked off the party line by %s (%s)", party_nick(), from, comment);
      console_send2snoopwindow("(000) *** %s has been kicked off the party line by %s (%s)", party_nick(), from,
					  comment);
    }
    else {
      if (assoc_channelname(channel)) {
	session_send2partychannel(channel, "*** %s has been kicked off channel \"%s\" by %s (%s)", party_nick(), 
				  assoc_channelname(channel), from, comment);
	botlist_send2all("chan %s %d %s has been kicked off channel \"%s\" by %s (%s)", botnick, channel, party_nick(), 
			 assoc_channelname(channel), from, comment);
	party_send2partychannel(channel, "*** %s has been kicked off channel \"%s\" by %s (%s)", party_nick(), 
				assoc_channelname(channel), from, comment);
	console_send2snoopwindow("(%03d) *** %s has been kicked off channel \"%s\" by %s (%s)", channel,
					    party_nick(), assoc_channelname(channel), from, comment);
      }
      else {
	session_send2partychannel(channel, "*** %s has been kicked off channel %d by %s (%s)", party_nick(), channel,
				  from, comment);
	botlist_send2all("chan %s %d %s has been kicked off channel %d by %s (%s)", botnick, channel, party_nick(), 
			 channel, from, comment);
	party_send2partychannel(channel, "*** %s has been kicked off channel %d by %s (%s)", party_nick(), channel,
				from, comment);
	console_send2snoopwindow("(%03d) *** %s has been kicked off channel %d by %s (%s)", channel,
					    party_nick(), channel, from, comment);
      }
    }
  }
  partlist.current->party = 0;
  close_and_delete_party();
}


int party_isonchannel(int channel) {

  int i;
 
  for (i=0; i<partlist.current->party; i++) {
    if (channel == partlist.current->party_channels[i])
      return 1;
  }
  return 0;
}

void party_send2partychannel(int channel, char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct partynode *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = partlist.current;
  partlist.current = partlist.head;
  while (partlist.current != NULL) {
    if (party_isonchannel(channel)) party_send2current("(%d) %s", channel, line);
    partlist.current = partlist.current->next;
  }
  partlist.current = tmp;
}

void party_send2partychannelexceptcurrent(int channel, char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct partynode *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = partlist.current;
  partlist.current = partlist.head;
  while (partlist.current != NULL) {
    if ((party_isonchannel(channel)) && (tmp != partlist.current)) party_send2current("(%d) %s", channel, line);
    partlist.current = partlist.current->next;
  }
  partlist.current = tmp;
}

void party_send2party(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct partynode *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = partlist.current;
  partlist.current = partlist.head;
  while (partlist.current != NULL) {
    if (partlist.current->party) party_send2current("%s", line);
    partlist.current = partlist.current->next;
  }
  partlist.current = tmp;
}

void party_leaveallchannels() {

  while (partlist.current->party) party_leavechannel(partlist.current->party_channels[0]);
}

void party_send2current(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  send_to_socketptr(partlist.current->socketptr, "%s\r", line);
}

void party_parse_input(char *text) {

  if (on_outside(party_nick(), text)) return;

  if (text[0] == '\0') return;
  if (text[0] == commandchar2) { 
    putlog(party_logfile, "%s@%s: %s", party_nick(), party_from(), &text[1]);
    console_send2snoopwindow("[spy] %s@%s: %s", party_nick(), party_from(), &text[1]);
    parse_party(&text[1]); 
    return;
  }
  if (partlist.current->party == 0) return;
  putlog(party_logfile, "channel %d: %s@%s: %s", partlist.current->party_channels[0], party_nick(), party_from(), text);
  session_send2partychannel(partlist.current->party_channels[0], "<%s> %s", party_nick(), text);
  botlist_send2all("chan %s@%s %d %s", party_nick(), botnick, partlist.current->party_channels[0], text);
  if (party_echo())
    party_send2partychannel(partlist.current->party_channels[0], "<%s> %s", party_nick(), text);
  else
    party_send2partychannelexceptcurrent(partlist.current->party_channels[0], "<%s> %s", party_nick(), text);
  console_send2snoopwindow("(%03d) <%s> %s", partlist.current->party_channels[0], party_nick(), text);
}

void party_sendwho2party(int channel) {

  struct partynode *localcurrent;
  struct partynode *savecurrent;
  char *nick, *from;

  localcurrent = partlist.current;
  if (party_findfirst()) {
    do {
      if (party_isonchannel(channel)) {
	savecurrent = partlist.current;
	nick = party_nick(); from = party_from();
	partlist.current = localcurrent;
	party_send2current("-%-11s %-10s %s", nick, botnick, from);
	partlist.current = savecurrent;
      }
    } while (party_next());
  }
  partlist.current = localcurrent;
}

char *party_currentchannels() {

  static char buf[100];
  int i;

  buf[0]='\0';
  for (i=0; i<partlist.current->party; i++) {
    sprintf(buf, "%s%d  ", buf, partlist.current->party_channels[i]);
  }
  return buf;
}
