
/*********************************************************/
/**                                                     **/
/**       server.c (c) 1996, 97 Emiel Bruijntjes        **/
/**                                                     **/
/** Server-database + sending mesages to current server **/
/**                                                     **/
/*********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdarg.h>
#include <stdlib.h>


#include "net.h"
#include "server.h"
#include "config.h"
#include "misc.h"
#include "channel.h"
#include "session.h"
#include "console.h"


static char   *serverhost[MAXSERVERS];     /* Array of server-hosts */
static int     portnumber[MAXSERVERS];     /* Array of server-ports */
static int     cycles;                     /* Number of times I have cycled through the server-list */
static int     currentserver=-1;           /* The server in server-array I'm currently connected to */
static int     nr_servers=0;               /* Number of servers in array */

static char    connection_state=0;         /* 0 not connected, 1 connecting, 2 connected */
static int     server_socket_ptr=-1;


extern int     maxcycles;
extern char   *login;
extern char   *ircname;
extern char    botnick[10];
extern char    current_nick[10];

extern time_t  time_null;

void leave_server();

void delete_all_servers() {

  int i;

  switch(connection_state) {
  case 2:
    console_send2ircwindow("*** Signoff: %s (Rehashing ...)", current_nick);
    send_now_to_server("QUIT :Rehashing ...");
    /* No break or return here ... */
  case 1:
    close_socketptr(server_socket_ptr);
    leave_server();
  }

  for (i=0; i < nr_servers; i++) {
    xfree(serverhost[i]);
    serverhost[i] = NULL;
  }
  nr_servers = 0;
  currentserver = -1;
  connection_state = 0;
  cycles = 0;
  server_socket_ptr = -1;
}

void save_servers(FILE *fp) {

  int i;

  fwriteint(fp, nr_servers);
  
  for (i=0; i<nr_servers; i++) {
    fwritestr2bin(serverhost[i], fp);
    fwriteint(fp, portnumber[i]);
  }
}

int  load_servers(FILE *fp) {

  int i;

  delete_all_servers();
  freadint(fp, &nr_servers);

  for(i=0; i<nr_servers; i++) {
    fgetstrfrombin(&serverhost[i], fp);
    freadint(fp, &portnumber[i]);
  }
  if (nr_servers > 0) gotoserver(0);
  return 1;
}

void delete_servermsgs();

void init_server() {

  connection_state = 2;
  send_now_to_server("USER %s null null :%s", login, ircname);
  send_now_to_server("NICK :%s", botnick);
  strcpy(current_nick, botnick);
  console_print_status();        /* Update the console */
}

void leave_server() {

  delete_all_channels();
  console_print_status();     /* Update the console */
  delete_all_sessions();
  delete_servermsgs();
}


int setserver(char *host, int port) {
 
  if (nr_servers >= MAXSERVERS)
    return 0;
 
  serverhost[nr_servers] = (char *) xmalloc (strlen(host) + 1);
  strcpy(serverhost[nr_servers], host);
  portnumber[nr_servers] = port;
  nr_servers = nr_servers + 1;
  return 1;
}


void nextserver() {
 
  currentserver = currentserver + 1;
  if (currentserver == nr_servers) {
    currentserver = 0;
    cycles = cycles + 1;
    if (cycles > maxcycles) {
      fprintf(stderr, "All server-connections failed.\n");
      exit(1);
    }
  }
  if (connection_state != 0) { close_socketptr(server_socket_ptr); server_socket_ptr = -1; }
  connection_state = 0;
  leave_server();
  switch (start_connecting(T_SERVER, &server_socket_ptr, serverhost[currentserver], portnumber[currentserver])) {
  case -1: nextserver(); return nextserver();
  case  0: connection_state = 1; return;
  case  1: connection_state = 2; init_server(); return;
  default: int_err("Strange return value."); nextserver(); return; 
  }
}

void server_lost() {

  connection_state = 0;
  server_socket_ptr = -1;
  leave_server();
  nextserver();
}


void printservers(void output(char *, ...)) {

  int i;

  for (i=0; i < nr_servers; i++) {
    output("%s%4d %40s %5d", currentserver == i ? "" : "", i, serverhost[i], portnumber[i]);
  }
}


int gotoserver(int serv) {

  if ((serv < 0) || (serv >= nr_servers)) return 0;
  currentserver = serv;
  if (connection_state != 0) {
    send_now_to_server("QUIT :Going to server %s", serverhost[serv]);
    console_send2ircwindow("*** Signoff: %s (Going to server %s)", current_nick, serverhost[serv]);
    close_socketptr(server_socket_ptr); 
    server_socket_ptr = -1;
  }
  connection_state = 0;
  leave_server();
  switch (start_connecting(T_SERVER, &server_socket_ptr, serverhost[currentserver], portnumber[currentserver])) {
  case -1: nextserver(); return 1;
  case  0: connection_state = 1; return 1;
  case  1: connection_state = 2; init_server(); return 1;
  default: int_err("Strange return value."); nextserver(); return 1;
  }
}


int delserver(int serv) {

  if ((serv == currentserver) || (serv < 0) || (serv >= nr_servers)) return 0;

  nr_servers = nr_servers - 1;
  xfree(serverhost[serv]);
  serverhost[serv] = NULL;
  if (nr_servers == serv) return 1;

  serverhost[serv] = (char *) xmalloc (strlen(serverhost[nr_servers]) + 1);
  strcpy(serverhost[serv], serverhost[nr_servers]);
  xfree(serverhost[nr_servers]);
  serverhost[nr_servers] = NULL;
  portnumber[serv] = portnumber[nr_servers];
  if (currentserver == nr_servers) currentserver = serv;
  return 1;
}

struct server_queue {
  struct server_msg *first;
  struct server_msg *last;
};

struct server_msg {
  char *string;
  char  fast;
  struct server_msg *next;
};

static time_t last_msg_sent_to_server=0;

static struct server_queue normal_queue = { NULL , NULL };
static struct server_queue   fast_queue = { NULL , NULL };


void send_to_server(char *format, ...) {

  time_t  cur_time;
  char    line[WAYTOBIG_OUT];
  va_list msg;

  if (connection_state != 2) return;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  cur_time = time_null;
  if ((cur_time - last_msg_sent_to_server > TIMEBETWEENMSG) && (normal_queue.first == NULL) && 
      (fast_queue.first == NULL)) {
    send_to_socketptr(server_socket_ptr, "%s", line);
    last_msg_sent_to_server = cur_time;
  }
  else {
    if (normal_queue.last) {
      normal_queue.last->next = (struct server_msg *) xmalloc (sizeof(struct server_msg));
      normal_queue.last = normal_queue.last->next;
    }
    else {
      normal_queue.first = normal_queue.last = (struct server_msg *) xmalloc (sizeof(struct server_msg));
    }
    normal_queue.last->string = (char *) xmalloc (strlen(line) + 1);
    strcpy(normal_queue.last->string, line);
    normal_queue.last->next = NULL;
  }
}


void send_fast_to_server(char *format, ...) {

  time_t  cur_time;
  char    line[WAYTOBIG_OUT];
  va_list msg;

  if (connection_state != 2) return;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  cur_time = time_null;
  if ((cur_time - last_msg_sent_to_server > TIMEBETWEENMSG) && (fast_queue.first == NULL)) {
    send_to_socketptr(server_socket_ptr, "%s", line);
    last_msg_sent_to_server = cur_time;
  }
  else {
    if (fast_queue.last) {
      fast_queue.last->next = (struct server_msg *) xmalloc (sizeof(struct server_msg));
      fast_queue.last = fast_queue.last->next;
    }
    else {
      fast_queue.first = fast_queue.last = (struct server_msg *) xmalloc (sizeof(struct server_msg));
    }
    fast_queue.last->string = (char *) xmalloc (strlen(line) + 1);
    strcpy(fast_queue.last->string, line);
    fast_queue.last->next = NULL;
  }
}


void send_now_to_server(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  if (connection_state != 2) return;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  send_to_socketptr(server_socket_ptr, "%s", line);
  last_msg_sent_to_server = time_null;
}

int sendnextmessage() {

  struct server_msg *tmp;

  if (fast_queue.first) {
    if (fast_queue.first == fast_queue.last)
      fast_queue.last = NULL;
    send_to_socketptr(server_socket_ptr, "%s", fast_queue.first->string);
    last_msg_sent_to_server = time_null;
    xfree (fast_queue.first->string);
    tmp = fast_queue.first->next;
    xfree (fast_queue.first);
    fast_queue.first = tmp;
    return 1;
  }
  if (normal_queue.first) {
    if (normal_queue.first == normal_queue.last)
      normal_queue.last = NULL;
    send_to_socketptr(server_socket_ptr, "%s", normal_queue.first->string);
    last_msg_sent_to_server = time_null;
    xfree (normal_queue.first->string);
    tmp = normal_queue.first->next;
    xfree (normal_queue.first);
    normal_queue.first = tmp;
    return 1;
  }
  return 0;
}

int server_has_queue() {

  return (int)(fast_queue.first || normal_queue.first);
}

void delete_servermsgs() {

  struct server_msg *tmp;

  while (fast_queue.first) {
    if (fast_queue.first == fast_queue.last)
      fast_queue.last = NULL;
    xfree (fast_queue.first->string);
    tmp = fast_queue.first->next;
    xfree (fast_queue.first);
    fast_queue.first = tmp;
  }
  while (normal_queue.first) {
    if (normal_queue.first == normal_queue.last)
      normal_queue.last = NULL;
    xfree (normal_queue.first->string);
    tmp = normal_queue.first->next;
    xfree (normal_queue.first);
    normal_queue.first = tmp;
  }
}
