
/*********************************************************/
/**                                                     **/
/**       sescmd.c (c) 1996, 97 Emiel Bruijntjes        **/
/**                                                     **/
/**      Parsing commands entered by user from IRC      **/
/**                                                     **/
/*********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <dirent.h>
#include <strings.h>


#include "sescmd.h"
#include "misc.h"
#include "session.h"
#include "lines.h"
#include "userlist.h"
#include "gencmd.h"
#include "config.h"
#include "partyline.h"
#include "bots.h"
#include "notes.h"
#include "channel.h"
#include "bottree.h"
#include "reg.h"
#include "chanuser.h"
#include "server.h"
#include "assoc.h"
#include "console.h"
#include "on.h"


extern char  *helpdir;
extern char  *whoisdir;
extern char   current_nick[10];
extern char  *su_password;
extern char   botnick[10];
extern int    default_irc_partyline;
extern char  *ftp_root_dir;
extern char  *motd_file;

void cmd_help(char *argv[], int argc) {

  char *file;
  int i;

  struct {
    char *command;
    char *file;
  } helpfile_argc1[] = {
    {"act", "help.act"},
    {"addbot", "help.addbot"},
    {"addchan", "help.addchan"},
    {"addscript", "help.addscript"},
    {"addserv", "help.addserv"},
    {"aline", "help.aline"},
    {"allow", "help.allow"},
    {"assoc", "help.assoc"},
    {"away", "help.away"},
    {"ban", "help.ban"},
    {"bankick", "help.bankick"},
    {"bkick", "help.bkick"},
    {"bline", "help.bline"},
    {"boot", "help.boot"},
    {"botac", "help.botac"},
    {"botaddr", "help.botaddr"},
    {"botinfo", "help.botinfo"},
    {"botlist", "help.botlist"},
    {"botnuh", "help.botnuh"},
    {"botpass", "help.botpass"},
    {"botprot", "help.botprot"},
    {"bots", "help.bots"},
    {"bottree", "help.bottree"},
    {"cd", "help.cd"},
    {"channels", "help.channels"},
    {"chat", "help.chat"},
    {"cheat", "help.cheat"},
    {"cmdreq", "help.cmdreq"},
    {"cmds", "help.cmds"},
    {"comment", "help.comment"},
    {"cons", "help.cons"},
    {"delbot", "help.delbot"},
    {"delchan", "help.delchan"},
    {"delline", "help.delline"},
    {"delscript", "help.delscript"},
    {"delserv", "help.delserv"},
    {"deop", "help.deop"},
    {"dir", "help.dir"},
    {"dline", "help.dline"},
    {"echo", "help.echo"},
    {"edit", "help.edit"},
    {"files", "help.files"},
    {"get", "help.get"},
    {"h", "help.h"},
    {"hangman", "help.hangman"},
    {"help", "help.help"},
    {"hour", "help.hour"},
    {"http", "help.http"},
    {"i", "help.i"},
    {"identify", "help.identify"},
    {"iline", "help.iline"},
    {"info", "help.info"},
    {"invite", "help.invite"},
    {"join", "help.join"},
    {"keep", "help.keep"},
    {"kick", "help.kick"},
    {"leave", "help.leave"},
    {"linepass", "help.linepass"},
    {"linereq", "help.linereq"},
    {"link", "help.link"},
    {"listen", "help.listen"},
    {"lock", "help.lock"},
    {"ls", "help.ls"},
    {"manip", "help.manip"},
    {"maskinfo", "help.maskinfo"},
    {"massdeop", "help.massdeop"},
    {"masskick", "help.masskick"},
    {"massop", "help.massop"},
    {"massunban", "help.massunban"},
    {"me", "help.me"},
    {"motd", "help.motd"},
    {"mycmd", "help.mycmd"},
    {"mycmds", "help.mycmds"},
    {"nline", "help.nline"},
    {"note", "help.note"},
    {"notes", "help.notes"},
    {"oline", "help.oline"},
    {"op", "help.op"},
    {"oper", "help.oper"},
    {"party", "help.party"},
    {"partycmd", "help.partycmd"},
    {"passwd", "help.passwd"},
    {"protkey", "help.protkey"},
    {"pwd", "help.pwd"},
    {"quit", "help.quit"},
    {"rehash", "help.rehash"},
    {"resassoc", "help.resassoc"},
    {"reset", "help.reset"},
    {"say", "help.say"},
    {"scripts", "help.scripts"},
    {"seen", "help.seen"},
    {"send", "help.send"},
    {"server", "help.server"},
    {"set", "help.set"},
    {"signoff", "help.signoff"},
    {"spy", "help.spy"},
    {"stats", "help.stats"},
    {"su", "help.su"},
    {"tcl", "help.tcl"},
    {"time", "help.time"},
    {"top", "help.top"},
    {"trace", "help.trace"},
    {"tryme", "help.tryme"},
    {"unban", "help.unban"},
    {"unkeep", "help.unkeep"},
    {"unlink", "help.unlink"},
    {"unlock", "help.unlock"},
    {"useradd", "help.useradd"},
    {"userdel", "help.userdel"},
    {"userlist", "help.userlist"},
    {"who", "help.who"},
    {"whoami", "help.whoami"},
    {"whois", "help.whois"},
    {"xline", "help.xline"},
    {"xwin", "help.xwin"},
    {NULL,   NULL }
  };
  struct {
    char *command1;
    char *command2;
    char *file;
  } helpfile_argc2[] = {
    {"allow", "upload", "help.allow.upload"},
    {"allow", "alwaysop", "help.allow.alwaysop"},
    {"allow", "ban", "help.allow.ban"},
    {"allow", "autoop", "help.allow.autoop"},
    {"allow", "getops", "help.allow.getops"},
    {"allow", "giveops", "help.allow.giveops"},
    {"allow", "unban", "help.allow.unban"},
    {"set", "userlist_length", "help.set.userlist_length"},
    {"set", "su_passwd", "help.set.su_passwd"},
    {"set", "help_dir", "help.set.help_dir"},
    {"set", "whois_dir", "help.set.whois_dir"},
    {"set", "maintainer", "help.set.maintainer"},
    {"set", "telnet_party_line", "help.set.telnet_party_line"},
    {"set", "http_dir", "help.set.http_dir"},
    {"set", "wordfile", "help.set.wordfile"},
    {"set", "root_dir", "help.set.root_dir"},
    {"set", "cycles", "help.set.cycles"},
    {"set", "irc_party_line", "help.set.irc_party_line"},
    {"set", "notes_file", "help.set.notes_file"},
    {"set", "warnings", "help.set.warnings"},
    {"set", "dcc_block_size", "help.set.dcc_block_size"},
    {"set", "login", "help.set.login"},
    {"set", "aline_timeout", "help.set.aline_timeout"},
    {"set", "bline_timeout", "help.set.bline_timeout"},
    {"set", "dline_timeout", "help.set.dline_timeout"},
    {"set", "hline_timeout", "help.set.hline_timeout"},
    {"set", "iline_timeout", "help.set.iline_timeout"},
    {"set", "oline_timeout", "help.set.oline_timeout"},
    {"set", "xline_timeout", "help.set.xline_timeout"},
    {"set", "nline_timeout", "help.set.nline_timeout"},
    {"set", "maxdccgetfilesize", "help.set.maxdccgetfilesize"},
    {"set", "commandchar1", "help.set.commandchar1"},
    {"set", "commandchar2", "help.set.commandchar2"},
    {"set", "commandchar3", "help.set.commandchar3"},
    {"set", "incoming_dir", "help.set.incoming_dir"},
    {"set", "ircname", "help.set.ircname"},
    {"set", "http_logfile", "help.set.http_logfile"},
    {"set", "irc_logfile", "help.set.irc_logfile"},
    {"set", "party_logfile", "help.set.party_logfile"},
    {"set", "session_logfile", "help.set.session_logfile"},
    {"set", "notefilesize", "help.set.notefilesize"},
    {"set", "notetimeout", "help.set.notetimeout"},
    {"set", "motd_file", "help.set.motd_file"},
    {"set", "xfont", "help.set.xfont"},
    {"set", "flood", "help.set.flood"},
    { NULL,    NULL, NULL }
  };

  if (argc == 1) {
    file = (char *)xmalloc(strlen(helpdir) + strlen("help") + 1);
    sprintf(file, "%shelp", helpdir);
    catfile(file, session_sendnotice, session_nick());
    xfree(file);
    return;
  }

  if (argc == 2) {
    for (i = 0; helpfile_argc1[i].command; i++) {
      if (strcasecmp(argv[1], helpfile_argc1[i].command) == 0) {
	file = (char *)xmalloc(strlen(helpdir) + strlen(helpfile_argc1[i].file) + 1);
	sprintf(file, "%s%s", helpdir, helpfile_argc1[i].file);
	catfile(file, session_sendnotice, session_nick());
	xfree(file);
	return;
      }
    }
    session_sendnotice("help: No such topic");
    return;
  }
  if (argc == 3) {
    for (i = 0; helpfile_argc2[i].command1; i++) {
      if ((strcasecmp(argv[1], helpfile_argc2[i].command1)==0) && (strcasecmp(argv[2], helpfile_argc2[i].command2)==0)) {
	file = (char *)xmalloc(strlen(helpdir) + strlen(helpfile_argc2[i].file) + 1);
	sprintf(file, "%s%s", helpdir, helpfile_argc2[i].file);
	catfile(file, session_sendnotice, session_nick());
	xfree(file);
	return;
      }
    }
    session_sendnotice("help: No such topic");
    return;
  }
  session_sendnotice("help: No such topic");
  return;
}

void cmd_userlist(char *argv[], int argc) {

  gen_cmd_userlist(argv, argc, session_sendnotice);
}

void cmd_top(char *argv[], int argc) {

  gen_cmd_top(argv, argc, session_sendnotice);
}


void cmd_spy(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: spy");
    return;
  }
  if (session_isspy()) {
    sendtospies("%s has stopped spying", session_nick());
    session_resetspy();
    return;
  }
  session_setspy();
  sendtospies("%s is now spying", session_nick());
}

void cmd_identify(char *argv[], int argc) {

  session_resetidentify();
  if (argc != 2) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: identify <password>");
    return;
  }
  if (!(userlist_findnick(session_nick()))) {
    sendtospies("No userlist entry");
    session_sendnotice("You do not have a userlist entry");
    return;
  }
  if (strcasecmp(userlist_password(), argv[1]) == 0) {
    sendtospies("Identification successful");
    sendtoadminspies("Password: %s", argv[1]);
    session_sendnotice("Identification successful. Protection is now active");
    session_setidentify();
    return;
  }
  sendtospies("Password incorrect");
  sendtoadminspies("Given password: %s Should be: %s", argv[1], userlist_password());
  session_sendnotice("Password incorrect");
}

void cmd_passwd(char *argv[], int argc) {

  if (argc != 2) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: identify <password>");
    return;
  }
  if (!(userlist_findnick(session_nick()))) {
    sendtospies("No userlist entry");
    session_sendnotice("You do not have a userlist entry");
    return;
  }
  if (strlen(argv[1]) > 9) {
    sendtospies("Password too long");
    session_sendnotice("A password may not contain more than nine characters");
    return;
  }
  if (index(argv[1], ':')) {
    sendtospies("Illegal character");
    session_sendnotice("\":\" is an illegal character");
    return;
  }
  userlist_setpassword(argv[1]);
  sendtospies("Password changed");
  sendtoadminspies("Password: %s", argv[1]);
  session_sendnotice("Password changed");
}

void cmd_keep(char *argv[], int argc) {

  gen_cmd_keep(argv, argc, session_sendnotice);
}

void cmd_unkeep(char *argv[], int argc) {

  gen_cmd_unkeep(argv, argc, session_sendnotice);
}

void cmd_manip(char *argv[], int argc) {

  gen_cmd_manip(argv, argc, session_sendnotice);
}

void cmd_stats(char *argv[], int argc) {

  gen_cmd_stats(argv, argc, session_sendnotice);
}

void cmd_oper(char *argv[], int argc) {

  int linenr;
  char type;

  session_resetoperator();
  if ((argc < 2) || (argc > 3)) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: oper [<line>] <password>");
    return;
  }
  if (argc == 2) {
    if (!(does_this_match('o', session_nickuserhost()))) {
      sendtospies("No matching line(s)");
      session_sendnotice("You are not matching with an oline");
      return;
    }
    if (strcasecmp(lines_comment('o'), argv[1]) == 0) {
      sendtospies("%s is now operator", session_nick());
      sendtoadminspies("Password: %s", argv[1]);
      session_sendnotice("You are now operator");
      session_setoperator();
      return;
    }
    sendtospies("Password incorrect");
    sendtoadminspies("Give password: %s Should be: %s", argv[1], lines_comment('o'));
    session_sendnotice("Password incorrect");
    return;
  }
  type = tolower(argv[1][0]);
  switch(type) {
  case 'a': case 'o':
    break;
  default:
    sendtospies("Illegal linetype");
    session_sendnotice("Illegal linetype: \"%c\"", toupper(type));
    return;
  }
  linenr = atoi(&argv[1][1]);
  if (linenr <= 0) {
    sendtospies("Illegal linenumber");
    session_sendnotice("Linenumbers should be greater than zero");
    return;
  }
  if (!(lines_findnr(type, linenr))) {
    sendtospies("%c%d: linenumber not found", toupper(type), linenr);
    session_sendnotice("Line %c%d does not exist", toupper(type), linenr);
    return;
  }
  if (!match(lines_mask(type), session_nickuserhost())) {
    sendtospies("User does not match with mask");
    session_sendnotice("Your nick!user@host does not match with %s", lines_mask(type));
    return;
  }
  if (strcasecmp(argv[2], lines_comment(type)) == 0) {
    switch(type) {
    case 'a':
      sendtospies("%s is now administrator. (line A%d)", session_nick(), linenr);
      sendtoadminspies("Password: %s", argv[2]);
      session_sendnotice("You are now administrator");
      session_setadministrator();
      return;
    case 'o':
      sendtospies("%s is now operator. (line O%d)", session_nick(), linenr);
      sendtoadminspies("Password: %s", argv[2]);
      session_sendnotice("You are now operator");
      session_setoperator();
      return;
    }
  }
  sendtospies("Password incorrect. (line %c%d)", toupper(type), linenr);
  sendtoadminspies("Given password: %s Should be: %s", argv[2], lines_comment(type));
  session_sendnotice("Password incorrect");
  return;
}
    
int check_addline_permission(char type) {

  switch(required_to_change(type)) {
  case 0: return 1;
  case 1: 
    if (session_operator()) return 1;
    session_sendnotice("You are not operator");
    return 0;
  case 2:
    if (session_administrator()) return 1;
    session_sendnotice("You are not administrator");
    return 0;
  case 3:
    if (session_superuser()) return 1;
    session_sendnotice("You are not superuser");
    return 0;
  default:
    session_sendnotice("Something seems to be wrong...");
    return 0;
  }
}


void cmd_aline(char *argv[], int argc) {

  int linenr;

  if (argc != 3) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: aline <nick!user@host mask> <password>");
    return;
  }
  if (!(check_addline_permission('a'))) return;
  linenr = addline('a', argv[1], argv[2], session_nickuserhost());
  sessions_update_added_line('a', argv[1]);
  sendtospies("A%d: new A-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  sendtoadminspies("Password: %-35s", lines_comment('a'));
  if (!(session_isspy())) {
    session_sendnotice("A%d: new A-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
    if (session_administrator())
      session_sendnotice("Password: %-35s", lines_comment('a'));
    else {
      if (strcasecmp(lines_comment('a'), argv[2]) == 0)
	session_sendnotice("Password: %-35s", argv[2]);
      else
	session_sendnotice("Password: %-35s", "<unchanged>");
    }
  }
  else {
    if (session_administrator()) return;
    if (strcasecmp(lines_comment('a'), argv[2]) == 0)
      session_sendnotice("Password: %-35s", argv[2]);
    else
      session_sendnotice("Password: %-35s", "<unchanged>");
  }
}

void cmd_oline(char *argv[], int argc) {

  int linenr;

  if (argc != 3) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: oline <nick!user@host mask> <password>");
    return;
  }
  if (!(check_addline_permission('o'))) return;
  linenr = addline('o', argv[1], argv[2], session_nickuserhost());
  sessions_update_added_line('o', argv[1]);
  sendtospies("O%d: new O-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  sendtoadminspies("Password: %-35s", lines_comment('o'));
  if (!(session_isspy())) {
    session_sendnotice("O%d: new O-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
    if (session_administrator())
      session_sendnotice("Password: %-35s", lines_comment('o'));
    else {
      if (strcasecmp(lines_comment('o'), argv[2]) == 0)
	session_sendnotice("Password: %-35s", argv[2]);
      else
	session_sendnotice("Password: %-35s", "<unchanged>");
    }
  }
  else {
    if (session_administrator()) return;
    if (strcasecmp(lines_comment('o'), argv[2]) == 0)
      session_sendnotice("Password: %-35s", argv[2]);
    else
      session_sendnotice("Password: %-35s", "<unchanged>");
  }
}
  
void cmd_bline(char *argv[], int argc) {

  int linenr;

  if (argc != 2) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: bline <nick!user@host mask>");
    return;
  }
  if (!(check_addline_permission('b'))) return;
  linenr = addline('b', argv[1], session_nickuserhost(), session_nickuserhost());
  sessions_update_added_line('b', argv[1]);
  sendtospies("B%d: new B-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  if (!(session_isspy())) {
    session_sendnotice("B%d: new B-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
  }
  return;
}

void cmd_dline(char *argv[], int argc) {

  int linenr;

  if (argc != 2) {
    session_sendnotice("Usage: dline <nick!user@host mask>");
    return;
  }
  if (!(check_addline_permission('d'))) return;
  linenr = addline('d', argv[1], session_nickuserhost(), session_nickuserhost());
  sessions_update_added_line('d', argv[1]);
  sendtospies("D%d: new D-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  if (!(session_isspy())) {
    session_sendnotice("D%d: new D-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
  }
  return;
}

void cmd_iline(char *argv[], int argc) {

  int linenr;

  if (argc != 2) {
    session_sendnotice("Usage: iline <nick!user@host mask>");
    return;
  }
  if (!(check_addline_permission('i'))) return;
  linenr = addline('i', argv[1], session_nickuserhost(), session_nickuserhost());
  sessions_update_added_line('i', argv[1]);
  sendtospies("I%d: new I-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  if (!(session_isspy())) {
    session_sendnotice("I%d: new I-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
  }
  return;
}

void cmd_xline(char *argv[], int argc) {

  int linenr;

  if (argc != 2) {
    session_sendnotice("Usage: xline <host mask>");
    return;
  }
  if (!(check_addline_permission('x'))) return;
  linenr = addline('x', argv[1], "", session_nickuserhost());
  sendtospies("X%d: new X-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  if (!(session_isspy())) {
    session_sendnotice("X%d: new X-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
  }
  return;
}

void cmd_nline(char *argv[], int argc) {

  int linenr;

  if (argc != 2) {
    session_sendnotice("Usage: nline <host mask>");
    return;
  }
  if (!(check_addline_permission('n'))) return;
  linenr = addline('n', argv[1], session_nickuserhost(), session_nickuserhost());
  sendtospies("N%d: new N-line", linenr);
  sendtospies("Mask:     %-35s", argv[1]);
  if (!(session_isspy())) {
    session_sendnotice("N%d: new N-line", linenr);
    session_sendnotice("Mask:     %-35s", argv[1]);
  }
  return;
}

void cmd_maskinfo(char *argv[], int argc) {

  gen_cmd_maskinfo(argv, argc, session_sendnotice, session_administrator());
}

void cmd_useradd(char *argv[], int argc) {

  char type;
  char *nickuserhost;
  char *nick;
  int i=0;
  char mask[MAXNICKUSERHOSTLEN];
  int linenr;

  if ((argc < 3) || (argc > 4)) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: useradd <nick> <lintypes> [<password>]");
    return;
  }
  nick = session_nick();
  if (!(session_findnick(argv[1]))) {
    session_findnick(nick);
    sendtospies("nick!user@host is not known");
    session_sendnotice("I do not know the nick!user@host of %s", argv[1]);
    return;
  }
  nickuserhost = session_nickuserhost();
  session_findnick(nick);
  maakbanmatch(nickuserhost, mask);
  while (argv[2][i] != '\0') {
    type = tolower(argv[2][i]);
    switch(type) {
    case 'a': case 'o':
      if (argc != 4) {
	sendtospies("No password for %c-line specified", toupper(type));
	session_sendnotice("Please specify a password when adding %c-lines", toupper(type));
	break;
      }
      if (does_this_match(type, nickuserhost)) {
	sendtospies("Already matching");
	session_sendnotice("%s is already matching with an %c-line", argv[1], toupper(type));
	break;
      }
      if (!(check_addline_permission(type))) break;
      linenr = addline(type, mask, argv[3], session_nickuserhost());
      sessions_update_added_line(type, mask);
      sendtospies("%c%d: new %c-line", toupper(type), linenr, toupper(type));
      sendtospies("Mask:     %-35s", mask);
      sendtoadminspies("Password: %-35s", lines_comment(type));
      if (!(session_isspy())) {
	session_sendnotice("%c%d: new %c-line", toupper(type), linenr, toupper(type));
	session_sendnotice("Mask:     %-35s", mask);
	if (session_administrator())
	  session_sendnotice("Password: %-35s", lines_comment(type));
	else {
	  if (strcasecmp(lines_comment(type), argv[3]) == 0)
	    session_sendnotice("Password: %-35s", argv[3]);
	  else
	    session_sendnotice("Password: %-35s", "<unchanged>");
	}
      }
      else {
	if (session_administrator()) break;
	if (strcasecmp(lines_comment(type), argv[3]) == 0)
	  session_sendnotice("Password: %-35s", argv[3]);
	else
	  session_sendnotice("Password: %-35s", "<unchanged>");
      }
      break;
    case 'b': case 'd': case 'i':
      if (does_this_match(type, nickuserhost)) {
	sendtospies("Already matching");
	session_sendnotice("%s is already matching with an %c-line", argv[1], toupper(type));
	break;
      }
      if (does_this_match('a', nickuserhost)) {
	sendtospies("%c-line for administrator. Failed", toupper(type));
	session_sendnotice("%s is matching with an A-line", argv[1]);
	break;
      }
      if (!(check_addline_permission(type))) break;
      linenr = addline(type, mask, session_nickuserhost(), session_nickuserhost());
      sessions_update_added_line(type, mask);
      sendtospies("%c%d: new %c-line", toupper(type), linenr, toupper(type));
      sendtospies("Mask:     %-35s", mask);
      if (!(session_isspy())) {
	session_sendnotice("%c%d: new %c-line", toupper(type), linenr, toupper(type));
	session_sendnotice("Mask:     %-35s", mask);
      }
      break;
    case 'h': case 'x': case 'n':
      session_sendnotice("It's not possible to create %c-lines with this command", toupper(type));
      break;
    default:
      session_sendnotice("Unknown line-type: \"%c\"", toupper(type));
      break; 
    }
    i++;
  }
}

int check_lock_permission(char type) {

  switch(required_to_change_when_locked(type)) {
  case 0: return 1;
  case 1: 
    if (session_operator()) return 1;
    session_sendnotice("You are not operator");
    return 0;
  case 2:
    if (session_administrator()) return 1;
    session_sendnotice("You are not administrator");
    return 0;
  case 3:
    if (session_superuser()) return 1;
    session_sendnotice("You are not superuser");
    return 0;
  default:
    session_sendnotice("Something seems to be wrong...");
    return 0;
  }
}


void cmd_lock(char *argv[], int argc) {

  char type;
  int linenr;

  if (argc != 2) {
    session_sendnotice("Usage: lock <linenumber>");
    return;
  }
  type = tolower(argv[1][0]);
  switch(type) {
  case 'a': case 'b': case 'd': case 'i': case 'o': case 'x': case 'n':
    linenr = atoi(&argv[1][1]);
    if (linenr < 1) {
      session_sendnotice("A linenumber should be at least one");
      return;
    }
    if (!(check_lock_permission(type))) return;
    if (!(lines_findnr(type, linenr))) {
      session_sendnotice("Line %c%d not found", toupper(type), linenr);
      return;
    }
    lines_setlock(type);
    sendtospies("Line %c%d is locked", toupper(type), linenr);
    sendtospies("Mask: %s", lines_mask(type));
    if (!(session_isspy())) {
      session_sendnotice("Line %c%d is locked", toupper(type), linenr);
      session_sendnotice("Mask: %s", lines_mask(type));
    }
    return;
  case 'h':
    session_sendnotice("You are not allowed to lock H-lines");
    return;
  default:
    session_sendnotice("Unknown linetype: \"%c\"", toupper(type));
    return;
  }
}

void cmd_unlock(char *argv[], int argc) {

  char type;
  int linenr;

  if (argc != 2) {
    session_sendnotice("Usage: unlock <linenumber>");
    return;
  }
  type = tolower(argv[1][0]);
  switch(type) {
  case 'a': case 'b': case 'd': case 'i': case 'o': case 'x': case 'n':
    linenr = atoi(&argv[1][1]);
    if (linenr < 1) {
      session_sendnotice("A linenumber should be at least one");
      return;
    }
    if (!(check_lock_permission(type))) return;
    if (!(lines_findnr(type, linenr))) {
      session_sendnotice("Line %c%d not found", toupper(type), linenr);
      return;
    }
    lines_setunlock(type);
    sendtospies("Line %c%d is unlocked", toupper(type), linenr);
    sendtospies("Mask: %s", lines_mask(type));
    if (!(session_isspy())) {
      session_sendnotice("Line %c%d is unlocked", toupper(type), linenr);
      session_sendnotice("Mask: %s", lines_mask(type));
    }
    return;
  case 'h':
    session_sendnotice("You are not allowed to unlock H-lines");
    return;
  default:
    session_sendnotice("Unknown linetype: \"%c\"", toupper(type));
    return;
  }
}

void cmd_comment(char *argv[], int argc) {

  char type;
  int  linenr;

  if (argc < 3) {
    session_sendnotice("Usage: comment <linenumber> <comment>");
    return;
  }
  type = tolower(argv[1][0]);
  switch(type) {
  case 'a': case 'o': case 'h': case 'x':
    session_sendnotice("You can't set the comment-field of %c-lines", toupper(type));
    return;
  case 'b': case 'd': case 'i': case 'n':
    linenr = atoi(&argv[1][1]);
    if (linenr < 1) {
      session_sendnotice("Linenumbers should be at least one");
      return;
    }
    if (!(lines_findnr(type, linenr))) {
      session_sendnotice("Line %c%d not found", toupper(type), linenr);
      return;
    }
    if (lines_islocked(type))
      if (!(check_lock_permission(type))) return;
    else
      if (!(check_addline_permission(type))) return;
    repairstring(&argv[2], argc - 2);
    lines_setcomment(type, argv[2]);
    sendtospies("New comment for line %c%d", toupper(type), linenr);
    sendtospies("Mask:    %s", lines_mask(type));
    sendtospies("Comment: %s", lines_comment(type));
    if (session_isspy()) return;
    session_sendnotice("New comment for line %c%d", toupper(type), linenr);
    session_sendnotice("Mask:    %s", lines_mask(type));
    session_sendnotice("Comment: %s", lines_comment(type));
    return;
  default:
    session_sendnotice("Unknown line-type: \"%c\"", toupper(type));
    return;
  }
}

void cmd_linepass(char *argv[], int argc) {

  char type;
  int  linenr;

  if (argc != 3) {
    session_sendnotice("Usage: linepass <linenumber> <password>");
    return;
  }
  type = tolower(argv[1][0]);
  switch(type) {
  case 'b': case 'd': case 'i': case 'h': case 'n':
    session_sendnotice("You can't set the password-field of %c-lines", toupper(type));
    return;
  case 'o': case 'a': case 'x':
    linenr = atoi(&argv[1][1]);
    if (linenr < 1) {
      session_sendnotice("Linenumbers should be at least one");
      return;
    }
    if (!(lines_findnr(type, linenr))) {
      session_sendnotice("Line %c%d not found", toupper(type), linenr);
      return;
    }
    if (lines_islocked(type))
      if (!(check_lock_permission(type))) return;
    else
      if (!(check_addline_permission(type))) return;
    lines_setcomment(type, argv[2]);
    sendtospies("New password for line %c%d", toupper(type), linenr);
    sendtospies("Mask:    %s", lines_mask(type));
    sendtoadminspies("Password: %s", lines_comment(type));
    if (session_isspy()) return;
    session_sendnotice("New password for line %c%d", toupper(type), linenr);
    session_sendnotice("Mask:    %s", lines_mask(type));
    session_sendnotice("Password: %s", lines_comment(type));
    return;
  default:
    session_sendnotice("Unknown line-type: \"%c\"", toupper(type));
    return;
  }
}

void cmd_userdel(char *argv[], int argc) {

  char *nickuserhost, *nick;
  char  type;
  int   i=0, linenr;

  if (argc != 3) {
    session_sendnotice("Usage: userdel <nick> <linetypes>");
    return;
  }
  nick = session_nick();
  if (!(session_findnick(argv[1]))) {
    session_findnick(nick);
    session_sendnotice("I do not know the nick!user@host of %s", argv[1]);
    return;
  }
  nickuserhost = session_nickuserhost();
  session_findnick(nick);
  while (argv[2][i] != '\0') {
    type = tolower(argv[2][i]);
    switch (type) {
    case 'a': case 'b': case 'd': case 'i': case 'o': case 'x': case 'n':
      linenr = 1;
      while ((linenr = next_match(type, nickuserhost, linenr)) != 0) {
	if (lines_islocked(type)) {
	  if (!(check_lock_permission(type))) {
	    linenr++;
	    continue;
	  }
	}
	else {
	  if (!(check_addline_permission(type))) {
	    linenr++;
	    continue;
	  }
	}
	sendtospies("Line %c%d has been moved to the H-lines", toupper(type), linenr);
	sendtospies("Mask: %s", lines_mask(type));
	if (!(session_isspy())) {
	  session_sendnotice("Line %c%d has been moved to the H-lines", toupper(type), linenr);
	  session_sendnotice("Mask: %s", lines_mask(type));
	}
	moveline('h', type, type == 'a' ? 0 : type == 'b' ? 1 : type == 'd' ? 2 : type == 'i' ? 3 : type == 'o' ? 4 : 
		 type == 'x' ? 5 : 6, session_nickuserhost());
	if (type != 'x' && type != 'n')
	  sessions_update_deleted_line(type, lines_mask('h'));
      }
      break;
    case 'h':
      session_sendnotice("You are not allowed to remove H-lines");
      break;
    default:
      session_sendnotice("Unknown linetype: \"%c\"", toupper(type));
      break;
    }
    i++;
  }
}

void cmd_delline(char *argv[], int argc) {

  char type;
  int  linenr;

  if (argc != 2) {
    session_sendnotice("Usage: delline <linenumber>");
    return;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'o': case 'b': case 'd': case 'i': case 'x': case 'n':
    if (linenr < 1) {
      session_sendnotice("Linenumbers should be at least one");
      return;
    }
    if (!(lines_findnr(type, linenr))) {
      session_sendnotice("Line %c%d not found", toupper(type), linenr);
      return;
    }
    if (lines_islocked(type)) {
      if (!(check_lock_permission(type))) return;
    }
    else {
      if (!(check_addline_permission(type))) return;
    }
    sendtospies("Line %c%d has been moved to the H-lines", toupper(type), linenr);
    sendtospies("Mask: %s", lines_mask(type));
    if (!(session_isspy())) {
      session_sendnotice("Line %c%d has been moved to the H-lines", toupper(type), linenr);
      session_sendnotice("Mask: %s", lines_mask(type));
    }
    moveline('h', type, type == 'a' ? 0 : type == 'b' ? 1 : type == 'd' ? 2 : type == 'i' ? 3 : type == 'o' ? 4 : 
	     type == 'x' ? 5 : 6, session_nickuserhost());
    if (type != 'x' && type != 'n')
      sessions_update_deleted_line(type, lines_mask('h'));
    return;
  case 'h':
    session_sendnotice("You are not allowed to remove H-lines");
    return;
  default:
    session_sendnotice("Unknown linetype: \"%c\"", toupper(type));
    return;
  }
}

void cmd_whois(char *argv[], int argc) {

  char *nick;
  char *nickuserhost;
  char *whoisfile;
  char tmpstr[12];
  char *chans;

  if (argc != 2) {
    session_sendnotice("Usage: whois <nick>");
    return;
  }
  if((strlen(argv[1]) > 9) || (argv[1][0] == '.') || (argv[1][0] == '/')) {
    session_sendnotice("Illegal nick");
    return;
  }
  if (userlist_findnick(argv[1])) {
    session_sendnotice("Nick: %36s %s", userlist_nick(), userlist_keepentry() ? "-=keeping=-" : "");
    session_sendnotice("Last seen: %31s", time2string(userlist_lastseen()));
    session_sendnotice("From: %36s", (userlist_lasthost()==NULL) ? "" : userlist_lasthost());
    session_sendnotice("-------------------------------------------------------");
    session_sendnotice("Publics: %8d Kicks: %10d Quits: %12d", userlist_publics(), 
		       userlist_kicks(), userlist_quits());
    session_sendnotice("Kicked: %9d Joins: %10d Modes: %12d", userlist_kickeds(),
		       userlist_joins(), userlist_modes());
    session_sendnotice("Leaves: %9d Nicks: %10d Ontime: %8d:%02d", userlist_leaves(),
		       userlist_nicks(), userlist_ontime()/60, userlist_ontime()%60);
    session_sendnotice("Manip: %10d Protection: %5d", userlist_manipulation(), userlist_protection());
  }
  else {
    session_sendnotice("Nick: %31s (Not in userlist)", argv[1]);
  }
  session_sendnotice("-------------------------------------------------------");
  nick = session_nick();
  if (session_findnick(argv[1])) {
    tmpstr[0]  = session_superuser();
    tmpstr[1]  = session_administrator();
    tmpstr[2]  = session_operator();
    tmpstr[3]  = session_identified();
    tmpstr[4]  = session_isspy();
    tmpstr[5]  = session_isaway() ? 1 : 0;
    tmpstr[6]  = session_chatstatus() == 3;
    tmpstr[7]  = session_sendstatus() == 2;
    tmpstr[8]  = session_getstatus()  == 2;
    tmpstr[9]  = (session_chatstatus() == 1) || (session_chatstatus() == 2);
    tmpstr[10] = session_sendstatus() == 1;
    tmpstr[11] = session_getstatus()  == 1;
    chans = session_currentchannels();
  }
  else {
    tmpstr[0] = tmpstr[1] = tmpstr[2] = tmpstr[3] = tmpstr[4] = tmpstr[5] = tmpstr[6] =
      tmpstr[7] = tmpstr[8] = tmpstr[9] = tmpstr[10] = tmpstr[11] = 0;
    chans=NULL;
  }
  session_findnick(nick);
  session_sendnotice("[%c] superuser     [%c] administrator [%c] operator", tmpstr[0] ? 'X' : ' ',
		     tmpstr[1] ? 'X' : ' ', tmpstr[2] ? 'X' : ' ');
  session_sendnotice("[%c] identified    [%c] spy           [%c] away", tmpstr[3] ? 'X' : ' ',
		     tmpstr[4] ? 'X' : ' ', tmpstr[5] ? 'X' : ' ');
  session_sendnotice("[%c] CHAT active   [%c] SEND active   [%c] GET active", tmpstr[6] ? 'X' : ' ',
		     tmpstr[7] ? 'X' : ' ', tmpstr[8] ? 'X' : ' ');
  session_sendnotice("[%c] CHAT request  [%c] SEND request  [%c] GET request", tmpstr[9] ? 'X' : ' ',
		     tmpstr[10] ? 'X' : ' ', tmpstr[11] ? 'X' : ' ');
  session_sendnotice("-------------------------------------------------------");
  tmpstr[0] = 0;
  if (user_has_note(argv[1])) {
    session_sendnotice("%s has unread note(s)", argv[1]);
    tmpstr[0] = 1;
  }
  if ((chans != NULL) && (chans[0] != '\0')) {
    session_sendnotice("Party channels: %s (IRC)", chans);
    tmpstr[0] = 1;
  }
  if ((party_findnick(argv[1])) && (party_currentchannel() != -1)) {
    session_sendnotice("Party channels: %s (outside)", party_currentchannels());
    session_sendnotice("From:           %s", party_from());
    tmpstr[0] = 1;
  }
  if (tmpstr[0])
    session_sendnotice("-------------------------------------------------------");
  nick = session_nick();
  if (session_findnick(argv[1])) {
    nickuserhost = session_nickuserhost();
    session_findnick(nick);
    session_sendnotice("Matching lines:");
    printmatchinglines('a', nickuserhost, session_sendnotice);
    printmatchinglines('b', nickuserhost, session_sendnotice);
    printmatchinglines('d', nickuserhost, session_sendnotice);
    printmatchinglines('h', nickuserhost, session_sendnotice);
    printmatchinglines('i', nickuserhost, session_sendnotice);
    printmatchinglines('o', nickuserhost, session_sendnotice);
    printmatchinglines('x', nickuserhost, session_sendnotice);
    printmatchinglines('n', nickuserhost, session_sendnotice);
    session_sendnotice("-------------------------------------------------------");
  }
  else {
    session_findnick(nick);
    if ((userlist_findnick(argv[1])) && (userlist_lasthost())) {
      nickuserhost = (char *)xmalloc(strlen(userlist_nick()) + strlen(userlist_lasthost()) + 2);
      sprintf(nickuserhost, "%s!%s", userlist_nick(), userlist_lasthost());
      session_sendnotice("Matching lines:");
      printmatchinglines('a', nickuserhost, session_sendnotice);
      printmatchinglines('b', nickuserhost, session_sendnotice);
      printmatchinglines('d', nickuserhost, session_sendnotice);
      printmatchinglines('h', nickuserhost, session_sendnotice);
      printmatchinglines('i', nickuserhost, session_sendnotice);
      printmatchinglines('o', nickuserhost, session_sendnotice);
      printmatchinglines('x', nickuserhost, session_sendnotice);
      printmatchinglines('n', nickuserhost, session_sendnotice);
      xfree(nickuserhost);
      session_sendnotice("-------------------------------------------------------");
    }
  }
  nick = (char *)xmalloc(strlen(argv[1])+1);
  strcpy(nick, argv[1]);
  hoofdletter(nick);
  whoisfile = (char *)xmalloc(strlen(whoisdir)+strlen(nick)+1);
  strcpy(whoisfile, whoisdir);
  strcat(whoisfile, nick);
  if (catfile(whoisfile, session_sendnotice, session_nick()))
    session_sendnotice("-------------------------------------------------------");
  xfree(whoisfile);
  xfree(nick);
}

void cmd_whoami(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: whoami");
    return;
  }
  argv[0]="whois";
  argv[1]=session_nick();
  cmd_whois(argv, 2);
}

void cmd_edit(char *argv[], int argc) {

  FILE *fp;
  char nick[10];
  int linenumber;
  char *filename;
  int i;
  char *message;

  if (argc < 1) {
    session_sendnotice("^BUsage:^B edit <number> [<message>]");
    return;
  }
  linenumber = atoi(argv[1]);
  if ((linenumber < 1) || (linenumber > 10)) {
    session_sendnotice("Linenumber out of range: must be 1-10");
    return;
  }
  strcpy(nick, session_nick());
  hoofdletter(nick);
  filename = (char *)xmalloc(strlen(nick)+strlen(whoisdir)+1);
  sprintf(filename, "%s%s", whoisdir, nick);

  if (argc == 2)
    message = "";
  else {
    repairstring(&argv[2], argc - 2);
    message = argv[2];
  }
  if (strlen(message) > MAXMSGSIZE - 1) {
    session_sendnotice("Text is too long");
    xfree(message);
    return;
  }

  if ((fp = fopen(filename, "r+")) == NULL) {
    if ((fp = fopen(filename, "w+")) == NULL) {
      session_sendnotice("I could not open %s", filename);
      xfree(filename);
      return;
    }
    for(i=0; i<(10*MAXMSGSIZE); i++) fputc('\0', fp);
    fseek(fp, 0L, SEEK_SET);
  }
  fseek(fp, ((linenumber-1) * MAXMSGSIZE), SEEK_SET);
  i=0;
  while (message[i] != '\0') {
    fputc(message[i], fp);
    i++;
  }
  while (i < MAXMSGSIZE) {
    fputc('\0', fp);
    i++;
  }
  fclose(fp);
  catfile(filename, session_sendnotice, session_nick());
  xfree(filename);
}

void cmd_seen(char *argv[], int argc) {

  if (argc != 2) {
    session_sendnotice("Usage: seen <nick>");
    return;
  }
  if((strlen(argv[1]) > 9) || (argv[1][0] == '.') || (argv[1][0] == '/')) {
    session_sendnotice("Illegal nick");
    return;
  }
  if (userlist_findnick(argv[1])) {
    session_sendnotice("Nick: %36s %s", userlist_nick(), userlist_keepentry() ? "-=keeping=-" : "");
    session_sendnotice("Last seen: %31s", time2string(userlist_lastseen()));
    session_sendnotice("From: %36s", (userlist_lasthost()==NULL) ? "" : userlist_lasthost());
    return;
  }
  session_sendnotice("%s does not have an userlist entry", argv[1]);
}

void cmd_time(char *argv[], int argc) {

  gen_cmd_time(argv, argc, session_sendnotice);
}

void cmd_say(char *argv[], int argc) {

  char *channel;
  char *message;

  if (argc < 2) {
    session_sendnotice("Usage: say [-c <channel>] <text>");
    return;
  }
  if (strcasecmp("-c", argv[1]) == 0) {
    if (argc < 4) {
      session_sendnotice("Usage: say [-c <channel>] <text>");
      return;
    }
    channel = argv[2];
    repairstring(&argv[3], argc - 3);
    message = argv[3];
  }
  else {
    if (!(is_on_a_channel(session_nick()))) {
      if (!(channellist_findfirst())) {
	session_sendnotice("I am not on a channel");
	return;
      }
    }
    channel = channellist_channelname();
    repairstring(&argv[1], argc - 1);
    message = argv[1];
  }
  gen_cmd_msg(channel, message, session_sendtoirc);
}

void cmd_me(char *argv[], int argc) {

  char *channel;
  char *message;

  if (argc < 2) {
    session_sendnotice("Usage: me [-c <channel>] <text>");
    return;
  }
  if (strcasecmp("-c", argv[1]) == 0) {
    if (argc < 4) {
      session_sendnotice("Usage: me [-c <channel>] <text>");
      return;
    }
    channel = argv[2];
    repairstring(&argv[3], argc - 3);
    message = argv[3];
  }
  else {
    if (!(is_on_a_channel(session_nick()))) {
      if (!(channellist_findfirst())) {
	session_sendnotice("I am not on a channel");
	return;
      }
    }
    channel = channellist_channelname();
    repairstring(&argv[1], argc - 1);
    message = argv[1];
  }
  gen_cmd_me(channel, message, session_sendtoirc);
}

void cmd_op(char *argv[], int argc) {

  char *channel;
  char *nick;
  char *tmp;

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: op [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: op [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: op [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(is_on_a_channel(nick))) {
      session_sendnotice("%s is not on the same channel as I am", nick);
      return;
    }
    channel = channellist_channelname();
  }
  if (!(isonchannel(channel, nick))) {
    session_sendnotice("%s is not on channel %s", nick, channel);
    return;
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  if (ischanneloperator(channel, nick)) {
    session_sendnotice("%s is already channel operator on channel %s", nick, channel);
    return;
  }
  tmp = session_nick();
  if (!(session_findnick(nick))) {
    session_findnick(tmp);
    session_sendnotice("%s does not have a sessionlist entry", nick);
    return;
  }
  if (session_ismatching('d') || (!session_isallowed(GETOPS_ALLOWED))) {
    session_findnick(tmp);
    session_sendnotice("I have to deop %s", nick);
    return;
  }
  session_findnick(tmp);
  session_sendtoirc("MODE %s +o %s", channel, nick);
}

void cmd_deop(char *argv[], int argc) {

  char *channel;
  char *nick;

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: deop [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: deop [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: deop [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(is_on_a_channel(nick))) {
      session_sendnotice("%s is not on the same channel as I am", nick);
      return;
    }
    channel = channellist_channelname();
  }
  if (strcasecmp(nick, current_nick) == 0) {
    session_sendnotice("I do not like your idea!");
    return;
  }
  if (botlist_findbot(nick) && botlist_protection() && part_of_botnet(nick)) {
    session_sendnotice("%s is a protected bot", nick);
    return;
  }
  if (!(isonchannel(channel, nick))) {
    session_sendnotice("%s is not on channel %s", nick, channel);
    return;
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  if (!(ischanneloperator(channel, nick))) {
    session_sendnotice("%s is not channel operator on channel %s", nick, channel);
    return;
  }
  if (session_better_protected(nick, session_nick())) {
    session_sendnotice("%s is protected", nick);
    return;
  }
  session_sendtoirc("MODE %s -o %s", channel, nick);
}

void cmd_ban(char *argv[], int argc) {

  char *channel;
  char *nick;
  char *tmp;
  char *nickuserhost;
  char banmatch[MAXNICKUSERHOSTLEN];
  int i;

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: ban [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: ban [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: ban [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(is_on_a_channel(nick))) {
      session_sendnotice("%s is not on the same channel as I am", nick);
      return;
    }
    channel = channellist_channelname();
  }
  if (strcasecmp(nick, current_nick) == 0) {
    session_sendnotice("I do not like your idea!");
    return;
  }
  if (botlist_findbot(nick) && botlist_protection() && part_of_botnet(nick)) {
    session_sendnotice("%s is a protected bot", nick);
    return;
  }
  if (!(isonchannel(channel, nick))) {
    session_sendnotice("%s is not on channel %s", nick, channel);
    return;
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  if (session_better_protected(nick, session_nick())) {
    session_sendnotice("%s is protected", nick);
    return;
  }
  tmp = session_nick();
  if (!(session_findnick(nick))) {
    session_findnick(tmp);
    session_sendnotice("%s does not have an entry in the sessionlist", nick);
    return;
  }
  nickuserhost = session_nickuserhost();
  session_findnick(tmp);
  for (i = 0; i < MAX_BANS; i++) {
    if (match(returnban(i) ? returnban(i) : "", nickuserhost)) {
      session_sendnotice("%s is already matching with ban %s", nick, returnban(i));
      return;
    }
  }
  maakbanmatch(nickuserhost, banmatch);
  session_sendtoirc("MODE %s %s+b %s %s", channel, ischanneloperator(channel, nick) ? "-o" : "", 
		    ischanneloperator(channel, nick) ? nick : "", banmatch);
}

void cmd_bankick(char *argv[], int argc) {
  
  char *channel;
  char *nick;
  char *tmp;
  char *nickuserhost;
  char banmatch[MAXNICKUSERHOSTLEN];
  int i;

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: bankick [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: bankick [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: bankick [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(is_on_a_channel(nick))) {
      session_sendnotice("%s is not on the same channel as I am", nick);
      return;
    }
    channel = channellist_channelname();
  }
  if (strcasecmp(nick, current_nick) == 0) {
    session_sendnotice("I do not like your idea!");
    return;
  }
  if (botlist_findbot(nick) && botlist_protection() && part_of_botnet(nick)) {
    session_sendnotice("%s is a protected bot", nick);
    return;
  }
  if (!(isonchannel(channel, nick))) {
    session_sendnotice("%s is not on channel %s", nick, channel);
    return;
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  if (session_better_protected(nick, session_nick())) {
    session_sendnotice("%s is protected", nick);
    return;
  }
  tmp = session_nick();
  if (!(session_findnick(nick))) {
    session_findnick(tmp);
    session_sendnotice("%s does not have an entry in the sessionlist", nick);
    return;
  }
  nickuserhost = session_nickuserhost();
  session_findnick(tmp);
  for (i = 0; i < MAX_BANS; i++) {
    if (match(returnban(i) ? returnban(i) : "", nickuserhost)) {
      session_sendtoirc("KICK %s %s :Kick requested by %s", channel, nick, session_nick());
      return;
    }
  }
  maakbanmatch(nickuserhost, banmatch);
  session_sendtoirc("MODE %s %s+b %s %s\nKICK %s %s :Kick requested by %s", channel, 
		    ischanneloperator(channel, nick) ? "-o" : "", ischanneloperator(channel, nick) ? nick : "", banmatch,
		    channel, nick, session_nick());
}

void cmd_kick(char *argv[], int argc) {
  
  char *channel;
  char *nick;

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: kick [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: kick [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: kick [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(is_on_a_channel(nick))) {
      session_sendnotice("%s is not on the same channel as I am", nick);
      return;
    }
    channel = channellist_channelname();
  }
  if (strcasecmp(nick, current_nick) == 0) {
    session_sendnotice("I do not like your idea!");
    return;
  }
  if (botlist_findbot(nick) && botlist_protection() && part_of_botnet(nick)) {
    session_sendnotice("%s is a protected bot", nick);
    return;
  }
  if (!(isonchannel(channel, nick))) {
    session_sendnotice("%s is not on channel %s", nick, channel);
    return;
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  if (session_better_protected(nick, session_nick())) {
    session_sendnotice("%s is protected", nick);
    return;
  }
  session_sendtoirc("KICK %s %s :Kick requested by %s", channel, nick, session_nick()); 
}

void cmd_unban(char *argv[], int argc) {

  char *channel;
  char *nick;
  char *tmp;
  char *nickuserhost;
  int i, j;
  char *bans[3];

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: unban [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: unban [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: unban [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(channellist_findfirst())) {
      session_sendnotice("I am not on a channel");
      return;
    }
    channel = channellist_channelname();
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  tmp = session_nick();
  if (!(session_findnick(nick))) {
    session_findnick(tmp);
    session_sendnotice("%s does not have an entry in the sessionlist", nick);
    return;
  }
  nickuserhost = session_nickuserhost();
  session_findnick(tmp);
  j = 0;
  for (i = 0; i < MAX_BANS; i++) {
    if (match(returnban(i) ? returnban(i) : "", nickuserhost)) {
      bans[j] = returnban(i);
      j++;
      if (j == 3) {
	session_sendtoirc_important("MODE %s -bbb %s %s %s", channel, bans[0], bans[1], bans[2]);
	j = 0;
      }
    }
  }
  if (j)
    session_sendtoirc_important("MODE %s -b%c %s %s", channel, j >= 2 ? 'b' : ' ', bans[0], j >= 2 ? bans[1] : "");
  session_sendnotice("Done");
}

void cmd_massunban(char *argv[], int argc) {

  char *channel;
  int   i, j;
  char *bans[3];

  if (argc > 2) {
    session_sendnotice("Usage: massunban [<channel>]");
    return;
  }
  if (argc == 2)
    channel = argv[1];
  else {
    if (!(channellist_findfirst())) {
      session_sendnotice("I am not on a channel");
      return;
    }
    channel = channellist_channelname();
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  j = 0;
  for (i = 0; i < MAX_BANS; i++) {
    if (returnban(i)) {
      bans[j] = returnban(i);
      j++;
      if (j == 3) {
	session_sendtoirc_important("MODE %s -bbb %s %s %s", channel, bans[0], bans[1], bans[2]);
	j = 0;
      }
    }
  }
  if (j)
    session_sendtoirc_important("MODE %s -b%c %s %s", channel, j >= 2 ? 'b' : ' ', bans[0], j >= 2 ? bans[1] : "");
  session_sendnotice("Done");
}

void cmd_massop(char *argv[], int argc) {

  char *channel;
  int   i;
  char *ops[3];
  char *tmp;

  if (argc > 2) {
    session_sendnotice("Usage: massop [<channel>]");
    return;
  }
  if (argc == 2)
    channel = argv[1];
  else {
    if (!(channellist_findfirst())) {
      session_sendnotice("I am not on a channel");
      return;
    }
    channel = channellist_channelname();
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  i = 0;
  tmp = session_nick();
  if (chanusers_findfirst(channellist_chanusers())) {
    do {
      if ((!(chanusers_ischanop(channellist_chanusers()))) && (session_findnick(chanusers_nick(channellist_chanusers()))) 
	  && (!session_ismatching('d')) && session_isallowed(GETOPS_ALLOWED)) {
	session_findnick(tmp);
	ops[i] = chanusers_nick(channellist_chanusers());
	i++;
	if (i == 3) {
	  session_sendtoirc_important("MODE %s +ooo %s %s %s", channel, ops[0], ops[1], ops[2]);
	  i = 0;
	}
      } 
    } while (chanusers_findnext(channellist_chanusers()));
  }
  session_findnick(tmp);
  if (i)
    session_sendtoirc_important("MODE %s +o%c %s %s", channel, i >= 2 ? 'o' : ' ', ops[0], i >= 2 ? ops[1] : "");
  session_sendnotice("Done");
}

void cmd_massdeop(char *argv[], int argc) {

  char *channel;
  int   i;
  char *ops[3];

  if (argc > 2) {
    session_sendnotice("Usage: massdeop [<channel>]");
    return;
  }
  if (argc == 2)
    channel = argv[1];
  else {
    if (!(channellist_findfirst())) {
      session_sendnotice("I am not on a channel");
      return;
    }
    channel = channellist_channelname();
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  i = 0;
  if (chanusers_findfirst(channellist_chanusers())) {
    do {
      if (chanusers_ischanop(channellist_chanusers()) == 0) continue;
      if (session_better_protected(chanusers_nick(channellist_chanusers()), session_nick())) continue;
      if (strcasecmp(chanusers_nick(channellist_chanusers()), current_nick) == 0) continue;
      if (botlist_findbot(chanusers_nick(channellist_chanusers())) && botlist_protection() &&
	  part_of_botnet(chanusers_nick(channellist_chanusers()))) continue;
      ops[i] = chanusers_nick(channellist_chanusers());
      i++;
      if (i == 3) {
	session_sendtoirc_important("MODE %s -ooo %s %s %s", channel, ops[0], ops[1], ops[2]);
	i = 0;
      } 
    } while (chanusers_findnext(channellist_chanusers()));
  }
  if (i)
    session_sendtoirc_important("MODE %s -o%c %s %s", channel, i >= 2 ? 'o' : ' ', ops[0], i >= 2 ? ops[1] : "");
  session_sendnotice("Done");
}

void cmd_masskick(char *argv[], int argc) {
  
  char *channel;
  int   i;
  char message[WAYTOBIG]="";

  if (argc > 2) {
    session_sendnotice("Usage: masskick [<channel>]");
    return;
  }
  if (argc == 2)
    channel = argv[1];
  else {
    if (!(channellist_findfirst())) {
      session_sendnotice("I am not on a channel");
      return;
    }
    channel = channellist_channelname();
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  i = 0;
  if (chanusers_findfirst(channellist_chanusers())) {
    do {
      if (session_better_protected(chanusers_nick(channellist_chanusers()), session_nick())) continue;
      if (strcasecmp(chanusers_nick(channellist_chanusers()), current_nick) == 0) continue;
      if (botlist_findbot(chanusers_nick(channellist_chanusers())) && botlist_protection() &&
	  part_of_botnet(chanusers_nick(channellist_chanusers()))) continue;
      strcat(message, chanusers_nick(channellist_chanusers()));
      strcat(message, ",");
      i++;
      if (i == 20) {
	session_sendtoirc_important("KICK %s %s :Masskick requested by %s", channel, message, session_nick());
	i = 0;
	strcpy(message, "");
      } 
    } while (chanusers_findnext(channellist_chanusers()));
  }
  if (i)
    session_sendtoirc_important("KICK %s %s :Masskick requested by %s", channel, message, session_nick());
  session_sendnotice("Done");
}

void cmd_invite(char *argv[], int argc) {

  char *nick;
  char *channel;

  if ((argc < 1) || (argc > 4)) {
    session_sendnotice("Usage: invite [-c <channel>] [<nick>]");
    return;
  }
  if ((argc > 1) && (strcasecmp(argv[1], "-c") == 0)) {
    if (argc < 3) {
      session_sendnotice("Usage: invite [-c <channel>] [<nick>]");
      return;
    }
    channel = argv[2];
    if (argc == 4)
      nick = argv[3];
    else
      nick = session_nick();
  }
  else {
    if (argc > 2) {
      session_sendnotice("Usage: invite [-c <channel>] [<nick>]");
      return;
    }
    if (argc == 2)
      nick = argv[1];
    else
      nick = session_nick();
    if (!(channellist_findfirst())) {
      session_sendnotice("I am not on a channel");
      return;
    }
    channel = channellist_channelname();
  }
  if (!(ischanneloperator(channel, current_nick))) {
    session_sendnotice("I am not channel operator on channel %s", channel);
    return;
  }
  if (isonchannel(channel, nick)) {
    session_sendnotice("%s is already on channel %s", nick, channel);
    return;
  }
  session_sendtoirc("INVITE %s %s", nick, channel);
  session_sendnotice("Done");
}

void cmd_su(char *argv[], int argc) {

  session_resetsuperuser();
  if (argc != 2) {
    sendtospies("Incorrect use of command");
    session_sendnotice("Usage: su <password>");
    return;
  }
  if (strcmp(argv[1], su_password) == 0) {
    sendtospies("%s is now superuser", session_nick());
    sendtospies("%s is now superuser", session_nick());
    sendtospies("%s is now superuser", session_nick());
    session_setsuperuser();
    session_sendnotice("You are now superuser");
    return;
  }
  else {
    sendtospies("Password incorrect");
    session_sendnotice("Sorry");
    return;
  }
}

void cmd_channels(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: channels");
    return;
  }
  printchans2watch(session_sendnotice);
}

void cmd_addchan(char *argv[], int argc) {

  gen_cmd_addchan(argv, argc, session_sendnotice, session_sendtoirc_important);
}

void cmd_delchan(char *argv[], int argc) {

  gen_cmd_delchan(argv, argc, session_sendnotice, session_sendtoirc_important);
}

void cmd_addserv(char *argv[], int argc) {

  int port;

  if ((argc < 2) || (argc > 3)) {
    session_sendnotice("Usage: addserv <host> [<port>]");
    return;
  }
  if (argc == 3)
    port = atoi(argv[2]);
  else
    port = 6667;

  if (!(setserver(argv[1], port))) {
    session_sendnotice("Could not add %s %d to the serverlist", argv[1], port);
    return;
  }
  else {
    session_sendnotice("Added %s %d to the serverlist", argv[1], port);
    return;
  }
}

void cmd_delserv(char *argv[], int argc) {

  gen_cmd_delserv(argv, argc, session_sendnotice);
}

void cmd_server(char *argv[], int argc) {

  gen_cmd_server(argv, argc, session_sendnotice);
}

void cmd_addbot(char *argv[], int argc) {
  
  gen_cmd_addbot(argv, argc, session_sendnotice);
}

void cmd_delbot(char *argv[], int argc) {

  gen_cmd_delbot(argv, argc, session_sendnotice);
}

void cmd_botpass(char *argv[], int argc) {

  gen_cmd_botpass(argv, argc, session_sendnotice);
}

void cmd_botnuh(char *argv[], int argc) {

  gen_cmd_botnuh(argv, argc, session_sendnotice);
}

void cmd_botaddr(char *argv[], int argc) {

  gen_cmd_botaddr(argv, argc, session_sendnotice);
}

void cmd_botprot(char *argv[], int argc) {

  gen_cmd_botprot(argv, argc, session_sendnotice);
}

void cmd_botac(char *argv[], int argc) {

  gen_cmd_botac(argv, argc, session_sendnotice);
}

void cmd_botlist(char *argv[], int argc) {

  gen_cmd_botlist(argv, argc, session_sendnotice);
}

void cmd_botinfo(char *argv[], int argc) {

  gen_cmd_botinfo(argv, argc, session_sendnotice, session_administrator());
}

void cmd_link(char *argv[], int argc) {

  gen_cmd_link(argv, argc, session_sendnotice, session_nick());
}

void cmd_party(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: party");
    return;
  }
  if (session_currentchannel() != -1) return;
  catfile(motd_file, session_sendnotice, session_nick());
  session_sendnotice("Welcome on the partyline of %s (%s)", botnick, LONGVERSION);
  session_joinchannel(default_irc_partyline);
}

void cmd_join(char *argv[], int argc) {

  int channel;

  if (session_currentchannel() == -1) {
    session_sendnotice("You are not on the partyline. (Use \"party\")");
    return;
  }
  if (argc == 1) {
    session_sendnotice("Usage: join <channelnumber>|<channelname>");
    return;
  }
  if (isdigit(argv[1][0])) { channel = atoi(argv[1]); }
  else {
    repairstring(&argv[1], argc - 1);
    if ((channel = assoc_channelnumber(argv[1])) == -1) {
      session_sendnotice("%s: No such channel", argv[1]);
      return;
    }
  }
  if ((channel < 0) || (channel > 99999)) {
    session_sendnotice("%d: No such channel", channel);
    return;
  }
  if (!(session_joinchannel(channel))) {
    session_sendnotice("You have joined too many channels");
    return;
  }
}

void cmd_leave(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    if ((channel = session_currentchannel()) == -1) {
      session_sendnotice("You are not on a channel");
      return;
    }
  }
  else {
    if (isdigit(argv[1][0])) { channel = atoi(argv[1]); }
    else {
      repairstring(&argv[1], argc - 1);
      if ((channel = assoc_channelnumber(argv[1])) == -1) {
	session_sendnotice("%s: No such channel", argv[1]);
	return;
      }
    }
    if ((channel < 0) || (channel > 99999)) {
      session_sendnotice("%d: No such channel", channel);
      return;
    }
  }
  session_leavechannel(channel);
}

void cmd_act(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    session_sendnotice("Usage: act <text>");
    return;
  }
  if ((channel = session_currentchannel()) == -1) {
    session_sendnotice("You are not on a channel");
    return;
  }
  repairstring(&argv[1], argc - 1);
  party_send2partychannel(channel, "* %s %s", session_nick(), argv[1]);
  session_send2partychannel(channel, "* %s %s", session_nick(), argv[1]);
  botlist_send2all("actchan %s@%s %d %s", session_nick(), botnick, channel, argv[1]);
  console_send2snoopwindow("(%03d) * %s %s", channel, session_nick(), argv[1]);
}

void cmd_assoc(char *argv[], int argc) {

  int   channel=0;
  char *channelname;

  if (argc == 1) {
    session_sendnotice("Number  Name");
    session_sendnotice("------  ---------------------------");
    for (channel = 1; channel < 100000; channel++) {
      if ((channelname = assoc_channelname(channel)) != NULL) {
	session_sendnotice("%6d  %s", channel, channelname);
      }
    }
    session_sendnotice("------  ---------------------------");
    return;
  }
  if ((channel = session_currentchannel()) == -1) {
    session_sendnotice("You are not on a channel");
    return;
  }
  if (channel == 0) {
    session_sendnotice("You can't change the name of channel 0, that's the main party line");
    return;
  }
  repairstring(&argv[1], argc - 1);
  if (argv[1][0] == '-') {
    delassoc(channel);
    botlist_send2all("assoc %d 0", channel);
    botlist_send2all("chan %s %d The name of channel %d has been removed by %s", botnick, channel, channel, 
		       session_nick());
    party_send2partychannel(channel, "*** The name of channel %d has been removed by %s", channel, session_nick());
    session_send2partychannel(channel, "*** The name of channel %d has been removed by %s", channel, session_nick());
    console_send2snoopwindow("(%03d) *** The name of channel %d has been removed by %s", channel, channel, session_nick());
    return;
  }
  if (isdigit(argv[1][0])) {
    session_sendnotice("*** First character of the channelname can't be a digit");
    return;
  }
  if (strlen(argv[1]) > 20) {
    session_sendnotice("*** Channelnames can't be longer than 20 characters");
    return;
  }
  addassoc(channel, argv[1]);
  botlist_send2all("assoc %d %s", channel, argv[1]);
  botlist_send2all("chan %s %d %s named this channel \"%s\"", botnick, channel, session_nick(), argv[1]);
  party_send2partychannel(channel, "*** %s has named this channel \"%s\"", session_nick(), argv[1]);
  session_send2partychannel(channel, "*** %s has named this channel \"%s\"", session_nick(), argv[1]);
  console_send2snoopwindow("(%03d) *** %s has named this channel \"%s\"", channel, session_nick(), argv[1]);
}

void cmd_quit(char *argv[], int argc) {

  session_leaveallchannels();
}

void cmd_who(char *argv[], int argc) {

  int channel;

  if (argc == 1) {
    if ((channel = session_currentchannel()) == -1) {
      session_sendnotice("*** You are not on a channel");
      return;
    }
  }
  else {
    if (isdigit(argv[1][0])) {
      channel = atoi(argv[1]);
    }
    else {
      repairstring(&argv[1], argc - 1);
      if ((channel = assoc_channelnumber(argv[1])) == -1) {
	session_sendnotice("*** %s: No such channel", argv[1]);
	return;
      }
    }
  }
  session_sendnotice("Nick         Bot        From");
  session_sendnotice("----------   ---------  ------------------------------");
  if (party_findfirst()) {
    do {
      if (party_isonchannel(channel)) {
	session_sendnotice("-%-11s %-10s %s", party_nick(), botnick, party_from());
      }
    } while (party_next());
  }
  session_sendwho2session(channel);
  botlist_send2all("who? %s@%s %d", session_nick(), botnick, channel);
}

void cmd_away(char *argv[], int argc) {

  if (argc == 1) {
    session_unaway();
    session_sendnotice("You are no longer marked as being away");
    return;
  }
 repairstring(&argv[1], argc - 1);
 session_setaway(argv[1]);
 session_sendnotice("You are now marked as being away");
}

void cmd_boot(char *argv[], int argc) {

  char *comment, *tmp, *bot;

  if (session_currentchannel() == -1) {
    session_sendnotice("You are not on the partyline");
    return;
  }
  if (argc < 2) {
    session_sendnotice("Usage: boot <nick>[@<bot>] [<comment>]");
    return;
  }
  if (argc >= 3) {
    repairstring(&argv[2], argc - 2);
    comment = argv[2];
  }
  else
    comment = session_nick();

  if (index(argv[1], '@')==NULL) {
  Idontlike2todothingstwice:
    tmp = session_nick();
    if (session_findnick(argv[1])) { 
      session_boot(tmp, comment);
      session_findnick(tmp);
      session_sendnotice("%s is not on the partyline now", argv[1]);
      return;
    }
    session_findnick(tmp);
    if (party_findnick(argv[1])) { 
      party_boot(tmp, comment);
      session_sendnotice("%s is not on the partyline now", argv[1]);
      return;
    }
    session_sendnotice("I do neither have an IRC- nor telnet-session with %s", argv[1]);
    return;
  }
  bot = index(argv[1], '@')+1;
  if (strcasecmp(bot, botnick)==0) {
    bot--; *bot='\0'; bot++;
    goto Idontlike2todothingstwice;
  }
  if (!(part_of_botnet(bot))) {
    session_sendnotice("%s is not part of the botnet", bot);
    return;
  }
  bottree_send2bot(bot, "reject %s@%s %s %s", session_nick(), botnick, argv[1], comment);
}

void cmd_bots(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: bots");
    return;
  }
  botlist_send2all("info? %s@%s", session_nick(), botnick);
}

void cmd_bottree(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: bottree");
    return;
  }
  printbottree(session_sendnotice);
}

void cmd_chat(char *argv[], int argc) {

  session_stopchat();
  session_startchat("");
}

void cmd_echo(char *argv[], int argc) {

  if (argc != 2) {
    session_sendnotice("Usage: echo [on|off]");
    return;
  }
  if (strcasecmp(argv[1], "on")==0) {
    session_echo_on();
    session_sendnotice("echo turned on");
    return;
  }
  if (strcasecmp(argv[1], "off")==0) {
    session_echo_off();
    session_sendnotice("echo turned off");
    return;
  }
  session_sendnotice("Usage: echo [on|off]");
}

void cmd_hangman(char *argv[], int argc) {

  if ((argc != 2) || (strlen(argv[1])>1) || (!(isalnum(argv[1][0])))) {
    session_sendnotice("Usage: hangman <alphanumeric character>");
    return;
  }
  if (!(session_hangman(argv[1][0]))) {
    session_sendnotice("It is not possible to play hangman");
    session_sendnotice("Probably the wordfile is missing...");
    return;
  }
}

void cmd_hour(char *argv[], int argc) {

  gen_cmd_hour(argv, argc, session_sendnotice);
}

void cmd_note(char *argv[], int argc) {

  char *tmp, *bot, *awayreason;

  if (argc < 3) {
    session_sendnotice("Usage: note <nick>[@<bot>] <message>");
    return;
  }
  repairstring(&argv[2], argc - 2);

  session_sendnotice("-> *%s* %s", argv[1], argv[2]);
  if (index(argv[1], '@')==NULL) {
  jumpincmd_note:
    tmp = session_nick();
    if ((session_findnick(argv[1])) && (session_has_chat())) {
      session_sendnotice("*%s* %s", tmp, argv[2]);
      if ((awayreason = session_isaway())) {
	session_findnick(tmp);
	session_sendnotice("%s is away: %s", argv[1], awayreason);
	if (addnote(argv[1], session_nick(), argv[2]))
	  session_sendnotice("Your note has been stored");
	else
	  session_sendnotice("Your note could not be stored");
	return;
      }
      return;
    }
    if (party_findnick(argv[1])) { 
      party_send2current("*%s* %s", tmp, argv[2]);
      return;
    }
    session_findnick(tmp);
    session_sendnotice("I do neither have an IRC- nor telnet-session with %s", argv[1]);
    if (addnote(argv[1], session_nick(), argv[2]))
      session_sendnotice("Your note has been stored");
    else
      session_sendnotice("Your note could not be stored");
    return;
  }
  bot = index(argv[1], '@')+1;
  if (strcasecmp(bot, botnick)==0) {
    bot--; *bot='\0'; bot++;
    goto jumpincmd_note;
  }
  if (!(part_of_botnet(bot))) {
    session_sendnotice("%s is not part of the botnet", bot);
    return;
  }
  bottree_send2bot(bot, "priv %s@%s %s %s", session_nick(), botnick, argv[1], argv[2]);
}

void cmd_notes(char *argv[], int argc) {

  char *from=NULL, *msg=NULL;

  if (argc != 1) {
    session_sendnotice("Usage: notes");
    return;
  }
  while (return_note(session_nick(), &from, &msg))
    session_sendnotice("*%s* %s", from, msg);
  session_sendnotice("End of notes");
}

void cmd_rehash(char *argv[], int argc) {

  gen_cmd_rehash(argv, argc, session_sendnotice);
}

void cmd_reset(char *argv[], int argc) {

  char *saved_nick;

  if (argc != 2) {
    session_sendnotice("Usage: reset <nick>");
    return;
  }
  saved_nick = session_nick();
  if (!(session_findnick(argv[1]))) {
    session_findnick(saved_nick);
    session_sendnotice("I do not have a session active with %s", argv[1]);
    return;
  }
  session_resetspy();
  session_resetidentify();
  session_resetoperator();
  session_resetsuperuser();
  session_leaveallchannels();
  session_unaway();
  session_stopchat();
  session_closesend();
  session_closeget();
  session_echo_on();
  session_findnick(saved_nick);
  session_sendnotice("Done");
}

void cmd_signoff(char *argv[], int argc) {

  char *file;

  if (argc != 1) {
    session_sendnotice("Usage: signoff");
    return;
  }
  file = (char *) xmalloc (strlen("Signoff requested by ")+strlen(session_nick())+1);
  sprintf(file, "Signoff requested by %s", session_nick());
  gen_cmd_signoff(file);
  xfree(file);
}

void cmd_trace(char *argv[], int argc) {

  if (argc != 2) {
    session_sendnotice("Usage: trace <bot>");
    return;
  }
  if (strcasecmp(argv[1], botnick)==0) {
    session_sendnotice("Trace result: %s", botnick);
    return;
  }
  if (!(part_of_botnet(argv[1]))) {
    session_sendnotice("%s is not part of the botnet", argv[1]);
    return;
  }
  bottree_send2bot(argv[1], "trace %s@%s %s %s", session_nick(), botnick, argv[1], botnick);
}

void cmd_set(char *argv[], int argc) {

  gen_cmd_set(argv, argc, session_sendnotice);
}

void cmd_cmdreq(char *argv[], int argc) {

  gen_cmd_cmdreq(argv, argc, session_sendnotice);
}

void cmd_protkey(char *argv[], int argc) {

  gen_cmd_protkey(argv, argc, session_sendnotice);
}

void cmd_listen(char *argv[], int argc) {

  gen_cmd_listen(argv, argc, session_sendnotice);
}

void cmd_http(char *argv[], int argc) {

  gen_cmd_http(argv, argc, session_sendnotice);
}

void cmd_linereq(char *argv[], int argc) {

  gen_cmd_linereq(argv, argc, session_sendnotice);
}

void cmd_mycmds(char *argv[], int argc);

void cmd_files(char *argv[], int argc) {
  
  char *cur_dir;
  DIR *directory;
  struct dirent *direntry;
  char *filename;
  struct stat file_info;

  if (argc != 1) {
    session_sendnotice("Usage: files");
    return;
  }

  cur_dir = (char *) xmalloc (strlen(ftp_root_dir) + strlen(session_ftp_directory()) + 1);
  sprintf(cur_dir, "%s%s", ftp_root_dir, session_ftp_directory());
  
  if ((directory = opendir(cur_dir)) == NULL) {
    session_sendnotice("I could not read %s", session_ftp_directory());
    xfree(cur_dir);
    return;
  }
  while ((direntry = readdir(directory)) != NULL) {
    if (direntry->d_name[0] != '.') {
      filename = (char *)xmalloc(strlen(cur_dir) + strlen(direntry->d_name) + 3);
      sprintf(filename, "%s/%s", cur_dir, direntry->d_name);
      if (stat(filename, &file_info)==0) {
        if (S_ISDIR(file_info.st_mode))
          session_sendnotice("%-29s <directory>", direntry->d_name);
        if (S_ISREG(file_info.st_mode))
          session_sendnotice("%-29s %7d bytes", direntry->d_name, file_info.st_size);
      }
      xfree(filename);
    }
  }
  closedir(directory);
  xfree(cur_dir);
}

void cmd_cd(char *argv[], int argc) {

  char *begin, *end;

  if (argc > 2) {
    session_sendnotice("Usage: cd [<directory>]");
    return;
  }
  if (argc == 1) {
    session_sendnotice("Your current working directory: /%s", session_ftp_directory());
    return;
  }
  begin = argv[1];
  if (argv[1][0] == '/') {
    session_change_dir("/");
    begin++;
  }
  while (1) {
    if ((end = index(begin, '/')) == NULL) {
      session_change_dir(begin);
      session_sendnotice("Your current working directory: /%s", session_ftp_directory());
      return;
    }
    *end = '\0';
    if (!(session_change_dir(begin))) {
      session_change_dir(begin);
      session_sendnotice("Your current working directory: /%s", session_ftp_directory());
      return;
    }
    begin = end + 1;
  }
}

void cmd_pwd(char *argv[], int argc) {

  if (argc != 1) {
    session_sendnotice("Usage: pwd");
    return;
  }
  session_sendnotice("Your current working directory: /%s", session_ftp_directory());
}

void cmd_send(char *argv[], int argc) {

  char *file;

  if (argc != 2) {
    session_sendnotice("Usage: send <file>");
    return;
  }
  file = (char *) xmalloc (strlen(ftp_root_dir) + strlen(session_ftp_directory()) + strlen(argv[1]) + 2);
  sprintf(file, "%s%s/%s", ftp_root_dir, session_ftp_directory(), argv[1]);
  if (!(session_sendfile(file)))
    session_sendnotice("I could not send you %s", argv[1]);
}

void cmd_allow(char *argv[], int argc) {

  gen_cmd_allow(argv, argc, session_sendnotice);
}

void cmd_info(char *argv[], int argc) {

  gen_cmd_info(argv, argc, session_sendnotice);
}

void cmd_partycmd(char *argv[], int argc) {

  gen_cmd_partycmd(argv, argc, session_sendnotice);
}

void cmd_cmds(char *argv[], int argc) {

  print_commands(session_sendnotice);
}

void cmd_unlink(char *argv[], int argc) {

  gen_cmd_unlink(argv, argc, session_sendnotice, session_nick());
}

void cmd_cons(char *argv[], int argc) {
  
  gen_cmd_cons(argv, argc, session_sendnotice);
}

void cmd_motd(char *argv[], int argc) {

  catfile(motd_file, session_sendnotice, session_nick());
}

void cmd_resassoc(char *argv[], int argc) {

  int channel;

  for (channel = 1; channel <= 10000; channel++) {
    if (assoc_channelname(channel)) {
      delassoc(channel);
      botlist_send2all("assoc %d 0", channel);
      botlist_send2all("chan %s %d The name of channel %d has been removed by %s", botnick, channel, channel, 
		       session_nick());
      party_send2partychannel(channel, "*** The name of channel %d has been removed by %s", channel, session_nick());
      session_send2partychannel(channel, "*** The name of channel %d has been removed by %s", channel, session_nick());
      console_send2snoopwindow("(%03d) *** The name of channel %d has been removed by %s", channel, channel, 
			       session_nick());
    }
  }
}


void cmd_tryme(char *argv[], int argc) {

  if (is_on_a_channel(session_nick()) && ischanneloperator(channellist_channelname(), current_nick)) {
    if (rand() % 2 == 0) {
      session_sendtoirc("KICK %s %s :You loose", channellist_channelname(), session_nick());
    }
    else {
      session_sendnotice("You win");
    }
  }
  else {
    session_sendnotice("You win");
  }
}
  

void cmd_cheat(char *argv[], int argc) {

  char *nick, *word;

  if (argc != 2) {
    session_sendnotice("Usage: cheat <user>");
    return;
  }
  nick = session_nick();
  if ((!session_findnick(argv[1])) || ((word = session_hangmanword()) == NULL)) {
    session_findnick(nick);
    session_sendnotice("%s is not playing hangman", argv[1]);
  }
  else {
    session_findnick(nick);
    session_sendnotice("%s is trying to guess \"%s\"", argv[1], word);
  }
}


void cmd_xwin(char *argv[], int argc) {

  gen_cmd_xwin(argv, argc, session_sendnotice);
}


void cmd_tcl(char *argv[], int argc) {

  gen_cmd_tcl(argv, argc, session_sendnotice);
}


void cmd_addscript(char *argv[], int argc) {

  gen_cmd_addscript(argv, argc, session_sendnotice);
}


void cmd_delscript(char *argv[], int argc) {

  gen_cmd_delscript(argv, argc, session_sendnotice);
}


void cmd_scripts(char *argv[], int argc) {
  
  gen_cmd_scripts(argv, argc, session_sendnotice);
}


struct {
  char *command;
  void (*function)(char *argv[], int argc);
  int  spywords;
  char required;
  char chat;
  int  cfgversion;
} parse_cmds[] = 
  {
    { "act",       cmd_act,        0, ALL,   0, 0  },
    { "addbot",    cmd_addbot,    40, SU,    0, 0  },
    { "addchan",   cmd_addchan,   40, SU,    0, 0  },
    { "addscript", cmd_addscript, 40, SU,    1, 45 },
    { "addserv",   cmd_addserv,   40, SU,    0, 0  },
    { "aline",     cmd_aline,      2, ADMIN, 1, 0  },
    { "allow",     cmd_allow,     40, SU,    0, 0  },
    { "assoc",     cmd_assoc,     40, ALL,   0, 0  },
    { "away",      cmd_away,      40, ALL,   0, 0  },
    { "ban",       cmd_ban,       40, OPER,  0, 0  },
    { "bankick",   cmd_bankick,   40, OPER,  0, 0  },
    { "bkick",     cmd_bankick,   40, OPER,  0, 38 },
    { "bline",     cmd_bline,     40, ADMIN, 1, 0  },
    { "boot",      cmd_boot,      40, ADMIN, 0, 0  },
    { "botac",     cmd_botac,     40, SU,    0, 32 },
    { "botaddr",   cmd_botaddr,   40, SU,    0, 0  },
    { "botinfo",   cmd_botinfo,   40, ALL,   1, 0  },
    { "botlist",   cmd_botlist,   40, ALL,   1, 0  },
    { "botnuh",    cmd_botnuh,    40, SU,    0, 0  },
    { "botpass",   cmd_botpass,    1, SU,    0, 0  },
    { "botprot",   cmd_botprot,   40, SU,    0, 0  },
    { "bots",      cmd_bots,      40, ALL,   1, 0  },
    { "bottree",   cmd_bottree,   40, ALL,   1, 0  },
    { "cd",        cmd_cd,        40, ALL,   0, 0  },
    { "channels",  cmd_channels,  40, ALL,   1, 0  },
    { "chat",      cmd_chat,      40, ALL,   0, 0  },
    { "cheat",     cmd_cheat,     40, ADMIN, 0, 40 },
    { "cmdreq",    cmd_cmdreq,    40, SU,    0, 0  },
    { "cmds",      cmd_cmds,      40, ALL,   1, 0  },
    { "comment",   cmd_comment,   40, ADMIN, 1, 0  },
    { "cons",      cmd_cons,      40, SU,    1, 34 },
    { "delbot",    cmd_delbot,    40, SU,    0, 0  },
    { "delchan",   cmd_delchan,   40, SU,    0, 0  },
    { "delline",   cmd_delline,   40, ADMIN, 1, 0  },
    { "delscript", cmd_delscript, 40, SU,    0, 45 },
    { "delserv",   cmd_delserv,   40, SU,    0, 0  },
    { "deop",      cmd_deop,      40, OPER,  0, 0  },
    { "dir",       cmd_files,     40, ALL,   1, 0  },
    { "dline",     cmd_dline,     40, ADMIN, 1, 0  },
    { "echo",      cmd_echo,      40, ALL,   0, 0  },
    { "edit",      cmd_edit,      40, IDENT, 1, 0  },
    { "files",     cmd_files,     40, ALL,   1, 0  },
    { "get",       cmd_send,      40, ALL,   1, 0  },
    { "h",         cmd_hangman,   40, ALL,   1, 0  },
    { "hangman",   cmd_hangman,   40, ALL,   1, 0  },
    { "help",      cmd_help,      40, ALL,   1, 0  },
    { "hour",      cmd_hour,      40, ADMIN, 0, 0  },
    { "http",      cmd_http,      40, SU,    0, 0  },
    { "i",         cmd_identify,   1, ALL,   0, 0  },
    { "identify",  cmd_identify,   1, ALL,   0, 0  },
    { "iline",     cmd_iline,     40, ADMIN, 1, 0  },
    { "info",      cmd_info,      40, ALL,   1, 0  },
    { "invite",    cmd_invite,    40, OPER,  0, 0  },
    { "join",      cmd_join,      40, ALL,   0, 0  },
    { "keep",      cmd_keep,      40, ADMIN, 0, 0  },
    { "kick",      cmd_kick,      40, OPER,  0, 0  },
    { "leave",     cmd_leave,     40, ALL,   0, 0  },
    { "linepass",  cmd_linepass,   2, ADMIN, 1, 0  },
    { "linereq",   cmd_linereq,   40, SU,    0, 0  },
    { "link",      cmd_link,      40, SU,    0, 0  },
    { "listen",    cmd_listen,    40, SU,    0, 0  },
    { "lock",      cmd_lock,      40, ADMIN, 1, 0  },
    { "ls",        cmd_files,     40, ALL,   1, 0  },
    { "manip",     cmd_manip,     40, ADMIN, 0, 0  },
    { "maskinfo",  cmd_maskinfo,  40, ALL,   1, 0  },
    { "massdeop",  cmd_massdeop,  40, OPER,  0, 0  },
    { "masskick",  cmd_masskick,  40, OPER,  0, 0  },
    { "massop",    cmd_massop,    40, OPER,  0, 0  },
    { "massunban", cmd_massunban, 40, OPER,  0, 0  },
    { "me",        cmd_me,        40, OPER,  0, 0  },
    { "motd",      cmd_motd,      40, ALL,   1, 35 },
    { "mycmd",     cmd_mycmds,    40, ALL,   1, 0  },
    { "mycmds",    cmd_mycmds,    40, ALL,   1, 0  },
    { "nline",     cmd_nline,     40, ADMIN, 1, 0  },
    { "note",      cmd_note,       0, ALL,   1, 0  },
    { "notes",     cmd_notes,     40, IDENT, 1, 0  },
    { "oline",     cmd_oline,      2, ADMIN, 1, 0  },
    { "op",        cmd_op,        40, OPER,  0, 0  },
    { "oper",      cmd_oper,       1, ALL,   0, 0  },
    { "party",     cmd_party,     40, ALL,   1, 0  },
    { "partycmd",  cmd_partycmd,  40, SU,    0, 0  },
    { "passwd",    cmd_passwd,     1, IDENT, 0, 0  },
    { "protkey",   cmd_protkey,   40, SU,    1, 0  },
    { "pwd",       cmd_pwd,       40, ALL,   0, 0  },
    { "quit",      cmd_quit,      40, ALL,   0, 0  },
    { "rehash",    cmd_rehash,    40, ADMIN, 0, 0  },
    { "resassoc",  cmd_resassoc,  40, SU,    0, 38 },
    { "reset",     cmd_reset,     40, ADMIN, 0, 0  },
    { "say",       cmd_say,       40, OPER,  0, 0  },
    { "scripts",   cmd_scripts,   40, ALL,   1, 45 },
    { "seen",      cmd_seen,      40, ALL,   1, 0  },
    { "send",      cmd_send,      40, ALL,   0, 0  },
    { "server",    cmd_server,    40, SU,    1, 0  },
    { "set",       cmd_set,        2, SU,    0, 0  },
    { "signoff",   cmd_signoff,   40, SU,    0, 0  },
    { "spy",       cmd_spy,       40, ALL,   1, 0  },
    { "stats",     cmd_stats,     40, ALL,   1, 0  },
    { "su",        cmd_su,         1, ALL,   0, 0  },
    { "tcl",       cmd_tcl,       40, SU,    1, 44 },
    { "time",      cmd_time,      40, ALL,   0, 0  },
    { "top",       cmd_top,       40, ALL,   1, 0  },
    { "trace",     cmd_trace,     40, ALL,   0, 0  },
    { "tryme",     cmd_tryme,     40, ALL,   0, 39 },
    { "unban",     cmd_unban,     40, OPER,  0, 0  },
    { "unkeep",    cmd_unkeep,    40, ADMIN, 0, 0  },
    { "unlink",    cmd_unlink,    40, SU,    0, 0  },
    { "unlock",    cmd_unlock,    40, ADMIN, 1, 0  },
    { "useradd",   cmd_useradd,    3, ADMIN, 1, 0  },
    { "userdel",   cmd_userdel,   40, ADMIN, 1, 33 },
    { "userlist",  cmd_userlist,  40, ALL,   1, 0  },
    { "who",       cmd_who,       40, ALL,   1, 0  },
    { "whoami",    cmd_whoami,    40, ALL,   1, 0  },
    { "whois",     cmd_whois,     40, ALL,   1, 0  },
    { "xline",     cmd_xline,     40, ADMIN, 1, 0  },
    { "xwin",      cmd_xwin,      40, SU,    0, 41 },
    { NULL,        NULL,           0, 0,     0, 0  }
  };

char required_sign(int i) {

  switch(parse_cmds[i].required) {
  case ALL:   return ' ';
  case OPER:  return '@';
  case ADMIN: return '+';
  case SU:    return '*';
  case IDENT: return 'i';
  default:    return '?';
  }
}

void print_commands(void output(char *, ...)) {

  int i = 0, j = 0;

  for (i=0; parse_cmds[i].command; i++) {
    j++;
    if (j == 5) {
      output("[%c]%-10s[%c]%-10s[%c]%-10s[%c]%-10s[%c]%-10s", required_sign(i-4), parse_cmds[i-4].command,
	     required_sign(i-3), parse_cmds[i-3].command, required_sign(i-2), parse_cmds[i-2].command,
	     required_sign(i-1), parse_cmds[i-1].command, required_sign(i), parse_cmds[i].command);
      j=0;
    }
  }
  switch(j) {
  case 4:
    output("[%c]%-10s[%c]%-10s[%c]%-10s[%c]%-10s", required_sign(i-4), parse_cmds[i-4].command, 
	   required_sign(i-3), parse_cmds[i-3].command, required_sign(i-2), parse_cmds[i-2].command, 
	   required_sign(i-1), parse_cmds[i-1].command);
    break;
  case 3:
    output("[%c]%-10s[%c]%-10s", required_sign(i-3), parse_cmds[i-3].command, required_sign(i-2), 
	   parse_cmds[i-2].command, required_sign(i-1), parse_cmds[i-1].command);
    break;
  case 2:
    output("[%c]%-10s[%c]%-10s", required_sign(i-2), parse_cmds[i-2].command, required_sign(i-1), 
	   parse_cmds[i-1].command);
    break;
  case 1:
    output("[%c]%-10s",  required_sign(i-1), parse_cmds[i-1].command);
    break;
  }
  output("");
  output(" (i) identify  (@) operator  (+) administrator  (*) superuser");
}

void cmd_mycmds(char *argv[], int argc) {

  int i = 0, session_mask = 1, j = 0;
  int matches[3];

  if (argc != 1) {
    session_sendnotice("Usage: mycmds");
    return;
  }
  if (session_superuser()) session_mask |= SU;
  if (session_administrator()) session_mask |= ADMIN; 
  if (session_operator()) session_mask |= OPER;
  if (session_identified()) session_mask |= IDENT;

  for (i=0; parse_cmds[i].command; i++) {
    if (parse_cmds[i].required & session_mask) {
      matches[j] = i; j++;
      if (j == 3) {
	session_sendnotice("   %-15s %-15s %-15s", parse_cmds[matches[0]].command, parse_cmds[matches[1]].command, 
			   parse_cmds[matches[2]].command);
	j = 0;
      }
    }
  }
  if (j)
    session_sendnotice("   %-15s %-15s %-15s", parse_cmds[matches[0]].command, j > 1 ? parse_cmds[matches[1]].command :
		       "", j > 2 ? parse_cmds[matches[2]].command : "");
}

int  change_cmd_require(char *command, int require) {

  int i;

  for (i=0; parse_cmds[i].command; i++) {
    if (strcasecmp(command, parse_cmds[i].command) == 0) {
      parse_cmds[i].required = require;
      return 1;
    }
  }
  return 0;
}

int  current_cmd_require(char *command) {

  int i;

  for (i=0; parse_cmds[i].command; i++) {
    if (strcasecmp(command, parse_cmds[i].command) == 0) {
      return parse_cmds[i].required;
    }
  }
  return 0;
}


int check_permission(char *command, char required) {

  if ((required == ALL) || session_superuser()) return 1;
  if ((required & IDENT) && (!(session_identified()))) {
    sendtospies("User is not identified");
    session_sendnotice("%s: You are not identified", command);
    return 0;
  }
  if ((required & OPER) && (!(session_operator()))) {
    sendtospies("User is not operator");
    session_sendnotice("%s: You are not operator", command);
    return 0;
  }
  if ((required & ADMIN) && (!(session_administrator()))) {
    sendtospies("User is not administrator");
    session_sendnotice("%s: You are not administrator", command);
    return 0;
  }
  if ((required & SU) && (!(session_superuser()))) {
    sendtospies("User is not superuser");
    session_sendnotice("%s: You are not superuser", command);
    return 0;
  }
  return 1;
}


int dcc_chat_connection_right(char *argv[], int argc, int chat) {

  if ((chat == 0) || session_has_chat()) return 1;

  repairstring(argv, argc);
  session_startchat(argv[0]);
  return 0;
}


void local_sendtospies(char *argv[], int argc) {

  char msg[WAYTOBIG]="";
  int i;

  for (i=0; i < argc; i++) {
    strcat(msg, argv[i]);
    strcat(msg, " ");
  }
  sendtospies("%s: %s", session_nickuserhost(), msg);
}


void parse_command(char *line) {

  int i=0;
  int argc=6;
  char *argv[6];

  while (isspace(*line)) line++;
  if (*line == '\0') return;

  if (on_irc_cmd(session_nick(), line)) return;

  split_string(line, argv, &argc);

  for (i = 0; parse_cmds[i].command; i++) {
    if (strcasecmp(parse_cmds[i].command, argv[0]) == 0) {
      local_sendtospies(argv, argc > parse_cmds[i].spywords ? parse_cmds[i].spywords : argc);
      if (!(check_permission(argv[0], parse_cmds[i].required))) return;
      if (!(dcc_chat_connection_right(argv, argc, parse_cmds[i].chat))) return;
      parse_cmds[i].function(argv, argc);
      return;
    }
  }
  local_sendtospies(argv, argc);
  session_sendnotice("%s: Command not found", argv[0]);
  return;
}

void save_sescmds(FILE *fp) {

  int i;

  for (i=0; parse_cmds[i].command; i++)
    fwritechar(fp, parse_cmds[i].required);
  fwriteint(fp, 733);
}

int load_sescmds(FILE *fp, int version) {

  int i;

  for (i=0; parse_cmds[i].command; i++) {
    if (version >= parse_cmds[i].cfgversion)
      freadchar(fp, &parse_cmds[i].required);
  }
  freadint(fp, &i);
  return (733 == i);
}

