
/*********************************************************/
/**                                                     **/
/**       session.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                     **/
/**  Database of user who have contact with me via IRC  **/
/**                                                     **/
/*********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>
#include <strings.h>
#include <sys/types.h>
#include <netinet/in.h>


#include "session.h"
#include "config.h"
#include "misc.h"
#include "server.h"
#include "net.h"
#include "sescmd.h"
#include "bots.h"
#include "partyline.h"
#include "assoc.h"
#include "lines.h"
#include "userlist.h"
#include "gencmd.h"
#include "channel.h"
#include "console.h"
#include "on.h"


struct session {
  char *nickuserhost;
  char ircnick[10];
  char identified;      
  char operator;
  char su;
  char spy;
  time_t last_irc_message;
  time_t last_revenge;
  int    revenge_warnings;
  struct session *prev;
  struct session *next;
  char dcc_chat_stat;     /* 0 = not active, 1 = connecting, 2 = listening, 3 = connection active */
  int  dcc_chat_sockptr;
  char *saved_command;
  time_t flood_last_checked;
  int    flood_messages[10];
  char party;
  int  party_channels[MAX_CHANNELS];
  char *away;
  char  echo;
  char *hangman_word;
  char *hangman_guessed;
  int   hangman_errors;
  char *ftp_directory;
  char  dcc_send_stat;   /* 0 = not active, 1 = listening, 2 = connection active */
  int   dcc_send_sockptr;
  FILE *dcc_send_fp;
  unsigned int dcc_send_bytes_sent;
  char  dcc_get_stat;    /* 0 = is not active, 1 = connecting, 2 = connection active */
  int   dcc_get_sockptr;
  FILE *dcc_get_fp;
  unsigned int dcc_get_bytes_read;
  unsigned int dcc_get_filesize;
  time_t notify;
  char  matching_lines[5];   /* a, b, d, i, o */
};


struct sessionlist {
  int length;
  struct session *current;
  struct session *head;
  struct session *tail;
} seslist = { 0 , NULL , NULL , NULL };


extern char *ftp_root_dir;
extern char *wordfile;
extern char *incoming_dir;
extern int   warnings;
extern int   dcc_block_size;
extern char  botnick[10];
extern char  commandchar2;
extern char *session_logfile;
extern int   flood;
extern time_t time_null;

static int   allow[8];

void init_allow() {

  setallow(ALLOW_ALL,   UPLOAD_ALLOWED   );
  setallow(ALLOW_OLINE, AUTOOP_ALLOWED   );
  setallow(ALLOW_ALL,   PLACEBANS_ALLOWED);
  setallow(ALLOW_SU,    ALWAYSOP_ALLOWED );
  setallow(ALLOW_ALL,   GETOPS_ALLOWED   );
  setallow(ALLOW_ALL,   GIVEOPS_ALLOWED  );
  setallow(ALLOW_ALL,   UNBAN_ALLOWED    );
}

void save_allow_settings(FILE *fp) {

  int i;

  for (i=0; i<8; i++)
    fwriteint(fp, allow[i]);
}

int  load_allow_settings(FILE *fp) {

  int i;

  for (i=0; i<8; i++)
    freadint(fp, &allow[i]);

  return i;
}

void setallow(int who, int what) { allow[what] = who; }
int  curallow(int what) { return allow[what]; }

int session_isallowed(int what) {

  switch(allow[what]) {
  case ALLOW_ALL:   return 1;
  case ALLOW_OLINE: return session_ismatching('o') || session_ismatching('a');
  case ALLOW_OPER:  return session_operator();
  case ALLOW_ALINE: return session_ismatching('a');
  case ALLOW_ADMIN: return session_administrator();
  case ALLOW_SU:    return session_superuser();
  }
  int_err("Bug in session_isalowed"); return 0;
}

#define SESSION_A_LINE 0
#define SESSION_B_LINE 1
#define SESSION_D_LINE 2
#define SESSION_I_LINE 3
#define SESSION_O_LINE 4

void create_session(char *nickuserhost) {

  char *ch;
  int i;

  seslist.current = (struct session *)xmalloc(sizeof(struct session));

  if (seslist.length == 0) {
    seslist.head = seslist.current;
    seslist.current->prev = NULL;
  } 
  else {
    seslist.tail->next = seslist.current;
    seslist.current->prev = seslist.tail;
  }

  seslist.current->next = NULL;
  seslist.tail = seslist.current;
  seslist.length = seslist.length + 1;

  seslist.current->nickuserhost = (char *) xmalloc (strlen(nickuserhost) + 1);
  strcpy(seslist.current->nickuserhost, nickuserhost);

  ch = index(nickuserhost, '!');
  *ch = '\0';
  strcpy(seslist.current->ircnick, nickuserhost);
  *ch = '!';
  seslist.current->identified = 0;
  seslist.current->operator   = 0;
  seslist.current->last_irc_message = 0;
  seslist.current->last_revenge     = 0;
  seslist.current->revenge_warnings = 0;
  seslist.current->dcc_chat_stat    = 0;
  seslist.current->dcc_chat_sockptr = -1;
  seslist.current->saved_command    = NULL;
  seslist.current->su = 0;
  for (i=0; i<10; i++)
    seslist.current->flood_messages[i] = 0;
  seslist.current->flood_last_checked = 0;
  seslist.current->spy              = 0;
  seslist.current->party            = 0;
  seslist.current->away             = NULL;
  seslist.current->echo             = 1;
  seslist.current->hangman_word     = NULL;
  seslist.current->hangman_guessed  = NULL;
  seslist.current->hangman_errors   = 0;
  if (party_findnick(seslist.current->ircnick)) party_nickcollision();
  seslist.current->ftp_directory = (char *) xmalloc (1);
  strcpy(seslist.current->ftp_directory, "");
  seslist.current->dcc_send_stat       = 0;
  seslist.current->dcc_send_sockptr    = -1;
  seslist.current->dcc_send_fp         = NULL;
  seslist.current->dcc_send_bytes_sent = 0;
  seslist.current->dcc_get_stat        = 0;
  seslist.current->dcc_get_sockptr     = -1;
  seslist.current->dcc_get_fp          = NULL;
  seslist.current->dcc_get_bytes_read  = 0;
  seslist.current->dcc_get_filesize    = 0;
  seslist.current->notify              = time_null;
  seslist.current->matching_lines[SESSION_A_LINE] = does_this_match('a', session_nickuserhost());
  seslist.current->matching_lines[SESSION_B_LINE] = does_this_match('b', session_nickuserhost());
  seslist.current->matching_lines[SESSION_D_LINE] = does_this_match('d', session_nickuserhost());
  seslist.current->matching_lines[SESSION_I_LINE] = does_this_match('i', session_nickuserhost());
  seslist.current->matching_lines[SESSION_O_LINE] = does_this_match('o', session_nickuserhost());
}


void delete_session() {

  if (seslist.current == NULL) return;

  session_leaveallchannels();
  session_closesend();
  session_closeget();
  xfree(seslist.current->nickuserhost);
  xfree(seslist.current->saved_command);
  xfree(seslist.current->away);
  xfree(seslist.current->hangman_word);
  xfree(seslist.current->hangman_guessed);
  xfree(seslist.current->ftp_directory);
  if (seslist.current->dcc_chat_stat != 0) { close_socketptr(seslist.current->dcc_chat_sockptr); }

  if (seslist.length == 1) {
    xfree(seslist.current);
    seslist.current = NULL;
    seslist.tail    = NULL;
    seslist.head    = NULL;
    seslist.length  = 0;
    return;
  }

  if (seslist.current->next == NULL) {
    seslist.tail = seslist.current->prev;
    seslist.tail->next = NULL;
    xfree(seslist.current);
    seslist.current = seslist.tail;
    seslist.length = seslist.length - 1;
    return;
  }

  if (seslist.current->prev == NULL) {
    seslist.head = seslist.current->next;
    seslist.head->prev = NULL;
    xfree(seslist.current);
    seslist.current = seslist.head;
    seslist.length = seslist.length - 1;
    return;
  }

  seslist.current->prev->next = seslist.current->next;
  seslist.current->next->prev = seslist.current->prev;
  xfree(seslist.current);
  seslist.length = seslist.length - 1;
  seslist.current = seslist.head;
  return;
}


void delete_all_sessions() {

  while (seslist.length > 0) {
    seslist.current = seslist.tail;
    delete_session();
  }
}


void session_newnick(char *nick) {

  char nickuserhost[MAXNICKUSERHOSTLEN];
  int i = 0;

  if (party_findnick(nick)) party_nickcollision();
  for (i=0; i<seslist.current->party; i++) {
    session_send2partychannel(seslist.current->party_channels[i], "*** %s is now known as %s", seslist.current->ircnick, 
			      nick);
    party_send2partychannel(seslist.current->party_channels[i], "*** %s is now known as %s", seslist.current->ircnick, 
			    nick);
    botlist_send2all("chan %s %d %s is now known as %s", botnick, seslist.current->party_channels[i],
		     seslist.current->ircnick, nick);
    console_send2snoopwindow("(%03d) *** %s is now known as %s", seslist.current->party_channels[i],
			     seslist.current->ircnick, nick);
  }
  strcpy(seslist.current->ircnick, nick);
  strcpy(nickuserhost, seslist.current->nickuserhost);
  i=0; while (nickuserhost[i] != '!') i++;
  xfree(seslist.current->nickuserhost);
  seslist.current->nickuserhost = (char *) xmalloc (strlen(nick) + strlen(&nickuserhost[i]) + 1);
  strcpy(seslist.current->nickuserhost, nick);
  strcat(seslist.current->nickuserhost, &nickuserhost[i]);
  seslist.current->identified = 0;

  return;
}


int session_findfirst() {
  
  if (seslist.length == 0) return 0;
  seslist.current = seslist.head;
  return 1;
}

int session_findlast() {
  
  if (seslist.length == 0) return 0;
  seslist.current = seslist.tail;
  return 1;
}


int session_next() {
  
  if ((seslist.current == NULL) || (seslist.current->next == NULL)) return 0;

  seslist.current = seslist.current->next;
  return 1;
}


int session_prev() {

  if ((seslist.current == NULL) || (seslist.current->prev == NULL)) return 0;

  seslist.current = seslist.current->prev;
  return 1;
}


int session_findnick(char *nick) {

  if ((seslist.current != NULL) && (strcasecmp(seslist.current->ircnick, nick) == 0)) return 1;

  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if (strcasecmp(seslist.current->ircnick, nick) == 0) return 1;
    seslist.current = seslist.current->next;
  }
  seslist.current = seslist.head;
  return 0;
}


int session_findchat(int sockptr) {

  if ((seslist.current != NULL) && (seslist.current->dcc_chat_stat != 0) && 
      (seslist.current->dcc_chat_sockptr == sockptr)) return 1;

  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if ((seslist.current->dcc_chat_stat != 0) && (seslist.current->dcc_chat_sockptr == sockptr)) return 1;
    seslist.current = seslist.current->next;
  }
  seslist.current = seslist.head;
  return 0;
}

int session_findsend(int sockptr) {

  if ((seslist.current != NULL) && (seslist.current->dcc_send_stat != 0) && 
      (seslist.current->dcc_send_sockptr == sockptr)) return 1;

  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if ((seslist.current->dcc_send_stat != 0) && (seslist.current->dcc_send_sockptr == sockptr)) return 1;
    seslist.current = seslist.current->next;
  }
  seslist.current = seslist.head;
  return 0;
}

int session_findget(int sockptr) {

  if ((seslist.current != NULL) && (seslist.current->dcc_get_stat != 0) && 
      (seslist.current->dcc_get_sockptr == sockptr)) return 1;

  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if ((seslist.current->dcc_get_stat != 0) && (seslist.current->dcc_get_sockptr == sockptr)) return 1;
    seslist.current = seslist.current->next;
  }
  seslist.current = seslist.head;
  return 0;
}

int session_findsockptr(char type, int sockptr) {

  switch(type) {
  case T_CHAT: return session_findchat(sockptr);
  case T_SEND: return session_findsend(sockptr);
  case T_GET:  return session_findget (sockptr);
  default: int_err("Error in session_findsockptr()"); return 0;
  }
}


char * session_nickuserhost() {

  if (seslist.current != NULL) return seslist.current->nickuserhost;
  return NULL;
}


char * session_nick() {

  if (seslist.current != NULL) return seslist.current->ircnick;
  return NULL;
}


int important_session() {

  return (seslist.current->identified || seslist.current->operator || seslist.current->dcc_chat_stat ||
	  seslist.current->su || seslist.current->dcc_send_stat || seslist.current->dcc_get_stat);
}


int  no_revenge_flood() {

  time_t tmp;

  tmp = seslist.current->last_revenge;
  if (tmp + 5 < time_null) seslist.current->last_revenge = time_null;
  return (tmp + 5 < time_null);
}


int revenge_should_be_hard() {

  if (seslist.current->revenge_warnings > warnings) return 1;
  seslist.current->revenge_warnings++;
  return (seslist.current->revenge_warnings > warnings);
}

int session_is_flooding() {

  int i, nr_msg=0;
  time_t now;

  now = time_null;
  if (now > seslist.current->flood_last_checked + 10) {
    for (i=0; i<9; i++)
      seslist.current->flood_messages[i] = 0;
    seslist.current->flood_messages[9] = 1;
    seslist.current->flood_last_checked = now;
    return 0;
  }
  else {
    for (i = 0; i < 10 - (now - seslist.current->flood_last_checked); i++)
      seslist.current->flood_messages[i] = seslist.current->flood_messages[i+(now - seslist.current->flood_last_checked)];
    for (i = 10 - (now - seslist.current->flood_last_checked); i < 10; i++)
      seslist.current->flood_messages[i] = 0;
    seslist.current->flood_last_checked++;
    seslist.current->flood_messages[9]++;
    for (i=0; i<10; i++)
      nr_msg += seslist.current->flood_messages[i];
    seslist.current->flood_last_checked = now;
    if (nr_msg >= flood) {
      for (i=0; i<10; i++) seslist.current->flood_messages[i] = 0;
      return 1;
    }
    else {
      return 0;
    }
  }
}


void increase_ontime_sessions() {

  if (!(session_findfirst())) return;
  do
    {
      if (!(session_ismatching('i'))) {
	if (!(userlist_findnick(session_nick()))) createunode(session_nick());
	userlist_incontime();
      }
    }
  while (session_next());
}

void random_delete_away_sessions() {

  if (!(session_findfirst())) return;
  do
    {
      if ((!(is_on_a_channel(session_nick()))) && ((rand() % 5) == 0) && (!(session_superuser()))) {
        delete_session();
      }
      break;
    }
  while (session_next());
}


void session_sendtoirc(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  if (seslist.current->last_irc_message + 2 >= time_null) return;
  seslist.current->last_irc_message = time_null;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  send_to_server("%s", line);
}


void session_sendtoirc_important(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  seslist.current->last_irc_message = time_null;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  send_to_server("%s", line);
}


void session_sendnotice(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  if (seslist.current->dcc_chat_stat == 3)
    send_to_socketptr(seslist.current->dcc_chat_sockptr, "%s", line);
  else
    gen_cmd_notice(seslist.current->ircnick, line, session_sendtoirc);
}

void session_connectchat(unsigned long int addr, unsigned int short port) {

  int ptr;

  switch (seslist.current->dcc_chat_stat) {
  case 0:
    switch(start_connecting2(T_CHAT, &ptr, addr, port)) {
    case -1: return;
    case  0: 
      seslist.current->dcc_chat_stat = 1;
      seslist.current->dcc_chat_sockptr = ptr;
      return;
    case  1:
      session_chat_connection_established(ptr);
      return;
    default:
      int_err("Error in session_connectchat()");
      return;
    }
  case 1: return;
  case 2:
    close_socketptr(seslist.current->dcc_chat_sockptr);
    switch(start_connecting2(T_CHAT, &ptr, addr, port)) {
    case -1: 
      seslist.current->dcc_chat_stat = 0;
      return;
    case  0: 
      seslist.current->dcc_chat_stat = 1;
      seslist.current->dcc_chat_sockptr = ptr;
      return;
    case  1:
      session_chat_connection_established(ptr);
      return;
    default:
      int_err("Error in session_connectchat()");
      return;
    }
  case 3: return;
  default:
    int_err("Error in session_connectchat()");
    return;
  }
}

void session_chat_connection_established(int sockptr) {

  seslist.current->dcc_chat_stat = 3;
  seslist.current->dcc_chat_sockptr = sockptr;
  
  if (seslist.current->saved_command != NULL) {
    session_parse_input(seslist.current->saved_command);  /* After parsing, it's possible that this session no longer */
  }							  /* exists. (reset, server, rehash) */
}						 


void session_chat_connection_lost() {

  session_leaveallchannels();
  seslist.current->dcc_chat_stat = 0;
  seslist.current->dcc_chat_sockptr = -1;
  seslist.current->spy = 0;
}


int session_has_chat() { return seslist.current->dcc_chat_stat == 3; }

void session_stopchat() {

  if (seslist.current->dcc_chat_stat != 0) {
    close_socketptr(seslist.current->dcc_chat_sockptr);
    seslist.current->dcc_chat_sockptr = -1;
    seslist.current->dcc_chat_stat = 0;
  }
}

void session_startchat(char *message) {

  switch(seslist.current->dcc_chat_stat) {
  case 0:
    seslist.current->dcc_chat_sockptr = start_random_listen(T_CHAT);
    if (seslist.current->dcc_chat_sockptr != -1) {
      seslist.current->dcc_chat_stat = 2;
      session_sendtoirc_important("PRIVMSG %s :DCC CHAT chat %lu %d", session_nick(), my_ip(), 
				  give_localport(seslist.current->dcc_chat_sockptr));
      xfree(seslist.current->saved_command);
      seslist.current->saved_command = (char *)xmalloc(strlen(message) + 1);
      strcpy(seslist.current->saved_command, message);
    }
    return;
  case 1: case 2:
    xfree(seslist.current->saved_command);
    seslist.current->saved_command = (char *)xmalloc(strlen(message) + 1);
    strcpy(seslist.current->saved_command, message);
    return;
  case 3:
    return;
  default:
    int_err("Error in session_startchat()");
    return;
  }
}


void session_parse_input(char *input) {

  char *tmp;

  if (on_irc_input(session_nick(), input)) return;

  if (seslist.current->party) {
    if (input[0] == commandchar2) {
      putlog(session_logfile, "%s: %s", session_nickuserhost(), input+1);
      console_send2snoopwindow("[spy] %s: %s", session_nickuserhost(), input+1);
      parse_command(&input[1]);
    }
    else {
      if (input[0] == '\0') return;
      if (strncasecmp(input, "ACTION", 7) == 0) {
	if ((input = index(input, ' ')) == NULL) return;
	input++;
	if ((tmp = index(input, '')) != NULL) *tmp = '\0';
	party_send2partychannel(session_currentchannel(), "* %s %s", session_nick(), input);
	botlist_send2all("actchan %s@%s %d %s", session_nick(), botnick, session_currentchannel(), input);
	if (session_echo())
	  session_send2partychannel(session_currentchannel(), "* %s %s", session_nick(), input);
	else
	  session_send2partychannelexceptcurrent(session_currentchannel(), "* %s %s", session_nick(), input);
	console_send2snoopwindow("(%03d) * %s %s", session_currentchannel(), session_nick(), input);
      }
      else {
	putlog(session_logfile, "channel %d: %s: %s", session_currentchannel(), session_nickuserhost(), input+1);
	party_send2partychannel(session_currentchannel(), "<%s> %s", session_nick(), input);
	botlist_send2all("chan %s@%s %d %s", session_nick(), botnick, session_currentchannel(), input);
	if (session_echo())
	  session_send2partychannel(session_currentchannel(), "<%s> %s", session_nick(), input);
	else
	  session_send2partychannelexceptcurrent(session_currentchannel(), "<%s> %s", session_nick(), input);
	console_send2snoopwindow("(%03d) <%s> %s", session_currentchannel(), session_nick(), input);
      }
    }
  }
  else {
    if (input[0] == commandchar2) input++;
    putlog(session_logfile, "%s: %s", session_nickuserhost(), input);
    console_send2snoopwindow("[spy] %s: %s", session_nickuserhost(), input);
    parse_command(input);
  }
}

int session_identified()    { return (seslist.current->identified  || seslist.current->su); }
int session_operator()      { return (seslist.current->operator || seslist.current->su); }
int session_administrator() { return ((seslist.current->operator == 2) || seslist.current->su); }
int session_superuser()     { return (seslist.current->su); }

void session_setidentify()      { seslist.current->identified = 1; }
void session_resetidentify()    { seslist.current->identified = 0; }
void session_setoperator()      { seslist.current->operator = 1; }
void session_setadministrator() { seslist.current->operator = 2; }
void session_resetoperator()    { seslist.current->operator = 0; }
void session_setsuperuser()     { seslist.current->su = 1; }
void session_resetsuperuser()   { seslist.current->su = 0; }

void session_setspy()    { seslist.current->spy = 1; }
void session_resetspy()  { seslist.current->spy = 0; }
int  session_isspy()     { return seslist.current->spy; }

char session_givechar() {

  if (session_superuser()) return '*';
  if (session_administrator()) return '+';
  if (session_operator()) return '@';
  return '-';
}

void sendtospies(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct session *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = seslist.current;
  if (session_findfirst()) {
    do {
      if (session_isspy())
	session_sendnotice("[spy] %s", line);
    } while (session_next());
  }
  seslist.current = tmp;
}

void int_err(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct session *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = seslist.current;
  if (session_findfirst()) {
    do {
      if (session_isspy())
	session_sendnotice("[error] %s", line);
    } while (session_next());
  }
  seslist.current = tmp;
}

void sendtoadminspies(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct session *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = seslist.current;
  if (session_findfirst()) {
    do {
      if (session_isspy() && session_administrator())
	session_sendnotice("[spy] %s", line);
    } while (session_next());
  }
  seslist.current = tmp;
}

int session_better_protected(char *nick1, char *nick2) {

  struct session *tmp;
  int prot;

  tmp = seslist.current;

  if (!(session_findnick(nick1))) {
    seslist.current = tmp;
    return 0;
  }
  if (!(session_identified())) {
    seslist.current = tmp;
    return 0;
  }
  if (!(userlist_findnick(nick1)))
    prot = 0;
  else
    prot = userlist_protection();

  if (!(session_findnick(nick2))) {
    seslist.current = tmp;
    return 1;
  }
  if (!(session_identified())) {
    seslist.current = tmp;
    return 1;
  }
  seslist.current = tmp;
  if (!(userlist_findnick(nick2)))
    return (prot > 0);
  else
    return (prot > userlist_protection());
}

int  session_joinchannel(int channel) {

  int i;

  if (session_isonchannel(channel)) {
    for (i = 0; i < seslist.current->party; i++) {
      if (channel == seslist.current->party_channels[i]) {
	seslist.current->party_channels[i] = seslist.current->party_channels[0];
	seslist.current->party_channels[0] = channel;
	session_sendnotice("You are now talking to channel %d", channel);
	return 1;
      }
    }
    session_sendnotice("There seems to be a bug in this bot");
  }
  
  if (seslist.current->party >= MAX_CHANNELS) return 0;
  
  if (seslist.current->party != 0)
    seslist.current->party_channels[(int)seslist.current->party] = seslist.current->party_channels[0];
  seslist.current->party_channels[0] = channel;
  seslist.current->party = seslist.current->party + 1;

  if (channel == 0) {
    session_send2partychannel(0, "*** %s has joined the party line", session_nick());
    botlist_send2all("chan %s 0 %s has joined the party line", botnick, session_nick());
    party_send2partychannel(0, "*** %s has joined the party line", session_nick());
    console_send2snoopwindow("(000) *** %s has joined the party line", session_nick());
    return 1;
  }
  if (assoc_channelname(channel)) {
    session_send2partychannel(channel, "*** %s has joined channel \"%s\"", session_nick(), assoc_channelname(channel));
    botlist_send2all("chan %s %d %s has joined channel \"%s\"", botnick, channel, session_nick(), 
		     assoc_channelname(channel));
    party_send2partychannel(channel, "*** %s has joined channel \"%s\"", session_nick(), assoc_channelname(channel));
    console_send2snoopwindow("(%03d) *** %s has joined channel \"%s\"", channel, session_nick(),
					assoc_channelname(channel));
    return 1;
  }
  session_send2partychannel(channel, "*** %s has joined channel %d", session_nick(), channel);
  botlist_send2all("chan %s %d %s has joined channel %d", botnick, channel, session_nick(), channel);
  party_send2partychannel(channel, "*** %s has joined channel %d", session_nick(), channel);
  console_send2snoopwindow("(%03d) *** %s has joined channel %d", channel, session_nick(), channel);
  return 1;
}

int  session_isonchannel(int channel) {

  int i;

  for (i=0; i<seslist.current->party; i++)
    if (seslist.current->party_channels[i] == channel) return 1;
  return 0;
}

void session_send2partychannel(int channel, char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct session *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = seslist.current;
  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if (session_isonchannel(channel)) session_sendnotice("(%d) %s", channel, line);
    seslist.current = seslist.current->next;
  }
  seslist.current = tmp;
}

void session_send2partychannelexceptcurrent(int channel, char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct session *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = seslist.current;
  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if ((session_isonchannel(channel)) && (tmp != seslist.current)) session_sendnotice("(%d) %s", channel, line);
    seslist.current = seslist.current->next;
  }
  seslist.current = tmp;
}


void session_send2party(char *format, ...) {

  char    line[WAYTOBIG_OUT];
  va_list msg;
  struct session *tmp;

  if(!format) return;
  va_start(msg, format);
  vsprintf(line, format, msg);

  line[WAYTOBIG_OUT-1] = '\0';         /* make sure it ends with 0 */
  va_end(msg);

  tmp = seslist.current;
  seslist.current = seslist.head;
  while (seslist.current != NULL) {
    if (seslist.current->party) session_sendnotice("%s", line);
    seslist.current = seslist.current->next;
  }
  seslist.current = tmp;
}

void session_leaveallchannels() {

  while (seslist.current->party) session_leavechannel(seslist.current->party_channels[0]);
}

void session_leavechannel(int channel) {

  int i;
  
  for (i=0; i<seslist.current->party; i++) {
    if (channel == seslist.current->party_channels[i]) {
      if (channel == 0) {
	session_send2partychannel(0, "*** %s has left the party line", session_nick());
	botlist_send2all("chan %s 0 %s has left the party line", botnick, session_nick());
	party_send2partychannel(0, "*** %s has left the party line", session_nick());
	console_send2snoopwindow("(000) *** %s has left the party line", session_nick());
      } 
      else {
	if (assoc_channelname(channel)) {
	  session_send2partychannel(channel, "*** %s has left channel \"%s\"", session_nick(), assoc_channelname(channel));
	  botlist_send2all("chan %s %d %s has left channel \"%s\"", botnick, channel, session_nick(), 
			   assoc_channelname(channel));
	  party_send2partychannel(channel, "*** %s has left channel \"%s\"", session_nick(), assoc_channelname(channel));
	  console_send2snoopwindow("(%03d) *** %s has left channel \"%s\"", channel, session_nick(), 
					      assoc_channelname(channel));
	}
	else {
	  session_send2partychannel(channel, "*** %s has left channel %d", session_nick(), channel);
	  botlist_send2all("chan %s %d %s has left channel %d", botnick, channel, session_nick(), channel);
	  party_send2partychannel(channel, "*** %s has left channel %d", session_nick(), channel);
	  console_send2snoopwindow("(%03d) *** %s has left channel %d",channel, session_nick(), channel);
	}
      }
      seslist.current->party_channels[i] = seslist.current->party_channels[seslist.current->party - 1];
      (seslist.current->party)--;
      return;
    }
  }
}

void session_boot(char *from, char *comment) {

  int i, channel;
  
  for (i=0; i<seslist.current->party; i++) {
    channel = seslist.current->party_channels[i];
    if (channel == 0) {
      session_send2partychannel(0, "*** %s has been kicked off the party line by %s (%s)", session_nick(), from, comment);
      botlist_send2all("chan %s 0 %s has been kicked off the party line by %s (%s)", botnick, session_nick(), from, 
		       comment);
      party_send2partychannel(0, "*** %s has been kicked off the party line by %s (%s)", session_nick(), from, comment);
      console_send2snoopwindow("(000) *** %s has been kicked off the party line by %s (%s)", session_nick(), 
					  from, comment);
    } 
    else {
      if (assoc_channelname(channel)) {
	session_send2partychannel(channel, "*** %s has been kicked off channel \"%s\" by %s (%s)", session_nick(), 
				  assoc_channelname(channel), from, comment);
	botlist_send2all("chan %s %d %s has been kicked off channel \"%s\" by %s (%s)", botnick, channel, session_nick(), 
			 assoc_channelname(channel), from, comment);
	party_send2partychannel(channel, "*** %s has been kicked off channel \"%s\" by %s (%s)", session_nick(), 
				assoc_channelname(channel), from, comment);
	console_send2snoopwindow("(%03d) *** %s has been kicked off channel \"%s\" by %s (%s)", channel, 
					    session_nick(), assoc_channelname(channel), from, comment);
      }
      else {
	session_send2partychannel(channel, "*** %s has been kicked off channel %d by %s (%s)", session_nick(), channel,
				  from, comment);
	botlist_send2all("chan %s %d %s has been kicked off channel %d by %s (%s)", botnick, channel, session_nick(), 
			 channel, from, comment);
	party_send2partychannel(channel, "*** %s has been kicked off channel %d by %s (%s)", session_nick(), channel,
				from, comment);
	console_send2snoopwindow("(%03d) *** %s has been kicked off channel %d by %s (%s)",channel, 
					    session_nick(), channel, from, comment);
      }
    }
  }
  seslist.current->party = 0;
}

int session_currentchannel() { return seslist.current->party ? seslist.current->party_channels[0] : -1; }
int session_isonpartyline()  { return seslist.current->party; }

char *session_currentchannels() {

  static char buf[100];
  int i;

  buf[0]='\0';
  for (i=0; i<seslist.current->party; i++) {
    sprintf(buf, "%s%d  ", buf, seslist.current->party_channels[i]);
  }
  return buf;
}

void session_sendwho2session(int channel) {

  struct session *localcurrent;
  struct session *savecurrent;
  char *nick, *from, ch;

  localcurrent = seslist.current;
  if (session_findfirst()) {
    do {
      if (session_isonchannel(channel)) {
	savecurrent = seslist.current;
	nick = session_nick(); from = session_nickuserhost(); ch = session_givechar();
	seslist.current = localcurrent;
	session_sendnotice("%c%-11s %-10s %s", ch, nick, botnick, index(from, '!')+1);
	seslist.current = savecurrent;
      }
    } while (session_next());
  }
  seslist.current = localcurrent;
}

void session_unaway() {

  int i;
  
  xfree(seslist.current->away);
  seslist.current->away = NULL;

  for (i=0; i<seslist.current->party; i++) {
    session_send2partychannel(seslist.current->party_channels[i], "* %s is BACK!", session_nick());
    party_send2partychannel(seslist.current->party_channels[i], "* %s is BACK!", session_nick());
    botlist_send2all("actchan %s@%s %d is BACK!", session_nick(), botnick, seslist.current->party_channels[i]);
    console_send2snoopwindow("(%03d) * %s is BACK!", seslist.current->party_channels[i], session_nick());
  }
}

void session_setaway(char *reason) {

  int i;

  xfree(seslist.current->away);
  seslist.current->away = (char *) xmalloc (strlen(reason) + 1);
  strcpy(seslist.current->away, reason);
  
  for (i=0; i<seslist.current->party; i++) {
    session_send2partychannel(seslist.current->party_channels[i], "* %s is away: %s", session_nick(), reason);
    party_send2partychannel(seslist.current->party_channels[i], "* %s is away: %s", session_nick(), reason);
    botlist_send2all("actchan %s@%s %d is away: %s", session_nick(), botnick, seslist.current->party_channels[i], reason);
    console_send2snoopwindow("(%03d) * %s is away: %s", seslist.current->party_channels[i], session_nick(), 
					reason);
  }
}

char * session_isaway() { return seslist.current->away; }

void  session_echo_on()  { seslist.current->echo = 1; }
void  session_echo_off() { seslist.current->echo = 0; }
int   session_echo()     { return seslist.current->echo; }


static int session_sethangmanword() {

  FILE *fp;
  unsigned int position;
  char word[100];
  int i;
  char ch;

  xfree(seslist.current->hangman_word);    seslist.current->hangman_word    = NULL;
  xfree(seslist.current->hangman_guessed); seslist.current->hangman_guessed = NULL;

  if ((fp = fopen(wordfile, "r")) == NULL) return 0;

  srand(time_null);
  position = rand();
  position = position << 8;
  position = position | rand();
  position = position << 8;
  position = position | rand();
  position = position << 8;
  position = position | rand();

  fseek(fp, 0L, SEEK_END);
  fseek(fp, position % ftell(fp), SEEK_SET);

  while (1) {
    ch = fgetc(fp);
    if (ch == EOF) { fclose(fp); return session_sethangmanword(); }
    if (ch == '\n') break;
  }

  fscanf(fp, "%99s", word);
  seslist.current->hangman_word    = (char *) xmalloc (strlen(word) + 1);
  seslist.current->hangman_guessed = (char *) xmalloc (strlen(word) + 1);
  strcpy(seslist.current->hangman_word, word);
  strcpy(seslist.current->hangman_guessed, word);
  for (i=0; word[i]!='\0'; i++) {
    seslist.current->hangman_guessed[i]='-';
  }
  fclose(fp);
  return 1;
}



int   session_hangman(char ch) {

  int i, found=0;

  if ((seslist.current->hangman_word == NULL) && (!(session_sethangmanword()))) return 0;
  
  for (i=0; seslist.current->hangman_word[i]!='\0'; i++) {
    if (tolower(seslist.current->hangman_word[i]) == tolower(ch)) { seslist.current->hangman_guessed[i] = ch; found = 1; }
  }
  if (found == 0) seslist.current->hangman_errors++;
  
  session_sendnotice("    ______");
  session_sendnotice("    |    |");
  session_sendnotice("    |    %c", seslist.current->hangman_errors > 0 ? 'o' : ' ');
  session_sendnotice("    |   %c%c%c", seslist.current->hangman_errors > 4 ? '/' : ' ', seslist.current->hangman_errors > 1 ? '|' : ' ',
		     seslist.current->hangman_errors > 5 ? '\\' : ' ');
  session_sendnotice("    |    %c", seslist.current->hangman_errors > 2 ? '|' : ' ');
  session_sendnotice("    |   %c %c", seslist.current->hangman_errors > 3 ? '/' : ' ', seslist.current->hangman_errors > 6 ? '\\' : ' ');
  session_sendnotice("  __|_____");
  session_sendnotice("  |      |___");
  session_sendnotice("  |_________|");
  session_sendnotice("");
  session_sendnotice(" Word: %s", seslist.current->hangman_guessed);
  if (seslist.current->hangman_errors > 6) {
    session_sendnotice(" Sorry, the word was \"%s\"", seslist.current->hangman_word);
    xfree(seslist.current->hangman_word);    seslist.current->hangman_word    = NULL;
    xfree(seslist.current->hangman_guessed); seslist.current->hangman_guessed = NULL;
    seslist.current->hangman_errors = 0;
    return 1;
  }
  if (strcasecmp(seslist.current->hangman_guessed, seslist.current->hangman_word) == 0) {
    session_sendnotice(" You got it!");
    xfree(seslist.current->hangman_word);    seslist.current->hangman_word    = NULL;
    xfree(seslist.current->hangman_guessed); seslist.current->hangman_guessed = NULL;
    seslist.current->hangman_errors = 0;
    return 1;
  }
  return 1;
}

char *session_hangmanword() { return seslist.current->hangman_word; }


char *session_ftp_directory() { return seslist.current->ftp_directory; }

int   session_change_dir(char *dir) {

  char *ch, *testdir;
  struct stat file_info;

  if (strcasecmp(dir, "") == 0) return 1;
  if (strcasecmp(dir, ".") == 0) return 1;
  if (strcasecmp(dir, "/") == 0) {
    seslist.current->ftp_directory[0] = '\0';
    seslist.current->ftp_directory = (char *) xrealloc (seslist.current->ftp_directory, 1);
    return 1;
  }
  if (strcasecmp(dir, "..") == 0) {
    if ((ch = rindex(seslist.current->ftp_directory, '/')) == NULL) {
      seslist.current->ftp_directory[0] = '\0';
      seslist.current->ftp_directory = xrealloc (seslist.current->ftp_directory, 1);
      return 1;
    }
    *ch = '\0';
    seslist.current->ftp_directory = (char *) xrealloc (seslist.current->ftp_directory, 
							strlen(seslist.current->ftp_directory) + 1);
    return 1;
  }
  testdir = (char *) xmalloc (strlen(ftp_root_dir)+strlen(seslist.current->ftp_directory)+strlen(dir)+2);
  sprintf(testdir, "%s%s/%s", ftp_root_dir, seslist.current->ftp_directory, dir);
  if ((stat(testdir, &file_info)==0) && (S_ISDIR(file_info.st_mode))) {
    xfree(testdir);
    seslist.current->ftp_directory = (char *) xrealloc (seslist.current->ftp_directory, 
							strlen(seslist.current->ftp_directory) + 1 + strlen(dir) + 1);
    if (seslist.current->ftp_directory[0] != '\0') strcat(seslist.current->ftp_directory, "/");
    strcat(seslist.current->ftp_directory, dir);
    return 1;
  }
  else {
    xfree(testdir);
    return 0;
  }
}

int  session_sendfile(char *file) {

  if (seslist.current->dcc_send_stat != 0) return 0;
  if ((seslist.current->dcc_send_fp = fopen(file, "r")) == NULL) return 0;
  if ((seslist.current->dcc_send_sockptr = start_random_listen(T_SEND)) == -1) { 
    fclose(seslist.current->dcc_send_fp);
    return 0;
  }
  seslist.current->dcc_send_stat = 1;
  seslist.current->dcc_send_bytes_sent = 0;
  fseek(seslist.current->dcc_send_fp, 0L, SEEK_END);
  session_sendtoirc_important("PRIVMSG %s :DCC SEND %s %lu %d %lu", session_nick(), 
			      rindex(file, '/') ? rindex(file, '/') + 1 : file, 
			      my_ip(), give_localport(seslist.current->dcc_send_sockptr), 
			      ftell(seslist.current->dcc_send_fp));
  fseek(seslist.current->dcc_send_fp, 0L, SEEK_SET);
  return 1;
}

void session_closesend() {

  if (seslist.current->dcc_send_stat == 0) return;
  close_socketptr(seslist.current->dcc_send_sockptr);
  fclose(seslist.current->dcc_send_fp);
  seslist.current->dcc_send_stat = 0;
  seslist.current->dcc_send_sockptr = -1;
  seslist.current->dcc_send_fp = NULL;
  seslist.current->dcc_send_bytes_sent = 0;
}

void session_sendestablished(int sockptr) {

  int nbytes;
  char buffer[WAYTOBIG];

  seslist.current->dcc_send_stat = 2;
  seslist.current->dcc_send_sockptr = sockptr;

  if ((nbytes = fread(buffer, sizeof(char), dcc_block_size, seslist.current->dcc_send_fp)) <= 0) {
    session_closesend();
    return;
  }

  send_to_socketptr_raw(seslist.current->dcc_send_sockptr, buffer, nbytes);
  seslist.current->dcc_send_bytes_sent = nbytes;
}

void session_parsesend(unsigned int bytes_received) {

  int nbytes;
  char buffer[WAYTOBIG];

  if (bytes_received > seslist.current->dcc_send_bytes_sent) {
    session_closesend();
    return;
  }
  if (bytes_received == seslist.current->dcc_send_bytes_sent) {
    if ((nbytes = fread(buffer, sizeof(char), dcc_block_size, seslist.current->dcc_send_fp)) <= 0) {
      session_closesend();
      return;
    }
    send_to_socketptr_raw(seslist.current->dcc_send_sockptr, buffer, nbytes);
    seslist.current->dcc_send_bytes_sent += nbytes;
  }
}

int  session_getfile(char *file, unsigned long int addr, unsigned int short port, unsigned int filelength) {

  int ptr;
  char *destfile;

  if (seslist.current->dcc_get_stat) return 0;

  destfile = (char *) xmalloc (strlen(incoming_dir)+strlen(file)+1);
  sprintf(destfile, "%s%s", incoming_dir, file);
  if ((seslist.current->dcc_get_fp = fopen(destfile, "w")) == NULL) {
    xfree(destfile);
    return 0;
  }
  xfree(destfile);
  seslist.current->dcc_get_filesize = filelength;

  switch(start_connecting2(T_GET, &ptr, addr, port)) {
  case -1:
    fclose(seslist.current->dcc_get_fp);
    seslist.current->dcc_get_filesize = 0;
    return 0;
  case  0:
    seslist.current->dcc_get_stat = 1;
    seslist.current->dcc_get_sockptr = ptr;
    return 1;
  case  1:
    seslist.current->dcc_get_sockptr = ptr;
    session_get_connection_established(ptr);
    return 1;
  default:
    int_err("Error in session_getfile()");
    return 0;
  }
}

void session_get_connection_established() {

  seslist.current->dcc_get_stat = 2;
}

void session_getlost() {

  seslist.current->dcc_get_stat = 0;
  seslist.current->dcc_get_sockptr = -1;
  fclose(seslist.current->dcc_get_fp);
  seslist.current->dcc_get_bytes_read = 0;
  seslist.current->dcc_get_filesize = 0;
}

void session_parseget(char *buffer, size_t size) {

  unsigned int tmpbytes;

  seslist.current->dcc_get_bytes_read += size;
  if (seslist.current->dcc_get_bytes_read > seslist.current->dcc_get_filesize) {
    close_socketptr(seslist.current->dcc_get_sockptr);
    session_getlost();
  }
  fwrite(buffer, sizeof(char), size, seslist.current->dcc_get_fp);
  tmpbytes = htonl(seslist.current->dcc_get_bytes_read);
  send_to_socketptr_raw(seslist.current->dcc_get_sockptr, (char *)&tmpbytes, sizeof(unsigned int));
}

void session_closeget() {
  
  if (seslist.current->dcc_get_stat) {
    close_socketptr(seslist.current->dcc_get_sockptr);
    session_getlost();
  }
}

int session_chatstatus() { return seslist.current->dcc_chat_stat; }
int session_sendstatus() { return seslist.current->dcc_send_stat; }
int session_getstatus() { return seslist.current->dcc_get_stat; }

int sessions_count(int function()) {

  struct session *tmp;
  int i=0;

  tmp = seslist.current;
  if (session_findfirst()) {
    do
      {
	if (function()) i++;
      }
    while (session_next());
  }
  seslist.current = tmp;
  return i;
}

int   session_length() { return seslist.length; }


void session_marknotify() { seslist.current->notify = time_null; }

void session_startnotify() {

  char *nick[5];
  int   counter=0;

  if (session_findfirst()) {
    do
      {
	if ((!(is_on_a_channel(session_nick()))) && (seslist.current->notify + 120 < time_null)) {
	  delete_session();
	  session_findfirst();
	}
      }
    while (session_next());
  }

  if (session_findfirst()) {
    do
      {
	if (!(is_on_a_channel(session_nick()))) {
	  nick[counter] = session_nick();
	  counter++;
	  if (counter >= 5) {
	    send_to_server("USERHOST %s %s %s %s %s", nick[0], nick[1], nick[2], nick[3], nick[4]);
	    counter = 0;
	  }
	}
      }
    while (session_next());
    if (counter) {
      send_to_server("USERHOST %s %s %s %s %s", nick[0], counter > 1 ? nick[1] : "", counter > 2 ? nick[2] : "", 
		     counter > 3 ? nick[3] : "", counter > 4 ? nick[4] : "");
    }
  }
}


static int convert_linechar(char ch) {

  switch (ch) {
  case 'a': return SESSION_A_LINE;
  case 'b': return SESSION_B_LINE;
  case 'd': return SESSION_D_LINE;
  case 'i': return SESSION_I_LINE;
  case 'o': return SESSION_O_LINE;
  default:  return -1;
  }
}

int session_ismatching(char linetype) { 

  int idx;

  if ((idx = convert_linechar(linetype)) < 0) return 0;
  return seslist.current->matching_lines[idx];
}

void sessions_update_deleted_line(char linetype, char *mask) {

  struct session *tmp;
  int idx;

  if ((idx = convert_linechar(linetype)) < 0) return;
  tmp = seslist.current;

  for (seslist.current = seslist.head; seslist.current != NULL; seslist.current = seslist.current->next) {
    if (seslist.current->matching_lines[idx] && match(mask, seslist.current->nickuserhost)) {
      seslist.current->matching_lines[idx] = does_this_match(linetype, seslist.current->nickuserhost);
    }
  }

  seslist.current = tmp;
}

void sessions_update_added_line(char linetype, char *mask) {

  struct session *tmp;
  int idx;

  if ((idx = convert_linechar(linetype)) < 0) return;
  tmp = seslist.current;

  for (seslist.current = seslist.head; seslist.current != NULL; seslist.current = seslist.current->next) {
    if (seslist.current->matching_lines[idx] == 0 && match(mask, seslist.current->nickuserhost)) {
      seslist.current->matching_lines[idx] = 1;
    }
  }

  seslist.current = tmp;
}
