
/*********************************************************/
/**                                                     **/
/**       session.h (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                     **/
/**  Database of user who have contact with me via IRC  **/
/**                                                     **/
/*********************************************************/


#ifndef SESSION_H_
#define SESSION_H_

#include <stdio.h>

void init_allow();
void save_allow_settings(FILE *);
int  load_allow_settings(FILE *);

#define ALLOW_ALL    1
#define ALLOW_OLINE  2
#define ALLOW_OPER   3
#define ALLOW_ALINE  4
#define ALLOW_ADMIN  5
#define ALLOW_SU     6

#define UPLOAD_ALLOWED    1
#define AUTOOP_ALLOWED    2
#define PLACEBANS_ALLOWED 3
#define ALWAYSOP_ALLOWED  4
#define GETOPS_ALLOWED    5
#define GIVEOPS_ALLOWED   6
#define UNBAN_ALLOWED     7

void setallow(int who, int what);
int  curallow(int what);
int  session_isallowed(int what);

/* Creating, deleting and updating sessions: */

void create_session(char *nickuserhost);     /* Create a session              */
void delete_session();                       /* Delete a session              */
void session_newnick(char *nick);            /* Changes the nick of a session */
void delete_all_sessions();                  /* Delete all sessions           */

/* Manipulating the current in the sessionlist: */

int session_findfirst();                     /* Places current on the first session                             */
int session_findlast();                      /* Makes current point to the last session                         */
int session_next();                          /* Moves current to the next session                               */
int session_prev();                          /* Moves current to the previous session                           */
int session_findnick(char *nick);            /* Places current on session with nick "nick"                      */
int session_findsockptr(char, int);          /* Current points to node with connection "char" and sockptr "int" */

/* Get information from the current session: */

char * session_nickuserhost();               /* Returns the nick!user@host of the current session */
char * session_nick();                       /* Returns the nick of the current session           */

int  important_session();                    /* Returns 0 when notify'ing is not needed */
int  session_identified();
int  session_operator();
int  session_administrator();
int  session_superuser();
char session_givechar();                     /* '-' for nothing, '@' for oper, '+' for admin and '*' for su */

void session_setspy();
void session_resetspy();
int  session_isspy();
void sendtospies(char *, ...);
void sendtoadminspies(char *, ...);
void int_err(char *, ...);

void session_setidentify();
void session_resetidentify();
void session_setoperator();
void session_setadministrator();
void session_resetoperator();
void session_setsuperuser();
void session_resetsuperuser();

int no_revenge_flood();                      /* Returns 1 when this session hasn't been punished for 5 secs */
int revenge_should_be_hard();                /* Returns 0 if the user should only be warned                 */
int session_is_flooding();                   /* Returns 1 when session is flooding                          */

int session_better_protected(char *nick1, char *nick2); /* returns 1 when nick1 is better protected, or 0 when that
							 * is not the case, current remains unchanged
							 */

void increase_ontime_sessions();             /* Increases the ontime of every session in the userlist */
void random_delete_away_sessions();          /* Could delete one or more session of users who are not on a channel */

void session_sendtoirc(char *, ...);            /* Sends message to IRC, message could be rejected to prevent flooding */
void session_sendtoirc_important(char *, ...);  /* Sends message to IRC                                                */
void session_sendnotice(char *, ...);           /* Send notice to this session to IRC, message could be rejected       */

void  session_connectchat(unsigned long int, unsigned short int);
void  session_chat_connection_established();
void  session_chat_connection_lost();
void  session_parse_input(char *);
int   session_has_chat();
void  session_startchat(char *);
void  session_stopchat();
int   session_chatstatus();

int   session_joinchannel(int);
void  session_leavechannel(int);
void  session_send2party(char *, ...);
void  session_send2partychannel(int, char *, ...);
void  session_send2partychannelexceptcurrent(int, char *, ...);
void  session_leaveallchannels();
void  session_boot(char *from, char *comment);
int   session_isonchannel(int);
int   session_currentchannel();
void  session_sendwho2session(int);
char *session_currentchannels();
int   session_isonpartyline();

void  session_setaway(char *reason);
void  session_unaway();
char *session_isaway();

void  session_echo_on();
void  session_echo_off();
int   session_echo();

int   session_hangman(char);
char *session_hangmanword();

char *session_ftp_directory();
int   session_change_dir(char *);
int   session_sendfile(char *);
void  session_closesend();
void  session_sendestablished(int);
void  session_parsesend(unsigned int);
int   session_sendstatus();

int   session_getfile(char *, unsigned long int, unsigned short int, unsigned int);
void  session_get_connection_established();
void  session_getlost();
void  session_parseget(char *, size_t);
void  session_closeget();
int   session_getstatus();

int   sessions_count(int());
int   session_length();

void  session_startnotify();
void  session_marknotify();

int   session_ismatching(char linetype);
/* session_ismatching('a') returns whether the current session matches with an aline. This is faster than calling
** does_this_match from lines.c, since there is no matching done in routine. Matching the session was done when the
** session was created or when on of the following functions were called:
*/

void sessions_update_deleted_line(char linetype, char *mask);
void sessions_update_added_line(char linetype, char *mask);
/* These functions should be called AFTER a line (a, b, d, i or o, other line-types are not useful for sessions)
** has been created or has been deleted (be moved to the H-line actually ..)
*/

#endif
