
/*********************************************************/
/**                                                     **/
/**       tcl_procs.c (c) 1997, Emiel Bruijntjes        **/
/**                                                     **/
/**   The procs that can be called from a tcl-script    **/
/**                                                     **/
/*********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#include "config.h"

#ifndef NO_TCL

#include <tcl.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>


#include "tcl_procs.h"
#include "xwin.h"
#include "assoc.h"
#include "bots.h"
#include "partyline.h"
#include "session.h"
#include "console.h"
#include "bottree.h"
#include "cfgfile.h"
#include "channel.h"
#include "misc.h"
#include "server.h"
#include "history.h"
#include "cmd_console.h"
#include "party_console.h"
#include "irc_console.h"
#include "gencmd.h"
#include "lines.h"
#include "notes.h"
#include "sescmd.h"
#include "userlist.h"
#include "on.h"
#include "values.h"


extern char *xfont;
extern char  botnick[10];
extern char  current_nick[10];


int tcl_xconsole(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"xconsole display\"", NULL);
    return TCL_ERROR;
  }
  switch (create_X_session(argv[1])) {
  case 0:
    Tcl_AppendResult(interp, "Connection to X-server failed or no authorization", NULL);
    return TCL_ERROR;
  case 1:
    Tcl_AppendResult(interp, "Connection to X-server failed: All connections in use", NULL);
    return TCL_ERROR;
  case 2:
    Tcl_AppendResult(interp, "Could not load font: ", xfont, NULL);
    return TCL_ERROR;
  case 3:
    return TCL_OK;
  }
  return TCL_OK;
}


int tcl_addassoc(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"addassoc channel name\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[1]);
  if (chan < 1) {
    Tcl_AppendResult(interp, "Channelnumbers must be greater than 0", NULL);
    return TCL_ERROR;
  }
  if (chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumbers may not be greater than 99999", NULL);
    return TCL_ERROR;
  }
  if (strlen(argv[2]) > 20) {
    Tcl_AppendResult(interp, "Channelnames may not be longer than 20 characters", NULL);
    return TCL_ERROR;
  }
  if (isdigit(argv[2][0]) || (argv[2][0] == '-')) {
    Tcl_AppendResult(interp, "First character of the channelname can't be a digit", NULL);
    return TCL_ERROR;
  }
  addassoc(chan, argv[2]);
  botlist_send2all("assoc %d %s", chan, argv[2]);
  botlist_send2all("chan %s %d This channel has been names \"%s\"", botnick, chan, argv[2]);
  party_send2partychannel(chan, "*** This channel has been named \"%s\"", argv[2]);
  session_send2partychannel(chan, "*** This channel has been named \"%s\"", argv[2]);
  console_send2snoopwindow("(%03d) *** This channel has been named \"%s\"", chan, argv[2]);
  return TCL_OK;
}


int tcl_delassoc(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"delassoc channel\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[1]);
  if (chan < 1) {
    Tcl_AppendResult(interp, "Channelnumbers must be greater than 0", NULL);
    return TCL_ERROR;
  }
  if (chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumbers may not be greater than 99999", NULL);
    return TCL_ERROR;
  }
  delassoc(chan);
  botlist_send2all("assoc %d 0", chan, argv[2]);
  botlist_send2all("chan %s %d The name of channel %d has been removed", botnick, chan, chan);
  party_send2partychannel(chan, "*** The name of channel %d has been removed", chan);
  session_send2partychannel(chan, "*** The name of channel %d has been removed", chan);
  console_send2snoopwindow("(%03d) *** The name of channel %d has been removed", chan, chan);
  return TCL_OK;
}


int tcl_assoc_name(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int   chan;
  char *name;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"assoc_name channel\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[1]);
  if (chan < 1) {
    Tcl_AppendResult(interp, "Channelnumbers must be greater than 0", NULL);
    return TCL_ERROR;
  }
  if (chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumbers may not be greater than 99999", NULL);
    return TCL_ERROR;
  }
  if ((name = assoc_channelname(chan)) == NULL) {
    Tcl_AppendResult(interp, "", NULL);
    return TCL_OK;
  }
  else {
    Tcl_AppendResult(interp, name, NULL);
    return TCL_OK;
  }
}


int tcl_assoc_number(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int nr;
  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"assoc_number channel\"", NULL);
    return TCL_ERROR;
  }
  if ((nr = assoc_channelnumber(argv[1])) < 0)
    strcpy(buf, "0");
  else
    sprintf(buf, "%d", nr);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_addbot(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"addbot nick host port\"", NULL);
    return TCL_ERROR;
  }
  if (botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, "There is already a bot called ", argv[1], NULL);
    return TCL_ERROR;
  }
  addbot(argv[1], argv[2], atoi(argv[3]));
  return TCL_OK;
}


int tcl_delbot(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"delbot bot\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  delbot(argv[1]);
  return TCL_OK;
}


int tcl_botlist(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botlist\"", NULL);
    return TCL_ERROR;
  }
  if (botlist_findfirst()) {
    do
      {
	Tcl_AppendElement(interp, botlist_nick());
      }
    while (botlist_findnext());
  }
  return TCL_OK;
}


int tcl_bothost(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"bothost bot ?host?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, botlist_host(), NULL);
  if (argc == 3) botlist_setaddress(argv[2], botlist_port());
  return TCL_OK;
}


int tcl_botport(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botport bot ?port?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", botlist_port());
  Tcl_AppendResult(interp, buf, NULL);
  if (argc == 3) botlist_setaddress(botlist_host(), atoi(argv[2]));
  return TCL_OK;
}


int tcl_botpasswd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botpasswd bot ?passwd?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, botlist_password(), NULL);
  if (argc == 3) botlist_setpassword(argv[2]);
  return TCL_OK;
}


int tcl_botprotection(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botprotection bot ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", botlist_protection() ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  if (argc == 3) {
    if (atoi(argv[2]))
      botlist_setprot();
    else
      botlist_resetprot();
  }
  return TCL_OK;
}


int tcl_botnickuserhost(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botnickuserhost bot ?nick!user@host?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, botlist_nickuserhost(), NULL);
  if (argc == 3) botlist_setnickuserhost(argv[2]);
  return TCL_OK;
}


int tcl_botversion(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botversion bot ?version?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, botlist_version(), NULL);
  if (argc == 3) botlist_setversion(argv[2]);
  return TCL_OK;
}


int tcl_botautoconnect(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botautoconnect bot ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!botlist_findbot(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", botlist_autoconnect() ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  if (argc == 3) {
    if (atoi(argv[2]))
      botlist_setautoconnect();
    else
      botlist_resetautoconnect();
  }
  return TCL_OK;
}


int tcl_send2botnet(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2botnet msg\"", NULL);
    return TCL_ERROR;
  }
  botlist_send2all("%s", argv[1]);
  return TCL_OK;
}


int tcl_link(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"link bot ?bot?\"", NULL);
    return TCL_ERROR;
  }
  if ((argc == 3) && (strcasecmp(argv[1], botnick) == 0)) {
    argc = 2;
    argv[1] = argv[2];
  }
  if (argc == 3) {
    if (part_of_botnet(argv[2])) {
      Tcl_AppendResult(interp, argv[2], " is already part of the botnet", NULL);
      return TCL_ERROR;
    }
    if (!(part_of_botnet(argv[1]))) {
      Tcl_AppendResult(interp, argv[1], " is not part of the botnet", NULL);
      return TCL_ERROR;
    }
    else {
      bottree_send2bot(argv[1], "link %s %s %s", botnick, argv[1], argv[2]);
      return TCL_OK;
    }
  }

  /* argc = 2 */
  if (part_of_botnet(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " is already part of the botnet", NULL);
    return TCL_ERROR;
  }
  if (!(botlist_findbot(argv[1]))) {
    Tcl_AppendResult(interp, argv[1], " not found in the botlist", argv[1]);
    return TCL_ERROR;
  }
  botlist_startconnection();
  return TCL_OK;
}


int tcl_unlink(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"unlink bot\"", NULL);
    return TCL_ERROR;
  }
  if (strcasecmp(argv[1], botnick) == 0) {
    Tcl_AppendResult(interp, "How should I unlink from myself?", NULL);
    return TCL_ERROR;
  }
  if (!(part_of_botnet(argv[1]))) {
    Tcl_AppendResult(interp, argv[1], " is not part of the botnet", NULL);
    return TCL_ERROR;
  }
  if (strcasecmp(geefparent(argv[1]), botnick) == 0) {
    if (!(botlist_findbot(argv[1]))) {
      Tcl_AppendResult(interp, argv[1], " not found. (This is probably a bug in the bot)", NULL);
      return TCL_ERROR;
    }
    botlist_breakconnection();
    return TCL_OK;
  }
  bottree_send2bot(argv[1], "unlink %s %s %s", botnick, geefparent(argv[1]), argv[1]);
  return TCL_OK;
}


int tcl_send2bot(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2bot bot msg\"", NULL);
    return TCL_ERROR;
  }
  if (!part_of_botnet(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " is not part of the botnet", NULL);
    return TCL_ERROR;
  }
  if (strcasecmp(botnick, argv[1]) == 0) {
    Tcl_AppendResult(interp, "I cannot send messages to myself", NULL);
    return TCL_ERROR;
  }
  bottree_send2bot(argv[1], "%s", argv[2]);
  return TCL_OK;
}


int tcl_partofbotnet(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"partofbotnet bot\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", part_of_botnet(argv[1]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_samebranche(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"samebranche bot1 bot2\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", same_branche(argv[1], argv[2]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_readconfigfile(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"readconfigfile\"", NULL);
    return TCL_ERROR;
  }
  read_cfg_file();
  return TCL_OK;
}


int tcl_writeconfigfile(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"writeconfigfile\"", NULL);
    return TCL_ERROR;
  }
  write_cfg_file();
  return TCL_OK;
}


int tcl_addchan(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char key[30];
  int  mode, limit;

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"addchan channel modes\"", NULL);
    return TCL_ERROR;
  }
  if (!(is_channel(argv[1]))) {
    Tcl_AppendResult(interp, "wrong # args: should be \"addchan #channel modes\"", NULL);
    return TCL_ERROR;
  }
  if (strlen(argv[1]) > MAXKANAALLEN - 1) {
    Tcl_AppendResult(interp, "channelname is too long", NULL);
    return TCL_ERROR;
  }
  if (!convert_mode_string(argv[2], &mode, key, &limit)) {
    Tcl_AppendResult(interp, "Illegal mode-string", NULL);
    return TCL_ERROR;
  }
  addchan2watch(argv[1], mode, key, limit);
  return TCL_OK;
}


int tcl_delchan(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"delchan channel\"", NULL);
    return TCL_ERROR;
  }
  delchan2watch(argv[1]);
  if (channellist_findchan(argv[1])) {
    send_to_server("PART :%s", channellist_channelname());
  }
  return TCL_OK;
}


int tcl_chanlist(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int i;

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"chanlist\"", NULL);
    return TCL_ERROR;
  }
  for (i=1; channel2watch(i); i++) {
    Tcl_AppendElement(interp, channel2watch(i));
  }
  return TCL_OK;
}


int tcl_activechannels(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"activechannels\"", NULL);
    return TCL_ERROR;
  }
  if (channellist_findfirst()) {
    do
      {
	Tcl_AppendElement(interp, channellist_channelname());
      }
    while (channellist_next());
  }
  return TCL_OK;
}


int tcl_activatechannel(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"activatechannel channel\"", NULL);
    return TCL_ERROR;
  }
  if (!channellist_findchan(argv[1])) {
    Tcl_AppendResult(interp, "I'm not on channel ", argv[1], NULL);
    return TCL_ERROR;
  }
  channellist_makefirst();
  return TCL_OK;
}


int tcl_numberofusers(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"numberofusers channel\"", NULL);
    return TCL_ERROR;
  }
  if (!channellist_findchan(argv[1])) {
    Tcl_AppendResult(interp, "I'm not on channel ", argv[1], NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", channellist_numberofusers());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_chanusers(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"chanusers channel\"", NULL);
    return TCL_ERROR;
  }
  if (!channellist_findchan(argv[1])) {
    Tcl_AppendResult(interp, "I'm not on channel ", argv[1], NULL);
    return TCL_ERROR;
  }
  if (chanusers_findfirst(channellist_chanusers())) {
    do
      {
	Tcl_AppendElement(interp, chanusers_nick(channellist_chanusers()));
      }
    while (chanusers_findnext(channellist_chanusers()));
  }
  return TCL_OK;
}


int tcl_isonchan(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"isonchan nick channel\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", isonchannel(argv[2], argv[1]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_ischanop(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"ischanop nick channel\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", ischanneloperator(argv[2], argv[1]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_banlist(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int i;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"banlist channel\"", NULL);
    return TCL_ERROR;
  }
  for (i=0; i<MAX_BANS; i++) {
    if (returnban(i)) {
      Tcl_AppendElement(interp, returnban(i));
    }
  }
  return TCL_OK;
}


int tcl_isbanned(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"isbanned nick!user@host channel\"", NULL);
    return TCL_ERROR;
  }
  if (!channellist_findchan(argv[2])) {
    Tcl_AppendResult(interp, "I'm not on channel ", argv[2], NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", is_banned(argv[1]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}
  

int tcl_delayop(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"delayop nick channel\"", NULL);
    return TCL_ERROR;
  }
  if (!channellist_findchan(argv[2])) {
    Tcl_AppendResult(interp, "I'm not on channel ", argv[2], NULL);
    return TCL_ERROR;
  }
  channellist_adduser2op(argv[1]);
  return TCL_OK;
}


int tcl_consoleinput(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"consoleinput msg\"", NULL);
    return TCL_ERROR;
  }
  add_history(argv[1]);
  switch (console_mode()) {
  case CONSOLE_IRC_MODE:
    parse_irc_console(argv[1]);
    break;
  case CONSOLE_PARTY_MODE:
    parse_party_console(argv[1]);
    break;
  case CONSOLE_COMMAND_MODE:
    parse_cmd_console(argv[1]);
    break;
  }
  return TCL_OK;
}


int tcl_putwindow1(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"putwindow1 msg\"", NULL);
    return TCL_ERROR;
  }
  console_send2snoopwindow(argv[1]);
  return TCL_OK;
}


int tcl_putwindow2(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"putwindow2 msg\"", NULL);
    return TCL_ERROR;
  }
  console_send2ircwindow(argv[1]);
  return TCL_OK;
}


int tcl_switchmode(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"switchmode\"", NULL);
    return TCL_ERROR;
  }
  switch (console_mode()) {
  case CONSOLE_IRC_MODE:
    console_setmode(CONSOLE_COMMAND_MODE);
    break;
  case CONSOLE_COMMAND_MODE:
    console_setmode(CONSOLE_PARTY_MODE);
    break;
  case CONSOLE_PARTY_MODE:
    console_setmode(CONSOLE_IRC_MODE);
    break;
  }
  return TCL_OK;
}


int tcl_currentmode(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"currentmode\"", NULL);
    return TCL_ERROR;
  }
  switch (console_mode()) {
  case CONSOLE_IRC_MODE:
    Tcl_AppendResult(interp, "0", NULL);
    break;
  case CONSOLE_COMMAND_MODE:
    Tcl_AppendResult(interp, "1", NULL);
    break;
  case CONSOLE_PARTY_MODE:
    Tcl_AppendResult(interp, "2", NULL);
    break;
  }
  return TCL_OK;
}


int tcl_signoff(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"signoff ?msg?\"", NULL);
    return TCL_ERROR;
  }
  if (argc == 2) {
    gen_cmd_signoff(argv[1]);
  }
  else {
    gen_cmd_signoff("Signoff from Tcl-script");
  }
  /* unreachable code */
  return TCL_OK;
}


int tcl_msg(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {
  
  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"msg nick/channel msg\"", NULL);
    return TCL_ERROR;
  }
  gen_cmd_msg(argv[1], argv[2], send_to_server);
  return TCL_OK;
}


int tcl_notice(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {
  
  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"notice nick/channel msg\"", NULL);
    return TCL_ERROR;
  }
  gen_cmd_notice(argv[1], argv[2], send_to_server);
  return TCL_OK;
}


int tcl_me(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {
  
  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"me nick/channel msg\"", NULL);
    return TCL_ERROR;
  }
  gen_cmd_me(argv[1], argv[2], send_to_server);
  return TCL_OK;
}


int tcl_sendfile2user(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"sendfile2user user filename\"", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1])) {
    catfile(argv[2], session_sendnotice, argv[1]);
    return TCL_OK;
  }
  if (party_findnick(argv[1])) {
    catfile(argv[2], party_send2current, argv[1]);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, argv[1], ": No such nickname", NULL);
  return TCL_ERROR;
}


int tcl_sendfile2window1(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"sendfile2window1 filename\"", NULL);
    return TCL_ERROR;
  }
  catfile(argv[2], console_send2snoopwindow, "Tcl-script");
  return TCL_OK;
}


int tcl_sendfile2window2(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"sendfile2window2 filename\"", NULL);
    return TCL_ERROR;
  }
  catfile(argv[2], console_send2ircwindow, "Tcl-script");
  return TCL_OK;
}


int tcl_linereq(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int normal, locked;

  if (argc != 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linereq type normal locked\"", NULL);
    return TCL_ERROR;
  }
  switch (tolower(argv[1][0])) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
  normal = atoi(argv[2]);
  locked = atoi(argv[3]);
  if (normal < 0 || normal > 3 || locked < 0 || locked > 3) {
    Tcl_AppendResult(interp, "Linereq out of range", NULL);
    return TCL_ERROR;
  }
  change_require_to_change(tolower(argv[1][0]), normal, locked);
  return TCL_OK;
}


int tcl_normalreq(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"normalreq linetype\"", NULL);
    return TCL_ERROR;
  }
  switch (tolower(argv[1][0])) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", required_to_change(tolower(argv[1][0])));
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_lockreq(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"lockreq linetype\"", NULL);
    return TCL_ERROR;
  }
  switch (tolower(argv[1][0])) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", required_to_change_when_locked(tolower(argv[1][0])));
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_addline(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];
  
  if (argc != 5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"addline linetype mask comment from\"", NULL);
    return TCL_ERROR;
  }
  switch (tolower(argv[1][0])) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", addline(tolower(argv[1][0]), argv[2], argv[3], argv[4]));
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_moveline(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"moveline linenr from\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'o': case 'b': case 'd': case 'i': case 'x': case 'n':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    moveline('h', type, type == 'a' ? 0 : type == 'b' ? 1 : type == 'd' ? 2 : type == 'i' ? 3 : type == 'o' ? 4 : 
	     type == 'x' ? 5 : 6, argv[2]);
    if (type != 'x' && type != 'n')
      sessions_update_deleted_line(type, lines_mask('h'));
    return TCL_OK;
  case 'h':
    Tcl_AppendResult(interp, "You are not allowed to move H-lines", NULL);
    return TCL_ERROR;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_delline(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;
  char *mask;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"delline linenr\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'o': case 'b': case 'd': case 'i': case 'x': case 'n': case 'h':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    mask = (char *) xmalloc (strlen(lines_mask(type)) + 1);
    strcpy(mask, lines_mask(type));
    delline(type);
    if (type != 'x' && type != 'n')
      sessions_update_deleted_line(type, mask);
    xfree(mask);
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_matchinglines(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int nr=1;
  char buf[20];

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"matchinglines linetype string\"", NULL);
    return TCL_ERROR;
  }
  switch (tolower(argv[1][0])) {
  case 'a': case 'o': case 'b': case 'd': case 'i': case 'x': case 'n': case 'h':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal linetype", NULL);
    return TCL_ERROR;
  }
  while ((nr = next_match(tolower(argv[1][0]), argv[2], nr)) != 0) {
    sprintf(buf, "%c%d", toupper(argv[1][0]), nr);
    Tcl_AppendElement(interp, buf);
  }
  return TCL_OK;
}


int tcl_linecomment(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linecomment linenr ?comment?\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'b': case 'd': case 'i': case 'n':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    Tcl_AppendResult(interp, lines_comment(type), NULL);
    if (argc == 3) lines_setcomment(type, argv[2]);
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_linepass(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linepass linenr ?password?\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'o': case 'x':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    Tcl_AppendResult(interp, lines_comment(type), NULL);
    if (argc == 3) lines_setcomment(type, argv[2]);
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_linemask(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linemask linenr\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    Tcl_AppendResult(interp, lines_mask(type), NULL);
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_linefrom(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linefrom linenr\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    Tcl_AppendResult(interp, lines_from(type), NULL);
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_linetime(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;
  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linetime linenr\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    sprintf(buf, "%lu", (unsigned long)lines_time(type));
    Tcl_AppendResult(interp, buf, NULL);
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_linelocked(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char type;
  int  linenr;

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linelocked linenr ?value?\"", NULL);
    return TCL_ERROR;
  }
  type = tolower(argv[1][0]);
  linenr = atoi(&argv[1][1]);
  switch(type) {
  case 'a': case 'b': case 'd': case 'h': case 'i': case 'o': case 'x': case 'n':
    if (linenr < 1) {
      Tcl_AppendResult(interp, "Linenumbers should be at least one", NULL);
      return TCL_ERROR;
    }
    if (!(lines_findnr(type, linenr))) {
      Tcl_AppendResult(interp, "Line ", argv[1], " not found", NULL);
      return TCL_ERROR;
    }
    Tcl_AppendResult(interp, lines_islocked(type) ? "1" : "0", NULL);
    if (argc == 3) {
      if (atoi(argv[2]))
	lines_setlock(type);
      else
	lines_setunlock(type);
    }
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal line-type", NULL);
    return TCL_ERROR;
  }
}


int tcl_numberoflines(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"numberoflines linetype\"", NULL);
    return TCL_ERROR;
  }
  switch(tolower(argv[1][0])) {
  case 'a': case 'o': case 'b': case 'd': case 'i': case 'x': case 'n': case 'h':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal linetype", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", numberoflines(tolower(argv[1][0])));
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_numberoflockedlines(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"numberoflockedlines linetype\"", NULL);
    return TCL_ERROR;
  }
  switch(tolower(argv[1][0])) {
  case 'a': case 'o': case 'b': case 'd': case 'i': case 'x': case 'n': case 'h':
    break;
  default:
    Tcl_AppendResult(interp, "Illegal linetype", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", numberoflockedlines(tolower(argv[1][0])));
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_linetimeout(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"linetimeout linetype ?timeout?\"", NULL);
    return TCL_ERROR;
  }
  switch(tolower(argv[1][0])) {
  case 'a': 
    sprintf(buf, "%d", alinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setalinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'b': 
    sprintf(buf, "%d", blinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setblinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'd': 
    sprintf(buf, "%d", dlinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setdlinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'h': 
    sprintf(buf, "%d", hlinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      sethlinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'i': 
    sprintf(buf, "%d", ilinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setilinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'o': 
    sprintf(buf, "%d", olinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setolinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'x': 
    sprintf(buf, "%d", xlinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setxlinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  case 'n': 
    sprintf(buf, "%d", nlinetimeout());
    Tcl_AppendResult(interp, buf, NULL);
    if (argc == 3) {
      setnlinetimeout(atol(argv[2]));
    }
    return TCL_OK;
  default:
    Tcl_AppendResult(interp, "Illegal linetype", NULL);
    return TCL_ERROR;
  }
}


int tcl_time2string(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"time2string time\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, time2string(atol(argv[1])), NULL);
  return TCL_OK;
}


int tcl_putlog(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"putlog file msg\"", NULL);
    return TCL_ERROR;
  }
  putlog(argv[1], "%s", argv[2]);
  return TCL_OK;
}


int tcl_addnote(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"addnote to from msg\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", addnote(argv[1], argv[2], argv[3]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_hasnote(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"hasnote user\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", user_has_note(argv[1]) ? 1 : 0);
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_outsidepartyusers(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"outsidepartyusers\"", NULL);
    return TCL_ERROR;
  }
  if (party_findfirst()) {
    do
      {
	Tcl_AppendElement(interp, party_nick());
      }
    while (party_next());
  }
  return TCL_OK;
}


int tcl_ircpartyusers(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"ircpartyusers\"", NULL);
    return TCL_ERROR;
  }
  if (session_findfirst()) {
    do
      {
	if (session_isonpartyline()) {
	  Tcl_AppendElement(interp, session_nick());
	}
      }
    while (session_next());
  }
  return TCL_OK;
}


int tcl_userfrom(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"userfrom user\"", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1])) {
    Tcl_AppendResult(interp, session_nickuserhost(), NULL);
    return TCL_OK;
  }
  if (party_findnick(argv[1])) {
    Tcl_AppendResult(interp, party_from(), NULL);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, argv[1], ": user not found", NULL);
  return TCL_ERROR;
}


int tcl_send2user(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2user user msg\"", NULL);
    return TCL_ERROR;
  }
  if (party_findnick(argv[1])) {
    party_send2current(argv[2]);
    return TCL_OK;
  }
  if (session_findnick(argv[1])) {
    session_sendnotice(argv[2]);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, argv[1], ": user not found", NULL);
  return TCL_ERROR;
}


int tcl_partyline_botmsg(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"partyline_botmsg msg\"", NULL);
    return TCL_ERROR;
  }
  console_send2snoopwindow("(all) *** %s", argv[1]);
  party_send2party("*** %s", argv[1]);
  session_send2party("*** %s", argv[1]);
  botlist_send2all("chat %s %s", botnick, argv[1]);
  return TCL_OK;
}


int tcl_partychan_botmsg(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"partychan_botmsg channr msg\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[1]);
  if (chan < 0 || chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumber out of range", NULL);
    return TCL_ERROR;
  }
  console_send2snoopwindow("(%03d) *** %s", chan, argv[2]);
  party_send2partychannel(chan, "*** %s", argv[2]);
  session_send2partychannel(chan, "*** %s", argv[2]);
  botlist_send2all("chan %s %d %s", botnick, chan, argv[2]);
  return TCL_OK;
}


int tcl_partychan_usermsg(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"partychan_usermsg channr user msg\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[1]);
  if (chan < 0 || chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumber out of range", NULL);
    return TCL_ERROR;
  }
  console_send2snoopwindow("(%03d) <%s> %s", chan, argv[2], argv[3]);
  party_send2partychannel(chan, "<%s> %s", argv[2], argv[3]);
  session_send2partychannel(chan, "<%s> %s", argv[2], argv[3]);
  botlist_send2all("chan %s@%s %d %s", argv[2], botnick, chan, argv[3]);
  return TCL_OK;
}


int tcl_partychan_useract(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"partychan_useract channr user msg\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[1]);
  if (chan < 0 || chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumber out of range", NULL);
    return TCL_ERROR;
  }
  console_send2snoopwindow("(%03d) * %s %s", chan, argv[2], argv[3]);
  party_send2partychannel(chan, "* %s %s", argv[2], argv[3]);
  session_send2partychannel(chan, "* %s %s", argv[2], argv[3]);
  botlist_send2all("actchan %s@%s %d %s", argv[2], botnick, chan, argv[3]);
  return TCL_OK;
}


int tcl_boot(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"boot user from comment\"", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1])) {
    session_boot(argv[2], argv[3]);
    return TCL_OK;
  }
  if (party_findnick(argv[1])) {
    party_boot(argv[2], argv[3]);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, argv[1], " is not on the partyline", NULL);
  return TCL_ERROR;
}


int tcl_identified(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"identified user ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, "0", NULL);
    return TCL_OK;
  }

  if (session_identified())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  if (argc == 3) {
    if (atoi(argv[2]))
      session_setidentify();
    else
      session_resetidentify();
  }
  return TCL_OK;
}


int tcl_operator(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"operator user ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, "0", NULL);
    return TCL_OK;
  }

  if (session_operator())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  if (argc == 3) {
    if (atoi(argv[2]))
      session_setoperator();
    else
      session_resetoperator();
  }
  return TCL_OK;
}


int tcl_administrator(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"administrator user ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, "0", NULL);
    return TCL_OK;
  }

  if (session_administrator())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  if (argc == 3) {
    if (atoi(argv[2]))
      session_setadministrator();
    else
      session_resetoperator();
  }
  return TCL_OK;
}


int tcl_superuser(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"superuser user ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, "0", NULL);
    return TCL_OK;
  }

  if (session_superuser())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  if (argc == 3) {
    if (atoi(argv[2]))
      session_setsuperuser();
    else
      session_resetsuperuser();
  }
  return TCL_OK;
}


int tcl_isspy(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"isspy user ?value?\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, "0", NULL);
    return TCL_OK;
  }

  if (session_isspy())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  if (argc == 3) {
    if (atoi(argv[2]))
      session_setspy();
    else
      session_resetspy();
  }
  return TCL_OK;
}


int tcl_send2spies(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2spies string\"", NULL);
    return TCL_ERROR;
  }
  sendtospies("%s", argv[1]);
  return TCL_OK;
}


int tcl_send2adminspies(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2adminspies string\"", NULL);
    return TCL_ERROR;
  }
  sendtoadminspies("%s", argv[1]);
  return TCL_OK;
}


int tcl_betterprotected(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"betterprotected user1 user2\"", NULL);
    return TCL_ERROR;
  }
  if (session_better_protected(argv[1], argv[2]))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_send2irc_by(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2irc_by nick string\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1]))
    send_to_server("%s", argv[2]);
  else
    session_sendtoirc("%s", argv[2]);
  return TCL_OK;
}


int tcl_dosend2irc_by(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"dosend2irc_by nick string\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1]))
    send_to_server("%s", argv[2]);
  else
    session_sendtoirc_important("%s", argv[2]);
  return TCL_OK;
}


int tcl_simulircinput(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"simulircinput nick string\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " does not have a session active", NULL);
    return TCL_ERROR;
  }
  session_parse_input(argv[2]);
  return TCL_OK;
}


int tcl_simuloutsideinput(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"simuloutsideinput nick string\"", NULL);
    return TCL_ERROR;
  }
  if (!party_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " is not connected to me", NULL);
    return TCL_ERROR;
  }
  party_parse_input(argv[2]);
  return TCL_OK;
}


int tcl_haschat(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"haschat user\"", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1]) && session_has_chat())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_startchat(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"startchat user ?string?\"", NULL);
    return TCL_ERROR;
  }
  if (!session_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], " not found", NULL);
    return TCL_ERROR;
  }
  session_startchat(argc == 3 ? argv[2] : "");
  return TCL_OK;
}

int tcl_joinparty(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"joinparty user channr\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[2]);
  if (chan < 0) {
    Tcl_AppendResult(interp, "Channelnumbers must be greater than 0", NULL);
    return TCL_ERROR;
  }
  if (chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumbers may not be greater than 99999", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1])) {
    session_joinchannel(chan);
    return TCL_OK;
  }
  if (party_findnick(argv[1])) {
    party_joinchannel(chan);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, argv[1], " not found", NULL);
  return TCL_ERROR;
}


int tcl_leaveparty(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int chan;

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"leaveparty user channr\"", NULL);
    return TCL_ERROR;
  }
  chan = atoi(argv[2]);
  if (chan < 0) {
    Tcl_AppendResult(interp, "Channelnumbers must be greater than 0", NULL);
    return TCL_ERROR;
  }
  if (chan > 99999) {
    Tcl_AppendResult(interp, "Channelnumbers may not be greater than 99999", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1])) {
    session_leavechannel(chan);
    return TCL_OK;
  }
  if (party_findnick(argv[1])) {
    party_leavechannel(chan);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, argv[1], " not found", NULL);
  return TCL_ERROR;
}


int tcl_hasline(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"hasline user linetype\"", NULL);
    return TCL_ERROR;
  }
  if (session_findnick(argv[1]) && session_ismatching(argv[2][0]))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_cmdreq(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"cmdreq command ?value?\"", NULL);
    return TCL_ERROR;
  }
  switch (current_cmd_require(argv[1])) {
  case ALL:
    Tcl_AppendResult(interp, "0", NULL);
    break;
  case IDENT:
    Tcl_AppendResult(interp, "1", NULL);
    break;
  case OPER:
    Tcl_AppendResult(interp, "2", NULL);
    break;
  case ADMIN:
    Tcl_AppendResult(interp, "3", NULL);
    break;
  case SU:
    Tcl_AppendResult(interp, "4", NULL);
    break;
  default:
    Tcl_AppendResult(interp, argv[1], ": command not found", NULL);
    return TCL_ERROR;
  }
  if (argc == 3) {
    switch(atoi(argv[2])) {
    case 0:
      change_cmd_require(argv[1], ALL);
      break;
    case 1:
      change_cmd_require(argv[1], IDENT);
      break;
    case 2:
      change_cmd_require(argv[1], OPER);
      break;
    case 3:
      change_cmd_require(argv[1], ADMIN);
      break;
    case 4:
      change_cmd_require(argv[1], SU);
      break;
    default:
      Tcl_AppendResult(interp, "Illegal value specified", NULL);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}


int tcl_userlistmaxlength(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"userlistmaxlength ?length?\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_maxlength());
  Tcl_AppendResult(interp, buf, NULL);
  if (argc == 2)
    userlist_setmaxlength(atoi(argv[1]));
  return TCL_OK;
}


int tcl_userlistlength(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"userlistlength\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_length());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_userlistkept(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"userlistkept\"", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_kept());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_inuserlist(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"inuserlist user\"", NULL);
    return TCL_ERROR;
  }
  if (userlist_findnick(argv[1]))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_incjoins(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incjoins user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incjoins();
  return TCL_OK;
}


int tcl_incleaves(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incleaves user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incleaves();
  return TCL_OK;
}


int tcl_incnicks(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incnicks user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incnicks();
  return TCL_OK;
}


int tcl_inckicks(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"inckicks user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_inckicks();
  return TCL_OK;
}


int tcl_inckickeds(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"inckickeds user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_inckickeds();
  return TCL_OK;
}


int tcl_incquits(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incquits user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incquits();
  return TCL_OK;
}


int tcl_incpublics(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incpublics user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incpublics();
  return TCL_OK;
}


int tcl_incmodes(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incmodes user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incmodes();
  return TCL_OK;
}


int tcl_incontime(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"incontime user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_incontime();
  return TCL_OK;
}


int tcl_lasthost(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char *bang, *at;

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"lasthost user ?nick!user@host?\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, userlist_lasthost(), NULL);
  if (argc == 3) {
    if (((bang = index(argv[2], '!')) == NULL) || ((at = index(bang, '@')) == NULL)) {
      Tcl_AppendResult(interp, "wrong # args: should be \"lasthost user ?nick!user@host?\"", NULL);
      return TCL_ERROR;
    }
    userlist_setlasthost(argv[2]);
  }
  return TCL_OK;
}


int tcl_upasswd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"upasswd user ?password?\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, userlist_password(), NULL);
  if (argc == 3) {
    if (index(argv[2], ' ') != NULL) {
      Tcl_AppendResult(interp, "Illegal password specified", NULL);
      return TCL_ERROR;
    }
    userlist_setpassword(argv[2]);
  }
  return TCL_OK;
}


int tcl_keep(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"keep user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  if (userlist_setkeepentry())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_unkeep(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"unkeep user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  userlist_resetkeepentry();
  return TCL_OK;
}


int tcl_iskept(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"iskept user\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  if (userlist_keepentry())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_manip(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc < 2 || argc > 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"manip user ?change?\"", NULL);
    return TCL_ERROR;
  }
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_manipulation());
  Tcl_AppendResult(interp, buf, NULL);
  if (argc == 3)
    userlist_setmanipulation(atoi(argv[2]));
  return TCL_OK;
}


int tcl_protection(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"protection user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_protection());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_publics(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_publics user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_publics());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_kicks(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_kicks user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_kicks());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_kickeds(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_kickeds user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_kickeds());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_joins(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_joins user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_joins());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_leaves(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_leaves user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_leaves());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_quits(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_quits user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_quits());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_nicks(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_nicks user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_nicks());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nr_modes(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nr_modes user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_modes());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_ontime(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"ontime user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%d", userlist_ontime());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_lastseen(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  char buf[20];

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"lastseen user\"", NULL);
    return TCL_ERROR;
  } 
  if (!userlist_findnick(argv[1])) {
    Tcl_AppendResult(interp, argv[1], ": not in userlist", NULL);
    return TCL_ERROR;
  }
  sprintf(buf, "%lu", (unsigned long)userlist_lastseen());
  Tcl_AppendResult(interp, buf, NULL);
  return TCL_OK;
}


int tcl_nextserver(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"nextserver\"", NULL);
    return TCL_ERROR;
  }
  nextserver();
  return TCL_OK;
}


int tcl_send2irc(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2irc string\"", NULL);
    return TCL_ERROR;
  }
  send_to_server("%s", argv[1]);
  return TCL_OK;
}


int tcl_send2irc_fast(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2irc_fast string\"", NULL);
    return TCL_ERROR;
  }
  send_fast_to_server("%s", argv[1]);
  return TCL_OK;
}


int tcl_send2irc_now(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"send2irc_now string\"", NULL);
    return TCL_ERROR;
  }
  send_now_to_server("%s", argv[1]);
  return TCL_OK;
}


int tcl_on(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  int type=0, i;
  struct {
    char *on_function;
    int   on_type;
  } on_msgs[] = {
    {"msg",           ON_MSG           },
    {"public",        ON_PUBLIC        },
    {"public_notice", ON_PUBLIC_NOTICE },
    {"notice",        ON_NOTICE        },
    {"join",          ON_JOIN          },
    {"part",          ON_PART          },
    {"nick",          ON_NICK          },
    {"kick",          ON_KICK          },
    {"quit",          ON_QUIT          },
    {"mode",          ON_MODE          },
    {"connect",       ON_CONNECT       },
    {"console",       ON_CONSOLE       },
    {"link",          ON_LINK          },
    {"unlink",        ON_UNLINK        },
    {"ctcp",          ON_CTCP          },
    {"ctcp_reply",    ON_CTCP_REPLY    },
    {"botnet",        ON_BOTNET        },
    {"outside",       ON_OUTSIDE       },
    {"irc_input",     ON_IRC_INPUT     },
    {"irc_cmd",       ON_IRC_CMD       },
    {"outside_cmd",   ON_OUTSIDE_CMD   },
    {"timer",         ON_TIMER         },
    {NULL,            0                },
  };

  if (argc < 4 || argc > 5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"on priority on_type mask ?proc?\"", NULL);
    return TCL_ERROR;
  }
  if (argv[1][0] == '-' && argc == 5) {
    Tcl_AppendResult(interp, "You cannot specify a proc when deleting on-functions", NULL);
    return TCL_OK;
  }
  if (argv[1][0] == '+' && argc == 4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"on +priority on_type mask proc\"", NULL);
    return TCL_ERROR;
  }
  if (argv[1][0] != '-' && argv[1][0] != '+') {
    Tcl_AppendResult(interp, "First character of priority should be \"-\" or \"+\"", NULL);
    return TCL_ERROR;
  }
  for (i=0; on_msgs[i].on_function; i++) {
    if (strcasecmp(on_msgs[i].on_function, argv[2]) == 0) {
      type = on_msgs[i].on_type;
      break;
    }
  }
  if (on_msgs[i].on_function == NULL) {
    Tcl_AppendResult(interp, argv[2], ": No such on-function", NULL);
    return TCL_ERROR;
  }
  if (argv[1][0] == '-')
    del_on(type, abs(atoi(argv[1])), argv[3]);
  else
    add_on(type, atoi(argv[1]), argv[3], argv[4]);
  return TCL_OK;
}


int tcl_match(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"match pattern string\"", NULL);
    return TCL_ERROR;
  }
  if (match(argv[1], argv[2]))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
  return TCL_OK;
}


int tcl_login(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"login ?login?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returnlogin(), NULL);
  if (argc == 2) {
    setbotlogin(argv[1]);
  }
  return TCL_OK;
}


int tcl_ircname(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"ircname ?ircname?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returnircname(), NULL);
  if (argc == 2) {
    setircname(argv[1]);
  }
  return TCL_OK;
}


int tcl_commandchar1(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"commandchar1 ?char?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returncommandchar1(), NULL);
  if (argc == 2) {
    setcommandchar1(argv[1]);
  }
  return TCL_OK;
}


int tcl_commandchar2(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"commandchar2 ?char?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returncommandchar2(), NULL);
  if (argc == 2) {
    setcommandchar2(argv[1]);
  }
  return TCL_OK;
}


int tcl_commandchar3(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"commandchar3 ?char?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returncommandchar3(), NULL);
  if (argc == 2) {
    setcommandchar3(argv[1]);
  }
  return TCL_OK;
}


int tcl_su_passwd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"su_passwd ?password?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returnsu_password(), NULL);
  if (argc == 2) {
    setsu_password(argv[1]);
  }
  return TCL_OK;
}


int tcl_maintainer(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc < 1 || argc > 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"maintainer ?string?\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, returnmaintainer(), NULL);
  if (argc == 2) {
    setmaintainer(argv[1]);
  }
  return TCL_OK;
}


int tcl_botnick(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"botnick\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, botnick, NULL);
  return TCL_OK;
}


int tcl_ircnick(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]) {

  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"ircnick\"", NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp, current_nick, NULL);
  return TCL_OK;
}


void tcl_initcommands(Tcl_Interp *interp) {

  Tcl_CreateCommand(interp, "xconsole", tcl_xconsole, NULL, NULL);
  Tcl_CreateCommand(interp, "addassoc", tcl_addassoc, NULL, NULL);
  Tcl_CreateCommand(interp, "delassoc", tcl_delassoc, NULL, NULL);
  Tcl_CreateCommand(interp, "assoc_name", tcl_assoc_name, NULL, NULL);
  Tcl_CreateCommand(interp, "assoc_number", tcl_assoc_number, NULL, NULL);
  Tcl_CreateCommand(interp, "addbot", tcl_addbot, NULL, NULL);
  Tcl_CreateCommand(interp, "delbot", tcl_delbot, NULL, NULL);
  Tcl_CreateCommand(interp, "botlist", tcl_botlist, NULL, NULL);
  Tcl_CreateCommand(interp, "bothost", tcl_bothost, NULL, NULL);
  Tcl_CreateCommand(interp, "botport", tcl_botport, NULL, NULL);
  Tcl_CreateCommand(interp, "botpasswd", tcl_botpasswd, NULL, NULL);
  Tcl_CreateCommand(interp, "botprotection", tcl_botprotection, NULL, NULL);
  Tcl_CreateCommand(interp, "botnickuserhost", tcl_botnickuserhost, NULL, NULL);
  Tcl_CreateCommand(interp, "botversion", tcl_botversion, NULL, NULL);
  Tcl_CreateCommand(interp, "botautoconnect", tcl_botautoconnect, NULL, NULL);
  Tcl_CreateCommand(interp, "send2botnet", tcl_send2botnet, NULL, NULL);
  Tcl_CreateCommand(interp, "link", tcl_link, NULL, NULL);
  Tcl_CreateCommand(interp, "unlink", tcl_unlink, NULL, NULL);
  Tcl_CreateCommand(interp, "send2bot", tcl_send2bot, NULL, NULL);
  Tcl_CreateCommand(interp, "partofbotnet", tcl_partofbotnet, NULL, NULL);
  Tcl_CreateCommand(interp, "samebranche", tcl_samebranche, NULL, NULL);
  Tcl_CreateCommand(interp, "readconfigfile", tcl_readconfigfile, NULL, NULL);
  Tcl_CreateCommand(interp, "writeconfigfile", tcl_writeconfigfile, NULL, NULL);
  Tcl_CreateCommand(interp, "addchan", tcl_addchan, NULL, NULL);
  Tcl_CreateCommand(interp, "delchan", tcl_delchan, NULL, NULL);
  Tcl_CreateCommand(interp, "chanlist", tcl_chanlist, NULL, NULL);
  Tcl_CreateCommand(interp, "activechannels", tcl_activechannels, NULL, NULL);
  Tcl_CreateCommand(interp, "activatechannel", tcl_activatechannel, NULL, NULL);
  Tcl_CreateCommand(interp, "numberofusers", tcl_numberofusers, NULL, NULL);
  Tcl_CreateCommand(interp, "chanusers", tcl_chanusers, NULL, NULL);
  Tcl_CreateCommand(interp, "isonchan", tcl_isonchan, NULL, NULL);
  Tcl_CreateCommand(interp, "ischanop", tcl_ischanop, NULL, NULL);
  Tcl_CreateCommand(interp, "banlist", tcl_banlist, NULL, NULL);
  Tcl_CreateCommand(interp, "isbanned", tcl_isbanned, NULL, NULL);
  Tcl_CreateCommand(interp, "delayop", tcl_delayop, NULL, NULL);
  Tcl_CreateCommand(interp, "consoleinput", tcl_consoleinput, NULL, NULL);
  Tcl_CreateCommand(interp, "putwindow1", tcl_putwindow1, NULL, NULL);
  Tcl_CreateCommand(interp, "putwindow2", tcl_putwindow2, NULL, NULL);
  Tcl_CreateCommand(interp, "switchmode", tcl_switchmode, NULL, NULL);
  Tcl_CreateCommand(interp, "currentmode", tcl_currentmode, NULL, NULL);
  Tcl_CreateCommand(interp, "signoff", tcl_signoff, NULL, NULL);
  Tcl_CreateCommand(interp, "msg", tcl_msg, NULL, NULL);
  Tcl_CreateCommand(interp, "notice", tcl_notice, NULL, NULL);
  Tcl_CreateCommand(interp, "me", tcl_me, NULL, NULL);
  Tcl_CreateCommand(interp, "sendfile2user", tcl_sendfile2user, NULL, NULL);
  Tcl_CreateCommand(interp, "sendfile2window1", tcl_sendfile2window1, NULL, NULL);
  Tcl_CreateCommand(interp, "sendfile2window2", tcl_sendfile2window2, NULL, NULL);
  Tcl_CreateCommand(interp, "linereq", tcl_linereq, NULL, NULL);
  Tcl_CreateCommand(interp, "normalreq", tcl_normalreq, NULL, NULL);
  Tcl_CreateCommand(interp, "lockreq", tcl_lockreq, NULL, NULL);
  Tcl_CreateCommand(interp, "addline", tcl_addline, NULL, NULL);
  Tcl_CreateCommand(interp, "moveline", tcl_moveline, NULL, NULL);
  Tcl_CreateCommand(interp, "delline", tcl_delline, NULL, NULL);
  Tcl_CreateCommand(interp, "matchinglines", tcl_matchinglines, NULL, NULL);
  Tcl_CreateCommand(interp, "linecomment", tcl_linecomment, NULL, NULL);
  Tcl_CreateCommand(interp, "linepass", tcl_linepass, NULL, NULL);
  Tcl_CreateCommand(interp, "linemask", tcl_linemask, NULL, NULL);
  Tcl_CreateCommand(interp, "linefrom", tcl_linefrom, NULL, NULL);
  Tcl_CreateCommand(interp, "linetime", tcl_linetime, NULL, NULL);
  Tcl_CreateCommand(interp, "linelocked", tcl_linelocked, NULL, NULL);
  Tcl_CreateCommand(interp, "numberoflines", tcl_numberoflines, NULL, NULL);
  Tcl_CreateCommand(interp, "numberoflockedlines", tcl_numberoflockedlines, NULL, NULL);
  Tcl_CreateCommand(interp, "linetimeout", tcl_linetimeout, NULL, NULL);
  Tcl_CreateCommand(interp, "time2string", tcl_time2string, NULL, NULL);
  Tcl_CreateCommand(interp, "putlog", tcl_putlog, NULL, NULL);
  Tcl_CreateCommand(interp, "addnote", tcl_addnote, NULL, NULL);
  Tcl_CreateCommand(interp, "hasnote", tcl_hasnote, NULL, NULL);
  Tcl_CreateCommand(interp, "outsidepartyusers", tcl_outsidepartyusers, NULL, NULL);
  Tcl_CreateCommand(interp, "ircpartyusers", tcl_ircpartyusers, NULL, NULL);
  Tcl_CreateCommand(interp, "userfrom", tcl_userfrom, NULL, NULL);
  Tcl_CreateCommand(interp, "send2user", tcl_send2user, NULL, NULL);
  Tcl_CreateCommand(interp, "partyline_botmsg", tcl_partyline_botmsg, NULL, NULL);
  Tcl_CreateCommand(interp, "partychan_botmsg", tcl_partychan_botmsg, NULL, NULL);
  Tcl_CreateCommand(interp, "partychan_usermsg", tcl_partychan_usermsg, NULL, NULL);
  Tcl_CreateCommand(interp, "boot", tcl_boot, NULL, NULL);
  Tcl_CreateCommand(interp, "identified", tcl_identified, NULL, NULL);
  Tcl_CreateCommand(interp, "operator", tcl_operator, NULL, NULL);
  Tcl_CreateCommand(interp, "administrator", tcl_administrator, NULL, NULL);
  Tcl_CreateCommand(interp, "superuser", tcl_superuser, NULL, NULL);
  Tcl_CreateCommand(interp, "spy", tcl_isspy, NULL, NULL);
  Tcl_CreateCommand(interp, "send2spies", tcl_send2spies, NULL, NULL);
  Tcl_CreateCommand(interp, "send2adminspies", tcl_send2adminspies, NULL, NULL);
  Tcl_CreateCommand(interp, "betterprotected", tcl_betterprotected, NULL, NULL);
  Tcl_CreateCommand(interp, "send2irc_by", tcl_send2irc_by, NULL, NULL);
  Tcl_CreateCommand(interp, "dosend2irc_by", tcl_dosend2irc_by, NULL, NULL);
  Tcl_CreateCommand(interp, "simulircinput", tcl_simulircinput, NULL, NULL);
  Tcl_CreateCommand(interp, "simuloutsideinput", tcl_simuloutsideinput, NULL, NULL);
  Tcl_CreateCommand(interp, "haschat", tcl_haschat, NULL, NULL);
  Tcl_CreateCommand(interp, "startchat", tcl_startchat, NULL, NULL);
  Tcl_CreateCommand(interp, "joinparty", tcl_joinparty, NULL, NULL);
  Tcl_CreateCommand(interp, "leaveparty", tcl_leaveparty, NULL, NULL);
  Tcl_CreateCommand(interp, "hasline", tcl_hasline, NULL, NULL);
  Tcl_CreateCommand(interp, "cmdreq", tcl_cmdreq, NULL, NULL);
  Tcl_CreateCommand(interp, "userlistmaxlength", tcl_userlistmaxlength, NULL, NULL);
  Tcl_CreateCommand(interp, "userlistlength", tcl_userlistlength, NULL, NULL);
  Tcl_CreateCommand(interp, "userlistkept", tcl_userlistkept, NULL, NULL);
  Tcl_CreateCommand(interp, "inuserlist", tcl_inuserlist, NULL, NULL);
  Tcl_CreateCommand(interp, "incjoins", tcl_incjoins, NULL, NULL);
  Tcl_CreateCommand(interp, "incleaves", tcl_incleaves, NULL, NULL);
  Tcl_CreateCommand(interp, "incnicks", tcl_incnicks, NULL, NULL);
  Tcl_CreateCommand(interp, "inckicks", tcl_inckicks, NULL, NULL);
  Tcl_CreateCommand(interp, "inckickeds", tcl_inckickeds, NULL, NULL);
  Tcl_CreateCommand(interp, "incquits", tcl_incquits, NULL, NULL);
  Tcl_CreateCommand(interp, "incpublics", tcl_incpublics, NULL, NULL);
  Tcl_CreateCommand(interp, "incmodes", tcl_incmodes, NULL, NULL);
  Tcl_CreateCommand(interp, "incontime", tcl_incontime, NULL, NULL);
  Tcl_CreateCommand(interp, "lasthost", tcl_lasthost, NULL, NULL);
  Tcl_CreateCommand(interp, "upasswd", tcl_upasswd, NULL, NULL);
  Tcl_CreateCommand(interp, "keep", tcl_keep, NULL, NULL);
  Tcl_CreateCommand(interp, "unkeep", tcl_unkeep, NULL, NULL);
  Tcl_CreateCommand(interp, "iskept", tcl_iskept, NULL, NULL);
  Tcl_CreateCommand(interp, "manip", tcl_manip, NULL, NULL);
  Tcl_CreateCommand(interp, "protection", tcl_protection, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_publics", tcl_nr_publics, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_kicks", tcl_nr_kicks, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_kickeds", tcl_nr_kickeds, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_joins", tcl_nr_joins, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_leaves", tcl_nr_leaves, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_quits", tcl_nr_quits, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_nicks", tcl_nr_nicks, NULL, NULL);
  Tcl_CreateCommand(interp, "nr_modes", tcl_nr_modes, NULL, NULL);
  Tcl_CreateCommand(interp, "ontime", tcl_ontime, NULL, NULL);
  Tcl_CreateCommand(interp, "lastseen", tcl_lastseen, NULL, NULL);
  Tcl_CreateCommand(interp, "nextserver", tcl_nextserver, NULL, NULL);
  Tcl_CreateCommand(interp, "send2irc", tcl_send2irc, NULL, NULL);
  Tcl_CreateCommand(interp, "send2irc_fast", tcl_send2irc_fast, NULL, NULL);
  Tcl_CreateCommand(interp, "send2irc_now", tcl_send2irc_now, NULL, NULL);
  Tcl_CreateCommand(interp, "on", tcl_on, NULL, NULL);
  Tcl_CreateCommand(interp, "match", tcl_match, NULL, NULL);
  Tcl_CreateCommand(interp, "login", tcl_login, NULL, NULL);
  Tcl_CreateCommand(interp, "ircname", tcl_ircname, NULL, NULL);
  Tcl_CreateCommand(interp, "commandchar1", tcl_commandchar1, NULL, NULL);
  Tcl_CreateCommand(interp, "commandchar2", tcl_commandchar2, NULL, NULL);
  Tcl_CreateCommand(interp, "commandchar3", tcl_commandchar3, NULL, NULL);
  Tcl_CreateCommand(interp, "su_passwd", tcl_su_passwd, NULL, NULL);
  Tcl_CreateCommand(interp, "botnick", tcl_botnick, NULL, NULL);
  Tcl_CreateCommand(interp, "ircnick", tcl_ircnick, NULL, NULL);
}

#endif
