
/*********************************************************/
/**                                                     **/
/**       toptree.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                     **/
/**               Sorting the userlist                  **/
/**                                                     **/
/*********************************************************/


#include <stdlib.h>

#include "toptree.h"
#include "misc.h"

void createtree(struct toptree **tree, char *nick, int aantal) {

  *tree = (struct toptree *) xmalloc (sizeof(struct toptree));
  strcpy((*tree)->nick, nick);
  (*tree)->aantal = aantal;
  (*tree)->left = NULL;
  (*tree)->right = NULL;
}

void deltree(struct toptree *tree) {

  if (tree == NULL) {
    return;
  }
  deltree(tree->left);
  deltree(tree->right);
  xfree(tree);
}

void add_to_tree(struct toptree *tree, char *nick, int aantal) {

  if (aantal > tree->aantal) {
    /* links toevoegen */
    if (tree->left == NULL) {
      createtree(&(tree->left), nick, aantal);
      return;
    }
    else {
      add_to_tree(tree->left, nick, aantal);
    }
  }
  else {
    /* rechts toevoegen */
    if (tree->right == NULL) {
      createtree(&(tree->right), nick, aantal);
      return;
    }
    else {
      add_to_tree(tree->right, nick, aantal);
    }
  }
}


void printtoptree(struct toptree *tree, int *positie, int min, int max, void output(char *, ...)) {

  if (tree->left != NULL) {
    printtoptree(tree->left, positie, min, max, output);
  }
  if ((*positie >= min) && (*positie <= max)) {
    output("%-8d %-9s %d", *positie, tree->nick, tree->aantal);
  }
  *positie = *positie + 1;
  if (tree->right != NULL) {
    printtoptree(tree->right, positie, min, max, output);
  }
}

void printtimetoptree(struct toptree *tree, int *positie, int min, int max, void output(char *, ...)) {

  if (tree->left != NULL) {
    printtimetoptree(tree->left, positie, min, max, output);
  }
  if ((*positie >= min) && (*positie <= max)) {
    output("%-8d %-9s %d:%02d", *positie, tree->nick, tree->aantal / 60, tree->aantal % 60);
  }
  *positie = *positie + 1;
  if (tree->right != NULL) {
    printtimetoptree(tree->right, positie, min, max, output);
  }
}

void printtoptree_html(struct toptree *tree, int *positie, int min, int max, void output(int, char *, ...), int s) {

  if (tree->left != NULL) {
    printtoptree_html(tree->left, positie, min, max, output, s);
  }
  if ((*positie >= min) && (*positie <= max)) {
    output(s, "<TR><TD>%d</TD><TH><A HREF=\"~%s\">%s</A><TD>%d</TD></TR>", *positie, tree->nick, tree->nick, 
	   tree->aantal);
  }
  *positie = *positie + 1;
  if (tree->right != NULL) {
    printtoptree_html(tree->right, positie, min, max, output, s);
  }
}

void printtimetoptree_html(struct toptree *tree, int *positie, int min, int max, void output(int, char *, ...), int s) {

  if (tree->left != NULL) {
    printtimetoptree_html(tree->left, positie, min, max, output, s);
  }
  if ((*positie >= min) && (*positie <= max)) {
    output(s, "<TR><TD>%d</TD><TH><A HREF=\"~%s\">%s</A><TD>%d:%02d</TD></TR>", *positie, tree->nick, tree->nick, 
	   tree->aantal / 60, tree->aantal % 60); 
  }
  *positie = *positie + 1;
  if (tree->right != NULL) {
    printtimetoptree_html(tree->right, positie, min, max, output, s);
  }
}

