
/*********************************************************/
/**                                                     **/
/**      userlist.c (c) 1996, 97 Emiel Bruijntjes       **/
/**                                                     **/
/**          List of nicks I have seen on IRC           **/
/**                                                     **/
/*********************************************************/


#ifdef SUNOS
#include "sunos.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "userlist.h"
#include "misc.h"
#include "config.h"

struct userlist {
  int length;
  int keeping;
  struct unode *current;
  struct unode *head;
  struct unode *tail;
} ulist = { 0 , 0 , NULL , NULL , NULL };

struct unode {
  unsigned int ontime;
  unsigned int kicks;
  unsigned int kickeds;
  unsigned int joins;
  unsigned int modes;
  unsigned int leaves;
  unsigned int nicks;
  unsigned int quits;
  unsigned int publics;
  int protmanip;
  char keep_entry;
  char *lasthost;
  time_t lastseen;

  char nick[10];
  char passwd[10];

  struct unode *prev;
  struct unode *next;
};

static int   maxlength=2000;

static int publicfactor =   1;
static int kickfactor   =  -5;
static int kickedfactor =  -1;
static int joinfactor   =   5;
static int leavefactor  = -10;
static int nickfactor   =  -5;
static int quitfactor   =  -4;
static int ontimefactor =   1;
static int modefactor   = -10;
static int manipfactor  =   1;


extern time_t time_null;

void save_userlist(FILE *fp) {

  int i=0;

  fwriteint(fp, maxlength);
  fwriteint(fp, publicfactor);
  fwriteint(fp, kickfactor);
  fwriteint(fp, kickedfactor);
  fwriteint(fp, joinfactor);
  fwriteint(fp, leavefactor);
  fwriteint(fp, nickfactor);
  fwriteint(fp, quitfactor);
  fwriteint(fp, ontimefactor);
  fwriteint(fp, modefactor);
  fwriteint(fp, manipfactor);
  
  fwriteint(fp, userlist_length());
  fwriteint(fp, userlist_kept());

  if (userlist_findfirst()) {
    do
      {
	fwriteint(fp, userlist_publics());
	fwriteint(fp, userlist_kicks());
	fwriteint(fp, userlist_kickeds());
	fwriteint(fp, userlist_joins());
	fwriteint(fp, userlist_leaves());
	fwriteint(fp, userlist_quits());
	fwriteint(fp, userlist_nicks());
	fwriteint(fp, userlist_modes());
	fwriteint(fp, userlist_ontime());
	fwriteint(fp, userlist_manipulation());
	fwritestr2bin(userlist_nick(), fp);
	fwritestr2bin(userlist_lasthost(), fp);
	fwritetime_t(fp, userlist_lastseen());
	fwritestr2bin(userlist_password(), fp);
	fwritechar(fp, userlist_keepentry());
	i++;
      }
    while (userlist_findnext());
  }
  fwriteint(fp, 38457);
}

int  load_userlist(FILE *fp, int version) {

  int i;
  char *nick=NULL, *passwd=NULL;

  freadint(fp, &maxlength);
  freadint(fp, &publicfactor);
  freadint(fp, &kickfactor);
  freadint(fp, &kickedfactor);
  freadint(fp, &joinfactor);
  freadint(fp, &leavefactor);
  freadint(fp, &nickfactor);
  freadint(fp, &quitfactor);
  freadint(fp, &ontimefactor);
  freadint(fp, &modefactor);
  freadint(fp, &manipfactor);

  deleteuserlist();
  freadint(fp, &ulist.length);
  freadint(fp, &ulist.keeping);

  for (i=0; i<ulist.length; i++) {
    ulist.current = (struct unode *) xmalloc (sizeof(struct unode));
    if (ulist.head == NULL) {
      ulist.head = ulist.current;
      ulist.current->prev = NULL;
    }
    else {
      ulist.tail->next = ulist.current;
      ulist.current->prev = ulist.tail;
    }
    ulist.current->next = NULL;
    ulist.tail = ulist.current;
    
    freadint(fp, &ulist.current->publics);
    freadint(fp, &ulist.current->kicks);
    freadint(fp, &ulist.current->kickeds);
    freadint(fp, &ulist.current->joins);
    freadint(fp, &ulist.current->leaves);
    freadint(fp, &ulist.current->quits); 
    freadint(fp, &ulist.current->nicks);
    freadint(fp, &ulist.current->modes);
    freadint(fp, &ulist.current->ontime);
    freadint(fp, &ulist.current->protmanip);
    
    fgetstrfrombin(&nick, fp);
    strcpy(ulist.current->nick, nick);
    xfree(nick); nick=NULL;
    ulist.current->lasthost = NULL;
    fgetstrfrombin(&ulist.current->lasthost, fp);
    if (version >= 36) {
      freadtime_t(fp, &ulist.current->lastseen);
    }
    fgetstrfrombin(&passwd, fp);
    strcpy(ulist.current->passwd, passwd);
    xfree(passwd); passwd=NULL;
    freadchar(fp, &ulist.current->keep_entry);
  }
  freadint(fp, &i);
  return (i == 38457);
}

int  userlist_maxlength() { return maxlength; }

void userlist_setmaxlength(int newlength) {

  if (newlength < 1) newlength = 1;
  maxlength = newlength;
  while (ulist.length > maxlength) {
    ulist.current = ulist.tail;
    deleteunode();
  }
}

void deleteuserlist() { while (userlist_findfirst()) deleteunode(); }

void createunode(char *nick) {

  struct unode *tmp;

  ulist.current = (struct unode *)xmalloc(sizeof(struct unode));
  ulist.current->prev = NULL;
  ulist.current->next = ulist.head;
  ulist.head = ulist.current;
  if (ulist.tail == NULL) {
    ulist.tail = ulist.current;
  }
  else {
    ulist.current->next->prev = ulist.current;
  }
  
  (ulist.length)++;
  
  ulist.current->ontime     = 0;
  ulist.current->kicks      = 0;
  ulist.current->kickeds    = 0;
  ulist.current->joins      = 0;
  ulist.current->leaves     = 0;
  ulist.current->nicks      = 0;
  ulist.current->publics    = 0;
  ulist.current->lasthost   = (char *) xmalloc (1);
  ulist.current->lasthost[0]= '\0';
  ulist.current->quits      = 0;
  ulist.current->modes      = 0;
  ulist.current->lastseen   = time_null;
  ulist.current->protmanip  = 0;
  ulist.current->keep_entry = 0;
  
  strcpy(ulist.current->nick, nick);
  strcpy(ulist.current->passwd, nick);

  if (ulist.length > maxlength) {
    tmp = ulist.current;
    ulist.current = ulist.tail;
    while (ulist.current->keep_entry) ulist.current = ulist.current->prev;
    deleteunode();
    ulist.current = tmp;
  }
}

void deleteunode() { 

  if (ulist.current == NULL) return;

  xfree(ulist.current->lasthost);
  if (ulist.current->keep_entry) (ulist.keeping)--;

  if (ulist.length == 1) {
    xfree(ulist.current);
    ulist.current = NULL;
    ulist.tail = NULL;
    ulist.head = NULL;
    ulist.length = 0;
    return;
  }

  if (ulist.current->next == NULL) {
    ulist.tail = ulist.current->prev;
    ulist.tail->next = NULL;
    xfree(ulist.current);
    ulist.current = ulist.head;
    ulist.length = ulist.length - 1;
    return;
  }

  if (ulist.current->prev == NULL) {
    ulist.head = ulist.current->next;
    ulist.head->prev = NULL;
    xfree(ulist.current);
    ulist.current = ulist.head;
    ulist.length = ulist.length - 1;
    return;
  }

  ulist.current->prev->next = ulist.current->next;
  ulist.current->next->prev = ulist.current->prev;
  xfree(ulist.current);
  ulist.length = ulist.length - 1;
  ulist.current = ulist.head;
}

    
int userlist_findfirst() {

  return (int)(ulist.current = ulist.head);
}

int userlist_findnext() {

  return (int)(ulist.current = ulist.current->next);
}

int userlist_findnick(char *nick) {
  
  if (userlist_findfirst()==0) {
    return 0;
  }
  while (ulist.current != NULL) {
    if (strcasecmp(nick, ulist.current->nick)==0) {
      if (ulist.current->prev == NULL) {
	return 1;
      }
      if (ulist.current->next == NULL) {
	ulist.tail = ulist.current->prev;
	ulist.tail->next = NULL;
	ulist.current->prev = NULL;
	ulist.current->next = ulist.head;
	ulist.head->prev = ulist.current;
	ulist.head = ulist.current;
	return 1;
      }
      ulist.current->next->prev = ulist.current->prev;
      ulist.current->prev->next = ulist.current->next;
      ulist.current->prev = NULL;
      ulist.current->next = ulist.head;
      ulist.head->prev = ulist.current;
      ulist.head = ulist.current;
      return 1;
    }
    ulist.current = ulist.current->next;
  }
  return 0;
}

void userlist_incjoins() { ulist.current->joins = ulist.current->joins + 1; }
void userlist_incleaves() { ulist.current->leaves = ulist.current->leaves + 1; }
void userlist_incnicks() { ulist.current->nicks = ulist.current->nicks + 1; }
void userlist_inckicks() { ulist.current->kicks = ulist.current->kicks + 1; }
void userlist_inckickeds() { ulist.current->kickeds = ulist.current->kickeds + 1; }
void userlist_incquits() { ulist.current->quits = ulist.current->quits + 1; }
void userlist_incpublics() { ulist.current->publics = ulist.current->publics + 1; }
void userlist_incmodes() { ulist.current->modes = ulist.current->modes + 1; }
void userlist_incontime() { ulist.current->ontime = ulist.current->ontime + 1; }
void userlist_setpassword(char *pw) { strcpy(ulist.current->passwd, pw); }
void userlist_setmanipulation(int m) { ulist.current->protmanip += m; }

int userlist_publics() { return ulist.current->publics; }
int userlist_joins() { return ulist.current->joins; }
int userlist_leaves() { return ulist.current->leaves; }
int userlist_nicks() { return ulist.current->nicks; }
int userlist_kicks() { return ulist.current->kicks; }
int userlist_kickeds() { return ulist.current->kickeds; }
int userlist_quits() { return ulist.current->quits; }
int userlist_modes() { return ulist.current->modes; }
int userlist_length() { return ulist.length; }
int userlist_kept() { return ulist.keeping; }
int userlist_ontime() { return ulist.current->ontime; }
int userlist_manipulation() { return ulist.current->protmanip; }
char * userlist_lasthost() { return ulist.current->lasthost; }
char * userlist_nick() { return ulist.current->nick; }
char * userlist_password() { return ulist.current->passwd; }
char userlist_keepentry() { return ulist.current->keep_entry; }
time_t userlist_lastseen() { return ulist.current->lastseen; }
int userlist_msgpminute() { return userlist_ontime() == 0 ? userlist_publics() : userlist_publics() / userlist_ontime(); }
int userlist_msgphour() {
  return userlist_ontime() < 60 ? userlist_publics() : userlist_publics() / (userlist_ontime() / 60);
}


void userlist_setlasthost(char *nickuserhost) {

  char userhost[MAXNICKUSERHOSTLEN];
  int i=0, j=0;

  if (ulist.current == NULL) return;
  while (nickuserhost[i] != '!') i++;
  i++;
  while (nickuserhost[i] != '\0') {
    userhost[j] = nickuserhost[i];
    if (userhost[j]==':')
      userhost[j]='=';
    i++; j++;
  }
  userhost[j] = '\0';

  xfree(ulist.current->lasthost);
  ulist.current->lasthost = (char *) xmalloc ((strlen(userhost) + 1));
  strcpy(ulist.current->lasthost, userhost);
  ulist.current->lastseen = time_null;
}

int protection_publicfactor() { return publicfactor; }
int protection_kickfactor()   { return kickfactor;   }
int protection_kickedfactor() { return kickedfactor; }
int protection_joinfactor()   { return joinfactor;   }
int protection_leavefactor()  { return leavefactor;  }
int protection_nickfactor()   { return nickfactor;   }
int protection_quitfactor()   { return quitfactor;   }
int protection_ontimefactor() { return ontimefactor; }
int protection_modefactor()   { return modefactor;   }
int protection_manipfactor()  { return manipfactor;  }

void protection_set_publicfactor(int value) { publicfactor = value; }
void protection_set_kickfactor(int value)   { kickfactor   = value; }
void protection_set_kickedfactor(int value) { kickedfactor = value; }
void protection_set_joinfactor(int value)   { joinfactor   = value; }
void protection_set_leavefactor(int value)  { leavefactor  = value; }
void protection_set_nickfactor(int value)   { nickfactor   = value; }
void protection_set_quitfactor(int value)   { quitfactor   = value; }
void protection_set_ontimefactor(int value) { ontimefactor = value; }
void protection_set_modefactor(int value)   { modefactor   = value; }
void protection_set_manipfactor(int value)  { manipfactor  = value; }

int userlist_protection() {

  int prot=0;
	
  prot = prot + (publicfactor * ulist.current->publics);
  prot = prot + (kickfactor   * ulist.current->kicks);
  prot = prot + (kickedfactor * ulist.current->kickeds);
  prot = prot + (joinfactor   * ulist.current->joins);
  prot = prot + (leavefactor  * ulist.current->leaves);
  prot = prot + (nickfactor   * ulist.current->nicks);
  prot = prot + (quitfactor   * ulist.current->quits);
  prot = prot + (ontimefactor * ulist.current->ontime);
  prot = prot + (modefactor   * ulist.current->modes);
  prot = prot + (manipfactor  * ulist.current->protmanip);

  return prot;
}


int  userlist_setkeepentry() {

  if (ulist.current->keep_entry) return 1;
  if (ulist.keeping > maxlength / 2) return 0;
  ulist.current->keep_entry = 1;
  (ulist.keeping)++;
  return 1;
}


void userlist_resetkeepentry() {

  if (!(ulist.current->keep_entry)) return;
  (ulist.keeping)--;
  ulist.current->keep_entry = 0;
}
