
/*********************************************************/
/**                                                     **/
/**       X_session.c (c) 1997, Emiel Bruijntjes        **/
/**                                                     **/
/**       Database of connections with X-servers        **/
/**                                                     **/
/*********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif

#ifdef NO_X

int create_X_session(char *display) { return 0; }
int X_session_findsockptr(int ptr)  { return 1; }
void read_X_events() { return; }
void init_X() { return; }
void X_send2window(char window, char *text) { return; }
void X_print_window_title() { return; }

int nr_X_consoles = 0;

#else /* not NO_X */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>


#include "xwin.h"
#include "misc.h"
#include "net.h"
#include "config.h"
#include "history.h"
#include "console.h"
#include "irc_console.h"
#include "cmd_console.h"
#include "party_console.h"
#include "channel.h"
#include "on.h"

struct X_session {
  struct X_session * prev;
  struct X_session * next;

  Display     *display;
  Window       root_window;
  Window       main_window;
  Window       snoop_window;
  Window       irc_window;
  Window       input_window;
  Window       stop_window;
  char         stop_window_pressed;
  GC           gc;
  XFontStruct *font_info;
  Colormap     colormap;
  int          screen_num;
  int          sockptr;
  char         input_buffer[WAYTOBIG];
  int          buffer_length;
  int          input_cursor;
  unsigned long colors[NR_OF_COLORS];
};

static struct {
  struct X_session * first;
  struct X_session * last;
  struct X_session * current;
  int length;
} X_seslist = { NULL , NULL , NULL , 0 };

static char * color_names[NR_OF_COLORS] = { "black", "white", "gray70", "gray50", "gray80", "red", "blue", "yellow",
					    "green", "brown", "pink", "purple", "orange", "cyan", "khaki" };

extern char  botnick[10];
extern char  current_nick[10];
extern char *xfont;

       int   nr_X_consoles;

static char *snoop_window_text[X_NR_LINES];
static char *irc_window_text[X_NR_LINES];

static void X_seslist_createnode() {

  X_seslist.current = (struct X_session *)xmalloc(sizeof(struct X_session));

  if (X_seslist.length == 0) {
    X_seslist.first = X_seslist.current;
    X_seslist.current->prev = NULL;
  } 
  else {
    X_seslist.last->next = X_seslist.current;
    X_seslist.current->prev = X_seslist.last;
  }

  X_seslist.current->next = NULL;
  X_seslist.last = X_seslist.current;
  X_seslist.length = X_seslist.length + 1;

  nr_X_consoles++;
}


static void X_seslist_deletenode() {

  int i;

  nr_X_consoles--;
  if (nr_X_consoles == 0) {
    for (i=0; i < X_NR_LINES; i++) {
      snoop_window_text[i] = xrealloc(snoop_window_text[i], 1);
      snoop_window_text[i][0] = '\0';
      irc_window_text[i]   = xrealloc(irc_window_text[i], 1);
      irc_window_text[i][0] = '\0';
    }
  }

  if (X_seslist.length == 1) {
    xfree(X_seslist.current);
    X_seslist.current = NULL;
    X_seslist.last    = NULL;
    X_seslist.first    = NULL;
    X_seslist.length  = 0;
    return;
  }

  if (X_seslist.current->next == NULL) {
    X_seslist.last = X_seslist.current->prev;
    X_seslist.last->next = NULL;
    xfree(X_seslist.current);
    X_seslist.current = X_seslist.last;
    X_seslist.length = X_seslist.length - 1;
    return;
  }

  if (X_seslist.current->prev == NULL) {
    X_seslist.first = X_seslist.current->next;
    X_seslist.first->prev = NULL;
    xfree(X_seslist.current);
    X_seslist.current = X_seslist.first;
    X_seslist.length = X_seslist.length - 1;
    return;
  }

  X_seslist.current->prev->next = X_seslist.current->next;
  X_seslist.current->next->prev = X_seslist.current->prev;
  xfree(X_seslist.current);
  X_seslist.length = X_seslist.length - 1;
  X_seslist.current = X_seslist.first;
  return;
}

int  X_session_findsockptr(int ptr) {

  for (X_seslist.current = X_seslist.first; X_seslist.current; X_seslist.current = X_seslist.current->next)
    if (X_seslist.current->sockptr == ptr) return 1;
  return 0;
}

void init_X() {

  int i;

  for (i=0; i<X_NR_LINES; i++) {
    irc_window_text[i]   = (char *) xmalloc (1);
    snoop_window_text[i] = (char *) xmalloc (1);
    strcpy(irc_window_text[i], "");
    strcpy(snoop_window_text[i], "");
  }
}

static void scrollwindow(char window) {

  int i;

  switch (window) {
  case X_SNOOP_WINDOW:
    xfree (snoop_window_text[0]);
    for (i=0; i < X_NR_LINES - 1; i++) {
      snoop_window_text[i] = snoop_window_text[i+1];
    }
    snoop_window_text[X_NR_LINES - 1] = NULL;
    break;
  case X_IRC_WINDOW:
    xfree (irc_window_text[0]);
    for (i=0; i < X_NR_LINES - 1; i++) {
      irc_window_text[i] = irc_window_text[i+1];
    }
    irc_window_text[X_NR_LINES - 1] = NULL;
    break;
  }
}

/*
#define CHAR_WIDTH     (X_seslist.current->font_info->max_bounds.rbearing - \
			X_seslist.current->font_info->min_bounds.lbearing)
*/
#define CHAR_WIDTH     (X_seslist.current->font_info->max_bounds.width)

#define CHAR_HEIGHT    (X_seslist.current->font_info->max_bounds.ascent + \
			X_seslist.current->font_info->max_bounds.descent)
#define WINDOW_WIDTH   ((CHAR_WIDTH) * X_NR_COLS)
#define WINDOW_HEIGHT  ((CHAR_HEIGHT) * X_NR_LINES)


/* Returns the length of the next word. Control characters are not counted */
static int X_strlen_nextword(char *string) {

  int i=0, j=0;

  while ((!isspace(string[i])) && (string[i] != '\0')) {
    if (isprint(string[i])) j++;
    i++;
  }
  return j;
}


#define A_BOLD      1
#define A_UNDERLINE 2
#define A_REVERSE   4


/* Local procedure to print something on one of the two windows. Use X_send2window to print something */
static void send2Xwindow_(char window, char *string, char attr, int startpos) {

  char *lastchar, *output;
  int   cursor, wordlength;

  scrollwindow(window);

  output = (char *) xmalloc (WAYTOBIG);
  strcpy(output, spaces(startpos));
  cursor = startpos;
  lastchar = index(output, '\0');
  if (attr & A_BOLD)      { strcpy(lastchar, ""); lastchar = index(lastchar, '\0'); }
  if (attr & A_UNDERLINE) { strcpy(lastchar, ""); lastchar = index(lastchar, '\0'); }
  if (attr & A_REVERSE)   { strcpy(lastchar, ""); lastchar = index(lastchar, '\0'); }

  while (1) {
    wordlength = X_strlen_nextword(string);

    if (wordlength < X_NR_COLS - cursor) {
      while (isgraph(*string)) {
	*lastchar = *string;
	lastchar++; string++; cursor++;
      }
      *lastchar = '\0';
    }
    else if (wordlength > X_NR_COLS - startpos) {
      while (isgraph(*string) && cursor < X_NR_COLS) {
	*lastchar = *string;
	lastchar++; string++; cursor++;
      }
      *lastchar = '\0';
      if (cursor >= X_NR_COLS) {
	output = (char *) xrealloc (output, strlen(output) + 1);
	if (window == X_SNOOP_WINDOW)
	  snoop_window_text[X_NR_LINES - 1] = output;
	else
	  irc_window_text[X_NR_LINES - 1] = output;
	send2Xwindow_(window, string, attr, startpos);
	return;
      }
    }
    else {
      *lastchar = '\0';
      output = (char *) xrealloc (output, strlen(output) + 1);
      if (window == X_SNOOP_WINDOW)
	snoop_window_text[X_NR_LINES - 1] = output;
      else
	irc_window_text[X_NR_LINES - 1] = output;
      send2Xwindow_(window, string, attr, startpos);
      return;
    }
    while (!isgraph(*string)) {
      switch (*string) {
      case ' ':
	if (startpos == 0) startpos = cursor + 1;
	*lastchar = *string;
	lastchar++; string++; cursor++;
	*lastchar = '\0';
	if (cursor >= X_NR_COLS) {
	  while (isspace(*string)) string++;
	  output = (char *) xrealloc (output, strlen(output) + 1);
	  if (window == X_SNOOP_WINDOW)
	    snoop_window_text[X_NR_LINES - 1] = output;
	  else
	    irc_window_text[X_NR_LINES - 1] = output;
	  send2Xwindow_(window, string, attr, startpos);
	  return;
	}
	break;
      case '':
	*lastchar = *string;
	lastchar++; string++;
	*lastchar = '\0';
	if (attr & A_BOLD)
	  attr &= ~A_BOLD;
	else
	  attr |= A_BOLD;
	break;
      case '':
	*lastchar = *string;
	lastchar++; string++;
	*lastchar = '\0';
	if (attr & A_REVERSE)
	  attr &= ~A_REVERSE;
	else
	  attr |= A_REVERSE;
	break;
      case '':
	*lastchar = *string;
	lastchar++; string++;
	*lastchar = '\0';
	if (attr & A_UNDERLINE)
	  attr &= ~A_UNDERLINE;
	else
	  attr |= A_UNDERLINE;
	break;
      case '\0':
	*lastchar = *string;
	output = (char *) xrealloc (output, strlen(output) + 1);
	if (window == X_SNOOP_WINDOW)
	  snoop_window_text[X_NR_LINES - 1] = output;
	else
	  irc_window_text[X_NR_LINES - 1] = output;
	return;
      default:
	string++;
      }
    }
  }
}


void X_writestringnumber2current(char window, int line_nr) {

  char     *string, *strptr;
  int       i=0, attr=0, cursor=0;
  Drawable  drawable;

  if (window == X_SNOOP_WINDOW) {
    string = snoop_window_text[line_nr];
    drawable = X_seslist.current->snoop_window;
  }
  else {
    string = irc_window_text[line_nr];
    drawable = X_seslist.current->irc_window;
  }

  strptr = string;

  XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_YELLOW]);
  while (1) {
    for (i=0; isprint(strptr[i]); i++) { /* do nothing */ }
    XDrawImageString(X_seslist.current->display, drawable, X_seslist.current->gc, cursor * CHAR_WIDTH,
		     X_seslist.current->font_info->ascent + (CHAR_HEIGHT * line_nr), strptr, i);
    cursor += i;
    for (; !isprint(strptr[i]); i++) {
      switch (strptr[i]) {
      case '':
	if (attr & A_BOLD) {
	  attr &= ~A_BOLD;
	} else {
	  attr |= A_BOLD;
	}
	break;
      case '':
	if (attr & A_REVERSE) {
	  attr &= ~A_REVERSE;
	} else {
	  attr |= A_REVERSE;
	}
	break;
      case '':
	if (attr & A_UNDERLINE) {
	  attr &= ~A_UNDERLINE;
	} else {
	  attr |= A_UNDERLINE;
	}
	break;
      case '\0':
	XDrawImageString(X_seslist.current->display, drawable, X_seslist.current->gc, cursor * CHAR_WIDTH,
			 X_seslist.current->font_info->ascent + (CHAR_HEIGHT * line_nr), spaces(X_NR_COLS - cursor), 
			 X_NR_COLS - cursor);
	return;
      }
    }
    switch (attr) {
    case 0:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_YELLOW]);
      break;
    case A_BOLD:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_RED]);
      break;
    case A_REVERSE:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_GREEN]);
      break;
    case A_UNDERLINE:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLUE]);
      break;
    case A_BOLD | A_REVERSE:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_PURPLE]);
      break;
    case A_BOLD | A_UNDERLINE:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_PINK]);
      break;
    case A_BOLD | A_UNDERLINE | A_REVERSE:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_ORANGE]);
      break;
    case A_REVERSE | A_UNDERLINE:
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_CYAN]);
      break;
    }
    strptr = &strptr[i];
  }
}


void X_send2window(char window, char *text) {

  int i;
  struct X_session *tmp;

  send2Xwindow_(window, text, 0, 0);

  tmp = X_seslist.current;
  for (X_seslist.current = X_seslist.first; X_seslist.current; X_seslist.current = X_seslist.current->next) {
    for (i=0; i<X_NR_LINES; i++)
      X_writestringnumber2current(window, i);
    XFlush(X_seslist.current->display);
  }
  X_seslist.current = tmp;
}


/**********************************************************************************/
/**                                                                              **/
/**                              Handling of input                               **/
/**                                                                              **/
/**********************************************************************************/


#define CURSOR_HOME   -1
#define CLEAR_TO_EOL  -2
#define CURSOR_MOTION -3

void X_putc(int ch, int curs_pos) {

  static int cursor=0;
  char outp;

  switch (ch) {
  case CLEAR_TO_EOL:
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc, 
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, spaces(X_NR_COLS - cursor), 
		     X_NR_COLS - cursor);
    cursor = X_NR_COLS;
    return;
  case CURSOR_HOME:   
    cursor = 0;
    return;
  case CURSOR_MOTION:
    outp = 
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLACK]);
    XSetBackground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    if (curs_pos) {
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
      XSetBackground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLACK]);
    }

  case '':
    outp = 'B';
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_RED]);
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc, 
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
    cursor++;
    return;
  case '':
    outp = 'V';
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_RED]);
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
    cursor++;
    return;
  case '':
    outp = '_';
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_RED]);
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
    cursor++;
    return;
  case '':
    outp = 'A';
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_RED]);
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
    cursor++;
    return;
  default:
    outp = ch;
    if (curs_pos) {
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLACK]);
      XSetBackground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
    }
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		     cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    if (curs_pos) {
      XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
      XSetBackground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLACK]);
    }
    cursor++;
    return;
  }
}


void X_restore_cursor() {
  
  char outp;

  outp = X_seslist.current->input_buffer[X_seslist.current->input_cursor] ?
    X_seslist.current->input_buffer[X_seslist.current->input_cursor] : ' ';

  XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLACK]);
  XSetBackground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);

  if (X_seslist.current->buffer_length > X_NR_COLS - 10) {
    if (X_seslist.current->input_cursor < X_NR_COLS - 10) {
      XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		       X_seslist.current->input_cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    }
    else {
      XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		       (X_NR_COLS - 10) * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
    }
  }
  else {
    XDrawImageString(X_seslist.current->display, X_seslist.current->input_window, X_seslist.current->gc,
		     X_seslist.current->input_cursor * CHAR_WIDTH, X_seslist.current->font_info->ascent, &outp, 1);
  }
  XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);
  XSetBackground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_BLACK]);
}


void X_print_input_buffer() {

  int i;

  XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_WHITE]);

  if (X_seslist.current->buffer_length > X_NR_COLS - 10) {
    X_putc(CURSOR_HOME, 0);
    if (X_seslist.current->input_cursor < X_NR_COLS - 10) {
      for (i=0; i<X_NR_COLS && i<X_seslist.current->buffer_length; i++) {
	X_putc(X_seslist.current->input_buffer[i], i == X_seslist.current->input_cursor);
      }
      if (X_seslist.current->buffer_length < X_NR_COLS) {
	X_putc(CLEAR_TO_EOL, 0);
      }
      X_restore_cursor();
    }
    else {
      for (i = X_seslist.current->input_cursor - (X_NR_COLS - 10); 
	   i - (X_seslist.current->input_cursor - (X_NR_COLS - 10)) < X_NR_COLS && i < X_seslist.current->buffer_length; 
	   i++) {
	X_putc(X_seslist.current->input_buffer[i], i == X_seslist.current->input_cursor);
      }
      if (X_seslist.current->buffer_length - X_seslist.current->input_cursor < 10) {
	X_putc(CLEAR_TO_EOL, 0);
      }
      X_restore_cursor();
    }
  }
  else {
    X_putc(CURSOR_HOME, 0);
    for (i=0; i<X_seslist.current->buffer_length; i++) {
      X_putc(X_seslist.current->input_buffer[i], i == X_seslist.current->input_cursor);
    }
    X_putc(CLEAR_TO_EOL, 0);
    X_restore_cursor();
  }
}


void X_input_addchar(char ch) {

  struct X_session *saved_X_session;

  if (isprint(ch) || ch == '' || ch == '' || ch == '' || ch == '') {
    if (X_seslist.current->buffer_length > 1000) return;
    if (X_seslist.current->input_cursor == X_seslist.current->buffer_length) {
      X_seslist.current->input_buffer[X_seslist.current->input_cursor] = ch;
      X_seslist.current->input_buffer[X_seslist.current->input_cursor + 1] = '\0';
      X_seslist.current->input_cursor++;
      X_seslist.current->buffer_length++;
      X_print_input_buffer();
    }
    else {
      memmove(&X_seslist.current->input_buffer[X_seslist.current->input_cursor + 1], 
	      &X_seslist.current->input_buffer[X_seslist.current->input_cursor], 
	      X_seslist.current->buffer_length - X_seslist.current->input_cursor + 1);
      X_seslist.current->input_buffer[X_seslist.current->input_cursor] = ch;
      X_seslist.current->input_cursor++;
      X_seslist.current->buffer_length++;
      X_print_input_buffer();
    }
  }
  else {
    switch (ch) {
    case '\r': case '\n':
      add_history(X_seslist.current->input_buffer);
      saved_X_session = X_seslist.current;
      if (!on_console(X_seslist.current->input_buffer)) {
	switch (console_mode()) {
	case CONSOLE_IRC_MODE:
	  parse_irc_console(X_seslist.current->input_buffer);
	  break;
	case CONSOLE_PARTY_MODE:
	  parse_party_console(X_seslist.current->input_buffer);
	  break;
	case CONSOLE_COMMAND_MODE:
	  parse_cmd_console(X_seslist.current->input_buffer);
	  break;
	}
      }
      X_seslist.current = saved_X_session;
      strcpy(X_seslist.current->input_buffer, "");
      X_seslist.current->buffer_length = 0;
      X_seslist.current->input_cursor = 0;
      X_print_input_buffer();
      break;
    case '\t':
      if (console_mode() == CONSOLE_IRC_MODE) {
	add_history(X_seslist.current->input_buffer);
	sprintf(X_seslist.current->input_buffer, "/msg %s ", msg_history_prev());
	X_seslist.current->buffer_length = X_seslist.current->input_cursor = strlen(X_seslist.current->input_buffer);
	X_print_input_buffer();
      }
      break;
    case '\b':
      if (X_seslist.current->input_cursor > 0) {
	memmove(&X_seslist.current->input_buffer[X_seslist.current->input_cursor - 1], 
		&X_seslist.current->input_buffer[X_seslist.current->input_cursor], 
		X_seslist.current->buffer_length - X_seslist.current->input_cursor + 1);
	X_seslist.current->input_cursor--;
	X_seslist.current->buffer_length--;
	X_print_input_buffer();
      }
      break;
    case '':
      switch (console_mode()) {
      case CONSOLE_IRC_MODE:
	console_setmode(CONSOLE_COMMAND_MODE);
	break;
      case CONSOLE_COMMAND_MODE:
	console_setmode(CONSOLE_PARTY_MODE);
	break;
      case CONSOLE_PARTY_MODE:
	console_setmode(CONSOLE_IRC_MODE);
	break;
      }
      break;
    }
  }
}


void X_print_window_title() {

  char             *buffer, *channel;
  XTextProperty     text_prop;
  struct X_session *tmp;

  buffer = (char *) xmalloc (250);
  switch (console_mode()) {
  case CONSOLE_IRC_MODE:
    if (channellist_findfirst())
      channel = channellist_channelname();
    else
      channel = "";
    sprintf(buffer, "%s: IRC-mode   (channel: %c%s)", botnick, 
	    ischanneloperator(channel, current_nick) ? '@' : ' ', channel);
    break;
  case CONSOLE_PARTY_MODE:
    sprintf(buffer, "%s: Party-mode   (channel: %s)", botnick, current_party_channel());
    break;
  case CONSOLE_COMMAND_MODE:
    sprintf(buffer, "%s: Command-mode", botnick);
    break;
  }

  if (XStringListToTextProperty(&buffer, 1, &text_prop) == 0) {
    fprintf(stderr, "Error in XStringListToTextProperty()\n");
    exit(1);
  }

  tmp = X_seslist.current;
  for (X_seslist.current = X_seslist.first; X_seslist.current; X_seslist.current = X_seslist.current->next) {
    XSetWMName(X_seslist.current->display, X_seslist.current->main_window, &text_prop);
    XFlush(X_seslist.current->display);
  }
  X_seslist.current = tmp;
  xfree(buffer);
}


int create_X_session(char *display_name) {

  XColor                xcolor;
  int                   i;
  XSetWindowAttributes  win_attr;
  XGCValues             GCValues;
  XSizeHints            sizehints;

  X_seslist_createnode();     /* Create the node in the list */

  strcpy(X_seslist.current->input_buffer, "");
  X_seslist.current->buffer_length = X_seslist.current->input_cursor = 0;

  if ((X_seslist.current->display = XOpenDisplay(display_name)) == NULL) {

    /* Attempt to make connection to X-server failed,
     * Remove the current and return 0 */

    X_seslist_deletenode();
    return 0;
  }

  if ((X_seslist.current->sockptr = add_X_socket(XConnectionNumber(X_seslist.current->display))) == -1) {

    /* All connections are in use, no more sockets could be added */

    XCloseDisplay(X_seslist.current->display);
    X_seslist_deletenode();
    return 1;
  }

  /* Connection to X-server established, we can now download the data-
   * structures we might need in future */

  X_seslist.current->screen_num  = DefaultScreen(X_seslist.current->display);      /* Store the screen_num */
  X_seslist.current->root_window = RootWindow(X_seslist.current->display, X_seslist.current->screen_num);   
  /* Store the root-window */

  /* set colors I want to use */
  
  X_seslist.current->colormap = DefaultColormap(X_seslist.current->display, X_seslist.current->screen_num);

  for (i=0; i<NR_OF_COLORS; i++) {
    if (XParseColor(X_seslist.current->display, X_seslist.current->colormap, color_names[i], &xcolor) == 0)
      X_seslist.current->colors[i] = XWhitePixel(X_seslist.current->display, X_seslist.current->screen_num);
    else {
      if (XAllocColor(X_seslist.current->display, X_seslist.current->colormap, &xcolor) == 0)
	X_seslist.current->colors[i] = XWhitePixel(X_seslist.current->display, X_seslist.current->screen_num);
      else
	X_seslist.current->colors[i] = xcolor.pixel;
    }
  }

  /* Colors are set, we can now create the window(s) */

  /* I do not know the size of a character. The window size depends on the character-size, so we must select the
  ** font now
  */
  
  if ((X_seslist.current->font_info = XLoadQueryFont(X_seslist.current->display, xfont)) == NULL) {
    XFreeColormap(X_seslist.current->display, X_seslist.current->colormap);
    XCloseDisplay(X_seslist.current->display);
    del_X_sockptr(X_seslist.current->sockptr);
    X_seslist_deletenode();
    return 2;
  }

  /* First the main-window */

  /* Set the attributes of the window */

  /* I want to have a gray background */
  win_attr.background_pixel = X_seslist.current->colors[COLORVALUE_GRAY];

  /* What events do I want to receive for this window?
  ** ConfigureNotify: this event announces actual changes to a window's configuration (size, position, border and
  ** atcking order).
  ** DestroyNotify: This event reports that a window has been destroyed
  ** Expose: this event is generated when a window becomes visible
  */

  win_attr.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask;

  /* Create the window. (Note that window is not visible until it is "mapped") */

  X_seslist.current->main_window = XCreateWindow(
						 /* The display, the connection to the X-server */
						 X_seslist.current->display,
						 /* Root-window of the main-window, the main X root in this case */
						 X_seslist.current->root_window,
						 /* (x, y) - Coordinates */
						 0, 0,
						 /* width */
						 20 + WINDOW_WIDTH,
						 /* height */
						 50 + (2 * WINDOW_HEIGHT) + (3 * CHAR_HEIGHT),
						 /* border width */
						 0,
						 /* depth */
						 CopyFromParent,
						 /* class */
						 InputOutput,
						 /* visual */
						 CopyFromParent,
						 /* valuemask, specifies which window attributes are defined in the
						 ** "attributes" argument. 
						 */
						 CWBackPixel | CWEventMask,
						 /* The actual attributes */
						 &win_attr);

  /* Now we can create the sub-windows (snoop-window, IRC-window and input-window) */
						 
  win_attr.background_pixel = X_seslist.current->colors[COLORVALUE_BLACK];
  win_attr.event_mask       = ExposureMask;

  X_seslist.current->snoop_window = XCreateWindow(X_seslist.current->display, X_seslist.current->main_window,
						  10, 10, WINDOW_WIDTH, WINDOW_HEIGHT, 0, 
						  CopyFromParent, InputOutput, CopyFromParent,
						  CWBackPixel | CWEventMask, &win_attr);
  X_seslist.current->irc_window   = XCreateWindow(X_seslist.current->display, X_seslist.current->main_window,
						  10, WINDOW_HEIGHT + 20, WINDOW_WIDTH, WINDOW_HEIGHT, 0, 
						  CopyFromParent, InputOutput, CopyFromParent,
						  CWBackPixel | CWEventMask, &win_attr);
  X_seslist.current->input_window = XCreateWindow(X_seslist.current->display, X_seslist.current->main_window,
						  10, 30 + (2 * WINDOW_HEIGHT), WINDOW_WIDTH, CHAR_HEIGHT, 0,
						  CopyFromParent, InputOutput, CopyFromParent,
						  CWBackPixel | CWEventMask, &win_attr);

  win_attr.background_pixel = X_seslist.current->colors[COLORVALUE_GRAY];
  win_attr.event_mask       = ExposureMask | ButtonPressMask | ButtonReleaseMask | LeaveWindowMask;

  X_seslist.current->stop_window  = XCreateWindow(X_seslist.current->display, X_seslist.current->main_window,
						  10, 40 + (2 * WINDOW_HEIGHT) + CHAR_HEIGHT, WINDOW_WIDTH, 
						  2 *CHAR_HEIGHT, 0, CopyFromParent, InputOutput, CopyFromParent,
						  CWBackPixel | CWEventMask, &win_attr);

  X_seslist.current->stop_window_pressed = 0;

  /* Windows are created */
  /* Create a graphical context */

  GCValues.foreground = X_seslist.current->colors[COLORVALUE_WHITE];
  GCValues.background = X_seslist.current->colors[COLORVALUE_BLACK];

  GCValues.line_width = 4;
  GCValues.line_style = LineSolid;
  GCValues.cap_style  = CapButt;
  GCValues.join_style = JoinMiter;
  GCValues.fill_style = FillSolid;
  
  GCValues.font = X_seslist.current->font_info->fid;

  X_seslist.current->gc = XCreateGC(X_seslist.current->display, X_seslist.current->main_window, GCForeground |
				    GCBackground | GCLineWidth | GCLineStyle | GCCapStyle | GCJoinStyle | GCFont | 
				    GCFillStyle, &GCValues);

  XSetIconName(X_seslist.current->display, X_seslist.current->main_window, VERSION);
  X_print_window_title();

  sizehints.width = sizehints.min_width = sizehints.max_width = 20 + WINDOW_WIDTH;
  sizehints.height = sizehints.min_height = sizehints.max_height = 50 + (2 * WINDOW_HEIGHT) + (3 * CHAR_HEIGHT);
  sizehints.flags = USSize | PSize | PMinSize | PMaxSize;

  XSetWMNormalHints(X_seslist.current->display, X_seslist.current->main_window, &sizehints);

  XMapWindow(X_seslist.current->display, X_seslist.current->main_window);
  XMapWindow(X_seslist.current->display, X_seslist.current->snoop_window);
  XMapWindow(X_seslist.current->display, X_seslist.current->irc_window);
  XMapWindow(X_seslist.current->display, X_seslist.current->input_window);
  XMapWindow(X_seslist.current->display, X_seslist.current->stop_window);

  XFlush(X_seslist.current->display);

  return 3;
}


void draw_stop_window_outline() {

  if (X_seslist.current->stop_window_pressed)
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_HIGHGRAY]);
  else
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_DEEPGRAY]);

  /* horizontal line, above stop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    (2 * WINDOW_HEIGHT) + CHAR_HEIGHT + 39, WINDOW_WIDTH + 11, (2 * WINDOW_HEIGHT) + CHAR_HEIGHT + 39);
  /* vertical line, left to stop window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    (2 * WINDOW_HEIGHT) + CHAR_HEIGHT + 39, 9, (2 * WINDOW_HEIGHT) + 39 + (3 * CHAR_HEIGHT));

  if (X_seslist.current->stop_window_pressed)
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_DEEPGRAY]);
  else
    XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_HIGHGRAY]);
  
  /* vertical line, right to stop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, WINDOW_WIDTH + 11,
	    (2 * WINDOW_HEIGHT) + CHAR_HEIGHT + 39, WINDOW_WIDTH + 11, (2 * WINDOW_HEIGHT) + 39 + (3 * CHAR_HEIGHT) + 1);
  /* horizontal line, below stop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    (2 * WINDOW_HEIGHT) + 39 + (3 * CHAR_HEIGHT) + 1, WINDOW_WIDTH + 11, (2 * WINDOW_HEIGHT) + 39 + 
	    (3 * CHAR_HEIGHT) + 1);
}


static void redraw_main_window() {

  XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_DEEPGRAY]);
  
  /* horizontal line above snoop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 9, WINDOW_WIDTH + 11, 9);
  /* vertical line, left to snoop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 9, 9, WINDOW_HEIGHT+11);
  /* horizontal line above irc-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, WINDOW_HEIGHT + 19, 
	    WINDOW_WIDTH + 11, WINDOW_HEIGHT + 19);
  /* vertical line, left to irc-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, WINDOW_HEIGHT + 19, 9,
	    WINDOW_HEIGHT + 19 + WINDOW_HEIGHT + 2);
  /* horizontal line, above input-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    (2 * WINDOW_HEIGHT) + 29, WINDOW_WIDTH + 11, (2 * WINDOW_HEIGHT) + 29);
  /* vertical line, left to input_window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    (2 * WINDOW_HEIGHT) + 29, 9, (2 * WINDOW_HEIGHT) + 29 + CHAR_HEIGHT);

  XSetForeground(X_seslist.current->display, X_seslist.current->gc, X_seslist.current->colors[COLORVALUE_HIGHGRAY]);

  /* vertical line, right to snoop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, WINDOW_WIDTH + 11, 9,
	    WINDOW_WIDTH + 11, WINDOW_HEIGHT+11);
  /* horizontal line, below snoop-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, WINDOW_HEIGHT+11,
	    WINDOW_WIDTH + 11, WINDOW_HEIGHT+11);
  /* vertical line, right to IRC-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, WINDOW_WIDTH + 11, 
	    WINDOW_HEIGHT + 19, WINDOW_WIDTH + 11, WINDOW_HEIGHT + 19 + WINDOW_HEIGHT + 2);
  /* horizontal line, below IRC-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    WINDOW_HEIGHT + 19 + WINDOW_HEIGHT + 2, WINDOW_WIDTH + 11, WINDOW_HEIGHT + 19 + WINDOW_HEIGHT + 2);
  /* vertical line, right to input-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, WINDOW_WIDTH + 11,
	    (2 * WINDOW_HEIGHT) + 29, WINDOW_WIDTH + 11, (2 * WINDOW_HEIGHT) + 29 + CHAR_HEIGHT + 1);
  /* horizontal line, below input-window */
  XDrawLine(X_seslist.current->display, X_seslist.current->main_window, X_seslist.current->gc, 9, 
	    (2 * WINDOW_HEIGHT) + 29 + CHAR_HEIGHT + 1, WINDOW_WIDTH + 11, (2 * WINDOW_HEIGHT) + 29 + CHAR_HEIGHT + 1);

  draw_stop_window_outline();
}


void read_X_events() {

  XEvent event;
  int    i;
  char   buffer[10];
  KeySym keysym;

  do
    {
      XNextEvent(X_seslist.current->display, &event);
      switch (event.type) {
      case Expose:
	/* Request to draw a (part of a) window */
	if (event.xexpose.window == X_seslist.current->main_window && event.xexpose.count == 0) {
	  redraw_main_window();
	}
	else if (event.xexpose.window == X_seslist.current->irc_window || 
		 event.xexpose.window == X_seslist.current->snoop_window) {
	  for (i=0; i<X_NR_LINES; i++) {
	    if (((event.xexpose.y >= i * CHAR_HEIGHT) && (event.xexpose.y < (i * CHAR_HEIGHT) + CHAR_HEIGHT)) ||
		((event.xexpose.y < CHAR_HEIGHT * i) && (event.xexpose.y + event.xexpose.height > (CHAR_HEIGHT * i) + 
							 CHAR_HEIGHT)) ||
		((event.xexpose.y + event.xexpose.height > i * CHAR_HEIGHT) &&
		 (event.xexpose.y + event.xexpose.height <= (i * CHAR_HEIGHT) + CHAR_HEIGHT))) {
	      X_writestringnumber2current(event.xexpose.window == X_seslist.current->snoop_window ? X_SNOOP_WINDOW :
					  X_IRC_WINDOW, i);
	    }
	  }
	}
	else if (event.xexpose.window == X_seslist.current->input_window && event.xexpose.count == 0) {
	  X_print_input_buffer();
	}
	break;
      case KeyPress:
	if (XLookupString(&event.xkey, buffer, 10, &keysym, NULL)) {
	  X_input_addchar(buffer[0]);
	}
	else {
	  switch (keysym) {
	  case XK_Left:
	    if (X_seslist.current->input_cursor > 0) {
	      X_seslist.current->input_cursor--;
	      X_print_input_buffer();
	    }
	    break;
	  case XK_Right:
	    if (X_seslist.current->input_cursor < X_seslist.current->buffer_length) {
	      X_seslist.current->input_cursor++;
	      X_print_input_buffer();
	    }
	    break;
	  case XK_Up:
	    strcpy(X_seslist.current->input_buffer, history_prev());
	    X_seslist.current->buffer_length = X_seslist.current->input_cursor = strlen(X_seslist.current->input_buffer);
	    X_print_input_buffer();
	    break;
	  case XK_Down:
	    strcpy(X_seslist.current->input_buffer, history_next());
	    X_seslist.current->buffer_length = X_seslist.current->input_cursor = strlen(X_seslist.current->input_buffer);
	    X_print_input_buffer();
	    break;
	  }
	}
	break;
      case DestroyNotify:
	XFreeColormap(X_seslist.current->display, X_seslist.current->colormap);
	XFreeFont(X_seslist.current->display, X_seslist.current->font_info);
	XFreeGC(X_seslist.current->display, X_seslist.current->gc);
	XCloseDisplay(X_seslist.current->display);
	del_X_sockptr(X_seslist.current->sockptr);
	X_seslist_deletenode();
	return;
      case ButtonPress:
	if (event.xbutton.window == X_seslist.current->stop_window) {
	  if (X_seslist.current->stop_window_pressed == 0) {
	    X_seslist.current->stop_window_pressed = 1;
	    draw_stop_window_outline();
	  }
	}
	break;
      case ButtonRelease:
	if (event.xbutton.window == X_seslist.current->stop_window) {
	  if (X_seslist.current->stop_window_pressed) {
	    X_seslist.current->stop_window_pressed = 0;
	    draw_stop_window_outline();
	    if (event.xbutton.same_screen) {
	      XFreeColormap(X_seslist.current->display, X_seslist.current->colormap);
	      XFreeFont(X_seslist.current->display, X_seslist.current->font_info);
	      XFreeGC(X_seslist.current->display, X_seslist.current->gc);
	      XCloseDisplay(X_seslist.current->display);
	      del_X_sockptr(X_seslist.current->sockptr);
	      X_seslist_deletenode();
	      return;
	    }
	  }
	}
	break;
      case LeaveNotify:
	if (event.xcrossing.window == X_seslist.current->stop_window) {
	  if (X_seslist.current->stop_window_pressed) {
	    X_seslist.current->stop_window_pressed = 0;
	    draw_stop_window_outline();
	  }
	}
      }
    }
  while (XQLength(X_seslist.current->display));
  XFlush(X_seslist.current->display);
}


#endif /* not NO_X */
